/*
 * $Id: epgfilter.cpp 399 2020-10-22 17:12:06Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation du filtre (interne) EPG.
 *
 *	Ce fichier implmente le filtre (interne) EPG, qui sert 
 *	rcuprer, dans le flux TS du multiplex courant, toutes les informations
 *	destines au guide lectronique des programmes (EPG).
 **/

#include "stdafx.h"
#include "epgfilter.h"
#include "..\..\channels.h"
#include "crc32.h"
#include "mpeg2defs.h"
#include "..\..\epg.h"

CEPGFilter::CEPGFilter(IUnknown *pUnk, HRESULT *phr) :
	CBaseRenderer(__uuidof(this), TEXT(EPG_FILTER_NAME), pUnk, phr)
{
	myprintf(TEXT("%") W2t TEXT(" construit") EOL, szFilterName);
}

CEPGFilter::~CEPGFilter()
{
	myprintf(TEXT("%") W2t TEXT(" detruit") EOL, szFilterName);
}

LPCWSTR CEPGFilter::szFilterName = WTXT(EPG_FILTER_NAME);

CUnknown * WINAPI CEPGFilter::CreateInstance(IUnknown *pUnk, HRESULT *phr) // static
{
	CEPGFilter *pNewFilter = new CEPGFilter(pUnk, phr);

	if (phr) {
		if (pNewFilter == NULL)
			*phr = E_OUTOFMEMORY;
		else
			*phr = S_OK;
	}

	return pNewFilter;
}

static bool parse_eit_entry(Chaine & cChan, int ixChaine, const SI_EIT::entry & sEitEn, BYTE nFlags)
{
	EmissionBase cEBase(sEitEn);

	SYSTEMTIME	sUtcTime;

	GetSystemTime(&sUtcTime);

	UINT64		nUtcTime	= SI_EIT::e_time(sUtcTime)();

	// Ne pas retenir les missions dont l'heure de fin est dpasse de plus de 10 minutes
	if (nUtcTime > cEBase.nUtcEnd10)
		return false;

	// Chercher si l'mission n'est pas dj mmorise, et trouver le point d'insertion, s'il y a lieu
	CAutoLock				cLock(cChan.crit_sec());	// Protection contre les modifications concurrentes
	ChanEventList::iterator	it	= cChan.begin();		// itrateur d'exploration

	while (it != cChan.end()) {
		Emission &	cEmis	= **it;

		if (nUtcTime > cEmis.nUtcEnd10) {
			// Descripteur d'mission prim: dtruire
			it = cChan.erase(it);
			continue;
		}
		if (cEBase.nEvtId == cEmis.nEvtId) {	// Vrifier si item existant
			cEmis.nFlags |= nFlags;
			return false;		// L'mission est dj mmorise
		}
		if (cEBase.nUtcBegT < cEmis.nUtcBegT) {	// Vrifier si on n'a pas dpass l'heure dans la liste
			break;
		}
		++it;
	}

	Emission * pEmis = new Emission(sEitEn, ixChaine, cChan.cle(), nFlags);

	// Chercher une nouvelle fois avec un test de comparaison plus complet
	for (ITERATE_LIST(cChan, EmissionPtr, it2)) {
		EmissionPtr	& pEmi2	= *it2;

		if (*pEmi2 == *pEmis) {
			myprintf(TEXT("Remplacement EPG %s (0x%04x remplace 0x%04x)") EOL,
				pEmis->getName().c_str(), pEmis->nEvtId, pEmi2->nEvtId);
			pEmi2.reset(pEmis);
			return false;		// L'mission est dj mmorise
		}
	}

	// Insertion d'un nouvel lment au point d'insertion
	cChan.insert(it, pEmis);
	//myprintf(TEXT("Ajout d'mission 0x%04x : %s") EOL,
	//	pEmis->nEvtId, pEmis->getName().c_str());
	return true;
}

/// \todo Vrifier souci probable en cas de rception multiple de la mme chane sur plusieurs frquences
static void parse_eit(const SI_EIT & eit, size_t size)
{
	bool		modif		= false;
	ChaineIDs	sIDs(eit.onid, eit.tsid, eit.service_id());
	int			ixChaine	= Canaux.trouve_par_Ids(sIDs);
	BYTE		nFlags		= (eit.table_id() > si_EIT_oth) ? DEF_IN_SCHEDULE : DEF_IN_ACTUAL;

	if (ixChaine < 0)
		return;

	Chaine & cChan = Canaux[ixChaine];

	// Ignorer chanes inactives
	if (cChan.etat == ec_inactive)
		return;

	for (SI_EIT::entryScanner se(eit.entryRange()); !se; ++se) {
		if (parse_eit_entry(cChan, ixChaine, se(), nFlags))
			modif = true;
	}

	// Passe par le thread principal pour mettre  jour l'affichage, sinon a plante
	if (modif)
		CEpgDialog::PostDlgMessage(WM_APP_EPG_UPDATED);

	if (modif && ixChaine == ixChaineCourante && (nFlags & DEF_IN_ACTUAL)!=0)
		PostMessage(hMainWnd, WM_APP_EPG_UPDATED, 0, 0);
}

HRESULT CEPGFilter::DoRenderSample(IMediaSample *pMediaSample)
{
	const SI_EIT *	peit;
	size_t			taille = pMediaSample->GetActualDataLength();

	pMediaSample->GetPointer((LPBYTE *)(&peit));

	if (
		(peit->table_id()>=si_EIT_act && peit->table_id()<=si_EIT_osf) &&
		taille >= peit->size()
		&& peit->check_crc32()
	) {
		parse_eit(*peit, taille);
	}

	return S_OK;
}

HRESULT CEPGFilter::CheckMediaType(const CMediaType * pmt)
{
	if ( *pmt->Type() == MEDIATYPE_MPEG2_SECTIONS) return S_OK;
	return S_FALSE;
}
