/*
 * $Id: pmtfilter.cpp 306 2011-09-28 09:28:04Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation du filtre (interne) PMT.
 *
 *	Ce fichier implmente le filtre (interne) PMT, qui sert 
 *	dtecter  la vole les changements de caractristiques survenant
 *	dans le flux TS de la chane courante.
 **/

#include "stdafx.h"
#include "pmtfilter.h"

#if USE_PMTFILTER

#include "..\..\parse.h"		/// \todo Supprimer cette dpendance !
#include "crc32.h"

#define INITIAL_GRACE_COUNT 20

CPMTFilter::CPMTFilter(IUnknown *pUnk, HRESULT *phr) :
	CBaseRenderer(__uuidof(this), TEXT(PMT_FILTER_NAME), pUnk, phr),
  #ifdef _DEBUG
	m_nPmtCount(0),
  #endif
	m_fnCallBack(NULL),
	m_PmtVersion(UINT(-1)),
	m_dwUserData(0),
	m_nGraceCount(INITIAL_GRACE_COUNT)
{
	myprintf(TEXT("%") W2t TEXT(" construit") EOL, szFilterName);
}

CPMTFilter::~CPMTFilter()
{
	myprintf(TEXT("%") W2t TEXT(" detruit") EOL, szFilterName);
}

LPCWSTR CPMTFilter::szFilterName = WTXT(PMT_FILTER_NAME);

//
// NonDelegatingQueryInterface
//
STDMETHODIMP CPMTFilter::NonDelegatingQueryInterface(REFIID riid, void** ppv)
{
	if (riid == __uuidof(IPMTMonitor))
		return GetInterface((IPMTMonitor *)this, ppv);
	return __super::NonDelegatingQueryInterface(riid, ppv);
}

CUnknown * WINAPI CPMTFilter::CreateInstance(IUnknown *pUnk, HRESULT *phr) // static
{
	CPMTFilter *pNewFilter = new CPMTFilter(pUnk, phr);

	if (phr) {
		if (pNewFilter == NULL)
			*phr = E_OUTOFMEMORY;
		else
			*phr = S_OK;
	}

	return pNewFilter;
}

HRESULT CPMTFilter::DoRenderSample(IMediaSample *pMediaSample)
{
	const SI_PMT *	ppmt;
	size_t			taille	= pMediaSample->GetActualDataLength();
	CAutoLock		cLock(&m_cPmtLock);

	pMediaSample->GetPointer((LPBYTE *)(&ppmt));
	//myprintf(TEXT("PMT DoRenderSample %i, cnxt=%u, version=%i") EOL, taille, ppmt->current_next_indicator(), ppmt->version_number());

	if (m_nGraceCount > 0) {
		--m_nGraceCount;
		return S_OK;
	}

	if (
		m_fnCallBack &&
		ppmt->current_next_indicator() &&
		ppmt->table_id()==si_PMT &&
		taille >= ppmt->size() &&
		ppmt->check_crc32()
	) {
		ChainePMT lpmt_e;

		if (ppmt->version_number() != m_PmtVersion) {
			m_PmtVersion = ppmt->version_number();
		}
		lpmt_e.pmt_pid = m_sExpectedPmt.pmt_pid;	// Rcupration pid PMT (pour que la comparaison n'choue pas systmatiquement)
		parse_pmt(*ppmt, false, lpmt_e);
  #ifdef _DEBUG
		++m_nPmtCount;		// Compter les tables PMT (a servira peut-tre un jour hors debugging)
		// Beep(400, 1);	// Debugging avanc : met un (affreux) son  chaque table PMT reue
  #endif

		if (lpmt_e != m_sExpectedPmt) {	// "operator !=" maintenant implment
			m_fnCallBack(m_sExpectedPmt, lpmt_e, m_dwUserData);
			m_sExpectedPmt = lpmt_e;	// Une seule fois !
		}
	}

	return S_OK;
}

STDMETHODIMP CPMTFilter::StartPmtMonitoring(const ChainePMT & sExpectedPmt, PmtMonitorCallback fnCallBack, DWORD_PTR dwUserData)
{
	CAutoLock cLock(&m_cPmtLock);

	m_sExpectedPmt	= sExpectedPmt;
	m_fnCallBack	= fnCallBack;
	m_PmtVersion	= UINT(-1);
	m_dwUserData	= dwUserData;
	m_nGraceCount	= INITIAL_GRACE_COUNT;
	return S_OK;
}

STDMETHODIMP CPMTFilter::StopPmtMonitoring()
{
	CAutoLock cLock(&m_cPmtLock);

	m_fnCallBack	= NULL;
	m_PmtVersion	= UINT(-1);
	return S_OK;
}

HRESULT CPMTFilter::CheckMediaType(const CMediaType * pmt)
{
	if ( *pmt->Type() == MEDIATYPE_MPEG2_SECTIONS)
		return S_OK;
	return S_FALSE;
}

#endif // #if USE_PMTFILTER
