/*
 * $Id: capture.h 348 2012-07-01 20:27:42Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions pour les enregistrements (TS, PS, multiplex)
 *
 *	Ce fichier dfinit les objets ncessaires  l'enregistrement en mode
 *	TS, PS et multiplex,  partir des chantillons fournis par l'interface
 *	"callback" du filtre "Sample Grabber".
 **/

#pragma once

#include "chanutils.h"
#include "sbuffer.h"
#include "mean_helper.h"

#ifdef _DEBUG

// Activation ou dsactivation de divers lments  tracer en cours de traitement
// (dcommenter pour activer, commenter pour dsactiver)

#define TRACE_CCAPTURE_CONSTR		// Construction (et destruction) CCapture

#endif	// #ifdef _DEBUG


// ----------------------------------------------------------------------------

class CCapture
{
	HANDLE hFile;

protected:

	/// Classe utilitaire pour amener temporairement le pointeur de fichier
	/// en arrire dans le fichier, puis le retaurer  la destruction.
	class BackSeek
	{
		CCapture &		cCapture;
		UINT64			nPosSave;
	public:
		BackSeek(CCapture * pCapture) throw(...) :
			cCapture(*pCapture),
			nPosSave(pCapture->filepos())
		{
		}

		~BackSeek() throw(...) {
			cCapture.seek(nPosSave); }
	};

	/**
	 * \brief Constructeur de base
	 * \param[in]	hFil	"Handle" de fichier ouvert en criture pour enregistrer le
	 *						rsultat de la capture
	 **/
	CCapture(HANDLE hFil);

	/**
	 * \brief Sortie d'un bloc de donnes vers le fichier
	 * \param[in]	pBuf	Pointeur sur le premier octet du bloc
	 * \param[in]	sizBuf	Longueur du bloc
	 * \return		0 si succs, code d'erreur systme si chec
	 **/
	DWORD ecrit_fichier(LPCVOID pBuf, DWORD sizBuf) const;

	/**
	 * \brief (Re)lecture d'un bloc de donnes depuis le fichier
	 * \param[in]	pBuf	Pointeur sur le premier octet du tampon de rception
	 * \param[in]	nToRead	Longueur  lire
	 * \return		0 si succs, code d'erreur systme si chec
	 **/
	DWORD	lit_fichier(LPVOID pBuf, DWORD nToRead) const;

	UINT64	seek(UINT64 nPos, DWORD dwMoveMethod = FILE_BEGIN) const throw (...);

	UINT64	filepos() const throw (...) {
				return seek (0, FILE_CURRENT); }

public:

	virtual DWORD traite_paquet(const TS_packet & tsp) {
		return 0; }

	/// Test si une "handle" de fichier valide existe pour enregistrer
	bool fichier_valide() const {
		return hFile!=INVALID_HANDLE_VALUE;}

	// UINT64 pos_fichier(void);

	/// Destructeur
	virtual ~CCapture();
};

// ----------------------------------------------------------------------------

class CCapture_TS : public CCapture, public Mpeg_TS_Input
{
	ChainePMT	pistes;

public:

	CCapture_TS(HANDLE hFil, const ChainePMT & lst_pistes);

	virtual DWORD traite_paquet(const TS_packet & tsp);
};

// ----------------------------------------------------------------------------

enum PS_buffer_type
{
	pst_undef,
	pst_Audio,
	pst_Video
};

// Obtenir une chane de caractres identifiant le type de contenu :
inline LPCSTR pbt2str(PS_buffer_type eBType)
{
	return EnumToString(eBType, "\x01\0audio\0vido\0");
}

// ----------------------------------------------------------------------------

class PS_buffer : public StreamBuffer
{
	friend class CCapture_PS;
	friend class PS_chunk;

	CCapture_PS &	cCapture;		//!< Rfrence sur l'objet de capture commun
	UINT16			pid;
	bool			valid;
	UINT8			continuity_counter;

public:
	PS_buffer_type	eBType;
	PS_buffer(CCapture_PS & cCapt, UINT16 used_pid, UINT8 maxChnk, PS_buffer_type eBTyp);

	DWORD traite_paquet(const TS_packet & tsp);

	/// Traitement d'un paquet reconstitu
	virtual void send(MemRng sMm) throw(...);

	virtual ~PS_buffer() {
		}
};

// ----------------------------------------------------------------------------

class Video_PS_buffer : public PS_buffer
{
	int		nPictCount;

public:
	Video_PS_buffer(CCapture_PS & cCapt, UINT16 used_pid) :
		PS_buffer(cCapt, used_pid, 128, pst_Video),
		nPictCount(0)
	{
	}
};

// ----------------------------------------------------------------------------

class Audio_PS_buffer : public PS_buffer
{
public:
	Audio_PS_buffer(CCapture_PS & cCapt, UINT16 used_pid) :
		PS_buffer(cCapt, used_pid, 32, pst_Audio)
	{
	}
};

typedef MeanHelper<UINT32, 0x10> RateHelper;

/// tat de construction d'une capture PS :
enum CaptureState {
	cs_idle,
	cs_error,
	cs_started,
	cs_first_pcr_set,								//!< au moins une rfrence d'horloge a t reue
	cs_running,										//!< tous les flux sont en cours de capture
	cs_stopping,									//!< arrt en cours
	cs_finished										//!< capture termine
};

// ----------------------------------------------------------------------------

class CCapture_PS : public CCapture, public Mpeg_TS_Input
{
	friend class PS_buffer;
	friend class PS_chunk;

	UINT64						first_pcr;			//!<  Mmorisation du 1er PCR pour recalage  zro des suivants
	UINT64						curr_pcr;			//!< PCR courant (interpolation)
	UINT64						last_pcr;			//!< Dernier PCR reu
	UINT32						curr_pcr_second;	//!< Seconde entire du PCR courant (pour tests statistiques)
	UINT32						pck_pcr_count;		//!< Comptage des paquets "TS" depuis le dernier PCR reu
	UINT32						pcr_rate;			//!< Taux d'incrmentation du PCR par paquet
	UINT32						byte_count;			//!< Comptage des octets mis en mode PS pour calcul "multiplex rate"
	UINT32						byte_mux_rate;		//!< "multiplex rate" courant, calcul en octets par seconde
	RateHelper					pcr_rate_helper;	//!< Assistance au calcul de l'intervalle de temps moyen entre paquets TS
	RateHelper					mux_rate_helper;	//!< Assistance au calcul du "multiplex rate"
	CaptureState				eState;				//!< tat de progression de la capture
	bool						sys_header_set;		//!< L'enregistrement "System Header" a t mis
	auto_ptr<Video_PS_buffer>	pVideoBuf;			//!< Objet de construction du flux vido
	auto_ptr<Audio_PS_buffer>	pAudioBuf;			//!< Objet de construction du flux audio

public:
	CCapture_PS(HANDLE hFil, UINT16 video, UINT16 audio);

	virtual DWORD	traite_paquet(const TS_packet & tsp);

	UINT64			adjust_pcr(UINT64 pcr) const {
						return pcr-first_pcr;}
	UINT64			get_dst_pcr() const {
						return adjust_pcr(curr_pcr);}

	virtual			~CCapture_PS();
};

// ----------------------------------------------------------------------------

class CCapture_Stream : public CCapture, public Mpeg_TS_Input	// -> Multiplex
{
public:
	CCapture_Stream(HANDLE hFil);

	virtual DWORD traite_paquet(const TS_packet & tsp);
};

// ----------------------------------------------------------------------------
