/*
 * $Id: console.h 370 2020-09-09 12:43:56Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinition pour la console de diagnostic
 *
 *	Ce fichier inclut les dfinitions pour l'affichage de la console de diagnostic qui
 *	est affiche (uniquement) dans les compilations en mode \p _DEBUG ou \p _DIAGNOSTIC.
 **/

#pragma once

// ----------------------------------------------------------------------------

#ifndef	 CONSOLE_DECLSPEC
	#define CONSOLE_DECLSPEC	__fastcall
#endif

// ----------------------------------------------------------------------------

/**
 * Affichage d'un message d'erreur non format dans la console
 *
 * L'affichage du message est prcd d'un "time stamp" en dbut de ligne,
 * indiquant l'heure exacte du message avec les secondes et les millisecondes.
 * Le message est dcoup selon les retours chariot (CR) qu'il contient, et
 * les lignes produites sont affiches une par une, la premire avec le
 * "time stamp", et les suivantes avec des espaces pour aligner verticalement
 * les lignes affiches.
 * \param[in]	msg		Message  afficher
 **/
void CONSOLE_DECLSPEC OutputMsg(LPCTSTR msg);

/**
 * Affichage d'un message format "vprintf" dans la console
 *
 * Le message format est ensuite affich avec la fontion \p OutputMsg ci-dessus.
 * \param[in]	pszFmt	Chane de formatage
 * \param[in]	argptr	Arguments style \p vprintf (voir aussi note)
 * \retval		\p true si un message a t valu
 * \retval		\p false si un code "%?" a provoqu la lecture d'une valeur fausse
 * \note		Si la chane de formatage dbute par "%?", alors le premier paramtre
 *				suivant est un boolen qui conditionne si l'affichage a lieu ou pas.
 *				Le "%?" n'est bien sr pas affich.
 **/
bool CONSOLE_DECLSPEC OutputVF(LPCTSTR pszFmt, va_list & argptr);

/**
 * Affichage d'un message format "printf" dans la console
 *
 * Le message format est ensuite affich avec la fontion \p OutputVF ci-dessus,
 * elle-mme utilisant la fonction \p OutputMsg.
 * \param[in]	pszFmt	Chane de formatage
 * \param[in]	...		Arguments style \p printf (voir aussi note)
 * \retval		\p true si un message a t valu
 * \retval		\p false si un code "%?" a provoqu la lecture d'une valeur fausse
 * \note		Si la chane de formatage dbute par "%?", alors le premier paramtre
 *				suivant est un boolen qui conditionne si l'affichage a lieu ou pas.
 *				Le "%?" n'est bien sr pas affich.
 **/
inline bool OutputF(LPCTSTR pszFmt, ...)
{
	va_list		argptr;

	va_start(argptr, pszFmt);

	bool bRes = OutputVF(pszFmt, argptr);

	va_end(argptr);
	return bRes;
}

// ----------------------------------------------------------------------------

/** \defgroup DefAnsi Macros de dfinitions de squences ANSI
 * Macros destines  dfinir les squences ANSI  insrer dans les affichages \p myprintf
 * @{ */

/// Caractre "fin de ligne"
#define EOL			TEXT("\n")

/// Caractre "tabulation"
#define TAB			TEXT("\t")

/// "myprintf" conditionnel ( son 1er argument)
#define CMyP		TEXT("%?")

#define CSI			TEXT("\x1b[")	//!< Dbut de squence

/// Effacement d'cran
#define CLS			CSI TEXT("2J")

// (noter que dbuter par un ";" quivaut  placer un 0 devant)

#define HBLACK		"30;1"			//!< Noir haute intensit (gris fonc, en fait)
#define HRED		"31;1"			//!< Rouge haute intensit
#define HGREEN		"32;1"			//!< Vert haute intensit
#define HYELLOW		"33;1"			//!< Jaune haute intensit
#define HBLUE		"34;1"			//!< Bleu haute intensit
#define HMAGENTA	"35;1"			//!< Magenta haute intensit
#define HCYAN		"36;1"			//!< Cyan haute intensit
#define HWHITE		";1"			//!< Blanc haute intensit (devrait tre "37;1", mais le blanc tant "par dfaut", la surbrillance suffit)

#define LBLACK		";30"			//!< Noir normal (= noir sur fond noir ...)
#define LRED		";31"			//!< Rouge normal
#define LGREEN		";32"			//!< Vert normal
#define LYELLOW		";33"			//!< Jaune normal (plutt brun/ocre, en fait)
#define LBLUE		";34"			//!< Bleu normal
#define LMAGENTA	";35"			//!< Magenta normal
#define LCYAN		";36"			//!< Cyan normal
#define LWHITE		"0"				//!< Blanc normal (tout par dfaut), =";37"

#define ATTPUSH		"<"				//!< Empiler l'attribut (non standard)
#define ATPOP		">"				//!< Dpiler l'attribut (non standard)
/** @} */

// ----------------------------------------------------------------------------

#if !defined(CONSOLE_WITHOUT_COLORS) || CONSOLE_WITHOUT_COLORS == 0

/** \defgroup SeqAnsi Macros de synthse de squences ANSI
 * Macros destines  insrer des squences ANSI dans les affichages \p myprintf
 * @{ */

/**	\brief "Select Graphic Rendition" (dfinit une couleur).
 *
 * \param[in]	clr	Couleur (suite de chiffres ANSI spars par des ";").
 **/
#define SGR(clr) CSI TEXT(clr) TEXT("m")

/**	\brief "Select Graphic Rendition" (dfinit une couleur) + empilage.
 *
 * \param[in]	clr	Couleur (suite de chiffres ANSI spars par des ";").
 **/
#define SGRP(clr) CSI TEXT(ATTPUSH) TEXT(";") TEXT(clr) TEXT("m")

/**	\brief Macro  utiliser pour coloriser des parties de texte.
 *
 * \param[in]	clr	Couleur (suite de chiffres ANSI spars par des ";").
 * \param[in]	msg	Texte  coloriser
 * \note Il est (dornavant) possible d'imbriquer les couleurs.
 **/
#define COLR(clr, msg) SGRP(clr) msg SGR(ATPOP)	//!< Couleur

///  utiliser dans la liste d'arguments si le paramtre \p clr ci-dessus est un argument printf "%s".
#define COLR_ARG(arg)	arg ,

/**	\brief Macro  utiliser pour faire ressortir des mots ou des messages en surbrillance.
 *
 * \param[in]	msg	Texte  mettre en surbrillance
 **/
#define	BOLD(msg) COLR(HWHITE, msg)

/**	\brief Message d'erreur surlign en magenta + EOL
 *
 * Macro  utiliser pour tout ce qui doit tre considr comme un message d'erreur.
 * Ce texte apparatra en couleur "magenta" dans la console (le rouge tait trop pnible  lire)
 * Le code de fin de ligne est inclus dans la macro.
 * \param[in]	msg	Texte du message d'erreur
 **/
#define ERRX(msg) COLR(HMAGENTA, msg) EOL

#define MSGP(pos, msg) CSI TEXT(pos) TEXT("@") TEXT(msg) TEXT("\r")	//!< Message positionn
/** @} */

#else

	/** \defgroup DummySeqAnsi Macros de substitution aux squences ANSI
	 * Versions utilises pour les messages d'erreur sans console (et donc sans formatage)
	 * @{ */

	#define SGR(seq)
	#define SGRP(seq)
	#define MSGP(pos, msg)
	#define COLR(clr, msg)	msg
	#define COLR_ARG(arg)
	#define	BOLD(msg)		msg
	#define ERRX(msg)		msg EOL

	/** @} */

#endif

// ----------------------------------------------------------------------------

enum StartMode
{	// Note : certains de ces modes peuvent ne pas tre encore implments
	sm_None,		//!< Rien ( part sortie fichier ventuelle)
	sm_Attach,		//!< Tenter de s'attacher  une console existante
	sm_Debug,		//!< Ne pas ouvrir de console, envoyer tous les messages vers \p OutputDebugString
	sm_Console,		//!< Ouvrir une nouvelle console
	sm_ConsoleFb	//!< Ouvrir une nouvelle console avec "fallback" vers \p OutputDebugString si celle-ci est ferme
};

/**
 * Initialisation de la console de debugging.
 * \param[in]	pszTitle	Titre de la console
 * \param[in]	eMode		Mode d'ouverture
 * \param[in]	bTimeStamps	Mettre \p true si un horodatage est souhait
 * \param[in]	pszLogFile	Nom de fichier journal optionnel (mettre NULL si non dsir)
 **/
void CONSOLE_DECLSPEC startConsole(LPCTSTR pszTitle, StartMode eMode, bool bTimeStamps=false, LPCTSTR pszLogFile=NULL);

/**
 * Initialisation de la console de debugging.
 * \param[in]	alloc		\p true si on veut ouvrir une nouvelle console,
 *							ou bien \p false si on veut s'attacher  la console existante
 * \param[in]	pszLogFile	Nom du fichier de journalisation
 * \param[in]	(LPCVOID)	Ancien paramtre pour position de fentre  restaurer (ignor)
 * \deprecated	Remplac par \p startConsole, avec les paramtres adquats
 **/
inline void startConsoleWin(bool alloc, LPCTSTR pszLogFile, LPCVOID = NULL)
{
	startConsole(TEXT("Diagnostic console"), alloc ? sm_Console : sm_Attach, true, pszLogFile);
}

/**
 * Initialisation d'une pseudo-console envoyant les messages sur la sortie de debugging
 * \deprecated	Remplac par \p startConsole, avec les paramtres adquats
 **/
inline void startConsoleDbg()
{
	startConsole(TEXT("(debug)"), sm_Debug);
}

// ----------------------------------------------------------------------------
