/*
 * $Id: mpeg2defs.cpp 365 2020-09-08 13:34:31Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentations de mthodes pour MPEG2 PS et TS
 *
 *	Implmentations de certaines mthodes pour extraire le contenu des flux
 *	MPEG2 - PS (Program Stream) et PS (Transport Stream).
 **/

#include "stdafx.h"
#include <string.h>
#include <stddef.h>
#include "mpeg2defs.h"

// ----------------------------------------------------------------------------

/**
 * Copie d'une chane de caractres avec encodage optionnel,  partir d'un pointeur sur dbut
 * et d'une longueur. La chane source n'est PAS (ncessairement) termine par zro.
 * Version pour adresse de destination type \p CHAR (caractres 8 bits) :
 *   Si un caractre d'encodage est prsent au dbut de la chane (code 0x01  0x1f), celui-ci
 *   est limin, et il est retourn par la fonction (qui autrement retourne 0).
 *   Aucune conversion n'est cependant effectue en relation avec ce code.
 * \param[in]	src		Pointeur sur la chane source
 * \param[in]	siz		Longueur de la chane source
 * \param[out]	dst		Pointeur sur un tampon destin  recevoir la chane de destination
 * \param[in]	bufsize	Longueur du tampon destin  recevoir la chane de destination
 * \retval		Code de page qui tait prsent au dbut de la chane source, ou bien 0 s'il n'y en avait pas.
 **/
UINT get_raw_str(LPCSTR src, UINT siz, LPSTR dst, UINT bufsize)
{
	UINT ixcp = 0;

	if (siz>=bufsize)
		siz=UINT(bufsize-1);
	if (siz > 0 && UINT8(*src) < 0x20) {
		ixcp = *src++;
		--siz;
	}
	memcpy(dst, src, siz);
	dst[siz] = 0;
	return ixcp;
}

/**
 * Copie d'une chane de caractres avec encodage optionnel,  partir d'un pointeur sur dbut
 * et d'une longueur. La chane source n'est PAS (ncessairement) termine par zro.
 * Version pour adresse de destination type \p WCHAR (caractres 16 bits) :
 *   - Si un caractre d'encodage est prsent au dbut de la chane (code 0x01  0x1f), celui-ci
 *     est limin, et il est retourn par la fonction (qui autrement retourne 0).
 *   - De plus, dans ce dernier cas, il est tent, dans la mesure du possible, d'utiliser ce code
 *     pour convertir la chane de caractres en chane Unicode.
 *   - Les codes de contrle ventuels (0x80  0x9f) sont convertis en codes de 0xe080  0xe09f.
 * \param[in]	src		Pointeur sur la chane source
 * \param[in]	siz		Longueur de la chane source
 * \param[out]	dst		Pointeur sur un tampon destin  recevoir la chane de destination
 * \param[in]	bufsize	Longueur du tampon destin  recevoir la chane de destination
 * \retval		Code de page qui tait prsent au dbut de la chane source, ou bien 0 s'il n'y en avait pas.
 *				Si au moins un code de contrle tait prsent dans la chane, alors la valeur 0x1000000 est ajoute
 **/
UINT get_raw_str(LPCSTR src, UINT siz, LPWSTR dst, UINT bufsize)
{
	static const UINT16 codepages[0x20] = {
		28591,	// 00 - ISO 8859-1 Latin I
		28595,	// 01 - ISO 8859-5 Cyrillic
		28596,	// 02 - ISO 8859-6 Arabic
		28597,	// 03 - ISO 8859-7 Greek
		28598,	// 04 - ISO 8859-8 Hebrew
		28599,	// 05 - ISO 8859-9 Latin 5 
		28600,	// 06 - ISO/IEC 8859-10 - Latin alphabet No. 6
		28601,	// 07 - ISO/IEC 8859-11 - Latin/Thai (draft only)
		28602,	// 08 - ISO/IEC 8859-12 - possibly reserved for Indian
		28603,	// 09 - ISO/IEC 8859-13 - Latin alphabet No. 7
		28604,	// 0a - ISO/IEC 8859-14 - Latin alphabet No. 8 (Celtic)
		28605,	// 0b - ISO 8859-15 Latin 9
		28591,	// 0c - rserv
		28591,	// 0d - rserv
		28591,	// 0e - rserv
		28591,	// 0f - rserv
		0,		// 10 - extension
		// Les valeurs 0x11  0x14 encodent pour des choses plus complexes  rsoudre plus tard
		// (si jamais un jour on diffuse notre code dans les pays asiatiques ... :-) )
		28591, 28591, 28591, 28591, 28591, 28591, 28591, 28591,
		28591, 28591, 28591, 28591, 28591, 28591, 28591
	};

	static const UINT16 codepages_x10[0x10] = {
		28591,	// 00 - reserved for future use
		28591,	// 01 - ISO/IEC 8859-1 [25] West European
		28592,	// 02 - ISO/IEC 8859-2 [26] East European
		28593,	// 03 - ISO/IEC 8859-3 [27] South European
		28594,	// 04 - ISO/IEC 8859-4 [28] North and North-East European
		28595,	// 05 - ISO/IEC 8859-5 [29] Latin/Cyrillic A.2
		28596,	// 06 - ISO/IEC 8859-6 [30] Latin/Arabic A.3
		28597,	// 07 - ISO/IEC 8859-7 [31] Latin/Greek A.4
		28598,	// 08 - ISO/IEC 8859-8 [32] Latin/Hebrew A.5
		28599,	// 09 - ISO/IEC 8859-9 [33] West European & Turkish A.6
		28600,	// 0a - ISO/IEC 8859-10 [34] North European A.7
		28601,	// 0b - ISO/IEC 8859-11 [35] Thai A.8
		28602,	// 0c - Reserved for future use
		28603,	// 0d - ISO/IEC 8859-13 [36] Baltic A.9
		28604,	// 0e - ISO/IEC 8859-14 [37] Celtic A.10
		28605	// 0f - ISO/IEC 8859-15 [38] West European
	};

	UINT	ixcp = 0;
	UINT	cp   = 28591;	// Par dfaut

	if (siz>=bufsize)
		siz=UINT(bufsize-1);
	if (siz > 0 && UINT8(*src) < 0x20) {
		ixcp = *src++;
		--siz;
		if (ixcp == 0x10 && siz >= 2) {
			UINT	ixc2 = (UINT8(src[0]) << 8) | UINT8(src[1]);

			ixcp = (ixcp << 16) | ixc2;
			src += 2;
			siz -= 2;
			if (ixc2 < _countof(codepages_x10))
				cp = codepages_x10[ixc2];
		} else {
			if (ixcp < _countof(codepages))
				cp = codepages[ixcp];
		}
	}

	UINT	sofs = 0;
	UINT8	ch;

	for (UINT i=0; i<siz; i++) {
		if ((ch = src[i]) >= DCTL_MIN_A && ch <= DCTL_MAX_A) {
			if (i>sofs) {
				UINT dsiz = MultiByteToWideChar(cp, MB_PRECOMPOSED, src+sofs, int(i-sofs), dst, bufsize);

				dst += dsiz;
				bufsize -= dsiz;
			}
			if (bufsize > 1) {
				*dst++ = DCTL_W_OFFSET | ch;
				--bufsize;
			}
			sofs = i+1;
			ixcp |= 0x1000000;
		}
	}
	if (siz>sofs)
		dst += MultiByteToWideChar(cp, MB_PRECOMPOSED, src+sofs, int(siz-sofs), dst, bufsize);
	*dst = 0;
	return ixcp;
}

// ----------------------------------------------------------------------------

/// Retourne un pointeur sur le dbut de l'espace "payload" s'il existe, ou bien
/// \p NULL s'il n'existe pas.
PUINT8	TS_hdr::get_payload() const
{
	switch (adaptation_field_control()) {

	case afcv_payload_only:
		return beg_of_fields();

	case afcv_afield_and_payload:
		return get_af().next();
	}

	return NULL;
}

// ----------------------------------------------------------------------------

/*
 * Dballage de tous les champs "data" optionnels du paquet 'hdr':
 * La structure est charge avec des pointeurs vers tous les champs
 * existants, ou des valeurs NULL pour tous les champs inexistants.
 * Le pointeur 'p_next' pointe l'espace encore libre qui suit, lequel correspond
 *  des octets de remplissage optionnels suivis des donnes du paquet.
 */
PS_data_unwrapper_full::PS_data_unwrapper_full(const PS_hdr_wd & hdr) :
	PS_data_unwrapper_base(hdr),
	p_escr(NULL),
	p_rate(NULL),
	p_trick(NULL),
	p_add_copy_info(NULL),
	p_prev_crc(NULL),
	p_ext1(NULL),
	p_priv_data(NULL),
	p_pack(NULL),
	p_seqcount(NULL),
	p_stdbuf(NULL),
	p_ext2(NULL)
{
	PUINT8	p_nxt = p_next;
	UINT8	flags = hdr.flags;

	if (flags & PES_ESCR_FLAG)
		get_and_skip(p_nxt, p_escr);
	if (flags & PES_ES_RATE_FLAG)
		get_and_skip(p_nxt, p_rate);
	if (flags & PES_DSM_TRICK_MODE_FLAG)
		get_and_skip(p_nxt, p_trick);
	if (flags & PES_ADDITIONAL_COPY_INFO_FLAG)
		get_and_skip(p_nxt, p_add_copy_info);
	if (flags & PES_CRC_FLAG)
		get_and_skip(p_nxt, p_prev_crc);
	if (flags & PES_EXTENSION_FLAG) {
		get_and_skip(p_nxt, p_ext1);
		flags = p_ext1->flags;
		if (flags & PES_PRIVATE_DATA_FLAG)
			get_and_skip(p_nxt, p_priv_data);
		if (flags & PES_PACK_HEADER_FIELD_FLAG)
			get_and_skip(p_nxt, p_pack);
		if (flags & PES_PROG_PACKET_SEQ_COUNT_FLAG)
			get_and_skip(p_nxt, p_seqcount);
		if (flags & PES_P_STD_BUFFER_FLAG)
			get_and_skip(p_nxt, p_stdbuf);
		if (flags & PES_EXTENSION_FLAG_2)
			get_and_skip(p_nxt, p_ext2);
	}
	p_next = p_nxt;
}

	bool TS_AF_unwrapper_base::parse(const TS_AF_hdr * paf)
{
	p_next = NULL;
	p_pcr = p_opcr = NULL;
	if (paf) {
		PUINT8	p_nxt = paf->beg_of_fields();
		UINT8	flags = paf->get_flags();

		if (flags & AF_PCR_FLAG)
			get_and_skip(p_nxt, p_pcr);
		if (flags & AF_OPCR_FLAG)
			get_and_skip(p_nxt, p_opcr);
		p_next = p_nxt;
	}
	return true;
}

bool TS_AF_unwrapper_full::parse(const TS_AF_hdr * paf)
{
	if (__super::parse(paf)) {
		p_splicing_point = NULL;
		p_priv_data = NULL;
		p_ext = NULL;
		p_ltw = NULL;
		p_piecewise_rate = NULL;
		p_seamless_splice = NULL;

		PUINT8	p_nxt = p_next;
		UINT8	flags = paf->get_flags();

		//memset(PUINT8(&p_splicing_point), 0,
		//	sizeof(TS_AF_unwrapper_full)-offsetof(TS_AF_unwrapper_full, p_splicing_point));

		if (flags & AF_SPLICEPT_FLAG)
			get_and_skip(p_nxt, p_splicing_point);
		if (flags & AF_PRIV_DATA_FLAG)
			get_and_skip(p_nxt, p_priv_data);
		if (flags & AF_EXTENSION_FLAG) {
			get_and_skip(p_nxt, p_ext);
			flags = p_ext->flags;

			if (flags & AF_LTW_FLAG)
				get_and_skip(p_nxt, p_ltw);
			if (flags & AF_PW_RATE_FLAG)
				get_and_skip(p_nxt, p_piecewise_rate);
			if (flags & AF_SLSPLICE_FLAG)
				get_and_skip(p_nxt, p_seamless_splice);
		}
		p_next = p_nxt;
		return true;
	}
	return false;
}

// ----------------------------------------------------------------------------

UINT16 MpegAudio_hdr::bit_rate() const
{
	#define BR(x) (x>>3)

	static const UINT8 tbl_bit_rates[2][3][15] =
	{
		{	// frquences rduites
			{	BR(0),		BR(8),		BR(16),		BR(24),		BR(32),		BR(40),		BR(48),		BR(56),
				BR(64),		BR(80),		BR(96),		BR(112),	BR(128),	BR(144),	BR(160)},	// Layer III
			{	BR(0),		BR(8),		BR(16),		BR(24),		BR(32),		BR(40),		BR(48),		BR(56),
				BR(64),		BR(80),		BR(96),		BR(112),	BR(128),	BR(144),	BR(160)},	// Layer II
			{	BR(0),		BR(32),		BR(48),		BR(56),		BR(64),		BR(80),		BR(96),		BR(112),
				BR(128),	BR(144),	BR(160),	BR(176),	BR(192),	BR(224),	BR(256)}	// Layer I
		},
		{	// frquences standard
			{	BR(0),		BR(32),		BR(40),		BR(48),		BR(56),		BR(64),		BR(80),		BR(96),
				BR(112),	BR(128),	BR(160),	BR(192),	BR(224),	BR(256),	BR(320)},	// Layer III
			{	BR(0),		BR(32),		BR(48),		BR(56),		BR(64),		BR(80),		BR(96),		BR(112),
				BR(128),	BR(160),	BR(192),	BR(224),	BR(256),	BR(320),	BR(384)},	// Layer II
			{	BR(0),		BR(32),		BR(64),		BR(96),		BR(128),	BR(160),	BR(192),	BR(224),
				BR(256),	BR(288),	BR(320),	BR(352),	BR(384),	BR(416),	BR(448)}	// Layer I
		}
	};

	MpegAudioLayer	lay		= layer();
	UINT8			br_idx	= bit_rate_index();

	if (br_idx>=15 || lay==mal_reserved || lay>mal_Layer_I)
		return uint16_UNDEF;

	return tbl_bit_rates[version()==mv_Mpeg1][lay-1][br_idx]<<3;
}

UINT16 MpegAudio_hdr::samp_frequency() const
{
	static const UINT16 t_samp_freq[4][4] =
	{
		{11025,			12000,			8000,			uint16_UNDEF},
		{uint16_UNDEF,	uint16_UNDEF,	uint16_UNDEF,	uint16_UNDEF},
		{22050,			24000,			16000,			uint16_UNDEF},
		{44100,			48000,			32000,			uint16_UNDEF}
	};

	return t_samp_freq[version()][samp_freq_index()];
}

// ----------------------------------------------------------------------------

/// Constructeur
millisec2str::millisec2str(ITIM_1k val1k)
{
	bool neg = false;

	if (val1k < 0) {
		val1k = -val1k;
		neg = true;
	}

	UINT32	millisec	= UINT32(val1k%1000);
	UINT8	sec			= UINT8((val1k/1000)%60);
	UINT8	min			= UINT8((val1k/60000)%60);
	UINT8	hr			= UINT8(val1k/3600000);

	_stprintf_s(str, _T("%2u:%02u:%02u.%03u"), hr, min, sec, millisec);
}

// ----------------------------------------------------------------------------
