/*
 * $Id: mpeg2defs.h 365 2020-09-08 13:34:31Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions, structures et mthodes pour MPEG2 PS et TS
 *
 *	Ensemble de dfinitions, structures et mthodes pour extraire le contenu des flux
 *	MPEG2 - PS (Program Stream) et TS (Transport Stream).
 **/

#pragma once

#include <mpeg2structs.h>
#include <mpeg2bits.h>
#include <mmsystem.h>
#include "utils.h"
#include "memrng.h"
#include "crc32.h"

#pragma warning(disable:4512) // warning C4512: assignment operator could not be generated

// ----------------------------------------------------------------------------

// Constantes utilises pour indiquer certaines valeurs comme indfinies
#define uint8_UNDEF		UINT8(-1)
#define uint16_UNDEF	UINT16(-1)
#define uint32_UNDEF	UINT32(-1)

// Macros utilises pour amliorer la prsentation des versions alternatives dans les structures
#define UNION_BEGIN		union { struct
#define UNION_ELSE		; struct
#define UNION_END		; }

/*#######################################################################################
  #	###################################################################################	#
  #	#																				  #	#
  #	#				DFINITIONS DES STRUCTURES DE DONNES CONTENUES					  #	#
  #	#						DANS LES FICHIERS ET FLOTS MPEG2						  #	#
  #	#																				  #	#
  #	###################################################################################	#
  #######################################################################################*/

/** \brief Assistance  l'exploration d'une suite squentielle de structures
 *
 * Classe d'assistance pour parcourir squentiellement une suite d'enregistrements
 * se succdant directement, dans un intervalle entre deux adresses mmoire dfinies,
 * mais dont les longueurs sont ingales (chaque objet de type T doit implmenter
 * un membre "next()" qui retourne un pointeur sur l'objet de type T qui le suit immdiatement).
 * La fin de la liste est atteinte lorsque l'adresse 'pend' est atteinte ; on s'attend normalement
 *  ce que le dernier objet se termine exactement  l'adresse de fin.
 **/
template <typename T>
	struct streamScanner
{
	const T * pbeg;
	const T * pend;

	streamScanner(const void * pb, const void * pe) :
		pbeg(reinterpret_cast<const T *>(pb)),
		pend(reinterpret_cast<const T *>(pe))
	{
	}

	const T & operator ++ () {
		pbeg = pbeg->next(); return *pbeg; }
	const T & operator ++ (int) {
		const T * res = pbeg; pbeg = pbeg->next(); return *res; }
	bool operator ! () const {
		return pbeg < pend; }
	operator const T & () const {
		return *pbeg; }
	operator const T * () const {
		return pbeg; }
	const T & operator () () const {
		return *pbeg; }
	UINT32 size() const {
		return UINT32(pend-pbeg); }
};

template <typename T>
	UINT8 unwrap_sequence(streamScanner<T> limits, const T * tbl[])
{
	UINT8 count = 0;

	for (;!limits; ++limits)
		tbl[count++] = limits.pbeg;
	tbl[count] = limits.pbeg;
	return count;
}

/// Rcupration d'un pointeur sur une structure de type quelconque (et de longueur variable)
/// dans un flux d'octet, et incrment du pointeur sur cette valeur de la taille de la structure
template <typename T>
	void get_and_skip(PUINT8 & p_next, T * & p_res)
{
	p_res = reinterpret_cast<T *>(p_next);
	p_next = p_res->next();	// l'objet point doit avoir une mthode "next" pour trouver la fin du champ
}

/// Rcupration d'une valeur de type quelconque dans un flux d'octet, et incrment du pointeur sur
/// cette valeur de la taille de la valeur
template <typename T>
	const T & get_and_skip(PCUINT8 & p_next)
{
	const T & res = *reinterpret_cast<const T *>(p_next);
	p_next += sizeof(T);
	return res;
}

/// Chargement d'un tableau de valeurs de type quelconque depuis un flux d'octet, et incrment du pointeur
/// sur ce tableau de la taille de la tableau
template <typename TS, typename TD, size_t S>
	void get_and_skip(PCUINT8 & p_next, TD (&a_res)[S])
{
	for (int i=0; i<S; ++i)
		a_res[i] = get_and_skip<TS>(p_next);
}

/// Reprsentation d'une valeur sur N bits au format "big endian"
template <typename T, typename D>
	struct BIGEND
{
	T	hi;
	T	lo;

	operator D() const {
		return D(T(hi))<<(sizeof(T)<<3) | T(lo); }
	D operator () () const {
		return BIGEND(*this); }
	BIGEND & operator = (D val) {
		hi = T(val>>(sizeof(T)<<3)); lo=T(val); return *this; }
	PUINT8 next() const {
		return PUINT8(this+1); }

	bool operator == (BIGEND v2) const {
			return hi == v2.hi && lo == v2.lo; }
	bool operator != (BIGEND v2) const {
			return hi != v2.hi || lo != v2.lo; }
	bool operator < (BIGEND v2) const {
			return hi != v2.hi ?  hi < v2.hi : lo < v2.lo; }
	bool operator > (BIGEND v2) const {
			return hi != v2.hi ?  hi > v2.hi : lo > v2.lo; }
};

typedef BIGEND<UINT8,UINT16>	UBE16;	/// big endian sur 16 bits
typedef BIGEND<UBE16,UINT32>	UBE32;	/// big endian sur 32 bits
typedef BIGEND<UBE32,UINT64>	UBE64;	/// big endian sur 64 bits

/// Modle pour reprsenter 9  15 bits au format "big endian"
template <int N> struct UBE : public UBE16
{
	UINT16 operator () () const {
		return UBE16::operator ()() & ((1<<N)-1); }
	UINT8 flags() const {
		return hi & (0xff << (N-8)); }
	void set_flags(UINT8 mask, UINT8 flags) {
		hi = (hi & ~mask) | flags; }
	UBE<N> & operator = (UINT16 val) {
		hi = UINT8(val>>8) | flags(); lo=UINT8(val); return *this; }
};

typedef UBE<12> LEN12;	// Longueur de champ sur 12 bits
typedef UBE<13> PID13;	// PID sur 13 bits
typedef UBE16 CRC_16;
typedef UBE32 CRC_32;

/// Reprsentation d'un octet avec mthodes pour les structures "unwrapper"
struct UINT8uw {
	UINT8	b;

	UINT8 operator()() const {
		return b; }
	PUINT8 next() const {
		return PUINT8(this+1); }
};

/// Dfinition d'une chane de caractres de longueur fixe, utilisable dans une expression
template <typename T, int N>
	struct CFStr
{
	T str[N];

	operator const T * () {
			return str; }

	const T * operator () () {
			return str; }

};

/// Dfinition d'une chane de caractres format Pascal, de longueur fixe, utilisable dans une expression
template <int N>
	struct PFStr
{
	UINT8	nLen;
	CHAR	str[N-1];

	tstring operator () () const {
		TCHAR	szWrk[N];
		for (UINT8 nInx=0; nInx < nLen; ++nInx)
			szWrk[nInx] = str[nInx];
		szWrk[nLen] = 0;
		return szWrk; }
};

/// Dfinition d'une chane de caractres de longueur 5, spcialise pour recevoir une
/// conversion des codes "FOURCC" plus un zro terminateur
struct CFstr5 : public CFStr<TCHAR, 5>
{
	/// Spcialisation pour FOURCC
	CFstr5(FOURCC fcc)
	{
		str[0] = BYTE(fcc);
		str[1] = BYTE(fcc >> 8);
		str[2] = BYTE(fcc >> 16);
		str[3] = BYTE(fcc >> 24);
		str[4] = 0;
	}
};

/****************************************************************************************
 *							Structures de rfrence d'horloge							*
 ****************************************************************************************/

/*
	Les rfrences d'horloge sont constitues d'chantillons d'un compteur pilot  27 MHz
	(donc s'incrmentant de 27000000 units toutes les secondes).

	Les chantillons sont diviss par 300, reprsentant donc un compteur pilot  90 kHz,
	et sauvegards dans un champ de 33 bits reprsentant les valeurs de 0  8589934591,
	soit environ 95443 secondes (1 jour, 2 heures, 30 minutes, 43 secondes).
	Le reste de la division est ventuellement conserv dans un champ de 9 bits dnomm "ext".
*/

// (dfini ainsi au moins 10 fois dans les en-ttes Microsoft, alors un de plus ou un de moins ...)
typedef LONGLONG	REFERENCE_TIME;

/// Frquence d'horloge du REFERENCE_TIME en Hz
#define RFT_FREQ	10000000I64

/// Frquence d'horloge du PCR en Hz
#define PCR_FREQ	27000000I64

/// Ratio entre PCR et PTS
#define PCRtoPTS	300

/// Frquence d'horloge du PTS en Hz (= PCR_FREQ / 300)
#define PTS_FREQ (PCR_FREQ / PCRtoPTS)	// = 90000

/// Frquence d'horloge du REFERENCE_TIME en kHz
#define RFT_FREQ_KHZ (RFT_FREQ / 1000)	// = 10000

/// Frquence d'horloge du PCR en kHz
#define PCR_FREQ_KHZ (PCR_FREQ / 1000)	// = 27000

/// Frquence d'horloge du PCR en kHz
#define PTS_FREQ_KHZ (PTS_FREQ / 1000)	// = 90

/// Valeur d'horloge  27 MHz
typedef UINT64	TIM_27M;
typedef INT64	ITIM_27M;	// version signe

/// Valeur d'horloge  90 kHz
typedef UINT64	TIM_90k;
typedef INT64	ITIM_90k;	// version signe

/// Valeur d'horloge  1 kHz (millisecondes)
typedef UINT64	TIM_1k;
typedef INT64	ITIM_1k;	// version signe

/// Valeur de bouclage de l'horloge  90 kHz
#define WRAP_90k	(1I64<<33)	// Le champ comporte 33 bits

/// Valeur de bouclage de l'horloge  27 MHz
#define WRAP_27M	(WRAP_90k * PCRtoPTS)

/// Conversion PCR en REFERENCE_TIME
inline REFERENCE_TIME PCR2RefTime(ITIM_27M val27M)
{
	return val27M * (RFT_FREQ_KHZ/1000) / (PCR_FREQ_KHZ/1000);	//  1000 = optimisation PGCD
}

/// Conversion PTS en REFERENCE_TIME
inline REFERENCE_TIME PTS2RefTime(ITIM_90k val90k)
{
	return val90k * (RFT_FREQ_KHZ/10) / (PTS_FREQ_KHZ/10);		//  10 = optimisation PGCD
}

inline ITIM_27M RefTime2PCR(REFERENCE_TIME rt)
{
	return rt * (PCR_FREQ_KHZ/1000) / (RFT_FREQ_KHZ/1000);		//  1000 = optimisation PGCD
}

inline ITIM_90k RefTime2PTS(REFERENCE_TIME rt)
{
	return rt * (PTS_FREQ_KHZ/10) / (RFT_FREQ_KHZ/10);			//  10 = optimisation PGCD
}

/// Construction d'une chane d'affichage d'une valeur de temps exprime en millisecondes
class millisec2str : public CFStr<TCHAR, 64>
{
public:
	/// Constructeur
	millisec2str(ITIM_1k val1k);
};

/// Construction d'une chane d'affichage d'une valeur de temps exprime en units PCR de 90 khz
class clk90k2str : public millisec2str
{
public:
	clk90k2str(ITIM_90k val90k) :
		millisec2str(val90k/(PTS_FREQ/1000))
	{
	}
};

/// Construction d'une chane d'affichage d'une valeur de temps exprime en units PCR de 27 MHz
class clk27M2str : public millisec2str
{
public:
	clk27M2str(ITIM_27M val27M) :
		millisec2str(val27M/(PCR_FREQ/1000))
	{
	}
};

/// Construction d'une chane d'affichage d'une valeur de temps exprime en units REFERENCE_TIME (100S) de Microsoft
class refTime2str : public millisec2str
{
public:
	refTime2str(REFERENCE_TIME rt) :
		millisec2str(rt/10000)
	{
	}
};

/** \brief Structure d'accs  la valeur "SCR" (System Reference Clock)
 *
 * Structure d'accs  la valeur "SCR" (System Reference Clock) utilise
 * dans le Pack Header Mpeg 1 en mode PS (Program Stream).
 *
 * Description du SCR :
 * <pre>
 * Rpartition des bits dans les 5 octets du SCR :
 *            [rfrence]
 *      3 33222222 22221111 11111100 00000000
 *      2 10987654 32109876 54321098 76543210
 *      a bbbbbbbb cccccccc dddddddd eeeeeeee

 *         [0]      [1]      [2]      [3]      [4]
 *      33333333 33222222 22221111 11111100 00000000
 *      98765432 10987654 32109876 54321098 76543210
 * SCR: 0010abb1 bbbbbbcc ccccccd1 ddddddde eeeeeee1
 * </pre>
 **/
struct PS_D_SCR_v1
{
	UINT8 b[5];

	PS_D_SCR_v1 & set(TIM_90k val90k) {
		UINT32	lo = UINT32(val90k);

		// Note : vrifier que les 4 bits de poids fort de b[0] sont gaux  0010
		// dans tous les enregistrements contenant PS_D_SCR_v1
		b[0] = UINT8(val90k >> 29) | 0x21;
		b[1] = UINT8(lo >> 22);
		b[2] = UINT8(lo >> 14) | 0x01;
		b[3] = UINT8(lo >> 7);
		b[4] = UINT8(lo << 1) | 0x01;
		return *this;
	}

	TIM_90k get90k() const {
		UINT32	lo =
			 b[1]			<< 22	|
			(b[2] &	0xfe)	<< 14	|
			 b[3]			<<  7	|
			 b[4]			>>  1;

		return  TIM_90k(b[0] & 0x0e)<<29 | TIM_90k(lo);
	}

	bool isV1() const {
			return (b[0] & 0xf0) == 0x20; }
	PS_D_SCR_v1 & operator = (TIM_27M val27M) {
		return set(val27M/PCRtoPTS); }
	TIM_27M get27M() const {
		return get90k()*PCRtoPTS; }
	operator TIM_27M() const {
		return get90k()*PCRtoPTS; }
	PUINT8	next() const {
		return PUINT8(this+1); }

	tstring	toString() const {
				return clk27M2str(get27M())(); }
};

/** \brief Structure d'accs  la valeur "SCR" (System Reference Clock)
 *
 * Structure d'accs  la valeur "SCR" (System Reference Clock) utilise
 * dans le Pack Header Mpeg 2 en mode PS (Program Stream).
 *
 * Description du SCR :
 * <pre>
 * Rpartition des bits dans les 6 octets du SCR :
 *            [rfrence]                            [ext]
 *      3 33222222 22221111 11111100 00000000      0 00000000
 *      2 10987654 32109876 54321098 76543210      8 76543210
 *      a bbbbbbbb cccccccc dddddddd eeeeeeee      x yyyyyyyy

 *         [0]      [1]      [2]      [3]      [4]      [5]
 *      44444444 33333333 33222222 22221111 11111100 00000000
 *      76543210 98765432 10987654 32109876 54321098 76543210
 * SCR: 01abb1bb bbbbcccc ccccd1dd dddddeee eeeee1xy yyyyyyy1
 * </pre>
 **/
struct PS_D_SCR_v2
{
	UINT8 b[6];

	PS_D_SCR_v2 & set(TIM_90k val90k, UINT16 ext=0) {
		UINT32	lo = UINT32(val90k);

		// Note : vrifier que les 2 bits de poids fort de b[0] sont gaux  01
		// dans tous les enregistrements contenant PS_D_SCR_v2
		b[0] = UINT8(val90k >> 27 & 0x38) | UINT8(lo >> 28 & 0x03) | 0x44;
		b[1] = UINT8(lo >> 20);
		b[2] = UINT8(lo >> 12 & 0xf8) | UINT8(lo >> 13 & 0x03) | 0x04;
		b[3] = UINT8(lo >> 5);
		b[4] = UINT8(lo << 3) | UINT8(ext >> 7 & 0x03) | 0x04;
		b[5] = UINT8(ext << 1) | 0x01;
		return *this;
	}

	TIM_90k get90k() const {
		UINT32	lo =
			(b[0] & 0x03)	<< 28 |
			 b[1]			<< 20 |
			(b[2] & 0xf8)	<< 12 |
			(b[2] & 0x03)	<< 13 |
			 b[3]			<<  5 |
			(b[4] & 0xf8)	>>  3;

		return  TIM_90k(b[0] & 0x38)<<27 | TIM_90k(lo);
	}

	bool isV2() const {
			return (b[0] & 0xc0) == 0x40; }
	UINT16 getExt() const {
		return (b[4]&0x03)<<7 | b[5]>>1; }
	PS_D_SCR_v2 & operator = (TIM_27M val27M) {
		return set(val27M/PCRtoPTS, UINT16(val27M%PCRtoPTS)); }
	TIM_27M get27M() const {
		return get90k()*PCRtoPTS + getExt(); }
	operator TIM_27M() const {
		return get90k()*PCRtoPTS + getExt(); }
	PUINT8	next() const {
		return PUINT8(this+1); }

	tstring	toString() const {
				return clk27M2str(get27M())(); }
};

/** \brief Description du Time Stamp utilis pour PTS ou DTS
 *
 *  Description du Time Stamp utilis, soit en tant que Presentation Time Stamp (PTS),
 *  soit en tant que Decoding Time Stamp (DTS), dans l'extension optionnelle des
 *  packets PES en mode PS (Program Stream) :
 * <pre>
 *  Rpartition des bits dans les 5 octets du PTS ou du DTS :
 *             [time stamp]
 *       3 33222222 22221111 11111100 00000000      xx = 01 (DTS)
 *       2 10987654 32109876 54321098 76543210           10 (PTS sans DTS)
 *       a bbbbbbbb cccccccc dddddddd eeeeeeee           11 (PTS avec DTS)
 *
 *          [0]      [1]      [2]      [3]      [4]
 *       33333333 33222222 22221111 11111100 00000000
 *       98765432 10987654 32109876 54321098 76543210
 *  SCR: 00xxabb1 bbbbbbcc ccccccd1 ddddddde eeeeeee1
 * </pre>
 **/
struct PS_D_TimeStamp
{
	UINT8 b[5];

	PS_D_TimeStamp & set(TIM_90k val90k, UINT8 mark) {
		UINT32 lo = UINT32(val90k);

		b[0] = UINT8(val90k >> 29 & 0x0e) | mark | 0x01;
		b[1] = UINT8(lo >> 22);
		b[2] = UINT8(lo >> 14) | 0x01;
		b[3] = UINT8(lo >>  7);
		b[4] = UINT8(lo <<  1) | 0x01;
		return *this;
	}

	PS_D_TimeStamp & set(TIM_90k val90k) {
		return set(val90k, b[0] & 0x30); }

	/// Constructeur pour cration
	PS_D_TimeStamp(TIM_27M val27M, UINT8 mark)
	{
		set(val27M/PCRtoPTS, mark);
	}

	TIM_90k get90k() const {
		UINT32 lo =
			 b[1]			<< 22 |
			(b[2] & 0xfe)	<< 14 |
			 b[3]			<<  7 |
			 b[4]			>>  1;

		return  TIM_90k(b[0] & 0x0e) << 29 | TIM_90k(lo);
	}

	PS_D_TimeStamp & operator = (TIM_27M val27M) {
		return set(val27M/PCRtoPTS); }
	TIM_27M get27M() const {
		return get90k()*PCRtoPTS; }
	operator TIM_27M() const {
		return get90k()*PCRtoPTS; }
	PUINT8 next() const {
		return PUINT8(this+1); }

	tstring	toString() const {
				return clk27M2str(get27M())(); }
};

typedef PS_D_TimeStamp MPEG_DTS;

/** \brief Description du PCR utilis dans les champs "Adaptation Field" en mode TS
 *
 *  Description du PCR (Program Reference Clock) utilis dans le champ
 *  d'adaptation (Adaptation Field) en mode TS (Transport Stream) :
 * <pre>
 *  Rpartition des bits dans les 6 octets du PCR :
 *             [rfrence]                            [ext]
 *       3 33222222 22221111 11111100 00000000      0 00000000
 *       2 10987654 32109876 54321098 76543210      8 76543210
 *       a bbbbbbbb cccccccc dddddddd eeeeeeee      x yyyyyyyy
 *
 *          [0]      [1]      [2]      [3]      [4]      [5]
 *       44444444 33333333 33222222 22221111 11111100 00000000
 *       76543210 98765432 10987654 32109876 54321098 76543210
 *  SCR: abbbbbbb bccccccc cddddddd deeeeeee e------x yyyyyyyy
 * </pre>
 **/
struct TS_PCR
{
	UINT8 b[6];

	TS_PCR & set(TIM_90k val90k, UINT16 ext=0) {
		UINT32	lo = UINT32(val90k);

		b[0] = UINT8(val90k >> 25);
		b[1] = UINT8(lo >> 17);
		b[2] = UINT8(lo >>  9);
		b[3] = UINT8(lo >>  1);
		b[4] = UINT8(lo <<  7) | UINT8(ext >> 8);
		b[5] = UINT8(ext);
		return *this;
	}

	TIM_90k get90k() const {
		UINT8	hi =
			 b[0] >> 7;
		UINT32	lo =
			b[0] << 25 |
			b[1] << 17 |
			b[2] <<  9 |
			b[3] <<  1 |
			b[4] >>  7;

		return TIM_90k(lo) | TIM_90k(hi)<<32;
	}

	UINT16 getExt() const {
		return MAKEWORD(b[5], b[4]) & 0x1ff; }
	TS_PCR & operator = (TIM_27M val27M) {
		return set(val27M/PCRtoPTS, UINT16(val27M%PCRtoPTS)); }
	TIM_27M get27M() const {
		return get90k()*PCRtoPTS + getExt(); }
	operator TIM_27M() const {
		return get90k()*PCRtoPTS + getExt(); }
	PUINT8 next() const {
		return PUINT8(this+1); }
};

/*
	Correspondance des bits entre les 6 octets du PCR et les 6 octets du SCR du pack header :
			[0]      [1]      [2]      [3]      [4]      [5]
		 33322222 22222111 11111110 00000000 0      0 00000000
		 21098765 43210987 65432109 87654321 0      8 76543210
	PCR: aaaaaaaa bbbbbbbb cccccccc dddddddd e------x yyyyyyyy
	SCR: 01aaa1aa aaabbbbb bbbcc1cc ccccdddd dddde1xy yyyyyyy1
*/

/****************************************************************************************/

/** \brief Description du "mux rate" dans le "Pack Header" en mode PS
 *
 * <pre>
 *  Rpartition des bits dans les 3 octets du mux rate :
 *                      [rate]
 *              221111 11111100 00000000
 *              109876 54321098 76543210
 *              aaaaaa bbbbbbbb cccccccc
 *
 *                [0]      [1]      [2]
 *              22221111 11111100 00000000
 *              32109876 54321098 76543210
 *  mux rate:   aaaaaabb bbbbbbcc cccccc11
 * </pre>
 **/
struct PS_PH_mux_rate
{
	UINT8 b[3];

	PS_PH_mux_rate & operator = (UINT32 rate) {
		b[0] = UINT8(rate >> 14);
		b[1] = UINT8(rate >> 6);
		b[2] = UINT8(rate << 2) | 0x03;
		return *this;
	}
	operator UINT32() const {
		return MAKELONG(MAKEWORD(b[2], b[1]), b[0] & 0x3f) >> 2; }
	PUINT8	next() const {
		return PUINT8(this+1); }
};

/****************************************************************************************/

/** \brief Description du "mux rate" dans les extensions PES
 *
 * <pre>
 *  Rpartition des bits dans les 3 octets du mux rate (PES extension) :
 *                      [rate]
 *              221111 11111100 00000000
 *              109876 54321098 76543210
 *              aaaaaa bbbbbbbb cccccccc
 *
 *                [0]      [1]      [2]
 *              22221111 11111100 00000000
 *              32109876 54321098 76543210
 *  mux rate:   1aaaaaab bbbbbbbc ccccccc1
 * </pre>
 **/
struct PS_D_mux_rate
{
	UINT8 b[3];

	PS_D_mux_rate & operator = (UINT32 rate) {
		b[0] = UINT8(rate >> 15) | 0x80;
		b[1] = UINT8(rate >> 7);
		b[2] = UINT8(rate << 1) | 0x01;
		return *this;
	}
	operator UINT() const {
		return MAKELONG(MAKEWORD(b[2], b[1]), b[0] & 0x7f) >> 1; }
	PUINT8 next() const {
		return PUINT8(this+1); }
};

/****************************************************************************************
 *										DESCRIPTOR										*
 ****************************************************************************************/

/**
 * Copie d'une chane de caractres avec encodage optionnel,  partir d'un pointeur sur dbut
 * et d'une longueur. La chane source n'est PAS (ncessairement) termine par zro.
 * Version pour adresse de destination type \p CHAR (caractres 8 bits) :
 *   Si un caractre d'encodage est prsent au dbut de la chane (code 0x01  0x1f), celui-ci
 *   est limin, et il est retourn par la fonction (qui autrement retourne 0).
 *   Aucune conversion n'est cependant effectue en relation avec ce code.
 * \param[in]	src		Pointeur sur la chane source
 * \param[in]	siz		Longueur de la chane source
 * \param[out]	dst		Pointeur sur un tampon destin  recevoir la chane de destination
 * \param[in]	bufsize	Longueur du tampon destin  recevoir la chane de destination
 * \retval		Code de page qui tait prsent au dbut de la chane source, ou bien 0 s'il n'y en avait pas.
 **/
UINT get_raw_str(LPCSTR src, UINT siz, LPSTR dst, UINT bufsize);

/**
 * Copie d'une chane de caractres avec encodage optionnel,  partir d'un pointeur sur dbut
 * et d'une longueur. La chane source n'est PAS (ncessairement) termine par zro.
 * Version pour adresse de destination type \p WCHAR (caractres 16 bits) :
 *   - Si un caractre d'encodage est prsent au dbut de la chane (code 0x01  0x1f), celui-ci
 *     est limin, et il est retourn par la fonction (qui autrement retourne 0).
 *   - De plus, dans ce dernier cas, il est tent, dans la mesure du possible, d'utiliser ce code
 *     pour convertir la chane de caractres en chane Unicode.
 *   - Les codes de contrle ventuels (0x80  0x9f) sont convertis en codes de 0xe080  0xe09f.
 * \param[in]	src		Pointeur sur la chane source
 * \param[in]	siz		Longueur de la chane source
 * \param[out]	dst		Pointeur sur un tampon destin  recevoir la chane de destination
 * \param[in]	bufsize	Longueur du tampon destin  recevoir la chane de destination
 * \retval		Code de page qui tait prsent au dbut de la chane source, ou bien 0 s'il n'y en avait pas.
 *				Si au moins un code de contrle tait prsent dans la chane, alors la valeur 0x100 est ajoute
 **/
UINT get_raw_str(LPCSTR src, UINT siz, LPWSTR dst, UINT bufsize);

// ----------------------------------------------------------------------------

namespace MDESCR
{

// ----------------------------------------------------------------------------

/*
		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	descriptor(){
		descriptor_tag                   8          +0
		descriptor_length                8          +1
		for (i=0;i<N;i++) {                         +2
			data                         8
		}
	}
*/
/*
type 2 : vido, sous_type :     0x11, taille 1 SDT descriptor

type 3 ou 4 : son, sous_type :  0x0A, taille 4 (langue 3 + 00) ISO639 Language descriptor

type 6 : tltexte, sous_type : 0x05, taille 4 Registration Descriptor
								0x0A, taille 4 (langue 3 + 00) ISO639 Language descriptor
								0x45, taille 8 VBI data descriptor
								0x52, taille 1 Stream Identifier Descriptor
								0x56, taille 5 (langue 3 + ??) teletext descriptor
								0x59, taille 8 (langue 3 + ?) Subtitling descriptor
								0x6A, taille 5 Dolby AC3
*/

template <typename T>
	struct CFStr4 : public CFStr<T, 4>
{
	CFStr4(CHAR c1, CHAR c2, CHAR c3)
	{
		str[0] = c1;
		str[1] = c2;
		str[2] = c3;
		str[3] = 0;
	}
	CFStr4(LPCSTR s)
	{
		str[0] = s[0];
		str[1] = s[1];
		str[2] = s[2];
		str[3] = 0;
	}
};

/// Reprsentation d'un code ISO de 3 caractres ASCII
struct C3CHAR {
	CHAR code[3];

	template <typename T> void get(T * str) const {		// str doit rserver au moins 4 caractres
		str[0] = code[0]; str[1] = code[1]; str[2] = code[2]; str[3] = 0; }
	operator FOURCC() const {
		return MAKEFOURCC(code[0], code[1], code[2], 0); }
	template <typename T> operator CFStr4<T>() const {
		return CFStr4<T>(code); }
};

typedef C3CHAR ISO639_LANG;		//!< Code de langue
typedef C3CHAR ISO3166_CTRY;	//!< Code de pays

enum D_tag : UINT8 {
	mdt_Video_Stream			= 0x02,
	mdt_Audio_Stream			= 0x03,
	mdt_Hierarchy				= 0x04,
	mdt_Registration			= 0x05,
	mdt_Data_Stream_Alignment	= 0x06,
	mdt_Target_Background_Grid	= 0x07,
	mdt_Video_Window			= 0x08,
	mdt_Conditional_Access		= 0x09,
	mdt_ISO_639_Language		= 0x0a,
	mdt_System_Clock			= 0x0b,
	mdt_Multiplex_Buffer_Util	= 0x0c,
	mdt_Copyright_Descriptor	= 0x0d,
	mdt_Maximum_Bitrate			= 0x0e,
	mdt_Private_Data_Indicator	= 0x0f,
	mdt_Smoothing_Buffer		= 0x10,
	mdt_STD						= 0x11,	// System Target Decoder ?
	mdt_IBP						= 0x12,
	//
	// http://www.coolstf.com/tsreader/descriptors.html ou 	ETSI EN 300 468
	mdt_Network_Name			= 0x40,
	mdt_Service_List			= 0x41,
	mdt_Stuffing				= 0x42,
	mdt_Sat_Delivery_System		= 0x43,
	mdt_Cable_Delivery_System	= 0x44,
	mdt_VBI_Data				= 0x45,
	mdt_VBI_Teletext			= 0x46,
	mdt_Bouquet_Name			= 0x47,
	mdt_Service					= 0x48,
	mdt_Country_Availability	= 0x49,
	mdt_Linkage					= 0x4a,
	mdt_NVOD_Reference			= 0x4b,
	mdt_Time_Shifted_Service	= 0x4c,
	mdt_Short_Event				= 0x4d,
	mdt_Extended_Event			= 0x4e,
	mdt_Time_Shifted_Event		= 0x4f,
	mdt_Component				= 0x50,
	mdt_Mosaic					= 0x51,
	mdt_Stream_Identifier		= 0x52,
	mdt_CA_Indentifier			= 0x53,
	mdt_Content					= 0x54,
	mdt_Parental_Rating			= 0x55,
	mdt_Teletext				= 0x56,
	mdt_Telephone				= 0x57,
	mdt_Local_Time_Offset		= 0x58,
	mdt_Subtitling				= 0x59,
	mdt_Terrestrial_Deliv_Sys	= 0x5a,
	mdt_Multiling_Network_Name	= 0x5b,
	mdt_Multiling_Bouquet_Name	= 0x5c,
	mdt_Multiling_Service_Name	= 0x5d,
	mdt_Multiling_Component		= 0x5e,
	mdt_Private_Data			= 0x5f,
	mdt_Service_Move			= 0x60,
	mdt_Short_Smoothing_Buffer	= 0x61,
	mdt_Frequency_List			= 0x62,
	mdt_Partial_TransportStream = 0x63,
	mdt_Data_Broadcast			= 0x64,
	mdt_Scrambling				= 0x65,
	mdt_Data_Broadcast_ID		= 0x66,
	mdt_Transport_Stream		= 0x67,
	mdt_DSNG					= 0x68,
	mdt_PDC						= 0x69,
	mdt_AC3_Audio				= 0x6a,
	mdt_Ancillary_Data			= 0x6b,
	mdt_Cell_List				= 0x6c,
	mdt_Cell_Frequency_Link		= 0x6d,
	mdt_Announcement_Support	= 0x6e,
	mdt_Application_Signaling	= 0x6f,
	mdt_Adatation_Field_Data	= 0x70,
	mdt_Service_Identifier		= 0x71,
	mdt_Service_Availability	= 0x72,
	mdt_Default_Authority		= 0x73,
	mdt_Related_Content			= 0x74,
	mdt_TVA_ID					= 0x75,
	mdt_Content_Identifier		= 0x76,
	mdt_Time_Slice_FEC_Ident	= 0x77,
	mdt_ECM_Repetition_Rate		= 0x78,
	mdt_S2_Sat_Delivery_System	= 0x79,
	mdt_Enhanced_AC3_Audio		= 0x7a,
	mdt_DTS_Audio				= 0x7b,
	mdt_AAC_Audio				= 0x7c,
	mdt_XAIT_Location			= 0x7d,
	mdt_FTA_Content_Management	= 0x7e,
	mdt_Extension				= 0x7f,
	//
	mdt_User_Defined_First		= 0x80,		// user defined 0x80 to 0xFE
	//
	mdt_Stuffing_2				= 0x80,
	mdt_AC3_Audio_Stream		= 0x81,		// DCII ou ATSC
	mdt_Logical_Channel			= 0x83,
	mdt_Caption_Service			= 0x86,
	mdt_Content_Advisory		= 0x87,
	mdt_HD_simcast_log_channel	= 0x88,
	//
	mdt_Extended_Channel_Name	= 0xa0,		// ATSC A/65
	mdt_Service_Location		= 0xa1,		// ATSC A/65
	mdt_Time_Shifted_Service_2	= 0xa2,		// ATSC A/65
	mdt_Component_Name			= 0xa3,		// ATSC A/65
	mdt_Data_Service			= 0xa4,		// ATSC
	mdt_Pid_Count				= 0xa5,		// ATSC
	mdt_Download				= 0xa6,		// ATSC
	mdt_Multiprotocol_Encap		= 0xa7,		// ATSC
	mdt_DCC_Departing_Request	= 0xa8,		// ATSC A/65
	mdt_DCC_Arriving_Request	= 0xa9,		// ATSC A/65
	mdt_Redistribution_Control	= 0xaa,		// ATSC A/65
	mdt_Genre					= 0xab,		// ATSC A/65
	mdt_ATSC_Private_Info		= 0xad,		// ATSC A/65
	mdt_Enhanced_Signaling		= 0xb2,		// ATSC
	mdt_Content_Identifier_2	= 0xb6,		// ATSC
	//
	mdt_User_Defined_Last		= 0xfe
};

/// **** ETSI EN 300 468 v1.11.1 section 6.2.8, table 26
enum StrmContent
{
	sc_MpegVideo				= 0x1,
	sc_MpegAudio				= 0x2,
	sc_Subtitles				= 0x3,
	sc_AC_3_AudioModes			= 0x4,
	sc_H264_Video				= 0x5
};

/// **** ETSI EN 300 468 v1.11.1 section 6.2.8, table 26
enum SC_03_Type : UINT8	// StrmContent = sc_Subtitles
{
	s03_undef,
	s03_EBUTeletextSubtitles	= 0x01,
	s03_AssociatedEBUTeletext	= 0x02,
	s03_VBI_data				= 0x03,
	s03_DVBSub_std				= 0x10,
	s03_DVBSub_std_for_4_3_ar	= 0x11,
	s03_DVBSub_std_for_16_9_ar	= 0x12,
	s03_DVBSub_std_for_221_ar	= 0x13,
	s03_DVBSub_std_for_HD		= 0x14,
	s03_DVBSub_hoh				= 0x20,
	s03_DVBSub_hoh_for_4_3_ar	= 0x21,
	s03_DVBSub_hoh_for_16_9_ar	= 0x22,
	s03_DVBSub_hoh_for_221_ar	= 0x23,
	s03_DVBSub_hoh_for_HD		= 0x24,
	s03_OpenSign_for_deaf		= 0x30,
	s03_ClosedSign_for_deaf		= 0x31,
	s03_VideoUpSampled			= 0x40
};

// Dfinitions des codes de contrle utiliss dans les descripteurs de texte :
#define	DCTL_MIN_A		0x80	// limite infrieure (attention au signe du CHAR lors des comparaisons !)
#define	DCTL_EMPH_ON_A	0x86	// emphasis on
#define	DCTL_EMPH_OFF_A	0x87	// emphasis off
#define DCTL_CRLF_A		0x8a	// CR/LF (saut de ligne)
#define	DCTL_MAX_A		0x9f	// limite suprieure

// Version 16 bits :
#define	DCTL_W_OFFSET	0xe000

#define	DCTL_MIN_W		(DCTL_W_OFFSET|DCTL_MIN_A)
#define	DCTL_EMPH_ON_W	(DCTL_W_OFFSET|DCTL_EMPH_ON_A)
#define	DCTL_EMPH_OFF_W	(DCTL_W_OFFSET|DCTL_EMPH_OFF_A)
#define DCTL_CRLF_W		(DCTL_W_OFFSET|DCTL_CRLF_A)
#define	DCTL_MAX_W		(DCTL_W_OFFSET|DCTL_MAX_A)

#ifdef _UNICODE
#define DCTL_(s) DCTL_##s##_W
#else
#define DCTL_(s) DCTL_##s##_A
#endif

/// Modle pour viter d'avoir  prendre la taille du tampon destination avec ::get_raw_str
template <class T, UINT S>
	inline UINT get_raw_str(LPCSTR src, UINT siz, T (&dst)[S])
{
	return get_raw_str(src, siz, dst, S);
}

/// Classe de base pour tous les descripteurs des tables PSI
struct D_base
{
	D_tag	tag;
	UINT8	length;

	PUINT8 data() const {
		return PUINT8(this+1); }
	D_base * next() const {
		return reinterpret_cast<D_base *>(data() + length); }
	template <class T> const T & as() const {	// Drivation vers un type de descripteur spcifique
		return *reinterpret_cast<const T *>(this); }

	tstring	toString() const {
				TCHAR	szStr[32];
				_stprintf_s(szStr, TEXT("tag=0x%02x, length=%u"), tag, length);
				return szStr; }
};

/// Modle pour les descripteurs contenant des suites rptitives
template <class T> struct D_list : public D_base
{
	typedef streamScanner<T> sdScanner;

	const T & operator () () const {
		return *reinterpret_cast<const T *>(this+1); }
	sdScanner subDescrRange() const {
		return sdScanner(data(), data() + length); }
};

/// \brief Sous-descripteur gnrique pour les chanes
/// au format longueur + chane (style PASCAL).
struct S_str
{
	UINT8	strlen;

	template <class T> UINT get(T dst, UINT bufsize) const {
		return ::get_raw_str(LPCSTR(this+1), strlen, dst, bufsize); }
	template <class T, size_t S> UINT get(T (&dst)[S]) const {
		return get(dst, S); }

	const S_str * next() const {
		return reinterpret_cast<const S_str *>(this+1+strlen); }
};

// Instantiation qui revient souvent :
typedef streamScanner<D_base> D_scanner;

/* ISO 639 Registration descriptor tag 0x05

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	registration_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		format_identifier                   32          2..5
		for (i=0; i<N; i++) {                           6
			additional_identification_info   8          +0
		}
	}
*/

struct S_05 : public UINT8uw
{
	const S_05 * next() const {
		return this+1; }
};

struct D_05 : public D_base	// tag = mdt_Registration
{
	UBE32	format_identifier;

	typedef streamScanner<S_05> sdScanner;

	sdScanner subDescrRange() const {
		return sdScanner(this+1, PUINT8(this+1) + length); }
};

/* Conditional access descriptor tag 0x09

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	CA_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		CA_system_ID                        16          2,3
		reserved                             3          4
		CA_PID                              13          4,5
		for (i = 0; i < N; i++) {                       6
			private_data_byte                8          +0
		}
	}
*/

struct S_09 : public UINT8uw
{
	const S_09 * next() const {
		return this+1; }
};

struct D_09 : public D_base	// tag = mdt_Conditional_Access
{
	UBE16	ca_system_id;
	PID13	ca_pid;

	typedef streamScanner<S_09> sdScanner;

	sdScanner subDescrRange() const {
		return sdScanner(this+1, PUINT8(this+1) + length); }
};

/* ISO 639 Language Descriptor (PMT) tag 0x0a

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	ISO_639_language_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		for (i=0; i<N; i++) {                           2
			ISO_639_language_code           24          +0..+2
			audio_type                       8          +3
		}
	}
*/

struct S_0a
{
	ISO639_LANG	language_code;
	UINT8		audio_type;

	const S_0a * next() const {
		return this+1; }
};

typedef D_list<S_0a> D_0a;	// tag = mdt_ISO_639_Language

/* Maximum bitrate (PMT) tag 0x0e

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	maximum_bitrate_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		reserved                             2          2
		maximum_bitrate                     22          2..4
	}
*/

struct D_0e : public D_base	// tag = mdt_Maximum_Bitrate
{
	UINT8	_mb[3];

	UINT32	maximum_bitrate() const {
		return ((_mb[0] & 0x3f)<<16)|(_mb[1]<<8)|(_mb[2]); }
};

/* Network Name (NIT) tag 0x40

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	network_name_descriptor(){
		descriptor_tag                       8          0
		descriptor_length                    8          1
		for (i=0;i<N;i++){                              2
			char                             8          +0
		}
	}
*/

struct D_40 : public D_base	// tag = mdt_Network_Name
{
	template <class T> UINT get_network_name(T dst, UINT bufsize) const {
		return ::get_raw_str(LPCSTR(this+1), length, dst, bufsize); }
	template <class T, size_t S> UINT get_network_name(T (&dst)[S]) const {
		return get_network_name(dst, S); }
};

/* Service List Descriptor (NIT) tag 0x41

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	service_list_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		for (i=0; i<N;i++){                             2
			service_id                      16          +0,+1
			service_type                     8          +2
		}
	}
*/

struct S_41
{
	UBE16	service_id;
	UINT8	service_type;
	//
	const S_41 * next() const {
		return this+1; }
};

typedef D_list<S_41> D_41;	// tag = mdt_Service_List

/* Service Descriptor (SDT) tag 0x48

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	service_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		service_type                         8          2
		service_provider_name_length         8          3
		for (i=0;i<N;I++) {                             4
			char                                        +0
		}
		service_name_length                  8          +0
		for (i=0;i<N;I++) {
			char                             8          +0
		}
	}
*/

struct D_48 : public D_base	// tag = mdt_Service
{
	UINT8 service_type;
	//
	const S_str & service_provider_name() const {
		return *reinterpret_cast<const S_str *>(this+1); }
	const S_str & service_name() const {
		return *service_provider_name().next(); }
};

/* Linkage descriptor (NIT) tag 0x4a

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	linkage_descriptor(){
		descriptor_tag                   8          0
		descriptor_length                8          1
		transport_stream_id             16          2,3
		original_network_id             16          4,5
		service_id                      16          6,7
		linkage_type                     8          8
		if (linkage_type!=0x08) {                   9
			for (i=0;i<N;i++){
				private_data_byte        8          +0
			}
		}
		if (linkage_type==0x08) {
			hand-over_type               4          +0
			reserved_future_use          3          +0
			origin_type                  1          +0
			if (hand-over_type==0x01
			|| hand-over_type==0x02
			|| hand-over_type==0x03){
				network_id              16          +1,+2
			}
			if (origin_type==0x00){
				initial_service_id      16          +0,+1
			}
			for (i=0;i<N;i++){
				private_data_byte       8           +0
			}
		}
	}
*/

struct D_4a : public D_base	// tag = mdt_Linkage
{
	UBE16	tsid;
	UBE16	onid;
	UBE16	service_id;
	UINT8	linkage_type;
};

struct D_4a_unwrapper {
	const D_4a &	descr;
	PUINT8					p_next;		// pointe les "private_data_byte" ventuels
	UINT8uw *				p_hand_over_and_origin;
	UBE16 *					p_network_id;
	UBE16 *					p_initial_service_id;

	D_4a_unwrapper(const D_4a & d) :
		descr(d),
		p_hand_over_and_origin(NULL),
		p_network_id(NULL),
		p_initial_service_id(NULL)
	{
		PUINT8	p_nxt = PUINT8(&d+1);
		if (descr.linkage_type==0x08) {
			get_and_skip(p_nxt, p_hand_over_and_origin);

			UINT8	hand_over_type = (*p_hand_over_and_origin)() & 0xf0;

			if (hand_over_type>=0x01 && hand_over_type<=0x03)
				get_and_skip(p_nxt, p_network_id);
			if ((*p_hand_over_and_origin)() & 0x01)
				get_and_skip(p_nxt, p_initial_service_id);
		}
		p_next = p_nxt;
	}
};

/* Short event descriptor (EIT) tag 0x4d

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	short_event_descriptor(){
		descriptor_tag                   8          0
		descriptor_length                8          1
		ISO_639_language_code           24          2..4
		event_name_length                8          5
		for (i=0;i<event_name_length;i++) {         6
			event_name_char              8          +0
		}
		text_length                      8          +0
		for (i=0;i<text_length;i++) {               +1
			text_char                    8          +0
		}
	}
*/

struct D_4d : public D_base	// tag = mdt_Short_Event
{
	ISO639_LANG	language_code;
	//
	const S_str & event_name() const {
		return *reinterpret_cast<const S_str *>(this+1); }
	const S_str & text() const {
		return *event_name().next(); }
};

/* Extended event descriptor (EIT) tag 0x4e

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	extended_event_descriptor(){
		descriptor_tag                   8          0
		descriptor_length                8          1
		descriptor_number                4          2
		last_descriptor_number           4          2
		ISO_639_language_code           24          3..5
		length_of_items                  8          6
		for ( i=0;i<N;i++) {                        7
			item_description_length      8          +0
			for (j=0;j<N;j++) {                     +1
				item_description_char    8          +0
			}
			item_length                  8          +0
			for (j=0;j<N;j++) {                     +1
				item_char                8          +0
			}
		}
		text_length                      8          +0
		for (i=0;i<N;i++){                          +1
			text_char                    8          +0
		}
	}
*/

struct S_4e
{
	const S_str & item_description() const {
		return *reinterpret_cast<const S_str *>(this); }
	const S_str & item() const {
		return *item_description().next(); }

	const S_4e * next() const {
		return reinterpret_cast<const S_4e *>(item().next()); }
};

struct D_4e : public D_base 	// tag = mdt_Extended_Event
{
	UINT8		descriptor_numbers;
	ISO639_LANG	language_code;
	UINT8		length_of_items;
	//
	UINT8 descriptor_number() const {
		return UINT8(descriptor_numbers >> 4); }
	UINT8 last_descriptor_number() const {
		return UINT8(descriptor_numbers & 0x0f); }
	streamScanner<S_4e> subDescrRange() const {
		return streamScanner<S_4e>(this+1, PUINT8(this+1) + length_of_items); }
	const S_str & text() const {
		return *reinterpret_cast<const S_str *>(PUINT8(this+1)+length_of_items); }
};

/* Component descriptor (EIT) tag 0x50

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	component_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		reserved_future_use              4          2
		stream_content                   4          2
		component_type                   8          3
		component_tag                    8          4
		ISO_639_language_code           24          5..7
		for (i=0;i<N;i++) {                         8
			text_char                    8          +0
		}
	}
*/

/// **** ETSI EN 300 468 v1.11.1 section 6.2.8, table 25
struct D_50 : public D_base 	// tag = mdt_Component
{
	UINT8		_stream_content;
	SC_03_Type	component_type;
	UINT8		component_tag;
	ISO639_LANG	language_code;

	StrmContent stream_content() const {
		return StrmContent(_stream_content & 0x0f); }
	template <class T> UINT get_text_char(T dst, UINT bufsize) const {
		return ::get_raw_str(LPCSTR(this+1), UINT(PUINT8(next())-PUINT8(this+1)), dst, bufsize); }
	template <class T, UINT S> UINT get_text_char(T (&dst)[S]) const {
		return get_text_char(dst, S); }
};

/* Stream identifier descriptor (PMT) tag 0x52

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	stream_identifier_descriptor(){
		descriptor_tag                   8          0
		descriptor_length                8          1
		component_tag                    8          2
	}
*/

struct D_52 : public D_base 	// tag = mdt_Stream_Identifier
{
	UINT8	component_tag;
};

/* Content descriptor (EIT) tag 0x54

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	content_descriptor(){
		descriptor_tag 8 uimsbf
		descriptor_length 8 uimsbf
		for (i=0;i<N;i++) {
			content_nibble_level_1 4 uimsbf
			content_nibble_level_2 4 uimsbf
			user_nibble 4 uimsbf
			user_nibble 4 uimsbf
		}
	}
*/

struct S_54
{
	UINT8	content_nibbles;
	UINT8	user_nibbles;

	UINT8 content_nibble_level_1() const {
		return UINT8(content_nibbles>>4); }
	UINT8 content_nibble_level_2() const {
		return UINT8(content_nibbles&0x0f); }
	UINT8 user_nibble_1() const {
		return UINT8(user_nibbles>>4); }
	UINT8 user_nibble_2() const {
		return UINT8(user_nibbles&0x0f); }

	const S_54 * next() const {
		return this+1; }
};

typedef D_list<S_54> D_54;	// tag = mdt_Content

/* Parental rating descriptor (EIT) tag 0x55

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	parental_rating_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		for (i=0;i<N;i++) {                         2
			country_code                24          +0..+2
			rating                       8          +3
		}
	}
*/

struct S_55
{
	ISO3166_CTRY	country_code;
	UINT8			rating;

	const S_55 * next() const {
		return this+1; }
};

typedef D_list<S_55> D_55;	// tag = mdt_Parental_Rating

/* Subtitling descriptor (PMT) tag 0x56

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	teletext_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		for (i=0;i<N;i++) {                         2
			ISO_639_language_code       24          +0..+2
			teletext_type                5          +3
			teletext_magazine_number     3          +3
			teletext_page_number         8          +4
		}
	}
*/

/// **** ETSI EN 300 468 v1.11.1 section 6.2.43
struct S_56
{
	enum Type : UINT8
	{
		t_undef,
		t_Initial,
		t_Subtitle,
		t_AdditionalInfo,
		t_ProgramSchedule,
		t_SubtitleForHearingImpaired
	};

	ISO639_LANG	language_code;
	UINT8		teletext_type_and_mag;
	UINT8		teletext_page_number;

	Type teletext_type() const {
		return Type(teletext_type_and_mag>>3); }
	UINT8 teletext_magazine_number() const {
		return UINT8(teletext_type_and_mag&0x07); }

	const S_56 * next() const {
		return this+1; }
};

typedef D_list<S_56> D_56;	// tag = mdt_Teletext

/* Subtitling descriptor (PMT) tag 0x59

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	subtitling_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		for (i=0;i<N;i++) {
			ISO_639_language_code       24          +0..+2
			subtitling_type              8          +3
			composition_page_id         16          +4,+5
			ancillary_page_id           16          +6,+7
		}                                           +8
	}
*/

/// **** ETSI EN 300 468 v1.11.1 section 6.2.41
struct S_59
{
	ISO639_LANG	language_code;
	SC_03_Type	subtitling_type;
	UBE16		composition_page_id;
	UBE16		ancillary_page_id;

	const S_59 * next() const {
		return this+1; }
};

typedef D_list<S_59> D_59;	// tag = mdt_Subtitling

/* Private Data Specifier tag 0x5a

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	terrestrial_delivery_system_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		centre_frequency                32          2..5
		bandwidth                        3          6
		reserved_future_use              5          6
		constellation                    2          7
		hierarchy_information            3          7
		code_rate-HP_stream              3          7
		code_rate-LP_stream              3          8
		guard_interval                   2          8
		transmission_mode                2          8
		other_frequency_flag             1          8
		reserved_future_use             32          9..12
	}                                               13
*/

struct D_5a : public D_base 	// tag = mdt_Terrestrial_Deliv_Sys
{
	UBE32	centre_frequency;
	UINT8	_bandwidth;
	UINT8	data1;	// constellation, hierarchy_information, code_rate-HP_stream
	UINT8	data2;	// code_rate-LP_stream, guard_interval, transmission_mode, other_frequency_flag
	UBE32	reserved2;

	UINT8 bandwidth() const {
		return UINT8(_bandwidth>>5); }
	UINT8 constellation() const {
		return UINT8(data1>>6); }
	UINT8 hierarchy_information() const {
		return UINT8((data1>>3)&0x07); }
	UINT8 code_rate_HP_stream() const {
		return UINT8(data1&0x07); }
	UINT8 code_rate_LP_stream() const {
		return UINT8(data2>>5); }
	UINT8 guard_interval() const {
		return UINT8((data2>>3)&0x03); }
	UINT8 transmission_mode() const {
		return UINT8((data2>>1)&0x03); }
	UINT8 other_frequency_flag() const {
		return UINT8(data2&0x01); }
};

/* Private Data Specifier tag 0x5f

		Syntax                      No. of bits   Offset
	----------------------------------------------------------
	private_data_specifier_descriptor() {
		descriptor_tag                   8          0
		descriptor_length                8          1
		private_data_specifier           32         2
	}
*/

struct D_5f : public D_base 	// tag = mdt_Stream_Identifier
{
	UBE32	private_data_specifier;
};

/*	**** ETSI EN 300 468 v1.11.1 section D.3.1
	Logical Channel Descriptor (PMT) tag 0x6a

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	AC3_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		component_type_flag                  1          2
		bsid_flag                            1          2
		mainid_flag                          1          2
		asvc_flag                            1          2
		reserved                             4          2
		if (component_type_flag)==1 {                   3
			AC3_type                         8          +0
		}
		if (bsid_flag)==1 {                             +0
			bsid                             8          +0
		}
		if (mainid_flag)==1 {                           +0
			mainid                           8          +0
		}
		if (asvc_flag)==1 {                             +0
			asvc                             8          +0
		}
		for (i=0;i<N;i++) {                             +0
			additional_info                  8          +0
		}
	}
*/

#define AC3_COMPONENT_TYPE_FLAG	0x80
#define AC3_BSID_FLAG			0x40
#define AC3_MAINID_FLAG			0x20
#define AC3_ASVC_FLAG			0x10

enum AC3_NumberOfChannels
{
	anc_Undef			= -1,
	anc_Mono			= 0,
	anc_1_plus_1,
	anc_Stereo,
	anc_DolbySurroundStereo,
	anc_Multi_MoreThan_2_0,
	anc_Multi_MoreThan_5_1,
	anc_Multiple_Independant,
	anc_Reserved
};

enum AC3_ServiceType
{
	ast_Undef			= -1,
	anc_Complete_Main	= 0,
	anc_MusicAndEffects,
	anc_VisuallyImpaired,
	anc_HearingImpaired,
	anc_Dialogue,
	anc_Commentary,
	anc_Emergency,
	anc_Voiceover,
	anc_Karaoke
};

struct D_6a : public D_base 	// tag = mdt_AC3_Audio
{
	UINT8	ac3_flags;

	UINT8 flags() const {
		return UINT8(ac3_flags & 0xf0); }
	AC3_NumberOfChannels number_of_channels() const {
		return ac3_flags & AC3_COMPONENT_TYPE_FLAG ? static_cast<AC3_NumberOfChannels>(*PUINT8(this+1) & 0x07) : anc_Undef; }
	AC3_ServiceType service_type() const {
		return ac3_flags & AC3_COMPONENT_TYPE_FLAG ? static_cast<AC3_ServiceType>((*PUINT8(this+1)>>3) & 0x07) : ast_Undef; }
};

struct D_6a_unwrapper
{
	const D_6a &	descr;
	PUINT8					p_next;
	UINT8uw *				p_ac3_type;
	UINT8uw *				p_bsid;
	UINT8uw *				p_mainid;
	UINT8uw *				p_asvc;

	D_6a_unwrapper(const D_6a & d) :
		descr(d),
		p_ac3_type(NULL),
		p_bsid(NULL),
		p_mainid(NULL),
		p_asvc(NULL)
	{
		PUINT8	p_nxt = PUINT8(&d+1);
		UINT8	flags = d.flags();

		if (flags & AC3_COMPONENT_TYPE_FLAG)
			get_and_skip(p_nxt, p_ac3_type);
		if (flags & AC3_BSID_FLAG)
			get_and_skip(p_nxt, p_bsid);
		if (flags & AC3_MAINID_FLAG)
			get_and_skip(p_nxt, p_mainid);
		if (flags & AC3_ASVC_FLAG)
			get_and_skip(p_nxt, p_asvc);
		p_next = p_nxt;
	}
};

/*	**** ETSI EN 300 468 v1.11.1 section D.5.1
		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	Enhanced_AC3_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		component_type_flag                  1          2
		bsid_flag                            1          2
		mainid_flag                          1          2
		asvc_flag                            1          2
		mixinfoexists                        1          2
		substream1_flag                      1          2
		substream2_flag                      1          2
		substream3_flag                      1          2
		if (component_type_flag)==1 {                   3
			component_type                   8          +0
		}
		if (bsid_flag)==1 {                             +0
			bsid                             8          +0
		}
		if (mainid_flag)==1 {                           +0
			mainid                           8          +0
		}
		if (asvc_flag)==1 {                             +0
			asvc                             8          +0
		}
		if (substream1_flag)==1 {                       +0
			substream1                       8          +0
		}
		if (substream2_flag)==1 {                       +0
			substream2                       8          +0
		}
		if (substream3_flag)==1 {                       +0
			substream3                       8          +0
		}
		for (i=0;i<n;i++) {                             +0
			additional_info [i]              8          +0
		}
	}
*/

#define AC3_MIXINFOEXISTS_FLAG		0x08
#define AC3_SUBSTREAM1_FLAG			0x04
#define AC3_SUBSTREAM2_FLAG			0x02
#define AC3_SUBSTREAM3_FLAG			0x01

struct D_7a : public D_6a 	// tag = mdt_Enhanced_AC3_Audio
{
	UINT8 flags() const {
		return ac3_flags; }
};

struct D_7a_unwrapper : public D_6a_unwrapper
{
	UINT8uw *				p_substream1;
	UINT8uw *				p_substream2;
	UINT8uw *				p_substream3;

	D_7a_unwrapper(const D_7a & d) :
		D_6a_unwrapper(d),
		p_substream1(NULL),
		p_substream2(NULL),
		p_substream3(NULL)
	{
		PUINT8	p_nxt = p_next;
		UINT8	flags = d.ac3_flags;

		if (flags & AC3_SUBSTREAM1_FLAG)
			get_and_skip(p_nxt, p_substream1);
		if (flags & AC3_SUBSTREAM2_FLAG)
			get_and_skip(p_nxt, p_substream2);
		if (flags & AC3_SUBSTREAM3_FLAG)
			get_and_skip(p_nxt, p_substream3);
		p_next = p_nxt;
	}
};

/* Data broadcast id descriptor (PMT) tag 0x66

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	data_broadcast_id_descriptor(){
		descriptor_tag                       8          0
		descriptor_length                    8          1
		data_broadcast_id                   16          2,3
		for(i=0;i<N;i++) {                              4
			id_selector_byte                 8          +0
		}
	}
*/

struct S_66 : public UINT8uw
{
	const S_66 * next() const {
		return this+1; }
};

struct D_66 : public D_base 	// tag = mdt_Data_Broadcast_ID
{
	UBE16	data_broadcast_id;

	streamScanner<S_66> subDescrRange() const {
		return streamScanner<S_66>(this+1, next()); }
};

/* AC-3 Audio Descriptor Syntax (ATSC) tag 0x81

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	AC-3_audio_stream_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		sample_rate_code                     3          2
		bsid                                 5          2
		bit_rate_code                        6          3
		surround_mode                        2          3
		bsmod                                3          4
		num_channels                         4          4
		full_svc                             1          4
		langcod                              8          5
		if(num_channels==0) // 1+1 mode
			langcod2                         8          +0
		if(bsmod<2) {
			mainid                           3          +0
			priority                         2          +0
			reserved = 111                 3          +0
		}
		else
			asvcflags                        8          +0
		textlen                              7          +0
		text_code                            1          +0
		for(i=0; i<M; i++) {
			text[i]                          8          +0
		}
		language_flag                        1          +0
		language_flag_2                      1          +0
		reserved = 111111                  6          +0
		if(language_flag==1) {
			language                         3*8        +0..2
		}
		if(language_flag_2==1) {
			language_2                       3*8        +0..2
		}
		for(i=0; i<N; i++) {
			additional_info[i]               N8        +0..N
		}
	}
*/

struct D_81 : public D_base 	// tag = mdt_AC3_Audio_Stream
{
	UINT8	data1;
	UINT8	data2;
	UINT8	data3;
	UINT8	langcod;

	UINT8 sample_rate_code() const {
		return UINT8(data1 >> 5); }
	UINT8 bsid() const {
		return UINT8(data1 & 0x1f); }
	UINT8 bit_rate_code() const {
		return UINT8(data2 >> 2); }
	UINT8 surround_mode() const {
		return UINT8(data2 & 0x03); }
	UINT8 bsmod() const {
		return UINT8(data3 >> 5); }
	UINT8 num_channels() const {
		return UINT8((data3 >> 1) & 0x0f); }
	UINT8 full_svc() const {
		return UINT8(data3 & 0x01); }
};

/* Logical Channel Descriptor (NIT) tag 0x83

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	Logical_channel_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		for (i=0; i<N; i++) {                           2
			service_id                      16          +0,+1
			reserved                         6          +2
			logical_channel_number          10          +2,+3
		}
	}
*/

struct S_83
{
	UBE16	service_id;
	UBE<10>	logical_channel_number;
	//
	const S_83 * next() const {
		return this+1; }
};

typedef D_list<S_83> D_83;	// tag = mdt_Logical_Channel

/* HD Simulcast Logical Channel Descriptor (NIT) tag 0x88

	http://www.csa.fr/upload/publication/CSATNT.pdf

		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	HD_simulcast_logical_channel_descriptor() {
		descriptor_tag                       8          0
		descriptor_length                    8          1
		for (i=0; i<N; i++) {                           2
			service_id                      16          +0,+1
			visible_service_flag             1          +2
			reserved                         5          +2
			logical_channel_number          10          +2, +3
		}
	}
*/

#define VISIBLE_SERVICE_FLAG	0x80

struct S_88 : public S_83 {

	bool visible_service_flag() const {
		return (logical_channel_number.hi & VISIBLE_SERVICE_FLAG)!=0; }
	const S_88 * next() const {
		return this+1; }
};

typedef D_list<S_88> D_88;	// tag = mdt_HD_simcast_log_channel

// ----------------------------------------------------------------------------

}	// namespace MDESCR

/****************************************************************************************
 *										AUDIO											*
 ****************************************************************************************/

/*
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	header() {
		syncword							12			0,1
		ID									 1			1
		layer								 2			1
		protection_bit						 1			1
		bitrate_index						 4			2
		sampling_frequency					 2			2
		padding_bit							 1			2
		private_bit							 1			2
		mode								 2			3
		mode_extension						 2			3
		copyright							 1			3
		original/home						 1			3
		emphasis							 2			3
	}
*/

enum MpegVersion : UINT8
{
	mv_Mpeg25,		// Mpeg 2.5 (non officiel)
	mv_reserved,
	mv_Mpeg2,		// Mpeg 2 (LSF, Low Sampling Frequencies)
	mv_Mpeg1		// Mpeg 1
};

enum MpegAudioLayer : UINT8
{
	mal_reserved,
	mal_Layer_III,
	mal_Layer_II,
	mal_Layer_I,
};

enum MpegAudioProtection : UINT8
{
	map_Redundancy,
	map_NoRedundancy
};

enum MpegAudioSamplingFreq : UINT8
{
	masf_44_1,
	masf_48,
	masf_32,
	masf_reserved
};

enum MpegAudioMode : UINT8
{
	mam_Stereo,
	mam_JointStereo,
	mam_DualChannel,
	mam_Mono
};

enum MpegAudioEmphasis : UINT8
{
	mae_None,
	mae_50_15,
	mae_reserved,
	mae_CCITT_J_17
};

struct MpegAudio_hdr
{
	UBE16	sync_fields;
	UINT8	rate_fields;
	UINT8	mode_fields;

	MpegVersion				version() const {
								return MpegVersion((sync_fields>>3) & 0x03); }

	/// Obtenir le type de layer Mpeg (I, II ou III)
	MpegAudioLayer			layer() const {
								return MpegAudioLayer((sync_fields>>1) & 0x03); }

	MpegAudioProtection		protection() const {
								return MpegAudioProtection(sync_fields & 0x01); }

	/// Obtenir l'index du "bit rate"
	UINT8					bit_rate_index() const {
								return (rate_fields>>4) & 0x0f; }

	/// Obtenir le nombre de kilobits par seconde
	UINT16					bit_rate() const;

	/// Obtenir l'index de la frquence d'chantillonnage
	MpegAudioSamplingFreq	samp_freq_index() const {
								return MpegAudioSamplingFreq((rate_fields>>2) & 0x03); }

	/// Obtenir la frquence d'chantillonnage
	UINT16					samp_frequency() const;

	/// Obtenir le nombre d'chantillons par trame
	UINT16					samples_per_frame() const {
								static const UINT16 spf[2][4] = {{0,576,1152,384},{0,1152,1152,384}};
								return spf[version()==mv_Mpeg1][layer()]; }

	bool					padding_bit() const {
								return bool((rate_fields>>1) & 0x01); }
	bool					private_bit() const {
								return bool(rate_fields & 0x01); }
	MpegAudioMode			mode() const{
								return MpegAudioMode((mode_fields>>6) & 0x03); }
	UINT8					mode_extension() const {
								return (mode_fields>>4) & 0x03; }
	bool					intensity_stereo() const {	// mode + mode_extension, Layer III seulement
								return (mode_fields & 0xf0)==((mam_JointStereo<<6)|0x10); }
	bool					ms_stereo() const {			// mode + mode_extension, Layer III seulement
								return (mode_fields & 0xf0)==((mam_JointStereo<<6)|0x20); }
	bool					copyright() const {
								return bool((mode_fields>>3) & 0x01); }
	bool					original() const {
								return bool((mode_fields>>2) & 0x01); }
	MpegAudioEmphasis		emphasis() const {
								return MpegAudioEmphasis(mode_fields & 0x03); }
	//
	bool 					isMpegAudio() const {
								return (sync_fields() & 0xffe0)==0xffe0 && rate_fields != 0xff; }
};

// ----------------------------------------------------------------------------

// En-tte ID3v2 prsent au dbut de certains fichiers MP3 :
struct ID3v2_hdr
{
	CHAR	id[3];
	UBE16	version;
	UINT8	flags;
	UINT8	siz_by[4];

	bool isID3v2() const {
		return id[0]=='I' && id[1]=='D' && id[2]=='3'; }
	UINT32 size() const {
		return (siz_by[0]<<21)|(siz_by[1]<<14)|(siz_by[2]<<7)|siz_by[3]; }
	PUINT8 next() const {
		return PUINT8(this+1)+size(); }
	const MpegAudio_hdr & next_hdr() const {
		return *reinterpret_cast<const MpegAudio_hdr *>(next()); }
};

/****************************************************************************************
 *									PROGRAM STREAM										*
 ****************************************************************************************/

/// Identificateurs de flots (streams)
/// ISO IEC 13818-1 section 2.4.3.7, table 2-18
enum StreamIDs : UINT8	// ou identifiant PES
{
	sid_undef = 0,
	//
	// Codes internes  l'image Mpeg2 - ISO IEC 13818-2 section 6.2.1, table 6-1
	sid_internal_Picture		= 0x00,		//!< dbut d'image
	sid_internal_Slice_first	= 0x01,		//!< premier code de tranche d'image
	// toutes les valeurs de 0x01  0xaf dfinissent les tranches d'images et le numro de chacune
	sid_internal_Slice_last		= 0xaf,		//!< dernier code de tranche d'image
	sid_reserved_1				= 0xb0,
	sid_reserved_2				= 0xb1,
	sid_internal_UserData		= 0xb2,
	sid_internal_SequenceHeader	= 0xb3,		//!< dbut de squence
	sid_internal_SequenceError	= 0xb4,
	sid_internal_Extension		= 0xb5,		//!< prfixe d'extension
	sid_reserved_3				= 0xb6,
	sid_internal_SequenceEnd	= 0xb7,		//!< fin de squence
	sid_internal_Group			= 0xb8,		//!< "group of pictures"
	//
	// Codes de paquets de contrle Mpeg2 PS
	sid_Program_End = 0xb9,		// 0xb9 - fin de programme (de fichier) - ISO IEC 13818-1 section 2.5.3.2
	sid_Pack_Header,			// 0xba - ISO IEC 13818-1 section 2.5.3.4
	sid_System_Header,			// 0xbb - ISO IEC 13818-1 section 2.5.3.6
	//
	// ISO IEC 13818-1 section 2.4.3.7, table 2-18
	sid_Program_Stream_Map,		// 0xbc - ISO IEC 13818-1 section 2.5.4.2
	sid_Private_Stream_1,		// 0xbd
	sid_Padding_Stream,			// 0xbe
	sid_Private_Stream_2,		// 0xbf
	//
	// Audio Streams (0xc0  0xdf) -  ISO IEC 13818-1 section 2.4.3.7, table 2-18
	sid_Audio_Stream__1, sid_Audio_Stream__2, sid_Audio_Stream__3, sid_Audio_Stream__4,
	sid_Audio_Stream__5, sid_Audio_Stream__6, sid_Audio_Stream__7, sid_Audio_Stream__8,
	sid_Audio_Stream__9, sid_Audio_Stream_10, sid_Audio_Stream_11, sid_Audio_Stream_12,
	sid_Audio_Stream_13, sid_Audio_Stream_14, sid_Audio_Stream_15, sid_Audio_Stream_16,
	sid_Audio_Stream_17, sid_Audio_Stream_18, sid_Audio_Stream_19, sid_Audio_Stream_20,
	sid_Audio_Stream_21, sid_Audio_Stream_22, sid_Audio_Stream_23, sid_Audio_Stream_24,
	sid_Audio_Stream_25, sid_Audio_Stream_26, sid_Audio_Stream_27, sid_Audio_Stream_28,
	sid_Audio_Stream_29, sid_Audio_Stream_30, sid_Audio_Stream_31, sid_Audio_Stream_32,
	//
	// Video Streams (0xe0  0xef) -  ISO IEC 13818-1 section 2.4.3.7, table 2-18
	sid_Video_Stream__1, sid_Video_Stream__2, sid_Video_Stream__3, sid_Video_Stream__4,
	sid_Video_Stream__5, sid_Video_Stream__6, sid_Video_Stream__7, sid_Video_Stream__8,
	sid_Video_Stream__9, sid_Video_Stream_10, sid_Video_Stream_11, sid_Video_Stream_12,
	sid_Video_Stream_13, sid_Video_Stream_14, sid_Video_Stream_15, sid_Video_Stream_16,
	//
	// ISO IEC 13818-1 section 2.4.3.7, table 2-18
	sid_ECM_Stream,				// 0xf0
	sid_EMM_Stream,				// 0xf1
	sid_DSMCC_Stream,			// 0xf2 - ITU-T Rec. H.222.0 | ISO/IEC 13818-1 Annex A or ISO/IEC 13818-6_DSMCC_stream
	sid_ISO_IEC_13522,			// 0xf3
	sid_H_222_1_type_A,			// 0xf4
	sid_H_222_1_type_B,			// 0xf5
	sid_H_222_1_type_C,			// 0xf6
	sid_H_222_1_type_D,			// 0xf7
	sid_H_222_1_type_E,			// 0xf8
	sid_Ancillary_Stream,		// 0xf9
	//
	sid_Program_Stream_Directory	= 0xff, // ISO IEC 13818-1 section 2.5.5.2

	// Spcifique PS_SystemHeader (ISO IEC 13818-1 section 2.5.3.6), conflits impossibles
	// avec sid_internal_Group et sid_Program_End :
	sid_SB_AllAudio				= 0xb8,
	sid_SB_AllVideo				= 0xb9
};

/// Codes d'extension (qui prcisent la nature des extensions "sid_internal_Extension")
/// \see ISO IEC 13818-2 section 6.3.1, table 6-2
enum ExtensionStartCode
{
	esc_UNDEF					= 0,
	esc_Sequence				= 1,
	esc_SequenceDisplay			= 2,
	esc_QuantMatrix				= 3,
	esc_Copyright				= 4,
	esc_SequenceScalable		= 5,
	esc_PictureDisplay			= 7,
	esc_PictureCoding			= 8,
	esc_PictureSpatialScalable	= 9,
	esc_PictureTemporalScalable	= 10
};

// Pour insertion dans un switch/case :
#define sid_Audio_Stream_range sid_Audio_Stream__1: \
	case sid_Audio_Stream__2: \
	case sid_Audio_Stream__3: \
	case sid_Audio_Stream__4: \
	case sid_Audio_Stream__5: \
	case sid_Audio_Stream__6: \
	case sid_Audio_Stream__7: \
	case sid_Audio_Stream__8: \
	case sid_Audio_Stream__9: \
	case sid_Audio_Stream_10: \
	case sid_Audio_Stream_11: \
	case sid_Audio_Stream_12: \
	case sid_Audio_Stream_13: \
	case sid_Audio_Stream_14: \
	case sid_Audio_Stream_15: \
	case sid_Audio_Stream_16: \
	case sid_Audio_Stream_17: \
	case sid_Audio_Stream_18: \
	case sid_Audio_Stream_19: \
	case sid_Audio_Stream_20: \
	case sid_Audio_Stream_21: \
	case sid_Audio_Stream_22: \
	case sid_Audio_Stream_23: \
	case sid_Audio_Stream_24: \
	case sid_Audio_Stream_25: \
	case sid_Audio_Stream_26: \
	case sid_Audio_Stream_27: \
	case sid_Audio_Stream_28: \
	case sid_Audio_Stream_29: \
	case sid_Audio_Stream_30: \
	case sid_Audio_Stream_31: \
	case sid_Audio_Stream_32

/// Teste si une valeur suppose tre un \p StreamIDs est de type audio
inline bool isAudio_sid(int eSid) {
			return eSid >= sid_Audio_Stream__1 && eSid <= sid_Audio_Stream_32; }

// Pour insertion dans un switch/case :
#define sid_Video_Stream_range sid_Video_Stream__1: \
	case sid_Video_Stream__2: \
	case sid_Video_Stream__3: \
	case sid_Video_Stream__4: \
	case sid_Video_Stream__5: \
	case sid_Video_Stream__6: \
	case sid_Video_Stream__7: \
	case sid_Video_Stream__8: \
	case sid_Video_Stream__9: \
	case sid_Video_Stream_10: \
	case sid_Video_Stream_11: \
	case sid_Video_Stream_12: \
	case sid_Video_Stream_13: \
	case sid_Video_Stream_14: \
	case sid_Video_Stream_15: \
	case sid_Video_Stream_16

/// Teste si une valeur suppose tre un \p StreamIDs est de type vido
inline bool isVideo_sid(int eSid) {
			return eSid >= sid_Video_Stream__1 && eSid <= sid_Video_Stream_16; }
/// Teste si une valeur suppose tre un \p StreamIDs n'est ni de type vido ni de type audio
inline bool isData_sid(int eSid) {
			return eSid >= sid_Audio_Stream__1 && eSid <= sid_Video_Stream_16; }

/// ISO/IEC 13818-1 section 2.4.4.9, table 2-29
/// + ITU-T H.222.0, Amendment 4 (12/2009)
enum StreamType : UINT8 {
	st_reserved			= 0x00,	// ITU-T | ISO/IEC Reserved
	st_MPEG_1_video		= 0x01,	// ISO/IEC 11172 Video
	st_MPEG_2_video		= 0x02,	// ITU-T Rec. H.262 | ISO/IEC 13818-2 Video or ISO/IEC 11172-2 constrained parameter video stream
	st_MPEG_1_audio		= 0x03,	// ISO/IEC 11172 Audio
	st_MPEG_2_audio		= 0x04,	// ISO/IEC 13818-3 Audio
	st_private			= 0x05,	// ITU-T Rec. H.222.0 | ISO/IEC 13818-1 private_sections
	st_PES_private		= 0x06,	// ITU-T Rec. H.222.0 | ISO/IEC 13818-1 PES packets containing private data
	st_07				= 0x07,	// ISO/IEC 13522 MHEG
	st_08				= 0x08,	// ITU-T Rec. H.222.0 | ISO/IEC 13818-1 Annex A DSM-CC
	st_09				= 0x09,	// ITU-T Rec. H.222.1
	st_0a				= 0x0a,	// ISO/IEC 13818-6 type A
	st_0b				= 0x0b,	// ISO/IEC 13818-6 type B
	st_0c				= 0x0c,	// ISO/IEC 13818-6 type C
	st_0d				= 0x0d,	// ISO/IEC 13818-6 type D
	st_0e				= 0x0e,	// ITU-T Rec. H.222.0 | ISO/IEC 13818-1 auxiliary
	st_AAC_audio		= 0x0f,	// ISO/IEC 13818-7 Audio with ADTS transport syntax
	st_10				= 0x10,	// ISO/IEC 14496-2 Visual
	st_AAC_audio_2		= 0x11,	// ISO/IEC 14496-3 Audio with the LATM transport syntax as defined in ISO/IEC 14496-3 / AMD 1
	st_12				= 0x12,	// ISO/IEC 14496-1 SL-packetized stream or FlexMux stream carried in PES packets
	st_13				= 0x13,	// ISO/IEC 14496-1 SL-packetized stream or FlexMux stream carried in ISO/IEC14496_sections.
	st_14				= 0x14,	// ISO/IEC 13818-6 Synchronized Download Protocol
	st_15				= 0x15,	// Metadata carried in PES packets
	st_16				= 0x16,	// Metadata carried in metadata_sections
	st_17				= 0x17,	// Metadata carried in ISO/IEC 13818-6 Data Carousel
	st_18				= 0x18, // Metadata carried in ISO/IEC 13818-6 Object Carousel
	st_19				= 0x19,	// Metadata carried in ISO/IEC 13818-6 Synchronized Download Protocol
	st_1a				= 0x1a,	// IPMP stream (defined in ISO/IEC 13818-11, MPEG-2 IPMP)
	st_H264_video		= 0x1b,	// AVC video stream conforming to one or more profiles defined in Annex A of ITU-T Rec. H.264 |
								//	ISO/IEC 14496-10 or AVC video sub-bitstream as defined in 2.1.78 or MVC base view sub-bitstream
								//	as defined in 2.1.85 or AVC video sub-bitstream of MVC as defined in 2.1.88
	st_AAC_audio_3		= 0x1c,	// ISO/IEC 14496-3 Audio, without using any additional transport syntax, such as DST, ALS and SLS
	st_1d				= 0x1d,	// ISO/IEC 14496-17 Text
	st_1e				= 0x1e,	// Auxiliary video stream as defined in ISO/IEC 23002-3
	st_1f				= 0x1f,	// SVC video sub-bitstream of an AVC video stream conforming to one or more profiles defined in
								//	Annex G of ITU-T Rec. H.264 | ISO/IEC 14496-10
	st_20				= 0x20,	// MVC video sub-bitstream of an AVC video stream conforming to one or more profiles defined in
								//	Annex H of ITU-T Rec. H.264 | ISO/IEC 14496-10

	// note : 0x21-0x7E ITU-T Rec. H.222.0 | ISO/IEC 13818-1 Reserved

	st_7F				= 0x7f,	// IPMP stream

	// ATSC (standard A/53):
	st_AC3_audio		= 0x81,
	st_EAC3_audio		= 0x87
};

/*
	Program Stream pack
	ISO IEC 13818-1 section 2.5.3.3, table 2-32

		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	pack() {
		pack_header()                                                   0
		while (nextbits()==packet_start_code_prefix) {
			PES_packet()                                                +0
		}
	}
*/

/// ISO IEC 13818-1 section 2.4.3.7
struct PS_hdr
{
	UINT8		start_code[3];
	StreamIDs	stream_id;		// ou PES id

	/// Retourne le "start code" du paquet (devrait toujours tre 0x000001 (little endian)
	UINT32 get_start_code() const {
		return MAKELONG(MAKEWORD(start_code[2], start_code[1]), start_code[0]); }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS
	bool isPS() const {
		return get_start_code()==0x000001; }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS de type \p strm_id
	bool isStream(StreamIDs	strm_id) const {
			return isPS() && stream_id == strm_id; }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS de type audio
	bool isAudioPES() const {
			return isPS() && isAudio_sid(stream_id); }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS de type vido
	bool isVideoPES() const {
			return isPS() && isVideo_sid(stream_id); }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS de type autre (ni audio ni vido)
	bool isDataPES() const {
			return isPS() && isData_sid(stream_id); }

	/// Vrifie (\p true si Ok) si l'en-tte est bien un en-tte de paquet PS (mais pas son contenu)
	bool isPES() const {
		return isPS() && stream_id >= sid_Program_End; }

	/// Vrifie si le paquet est d'un type ligible au dcodage des champs tendus (ISO IEC 13818-1 section 2.4.3.7, table 2-17)
	/// (correspondant  la structure drive \p PS_hdr_wd).
	bool isTier1Stream() const {
		return
			(stream_id < sid_ECM_Stream && !(stream_id == sid_Program_Stream_Map || stream_id == sid_Padding_Stream || stream_id == sid_Private_Stream_2)) ||
			(stream_id > sid_DSMCC_Stream && !(stream_id == sid_H_222_1_type_E || stream_id == sid_Program_Stream_Directory)); }

	/// Drivation vers un type de descripteur spcifique
	template <class T> const T & as() const {
		return *reinterpret_cast<const T *>(this); }

	PUINT8 next() const {
		return PUINT8(this+1); }

	//operator PCUINT8 () const {
	//	return &as<UINT8>(); }

	/// Taille de l'en-tte PS, gnrique
	UINT header_size() const;

	/// Taille du paquet PS, gnrique
	UINT packet_size() const;

	/// Constructeur par dfaut
	PS_hdr()
	{
	}

	/// Constructeur avec initialisation
	PS_hdr(StreamIDs id) :
		stream_id(id)
	{
		start_code[1] = start_code[0] = 0x00;
		start_code[2] = 0x01;
	}

	tstring	toString() const {
				TCHAR	szStr[32];
				if (stream_id == 0)
					strcpy_T(szStr, TEXT("Picture"));
				else if (stream_id >= sid_internal_Slice_first && stream_id <= sid_internal_Slice_last)
					_stprintf_s(szStr, TEXT("Slice %u"), stream_id);
				else if (stream_id >= sid_reserved_1 && stream_id <= sid_Private_Stream_2)
					strcpy_T(szStr,
						EnumToString(stream_id - sid_reserved_1,
							"\1\0\1\0User Data\0Sequence Header\0Sequence Error\0Extension\0\1\0Sequence End\0Group of Pictures\0"
							"Program End\0Pack header\0System Header\0Program Stream Map\0Private 1\0Padding\0Private 2\0"));
				else if (stream_id >= sid_Audio_Stream__1 && stream_id <= sid_Audio_Stream_32)
					_stprintf_s(szStr, TEXT("Audio 0x%02x"), stream_id);
				else if (stream_id >= sid_Video_Stream__1 && stream_id <= sid_Video_Stream_16)
					_stprintf_s(szStr, TEXT("Video 0x%02x"), stream_id);
				else if (stream_id >= sid_ECM_Stream && stream_id < sid_Program_Stream_Directory)
					strcpy_T(szStr,
						EnumToString(stream_id - sid_reserved_1,
							"ECM Stream\0EMM Stream\0ISO IEC 13818\0ISO IEC 13522\0H 222 1 type A\0"
							"H 222 1 type B\0H 222 1 type C\0H 222 1 type D\0H 222 1 type E\0Ancillary Stream\0"));
				else if (stream_id == sid_Program_Stream_Directory)
					strcpy_T(szStr, TEXT("Stream Directory"));
				return szStr; }
};

/// ISO IEC 13818-1 section 2.4.3.7
struct PS_hdr_wl : public PS_hdr	// wl="With Length"
{
	UBE16	hdr_length;

	PUINT8 base_of_data_size() const {
		return PUINT8(this+1); }
	void set_data_size(PUINT8 adr) {
		hdr_length = UINT16(adr-base_of_data_size()); }
	UINT32 packet_size() const {
		return sizeof(*this) + hdr_length(); }
	PUINT8 next() const {
		return PUINT8(this) + packet_size(); }

	/// Constructeur par dfaut
	PS_hdr_wl()
	{
	}

	PS_hdr_wl(StreamIDs id) :
		PS_hdr(id)
	{
		hdr_length = 0;
	}

	tstring	toString() const {
				TCHAR	szStr[16];
				_stprintf_s(szStr, TEXT(" (l=%u)"), packet_size());
				return __super::toString() + szStr; }
};

/*
	Program Stream pack header (MPEG1)
	ISO IEC 11172-1 section 2.4.3.2

		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	pack_header() {
		start_code (0x00, 0x00, 0x01)                       24          0..2
		stream_id (0xBA)                                     8          3
		'0010'                                               4          4
		system_clock_reference_base [32..30]                 3          4
		marker_bit '1'                                       1          4
		system_clock_reference_base [29..15]                15          5..6
		marker_bit '1'                                       1          6
		system_clock_reference_base [14..0]                 15          7..8
		marker_bit '1'                                       1          8
		marker_bit '1'                                       1          9
		program_mux_rate                                    22          9..11
		marker_bit '1'                                       2          11
	}
*/

struct PS_PackHeader_v1 : public PS_hdr
{
	PS_D_SCR_v1		scr;	// System Reference Clock
	PS_PH_mux_rate	mux_rate;

	bool isV1() const {
		return scr.isV1(); }
	UINT32 packet_size() const {
		return sizeof(*this); }
	PUINT8 end_of_header() const {
		return PUINT8(this+1); }
	PUINT8 end_of_packet() const {
		return end_of_header(); }

	// Constructeur
	PS_PackHeader_v1() :
		PS_hdr(sid_Pack_Header)
	{
	}

	bool is_valid() const {
		return isStream(sid_Pack_Header) && isV1(); }

	tstring	toString() const {
				TCHAR	szStr[64];
				_stprintf_s(szStr, TEXT(": scr=%s, mux rate=%u"), scr.toString().c_str(), (UINT)mux_rate);
				return __super::toString() + szStr; }
};

/*
	Program Stream pack header (MPEG2)
	Voir aussi : http://dvd.sourceforge.net/dvdinfo/packhdr.html
	ISO IEC 13818-1 section 2.5.3.2, table 2-33

		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	pack_header() {
		start_code (0x00, 0x00, 0x01)                       24          0..2
		stream_id (0xBA)                                     8          3
		'01'                                                 2          4
		system_clock_reference_base [32..30]                 3          4
		marker_bit '1'                                       1          4
		system_clock_reference_base [29..15]                15          4..6
		marker_bit '1'                                       1          6
		system_clock_reference_base [14..0]                 15          6..8
		marker_bit '1'                                       1          8
		system_clock_reference_extension                     9          8, 9
		marker_bit '1'                                       1          9
		program_mux_rate                                    22          10..12
		marker_bits '11'                                     2          12
		reserved                                             5          13
		pack_stuffing_length                                 3          13
		for (i=0; i<pack_stuffing_length; i++) {
			stuffing_byte                                    8          +0
		}
		if (nextbits()==system_header_start_code) {
			system_header()                                             +0
		}
	}
*/

struct PS_PackHeader_v2 : public PS_hdr
{
	PS_D_SCR_v2		scr;	// System Reference Clock
	PS_PH_mux_rate	mux_rate;
	UINT8			length;	// 3 bits de poids faible seulement

	bool isV2() const {
		return scr.isV2(); }
	UINT32 stuffing_length() const {
		return length & 0x07; }
	UINT32 packet_size() const {
		return sizeof(*this)+stuffing_length(); }
	PUINT8 end_of_header() const {
		return PUINT8(this+1); }
	PUINT8 end_of_packet() const {
		return end_of_header()+stuffing_length(); }

	// Constructeur
	PS_PackHeader_v2() :
		PS_hdr(sid_Pack_Header)
	{
	}

	bool is_valid() const {
		return isStream(sid_Pack_Header) && isV2(); }

	tstring	toString() const {
				TCHAR	szStr[64];
				_stprintf_s(szStr, TEXT(": scr=%s, mux rate=%u"), scr.toString().c_str(), (UINT)mux_rate);
				return __super::toString() + szStr; }
};

#define PES_SB_BUFFERSCALE_FLAG 0x20

struct PS_D_StdBuff
{
	UBE<13>	buffer_size;

	UINT16 buffer_scale() const {
		return buffer_size.flags() & PES_SB_BUFFERSCALE_FLAG ? 0x400 : 0x80; }
	void set_size(bool scale, UINT16 size) {
		buffer_size.hi = 0xc0;	// markers
		buffer_size = size; buffer_size.set_flags(PES_SB_BUFFERSCALE_FLAG, scale ? PES_SB_BUFFERSCALE_FLAG : 0); }
	UINT actual_buffer_size() const {
		return buffer_size()*buffer_scale(); }
	PUINT8 next() const {
		return PUINT8(this+1); }

	tstring	toString() const {
				TCHAR	szStr[48]; UINT sz=buffer_size(), sc=buffer_scale();
				_stprintf_s(szStr, TEXT("%u (%u*%u)"), sz*sc, sz, sc);
				return szStr; }
};

/*
	Program Stream system header
	Voir aussi : ISO IEC 13818-1 section 2.5.3.5, table 2-34

		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	system_header () {
		start_code (0x00, 0x00, 0x01)                       24          0..2
		stream_id (0xBB)                                     8          3
		packet_length                                       16          4, 5
		marker_bit                                           1          6
		rate_bound                                          22          6..8
		marker_bit                                           1          8
		audio_bound                                          6          9
		fixed_flag                                           1          9
		CSPS_flag                                            1          9
		system_audio_lock_flag                               1          10
		system_video_lock_flag                               1          10
		marker_bit                                           1          10
		video_bound                                          5          10
		packet_rate_restriction_flag                         1          11
		reserved_bits                                        7          11
		while (nextbits()=='1') {
			stream_id                                        8          +0
			'11'                                             2          +1
			P-STD_buffer_bound_scale                         1          +1
			P-STD_buffer_size_bound                         13          +1, +2
		}
	}
*/

// ** Flags dans les champs de PS_SystemHeader :
// audio_bound_and_flags :
#define SH_FIXED_FLAG					0x02
#define SH_CSPS_FLAG					0x01
// video_bound_and_flags :
#define SH_SYSTEM_AUDIO_LOCK_FLAG		0x80
#define SH_SYSTEM_VIDEO_LOCK_FLAG		0x40
// packet_rate_restriction_flag :
#define SH_PACKET_RATE_RESTRICTION_FLAG	0x80

struct PS_SystemHeader : public PS_hdr_wl
{
	struct s_entry
	{
		StreamIDs		stream_id;
		PS_D_StdBuff	bound_buf;

		bool isEot() const {
			return (stream_id & 0x80)==0; }
		const s_entry * next() const {
			return this+1; }

		tstring	toString() const {
					TCHAR	szStr[16];
					_stprintf_s(szStr, TEXT("0x%02x scale = "), stream_id);
					return szStr + bound_buf.toString(); }
	};

	PS_D_mux_rate	rate_bound;						// MUX rate maximal de l'ensemble du fichier PS
	UINT8			audio_bound_and_flags;			// nombre maximal de flux audio (+ 2 flags)
	UINT8			video_bound_and_flags;			// nombre maximal de flux vido (+ 3 flags)
	UINT8			packet_rate_restriction_flag;	// + 7 bits rservs

	UINT8 audio_bound() const {
		return audio_bound_and_flags >> 2; }
	UINT8 video_bound() const {
		return video_bound_and_flags & 0x1f; }
	bool fixed() const {
		return (audio_bound_and_flags & SH_FIXED_FLAG)!=0; }
	bool CSPS() const {
		return (audio_bound_and_flags & SH_CSPS_FLAG)!=0; }
	bool system_audio_lock() const {
		return (video_bound_and_flags & SH_SYSTEM_AUDIO_LOCK_FLAG)!=0; }
	bool system_video_lock() const {
		return (video_bound_and_flags & SH_SYSTEM_VIDEO_LOCK_FLAG)!=0; }
	bool packet_rate_restriction() const {
		return (packet_rate_restriction_flag & SH_PACKET_RATE_RESTRICTION_FLAG)!=0; }
	//
	// Note : si, dans un lment, la fonction PS_SH_entry::isEot() retourne 'false', il se pourrait
	// qu'on doive comprendre que les lments suivants ne sont pas valides, mme si 'count' n'est pas puis.
	int count() const {
		return int(INT_PTR(next()-INT_PTR(this+1))/sizeof(s_entry)); }
	s_entry & operator [] (int index) {
		return reinterpret_cast<s_entry *>(this+1)[index]; }
	const s_entry & operator [] (int index) const {
		return reinterpret_cast<const s_entry *>(this+1)[index]; }
	streamScanner<s_entry> entryRange() const {
		return streamScanner<s_entry>(this+1, PUINT8(this+1) + hdr_length()); }

	// Constructeur
	PS_SystemHeader() :
		PS_hdr_wl(sid_System_Header)
	{
	}

	tstring	toString() const {
				TCHAR	szStr[64];
				_stprintf_s(szStr, TEXT(": rb=%u, ab=%u, vb=%u"), (UINT)rate_bound, audio_bound(), video_bound());
				tstring	strWrk =  __super::toString() + szStr;
				for (streamScanner<s_entry> se(entryRange()); !se && !se().isEot(); ++se)
					strWrk += TEXT("\n\t") + se().toString();
				return strWrk; }
};

/*
	Program Stream Map
	Voir aussi : ISO IEC 13818-1 section 2.5.4.1, table 2-35

		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	program_stream_map() {
		packet_start_code_prefix                            24          0..2
		map_stream_id                                        8          3
		program_stream_map_length                           16          4, 5
		current_next_indicator                               1          6
		reserved                                             2          6
		program_stream_map_version                           5          6
		reserved                                             7          7
		marker_bit                                           1          7
		program_stream_info_length                          16          8, 9
		for (i = 0; i < N; i++) {
			descriptor()
		}
		elementary_stream_map_length                        16          +0, +1
		for (i = 0; i < N1; i++) {
			stream_type                                      8          +0
			elementary_stream_id                             8          +1
			elementary_stream_info_length                   16          +2, +3
			for (i = 0; i < N2; i++) {
				descriptor()
			}
		}
		CRC_32                                              32          +0
	}
*/

struct PS_ProgramStreamMap : public PS_hdr_wl
{
	// Descripteur de flot lmentaire "elementary stream", partie de PS_ProgramStreamMap.
	// Ce type de champ peut tre rpt pour former une table, donc chaque lment peut tre
	// obtenu  partir du prcdent en appelant la mthode "next" :
	struct s_map_entry
	{
		StreamType	type;
		StreamIDs	id;
		UBE16		dloop_length;

		MDESCR::D_scanner descrRange() const {
			return MDESCR::D_scanner(this+1, PUINT8(this+1) + dloop_length()); }
		s_map_entry * next() const {
			return reinterpret_cast<s_map_entry *>(PUINT8(this+1) + dloop_length()); }

		tstring	toString() const {
					TCHAR	szStr[32];
					_stprintf_s(szStr, TEXT("type=0x%02x, id=0x%02x, lng=%u"), type, id, dloop_length());
					return szStr; }
	};

	// Seconde partie de PS_ProgramStreamMap, contenant la longueur de la table
	// des flots lmentaires "elementary stream" :
	struct s_map_hdr
	{
		UBE16	eloop_length;

		streamScanner<s_map_entry> entryRange() const {
			return streamScanner<s_map_entry>(this+1, PUINT8(this+1) + eloop_length()); }
	};

	UINT8	version_;
	UINT8	reserved;
	UBE16	dloop_length;

	UINT8	version() const {
				return version_ & 0x1f; }

	// Accs  la table de descripteurs incluse :
	MDESCR::D_scanner descrRange() const {
		return MDESCR::D_scanner(this+1, PUINT8(this+1) + dloop_length()); }
	// Accs aux champs suivant :
	s_map_hdr & map_hdr() const {
		return *reinterpret_cast<s_map_hdr *>(PUINT8(this+1) + dloop_length()); }
	CRC_32 & crc32() const {
		return *reinterpret_cast<CRC_32 *>(next() - sizeof(CRC_32)); }

	// Constructeur :
	PS_ProgramStreamMap() :
		PS_hdr_wl(sid_Program_Stream_Map)
	{
	}

	tstring	toString() const {
				TCHAR	szStr[32];
				_stprintf_s(szStr, TEXT(": v=%u"), version());
				tstring	strWrk = __super::toString() + szStr;
				for (MDESCR::D_scanner sd(descrRange()); !sd; ++sd)
					strWrk += TEXT("\n\tinf: ") + sd().toString();
				for (streamScanner<s_map_entry> se(map_hdr().entryRange()); !se; ++se)
					strWrk += TEXT("\n\tmap: ") + se().toString();
				_stprintf_s(szStr, TEXT("\n\tcrc32=0x%08x"), (UINT32)crc32());
				return strWrk + szStr; }
};

/*
	Program Stream PES packet
	Voir aussi : http://dvd.sourceforge.net/dvdinfo/pes-hdr.html
	ISO IEC 13818-1 section 2.4.3.7, table 2-17
		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	PES_packet() {
		start_code (0x00, 0x00, 0x01)                       24          0..2
		stream_id                                            8          3
		packet_length                                       16          4, 5
		if (stream_id!=program_stream_map
			&& stream_id!=padding_stream
			&& stream_id!=private_stream_2
			&& stream_id!=ECM
			&& stream_id!=EMM
			&& stream_id!=program_stream_directory
			&& stream_id!=DSMCC_stream
			&& stream_id!=ITU-T Rec.H.222.1 type E stream
		) {
			'10'                                             2          6
			PES_scrambling_control                           2          6
			PES_priority                                     1          6
			data_alignment_indicator                         1          6
			copyright                                        1          6
			original_or_copy                                 1          6
			PTS_DTS_flags                                    2          7
			ESCR_flag                                        1          7
			ES_rate_flag                                     1          7
			DSM_trick_mode_flag                              1          7
			additional_copy_info_flag                        1          7
			PES_CRC_flag                                     1          7
			PES_extension_flag                               1          7
			PES_header_data_length                           8          8
			if (PTS_DTS_flags=='10') {
				'0010'                                       4          9
				PTS [32..30]                                 3          9
				marker_bit                                   1          9
				PTS [29..15]                                15          10, 11
				marker_bit                                   1          11
				PTS [14..0]                                 15          12, 13
				marker_bit                                   1          13
			}
			if (PTS_DTS_flags=='11') {
				'0011'                                       4          9
				PTS [32..30]                                 3          9
				marker_bit                                   1          9
				PTS [29..15]                                15          10, 11
				marker_bit                                   1          11
				PTS [14..0]                                 15          12, 13
				marker_bit                                   1          13
				'0001'                                       4          14
				DTS [32..30]                                 3          14
				marker_bit                                   1          14
				DTS [29..15]                                15          15, 16
				marker_bit                                   1          16
				DTS [14..0]                                 15          17, 18
				marker_bit                                   1          18
			}
			if (ESCR_flag=='1') {
				reserved                                     2          +0
				ESCR_base[32..30]                            3          +0
				marker_bit                                   1          +0
				ESCR_base[29..15]                           15          +0..+2
				marker_bit                                   1          +2
				ESCR_base[14..0]                            15          +2..+4
				marker_bit                                   1          +4
				ESCR_extension                               9          +4, +5
				marker_bit                                   1          +5
			}
			if (ES_rate_flag=='1') {
				marker_bit                                   1          +0
				ES_rate                                     22          +0..+2
				marker_bit                                   1          +2
			}
			if (DSM_trick_mode_flag=='1') {
				trick_mode_control                           3          +0
				if (trick_mode_control==fast_forward) {
					field_id                                 2          +0
					intra_slice_refresh                      1          +0
					frequency_truncation                     2          +0
				} else if (trick_mode_control==slow_motion) {
					rep_cntrl                                5          +0
				} else if (trick_mode_control==freeze_frame) {
					field_id                                 2          +0
					reserved                                 3          +0
				} else if (trick_mode_control==fast_reverse) {
					field_id                                 2          +0
					intra_slice_refresh                      1          +0
					frequency_truncation                     2          +0
				} else if (trick_mode_control==slow_reverse) {
					rep_cntrl                                5          +0
				} else
					reserved                                 5          +0
			}
			if (additional_copy_info_flag=='1') {
				marker_bit                                   1          +0
				additional_copy_info                         7          +0
			}
			if (PES_CRC_flag=='1') {
				previous_PES_packet_CRC                     16          +0, +1
			}
			if (PES_extension_flag=='1') {
				PES_private_data_flag                        1          +0
				pack_header_field_flag                       1          +0
				program_packet_sequence_counter_flag         1          +0
				P-STD_buffer_flag                            1          +0
				reserved                                     3          +0
				PES_extension_flag_2                         1          +0
				if (PES_private_data_flag=='1') {
					PES_private_data                       128          +0..+15
				}
				if (pack_header_field_flag=='1') {
					pack_field_length                        8          +0
					pack_header()                                       +1
				}
				if (program_packet_sequence_counter_flag=='1') {
					marker_bit                               1          +0
					program_packet_sequence_counter          7          +0
					marker_bit                               1          +1
					MPEG1_MPEG2_identifier                   1          +1
					original_stuff_length                    6          +1
				}
				if (P-STD_buffer_flag=='1') {
					'01'                                     2          +0
					P-STD_buffer_scale                       1          +0
					P-STD_buffer_size                       13          +0, +1
				}
				if (PES_extension_flag_2=='1') {
					marker_bit                               1          +0
					PES_extension_field_length               7          +1
					for (i=0; i<PES_extension_field_length; i++) {
						reserved                             8          +0
					}
				}
			}
			for (i = 0; i < N1; i++) {
				stuffing_byte                                8          +0
			}
			for (i = 0; i < N2; i++) {
				PES_packet_data_byte                         8          +0
			}
		} else if (stream_id==program_stream_map
			|| stream_id==private_stream_2
			|| stream_id==ECM
			|| stream_id==EMM
			|| stream_id==program_stream_directory
			|| stream_id==DSMCC_stream
			|| stream_id==ITU-T Rec.H.222.1 type E stream
		) {
			for (i=0; i<PES_packet_length; i++) {
				PES_packet_data_byte                         8          +0
			}
		} else if (stream_id==padding_stream) {
			for (i=0; i<PES_packet_length; i++) {
				padding_byte                                 8          +0
			}
		}
	}
*/

// Dfinition des bits de "PS_hdr_wd::info" :
#define PES_PRIORITY_FLAG				0x08
#define DATA_ALIGNMENT_INDICATOR		0x04	// 1=video start code or audio syncword follows
#define COPYRIGHT_FLAG					0x02	// 1=copyrighted material
#define ORIGINAL_FLAG					0x01	// 1=original; 0=copy

// Dfinition des bits de "PS_hdr_wd::flags" :
#define	PES_PTS_FLAG					0x80
#define	PES_DTS_FLAG					0x40
#define	PES_ESCR_FLAG					0x20
#define	PES_ES_RATE_FLAG				0x10
#define	PES_DSM_TRICK_MODE_FLAG			0x08
#define	PES_ADDITIONAL_COPY_INFO_FLAG	0x04
#define	PES_CRC_FLAG					0x02
#define	PES_EXTENSION_FLAG				0x01

// Dfinition des flags de "PS_D_extension1" :
#define	PES_PRIVATE_DATA_FLAG			0x80
#define	PES_PACK_HEADER_FIELD_FLAG		0x40
#define	PES_PROG_PACKET_SEQ_COUNT_FLAG	0x20
#define	PES_P_STD_BUFFER_FLAG			0x10
// flags 0x08, 0x04, 0x02 rservs
#define	PES_EXTENSION_FLAG_2			0x01

/// ISO IEC 13818-1 section 2.4.3.7
struct PS_hdr_wd : public PS_hdr_wl	// wd = "With Data"
{
	UINT8	info;
	UINT8	flags;
	UINT8	data_length;

	UINT32 header_size() const {
		return sizeof(*this)+data_length; }
	PUINT8 base_of_data() const {
		return PUINT8(this)+header_size(); }
	PUINT8 end_of_header() const {
		return PUINT8(this)+header_size(); }
	PUINT8 beg_of_fields() const {
		return PUINT8(this+1); }

	UINT8 scrambling_control() const {
		return UINT8((info >> 4) & 0x03); }
	bool priority_flag() const {
		return (info & PES_PRIORITY_FLAG) != 0; }
	bool data_alignment() const {
		return (info & DATA_ALIGNMENT_INDICATOR) != 0; }
	bool copyrighted() const {
		return (info & COPYRIGHT_FLAG) != 0; }
	bool original() const {
		return (info & ORIGINAL_FLAG) != 0; }

	UINT8 pts_dts_flags_value() const {
		return flags >> 6; }
	bool pts_flag() const {
		return (flags & PES_PTS_FLAG) != 0; }
	bool dts_flag() const {
		return (flags & PES_DTS_FLAG) != 0; }
	bool escr_flag() const {
		return (flags & PES_ESCR_FLAG) != 0; }
	bool es_rate_flag() const {
		return (flags & PES_ES_RATE_FLAG) != 0; }
	bool dsm_trick_mode_flag() const {
		return (flags & PES_DSM_TRICK_MODE_FLAG) != 0; }
	bool additional_copy_info_flag() const {
		return (flags & PES_ADDITIONAL_COPY_INFO_FLAG) != 0; }
	bool crc_flag() const {
		return (flags & PES_CRC_FLAG) != 0; }
	bool extension_flag() const {
		return (flags & PES_EXTENSION_FLAG) != 0; }

	struct Ext1
	{
		UINT8	flags;

		PUINT8	begin() const {
			return PUINT8(this+1); }
		PUINT8	next() const {
			return PUINT8(this+1); }
	};

	struct PrivDt
	{
		UINT8	data[16];

		PUINT8	next() const {
			return PUINT8(this+1); }
	};

	struct Ext2
	{
		UINT8	length;

		PUINT8	data() const {
			return PUINT8(this+1); }
		PUINT8	next() const {
			return data() + length; }
	};

	struct PackHd
	{
		UINT8				length;
		PS_PackHeader_v2	pack_header;

		PUINT8	next() const {
			return pack_header.end_of_packet(); }
	};

	struct SeqC
	{
		UINT8	program_packet_sequence_counter;
		UINT8	original_stuff_length_and_id;	// + MPEG1_MPEG2_identifier

		UINT8	sequence_counter() const {
			return program_packet_sequence_counter & 0x7f; }
		UINT8	mpeg1_mpeg2_identifier() const {
			return (original_stuff_length_and_id & 0x40) >> 6; }
		UINT8	original_stuff_length() const {
			return original_stuff_length_and_id & 0x3f; }
		PUINT8	next() const {
			return PUINT8(this+1); }
	};

	// Ajout d'un champ optionnel. Tout les champs doivent tre ajouts dans
	// l'ordre de la spcification ISO/IEC 13818-1, section 2.4.3.7
	template <typename T>
		void addfield(UINT8 flg, const T & field) {
			flags |= flg;
			reinterpret_cast<T &>(PUINT8(this+1)[data_length]) = field;
			data_length += sizeof(T); }

	// Ajout d'un premier champ en tant que champ PTS
	void	addPts(TIM_27M pts) {
				addfield(PES_PTS_FLAG, PS_D_TimeStamp(pts, 0x20)); }

	// Ajout des deux premiers champs en tant que champ PTS avec champ DTS associ
	void	addPtsDts(TIM_27M pts, TIM_27M dts) {
				addfield(PES_PTS_FLAG, PS_D_TimeStamp(pts, 0x30));
				addfield(PES_DTS_FLAG, PS_D_TimeStamp(dts, 0x10)); }

	// Ajout d'octets de remplissage ( effectuer en dernier)
	void	addStuffing(UINT8 val, UINT8 size) {
				memset(&PUINT8(this+1)[data_length], val, size);
				data_length = data_length + size; }

	/// Constructeur
	PS_hdr_wd()							// neutre
	{
	}

	/// Constructeur pour cration
	PS_hdr_wd(StreamIDs id, UINT8 inf, UINT8 flgs) :
		PS_hdr_wl(id),
		info(inf),
		flags(flgs),
		data_length(0)
	{
	}

	PS_hdr_wd(const PS_hdr_wd & src) :	// copie
		PS_hdr_wl(src.stream_id)
	{
		memcpy(this, &src, src.header_size());
	}
};

/// Sous-ensemble de la classe 'PS_data_unwrapper_full'
/// (seulement les deux champs couramment utiliss contenant des informations de timing)
struct PS_data_unwrapper_base
{
	const PS_hdr_wd &	hdr;
	PUINT8				p_next;
	PS_D_TimeStamp *	p_pts;
	PS_D_TimeStamp *	p_dts;

	/// Constructeur pour dballage
	PS_data_unwrapper_base(const PS_hdr_wd & h) :
		hdr(h),
		p_next(h.beg_of_fields()),
		p_pts(NULL),
		p_dts(NULL)
	{
		PUINT8	p_nxt = p_next;
		UINT8	flags = h.flags;

		if (flags & PES_PTS_FLAG) {
			get_and_skip(p_nxt, p_pts);
			if (flags & PES_DTS_FLAG)
				get_and_skip(p_nxt, p_dts);
		}
		p_next = p_nxt;
	}
};

/*
 * Dballage de tous les champs "data" optionnels du paquet 'hdr':
 * La structure est charge avec des pointeurs vers tous les champs
 * existants, ou des valeurs NULL pour tous les champs inexistants.
 * Le pointeur 'p_next' pointe l'espace encore libre qui suit, lequel correspond
 *  des octets de remplissage optionnels suivis des donnes du paquet.
 */
struct PS_data_unwrapper_full : public PS_data_unwrapper_base
{
	PS_D_SCR_v2 *		p_escr;
	PS_D_mux_rate *		p_rate;
	UINT8uw *			p_trick;
	UINT8uw *			p_add_copy_info;
	CRC_16 *			p_prev_crc;
	//
	PS_hdr_wd::Ext1 *	p_ext1;		// pointeur dbut premier champ d'extension optionnel
	PS_hdr_wd::PrivDt *	p_priv_data;
	PS_hdr_wd::PackHd *	p_pack;
	PS_hdr_wd::SeqC *	p_seqcount;
	PS_D_StdBuff *		p_stdbuf;
	PS_hdr_wd::Ext2 *	p_ext2;		// pointeur dbut second champ d'extension optionnel

	PS_data_unwrapper_full(const PS_hdr_wd & hdr);
};

/// Taille du paquet PS, gnrique
inline UINT PS_hdr::packet_size() const
{
	switch (stream_id) {

	case sid_Pack_Header:
		return as<PS_PackHeader_v2>().isV2() ?
			as<PS_PackHeader_v2>().packet_size() :
			as<PS_PackHeader_v1>().packet_size();

	case sid_Program_End:
		return sizeof(PS_hdr);

	default:
		return as<PS_hdr_wl>().packet_size();
	}
}

	/// Taille de l'en-tte PS, gnrique
inline UINT PS_hdr::header_size() const
{
	switch (stream_id) {

	case sid_Pack_Header:
	case sid_System_Header :
	case sid_Program_End:
		return packet_size();
	}

	return isTier1Stream() ? as<PS_hdr_wd>().header_size() : sizeof(PS_hdr_wl);
}

/****************************************************************************************
 *								SOUS-TITRAGE DVB										*
 ****************************************************************************************/

/*
	**** ETSI EN 300 743 V1.3.1 section 7.1
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	PES_data_field() {
		data_identifier						 8			0
		subtitle_stream_id					 8			1
		while nextbits() == '0000 1111' {				2
			Subtitling_segment()						+0
		}
		end_of_PES_data_field_marker		 8			+0
	}
*/

/// ETSI EN 300 743 V1.3.1 section 7.1 (& ISO IEC 13818-1 section 2.4.3.7, "private_stream_1")
struct PES_data_field_subtitle
{
	UINT8	data_identifier;
	UINT8	subtitle_stream_id;

	PCUINT8	data() const {
				return PCUINT8(this+1); }
};

/*
	**** ETSI EN 300 743 V1.3.1 section 7.2
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	Subtitling_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		segment_data_field()							6
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.1
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	display_definition_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		dds_version_number					 4			6
		display_window_flag					 1			6
		reserved							 3			6
		display_width						16			7
		display_height						16			9
		if (display_window_flag == 1) {
			display_window_horizontal_position_minimum
											16			11
			display_window_horizontal_position_maximum
											16			13
			display_window_vertical_position_minimum
											16			15
			display_window_vertical_position_maximum
											16			17
		}
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.2
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	page_composition_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		page_time_out						 8			6
		page_version_number					 4			7
		page_state							 2			7
		reserved							 2			7
		while (processed_length < segment_length) {
			region_id						 8			+0
			reserved						 8			+1
			region_horizontal_address		16			+2
			region_vertical_address			16			+4
		}
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.3
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	region_composition_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		region_id							 8			6
		region_version_number				 4			7
		region_fill_flag					 1			7
		reserved							 3			7
		region_width						16			8
		region_height						16			10
		region_level_of_compatibility		 3			12
		region_depth						 3			12
		reserved							 2			12
		CLUT_id								 8			13
		region_8-bit_pixel_code				 8			14
		region_4-bit_pixel-code				 4			15
		region_2-bit_pixel-code				 2			15
		reserved							 2			15
		while (processed_length < segment_length) {
			object_id						16			+0
			object_type						 2			+2
			object_provider_flag			 2			+2
			object_horizontal_position		12			+2, +3
			reserved						 4			+4
			object_vertical_position		12			+4, +5
			if (object_type == 0x01 or object_type == 0x02) {
				foreground_pixel_code		 8 bslbf	+6
				background_pixel_code		 8 bslbf	+7
			}
		}
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.4
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	CLUT_definition_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		CLUT_id								 8			6
		CLUT_version_number					 4			7
		reserved							 4			7
		while (processed_length < segment_length) {
			CLUT_entry_id					 8			+0
			2-bit/entry_CLUT_flag			 1			+1
			4-bit/entry_CLUT_flag			 1			+1
			8-bit/entry_CLUT_flag			 1			+1
			reserved						 4			+1
			full_range_flag					 1			+1
			if full_range_flag =='1' {
				Y-value						 8			+2
				Cr-value					 8			+3
				Cb-value					 8			+4
				T-value						 8			+5
			} else {
				Y-value						 6			+2
				Cr-value					 4			+2, +3
				Cb-value					 4			+3
				T-value						 2			+3
			}
		}
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.5
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	object_data_segment() {
		sync_byte							 8			0
		segment_type						 8			1
		page_id								16			2
		segment_length						16			4
		object_id							16			6
		object_version_number				 4			8
		object_coding_method				 2			8
		non_modifying_colour_flag			 1			8
		reserved							 1			8
		if (object_coding_method == '00') {
			top_field_data_block_length		16			9
			bottom_field_data_block_length	16			11
			while(processed_length < top_field_data_block_length)
				pixel-data_sub-block()
			while (processed_length < bottom_field_data_block_length)
				pixel-data_sub-block()
			if (!wordaligned())
				8_stuff_bits				 8			+0
		}
		if (object_coding_method == '01') {
			number of codes					 8			9
			for (i == 1, i <= number of codes, i ++)
				character_code				16			+0
		}
	}

	**** ETSI EN 300 743 V1.3.1 section 7.2.5.1
		Syntax							No.	of bits	  Offset
	----------------------------------------------------------
	pixel-data_sub-block() {
		data_type							 8			0
		if data_type == '0x10' {
			repeat {
				2-bit/pixel_code_string()
			} until (end of 2-bit/pixel_code_string)
		while (!bytealigned())
			2_stuff_bits					 2			+0
		if data_type == '0x11' {
			repeat {
				4-bit/pixel_code_string()
			} until (end of 4-bit/pixel_code_string)
			if (!bytealigned())
				4_stuff_bits				 4			+0
		}
		if data_type == '0x12' {
			repeat {
				8-bit/pixel_code_string()
			} until (end of 8-bit/pixel_code_string)
		}
		if data_type == '0x20'
			2_to_4-bit_map-table			16			+0
		if data_type == '0x21'
			2_to_8-bit_map-table			32			+0
		if data_type == '0x22'
			4_to_8-bit_map-table		   128			+0
	}
*/

/// ETSI EN 300 743 V1.3.1 section 7.2
struct Subtitling_segment
{
	enum Sync : UINT8
	{
		s_Sync = 0x0f,
		s_End  = 0x3f	//!< end_of_PES_data_field_marker
	};

	/// ETSI EN 300 743 V1.3.1 section 7.2
	enum Type : UINT8
	{
		t_PageComposition	= 0x10,
		t_RegionComposition	= 0x11,
		t_ClutDefinition	= 0x12,
		t_ObjectData		= 0x13,
		t_DisplayDefinition	= 0x14,
		t_EndOfDisplaySet	= 0x80,
		t_PrivateDataBegin	= 0x81,
		t_PrivateDataEnd	= 0xef,
		t_Stuffing			= 0xff
	};

	Sync	sync_byte;
	Type	segment_type;
	UBE16	page_id;
	UBE16	segment_length;

	template <class T> const T&	as() const {	// Drivation vers un type de descripteur spcifique
									return *reinterpret_cast<const T *>(this); }
	UINT16						segment_size() const {
									return sizeof(*this) + segment_length; }
	PUINT8						data() const {
									return PUINT8(this+1); }
	PUINT8						end() const {
									return PUINT8(this) + segment_size(); }
	const Subtitling_segment *	next() const {
									return reinterpret_cast<Subtitling_segment *>(end()); }
};

/// ETSI EN 300 743 V1.3.1 section 7.2.1
struct display_definition_segment : public Subtitling_segment
{
	UINT8	dds_version_and_flag;
	UBE16	display_width;		// -1 !
	UBE16	display_height;		// -1 !

	UINT8	dds_version_number() const {
				return dds_version_and_flag >> 4; }
	bool	display_window_flag() const {
				return (dds_version_and_flag & 0x08) != 0; }

	// if (display_window_flag()) :
	UBE16	display_window_horizontal_position_minimum;
	UBE16	display_window_horizontal_position_maximum;
	UBE16	display_window_vertical_position_minimum;
	UBE16	display_window_vertical_position_maximum;
};

/// ETSI EN 300 743 V1.3.1 section 7.2.2
struct page_composition_segment : public Subtitling_segment
{
	enum State
	{
		s_NormalCase,
		s_AcquisitionPoint,
		s_ModeChange,
		s_Reserved
	};

	UINT8		page_time_out;
	UINT8		page_version_number_and_state;

	UINT8 page_version_number() const {
			return page_version_number_and_state >> 4; }
	State page_state() const {
			return State((page_version_number_and_state >> 2) & 0x03); }

	struct region
	{
		UINT8	region_id;
		UINT8	reserved;
		UBE16	region_horizontal_address;
		UBE16	region_vertical_address;

		const region * next() const {
				return this+1; }
	};

	typedef streamScanner<region> regionScanner;
	regionScanner	regionRange() const {
								return regionScanner(this+1, next()); }
};

/// ETSI EN 300 743 V1.3.1 section 7.2.3
struct region_composition_segment : public Subtitling_segment
{
	enum Depth : UINT8
	{
		d_reserved0,
		d_2bit,
		d_4bit,
		d_8bit,
		d_reserved4,
		d_reserved5,
		d_reserved6,
		d_reserved7,
	};

	UINT8	region_id;
	UINT8	region_version_and_flag;

	UINT8	region_version_number() const {
				return region_version_and_flag >> 4; }
	bool	region_fill_flag() const {
				return (region_version_and_flag & 0x08) != 0; }

	UBE16	region_width;
	UBE16	region_height;
	UINT8	region_lev_compat_and_depth;

	Depth	region_level_of_compatibility() const {
				return Depth(region_lev_compat_and_depth >> 5); }
	Depth	region_depth() const {
				return Depth((region_lev_compat_and_depth >> 2) & 0x07); }

	UINT8	CLUT_id;
	UINT8	region_8bit_pixel_code;
	UINT8	region_4bit_2bit_pixel_code;

	UINT8	region_4bit_pixel_code() const {
				return region_4bit_2bit_pixel_code >> 4; }
	UINT8	region_2bit_pixel_code() const {
				return (region_4bit_2bit_pixel_code >> 2) & 0x03; }
	UINT8	region_pixel_code() const {
				switch(region_depth()) {
				case d_2bit: return region_2bit_pixel_code();
				case d_4bit: return region_4bit_pixel_code();
				case d_8bit: return region_8bit_pixel_code;
				} return 0;	}

	struct object
	{
		enum Type : UINT8
		{
			t_bitmap,		//!< basic_object, bitmap
			t_character,	//!< basic_object, character
			t_string,		//!< composite_object, string of characters
			t_reserved		//!< reserved
		};

		enum Flag
		{
			f_inStream,		//!< provided in the subtitling stream
			f_inRom,		//!< provided by a ROM in the IRD
			f_reserved1,	//!< reserved
			f_reserved2		//!< reserved
		};

		UBE16	object_id;
		UBE16	object_horizontal_position_and_tf;	// +type & provider flag
		UBE16	object_vertical_position_and_rsvd;

		Type	object_type() const {
					return Type(object_horizontal_position_and_tf.hi >> 6); }
		Flag	object_provider_flag() const {
					return Flag((object_horizontal_position_and_tf.hi >> 4) && 0x03); }
		UINT16	object_horizontal_position() const {
					return object_horizontal_position_and_tf & 0x0fff; }
		UINT16	object_vertical_position() const {
					return object_vertical_position_and_rsvd & 0x0fff; }

		UINT8	foreground_pixel_code;
		UINT8	background_pixel_code;

		const object *	next() const {
							Type t = object_type();
							return (t==t_character || t==t_string) ?
								this+1 : reinterpret_cast<const object *>(&foreground_pixel_code); }
	};

	PCUINT8	objects_begin() const {
				return PCUINT8(this+1); }

	PCUINT8	end() const {
					return &region_id + segment_length; }

	streamScanner<object> objectRange() const {
		return streamScanner<object>(objects_begin(), end()); }
};

/// ETSI EN 300 743 V1.3.1 section 7.2.4
struct CLUT_definition_segment : public Subtitling_segment
{
	UINT8	CLUT_id;
	UINT8	CLUT_version_number_and_rsvd;

	UINT8	CLUT_version_number() const {
				return CLUT_version_number_and_rsvd >> 4; }

	struct entry
	{
		UINT8		CLUT_entry_id;
		UINT8		flags;

		bool		_2bit_entry_CLUT_flag() const {
						return (flags & 0x80) != 0; }
		bool		_4bit_entry_CLUT_flag() const {
						return (flags & 0x40) != 0; }
		bool		_8bit_entry_CLUT_flag() const {
						return (flags & 0x20) != 0; }
		bool		full_range_flag() const {
						return (flags & 0x01) != 0; }

		struct CLUT
		{
			UINT8	Y_value;
			UINT8	Cr_value;
			UINT8	Cb_value;
			UINT8	T_value;

			PCUINT8	next() const {
						return PCUINT8(this+1); }
		};

		struct SCLUT : public UBE16
		{
			operator CLUT() const {
						CLUT sc = {UINT8(hi & 0xfc), UINT8((*this >> 2) & 0xf0), UINT8((lo << 2) & 0xf0), UINT8(lo << 6)};
						return sc; }

			PCUINT8	next() const {
						return PCUINT8(this+1); }
		};

		union {
			CLUT	c;
			SCLUT	s;
		};

		operator CLUT() const {
					return full_range_flag() ? c : s; }

		const entry *
				next() const {
					return reinterpret_cast<const entry *>(full_range_flag() ? c.next() : s.next()); }
	};

	typedef streamScanner<entry>	entryScanner;

	entryScanner	entryRange() const {
						return entryScanner(this+1, next()); }
};

/// ETSI EN 300 743 V1.3.1 section 7.2.5
struct object_data_segment : public Subtitling_segment
{
	enum Method	//!< Coding Method
	{
		m_pixels,
		m_string,
		m_reserved1,
		m_reserved2
	};

	UBE16	object_id;
	UINT8	object_version_number_and_coding_method;

	UINT8	object_version_number() const {
				return object_version_number_and_coding_method >> 4; }
	Method	object_coding_method() const {
				return Method((object_version_number_and_coding_method >> 2) & 0x03); }
	bool	non_modifying_colour_flag() const {
				return (object_version_number_and_coding_method & 0x02) != 0; }

	enum ObjectType : UINT8
	{
		t_null,
		t_2_bit_pixel_code_string	= 0x10,
		t_4_bit_pixel_code_string	= 0x11,
		t_8_bit_pixel_code_string	= 0x12,
		t_2_to_4_bit_map_table_data	= 0x20,
		t_2_to_8_bit_map_table_data	= 0x21,
		t_4_to_8_bit_map_table_data	= 0x22,
		t_end_of_object_line_code	= 0xf0
	};

	/// ETSI EN 300 743 V1.3.1 section 7.2.5.1
	//struct pixel_data_sub_block
	//{
	//	Type	data_type;

	//	PCUINT8	data() const {
	//				return PCUINT8(this+1); }
	//	const pixel_data_sub_block * next() const {
	//			__debugbreak();
	//			return this+1; }
	//};

	UNION_BEGIN {
		UBE16	top_field_data_block_length;
		UBE16	bottom_field_data_block_length;
	} UNION_ELSE {
		UINT8	number_of_codes;
	} UNION_END;

	PCUINT8	top_field_data_block_begin() const {
				return PCUINT8(this+1); }
	PCUINT8	top_field_data_block_end() const {
				return top_field_data_block_begin() + top_field_data_block_length; }
	PCUINT8	bottom_field_data_block_begin() const {
				return top_field_data_block_end(); }
	PCUINT8	bottom_field_data_block_end() const {
				return bottom_field_data_block_begin() + bottom_field_data_block_length; }

	MemRng	topFieldBlockRange() const {
		return MemRng(top_field_data_block_begin(), top_field_data_block_end()); }
	MemRng	bottomFieldBlockRange() const {
		return MemRng(bottom_field_data_block_begin(), bottom_field_data_block_end()); }
};

/****************************************************************************************
 *						PROGRAM STREAM, donnes vido partielles						*
 ****************************************************************************************/

/// Obtenir un champ de bits (max 32)  cheval sur 1  5 octets,
/// positions connues  la compilation
template <typename T, int O, int N, typename U>
	T GetNBits(const U * p)
{
	// Dsactiver warnings : "l'expression conditionnelle est une constante",
	// "compteur de dcalage ngatif ou trop important, comportement non dfini"
	// et "dcalage vers la droite trop important, perte de donnes"
	#pragma warning( disable : 4127 4293 4333 )
	#define SHFT			((sizeof(T) << 3)-(8-(O&7)))	// Dcalage du premier octet
	#define XSHFT(b, sb)	SHFT > sb ? b << (SHFT-sb) : b >> (sb-SHFT)

	T			tmp	= p[O>>3] << SHFT;

	if (O+N >= 8) {
		tmp |= XSHFT(p[1+(O>>3)], 8);
		if (O+N >= 16) {
			tmp |= XSHFT(p[2+(O>>3)], 16);
			/* //-> on se limite  24 bits pour le moment
			if (O+N >= 24) {
				tmp |= XSHFT(p[3+(O>>3)], 24);
				if (O+N >= 32)
					tmp |= XSHFT(p[4+(O>>3)], 32);
			}
			*/
		}
	}

	// (la division devrait garantir un report correct du signe si T est sign)
	return tmp / (1 << ((sizeof(T) << 3)-N));

	#undef	SHFT
	#undef	XSHFT
	#pragma warning( default : 4127 4293 4333 )
}

// ISO IEC 13818-2, 6.2.2.6
struct PS_video_GopHeader : public PS_hdr
{
	UINT8	data[4];	// fhhhhhmm mmmm1sss sssppppp p.......

	bool	isGopHeader() const {
				return isStream(sid_internal_Group); }
	bool	drop_frame_flag() const {
				return GetNBits<UINT8, 0, 1>(data) != 0; }
	UINT8	time_code_hours() const {
				return GetNBits<UINT8, 1, 5>(data); }
	UINT8	time_code_minutes() const {
				return GetNBits<UINT8, 6, 6>(data); }
	UINT8	time_code_seconds() const {
				return GetNBits<UINT8, 13, 6>(data); }
	UINT8	time_code_pictures() const {
				return GetNBits<UINT8, 19, 6>(data); }
};

// ISO IEC 13818-2, 6.2.2.1
struct PS_video_SequenceHeader : public PS_hdr
{
	UINT8	horizontal_and_vertical_size[3];
	UINT8	aspect_ratio_and_frame_rate;
	UINT8	bit_rate_and_buffer_size[4];

	bool	isSequenceHeader() const {
				return isStream(sid_internal_SequenceHeader); }
	UINT16	horizontal_size_value() const {
				return GetNBits<UINT16,  0, 12>(horizontal_and_vertical_size); }
	UINT16	vertical_size_value() const {
				return GetNBits<UINT16, 12, 12>(horizontal_and_vertical_size); }
	UINT8	aspect_ratio_information() const {
				return aspect_ratio_and_frame_rate >> 4; }
	UINT8	frame_rate_code() const {
				UINT8 fr = aspect_ratio_and_frame_rate & 0x0f;
				/// "the '8' bit is the scramble flag used by sigma designs - ignore"
				/// \see http://archive.msdn.microsoft.com/mpeg2demux, file tsctrlr.cpp, line 616
				if (fr > 0x8) fr &= 0x07;
				return fr; }
	UINT32	bit_rate_value() const {
				return GetNBits<UINT32, 0, 18>(bit_rate_and_buffer_size); }
	bool	marker_bit() const {
				return GetNBits<UINT8, 18,  1>(bit_rate_and_buffer_size) != 0; }
	UINT16	vbv_buffer_size_value() const {
				return GetNBits<UINT16, 19,  10>(bit_rate_and_buffer_size); }
	bool	constrained_parameters_flag() const {
				return GetNBits<UINT8, 29,  1>(bit_rate_and_buffer_size) != 0; }

	UINT	size() const {	// Sequence header size
				PUINT8 pf = PUINT8(this+1);
				if (pf[-1] & 0x02) pf += 64;	// load_intra_quantiser_matrix
				if (pf[-1] & 0x01) pf += 64;	// load_non_intra_quantiser_matrix
				return UINT(pf - PUINT8(this)); }
};

// ISO IEC 13818-2, 6.2.2.3
struct PS_video_SequenceExtension : public PS_hdr
{
	UINT8	start_code_and_bits;
	UINT8	misc_extensions[3];		/// Multiple
	UINT8	vbv_buffer_size_extension;
	UINT8	delay_and_frame_rate;

	UINT8	profile_and_level_indication() const {
				return ((start_code_and_bits & 0x0f) << 4) | misc_extensions[0] >> 4; }
	bool	escape_bit() const {				// ISO IEC 13818-2, table 8.1
				return (start_code_and_bits & 0x08) != 0; }
	UINT8	profile_identification() const {	// ISO IEC 13818-2, table 8.1
				return start_code_and_bits & 0x07; }
	UINT8	level_identification() const {		// ISO IEC 13818-2, table 8.1
				return GetNBits<UINT8, 0,  4>(misc_extensions); }
	bool	progressive_sequence() const {
				return GetNBits<UINT8, 4,  1>(misc_extensions) != 0; }
	UINT8	chroma_format() const {
				return GetNBits<UINT8, 5,  2>(misc_extensions); }
	UINT8	horizontal_size_extension() const {
				return GetNBits<UINT8, 7,  2>(misc_extensions); }
	UINT8	vertical_size_extension() const {
				return GetNBits<UINT8, 9,  2>(misc_extensions); }
	UINT16	bit_rate_extension() const {
				return GetNBits<UINT16, 11, 12>(misc_extensions); }
	bool	marker_bit() const {
				return GetNBits<UINT8, 23, 1>(misc_extensions) != 0; }
	bool	low_delay() const {
				return (delay_and_frame_rate & 0x80) != 0; }
	UINT8	frame_rate_extension_n() const {
				return (delay_and_frame_rate >> 5) & 0x03; }
	UINT8	frame_rate_extension_d() const {
				return delay_and_frame_rate & 0x1f; }

	bool	is_valid() const {
				return isStream(sid_internal_Extension) && (start_code_and_bits >> 4) == esc_Sequence; }
};

// ISO IEC 13818-2, 6.2.2.4
struct PS_video_SequenceDisplayExtension : public PS_hdr
{
	UINT8	start_code_and_bits;

	UINT8	video_format() const {
				return (start_code_and_bits >> 1) & 0x07; }
	bool	colour_description() const {
				return (start_code_and_bits & 0x01) != 0; }

	PUINT8	display_sizes_ptr() const {
				return PUINT8(this+1) + (colour_description() ? 3 : 0); }

	UINT16	display_horizontal_size() const {
				return GetNBits<UINT16, 0, 14>(display_sizes_ptr()); }
	UINT16	display_vertical_size() const {
				return GetNBits<UINT16, 15, 14>(display_sizes_ptr()); }

	bool	is_valid() const {
				return isStream(sid_internal_Extension) && (start_code_and_bits >> 4) == esc_SequenceDisplay; }
};

// ISO IEC 13818-2, 6.2.3
struct PS_video_PictureHeader : public PS_hdr
{
	UINT8	data[2];

	bool	isPictureHeader() const {
				return isStream(sid_internal_Picture); }
	UINT16	temporal_reference() const {
				return GetNBits<UINT16,  0, 10>(data); }
	UINT8	picture_coding_type() const {
				return GetNBits<UINT8,	10,  3>(data); }
	bool	is_keyframe() const {
				return picture_coding_type() == 1; }
};

/****************************************************************************************
 *											RTP											*
 ****************************************************************************************/

/** Encapsulation RTP des flux TS rencontre sur certains flux de diffusion
 * \see http://en.wikipedia.org/wiki/Real-time_Transport_Protocol
 * \see http://rfc-ref.org/RFC-TEXTS/3550/chapter4.html
 * \test exprimental
 **/

struct RTP_hdr
{
	UINT8	flags;
	UINT8	payload_type_data;
	UBE16	sequence_number;
	UBE32	timestamp;
	UBE32	synchronization_source;

	UINT8	version() const {
				return UINT8(flags >> 6); }
	bool	padding() const {
				return (flags & 0x20) != 0; }
	bool	extension() const {
				return (flags & 0x10)!= 0; }
	UINT8	CSRC_count() const {
				return UINT8(flags & 0x0f); }
	bool	marker() const {
				return (payload_type_data & 0x80)!= 0; }
	UINT8	payload_type() const {
				return UINT8(payload_type_data & 0x7f); }

	bool	isRTP() const {
				return (flags & 0xc0) == 0x80; }

	PUINT8	next() const {
				return PUINT8(this+1) + CSRC_count() * sizeof(UBE32); }
};

/****************************************************************************************
 *									TRANSPORT STREAM									*
 ****************************************************************************************/

#define TS_SIZE 188			// Taille (fixe) des paquets TS

// Codes de PID prffinis :
enum Predefined_PID {
	p_pid_PAT	= 0x000,	// Program Association Table
	p_pid_CAT	= 0x001,	// Conditional Access Table
	p_pid_TSDT	= 0x002,	// Transport Stream Description Table
	// (codes 0x003  0x00f spcifis comme "rservs")
	p_pid_NIT	= 0x010,	// Network Identification Table
	p_pid_BAT	= 0x011,	// Bouquet Association Table ou ...
	p_pid_SDT	= 0x011,	// Service Description Table
	p_pid_EIT	= 0x012,	// Event Information Table (guide des programmes ...)
	p_pid_RST	= 0x013,	// Running Status Tection
	p_pid_TDT	= 0x014,	// Time and Date Table ou ...
	p_pid_TOT	= 0x014,	// Time Offset Table
	p_pid_SFN	= 0x015,	// SFN/MIP synchronisation
	p_pid_DIT	= 0x01e,	// Discontinuity Information Table
	p_pid_SIT	= 0x01f,	// Selection Information Table
	//
	p_pid_NULL	= 0x1fff	// Null packet
};

/*
	Description Transport stream packet
	**** ISO IEC 13818-1 section 2.4.3.2 table 2.2
		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	transport_packet() {
		sync_byte                                            8          0
		transport_error_indicator                            1          1
		payload_unit_start_indicator                         1          1
		transport_priority                                   1          1
		PID                                                 13          1, 2
		transport_scrambling_control                         2          3
		adaptation_field_control                             2          3
		continuity_counter                                   4          3
		if(adaptation_field_control=='10'||adaptation_field_control=='11') {
			adaptation_field()                                          4
		}
		if (adaptation_field_control=='01'||adaptation_field_control=='11') {
			for (i=0;i<N;i++) {
				data_byte                                    8          x..187
			}
		}
	}
*/

enum ScramblingCtlValues
{
	scv_not_scrambled,
	scv_user_defined_1,
	scv_user_defined_2,
	scv_user_defined_3
};

enum AdaptationFieldCtlValues
{
	afcv_reserved,
	afcv_payload_only,
	afcv_afield_only,
	afcv_afield_and_payload
};

struct TS_hdr
{
	UINT8	sync_byte;
	PID13	pid;
	UINT8	control_and_ccounter;

	bool	transport_error_indicator() const {
				return (pid.flags() & 0x80)!=0; }

	/// Dterminer si le champ "payload" contient une structure qui commence ici
	bool	payload_unit_start_indicator() const {
				return (pid.flags() & 0x40)!=0; }

	bool	transport_priority() const {
				return (pid.flags() & 0x20)!=0; }

	ScramblingCtlValues transport_scrambling_control() const {
				return ScramblingCtlValues((control_and_ccounter & 0xc0)>>6); }

	/// Dterminer quelles zones existent dans le paquet (parmi "adaptation field" et "payload")
	AdaptationFieldCtlValues adaptation_field_control() const {
				return AdaptationFieldCtlValues((control_and_ccounter & 0x30)>>4); }

	bool	has_af() const {
				return adaptation_field_control() >= afcv_afield_only; }

	/// Obtenir un pointeur sur le champ "adaptation field" (une fois qu'on a vrifi
	/// que celui-ci existe avec "has_af").
	const struct TS_AF_hdr & get_af() const {
				return *reinterpret_cast<const TS_AF_hdr *>(beg_of_fields()); }

	bool	has_payload() const {
				AdaptationFieldCtlValues afv = adaptation_field_control();
				return afv == afcv_payload_only || afv == afcv_afield_and_payload; }

	/// Obtenir le compteur de continuit
	UINT8	continuity_counter() const {
				return control_and_ccounter & 0x0f; }

	PUINT8	beg_of_fields() const {
				return PUINT8(this+1); }

	/// Retourne un pointeur sur le dbut de l'espace "payload" s'il existe, ou bien
	/// \p NULL s'il n'existe pas.
	PUINT8	get_payload() const;
};

/*
	Adaptation Field
	**** ISO IEC 13818-1 section 2.4.3.5 table 2.6
		Syntax                                          No. of bits   Offset
	--------------------------------------------------------------------------
	adaptation_field() {
		adaptation_field_length                              8          0
		if (adaptation_field_length>0) {
			discontinuity_indicator                          1          1
			random_access_indicator                          1          1
			elementary_stream_priority_indicator             1          1
			PCR_flag                                         1          1
			OPCR_flag                                        1          1
			splicing_point_flag                              1          1
			transport_private_data_flag                      1          1
			adaptation_field_extension_flag                  1          1
			if (PCR_flag=='1') {
				program_clock_reference_base                33          2..6
				reserved                                     6          6
				program_clock_reference_extension            9          6,7
			}
			if (OPCR_flag=='1') {
				original_program_clock_reference_base       33          +0..+4
				reserved                                     6          +4
				original_program_clock_reference_extension   9          +4,+5
			}
			if (splicing_point_flag=='1') {
				splice_countdown                             8          +0
			}
			if (transport_private_data_flag=='1') {
				transport_private_data_length                8          +0
				for (i=0; i<transport_private_data_length; i++) {
					private_data_byte                        8          +0
				}
			}
			if (adaptation_field_extension_flag=='1') {
				adaptation_field_extension_length            8          +0
				ltw_flag                                     1          +1
				piecewise_rate_flag                          1          +1
				seamless_splice_flag                         1          +1
				reserved                                     5          +1
				if (ltw_flag=='1') {
					ltw_valid_flag                           1          +2
					ltw_offset                              15          +2,+3
				}
				if (piecewise_rate_flag=='1') {
					reserved                                 2          +0
					piecewise_rate                          22          +0..+2
				}
				if (seamless_splice_flag=='1') {
					splice_type                              4          +0
					DTS_next_AU[32..30]                      3          +0
					marker_bit                               1          +0
					DTS_next_AU[29..15]                     15          +1,+2
					marker_bit                               1          +2
					DTS_next_AU[14..0]                      15          +3,+4
					marker_bit                               1          +4
				}
				for (i=0; i<N; i++) {
					reserved                                 8          +0
				}
			}
			for (i=0; i<N; i++) {
				stuffing_byte                                8          +0
			}
		}
	}
*/

// Dfinition des flags de "TS_AF_hdr" :
#define AF_DISCONTINUITY_INDICATOR				0x80
#define AF_RANDOM_ACCESS_INDICATOR				0x40
#define AF_ELEMENTARY_STREAM_PRIORITY_INDICATOR	0x20
#define	AF_PCR_FLAG								0x10
#define	AF_OPCR_FLAG							0x08
#define	AF_SPLICEPT_FLAG						0x04
#define	AF_PRIV_DATA_FLAG						0x02
#define	AF_EXTENSION_FLAG						0x01

// Dfinition des flags de "TS_AF_extension" :
#define	AF_LTW_FLAG								0x80
#define	AF_PW_RATE_FLAG							0x40
#define	AF_SLSPLICE_FLAG						0x20

/// Accs aux valeurs d'un champ "adaptation field" (incomplet)
struct TS_AF_hdr
{
	UINT8	adaptation_field_length;
	UINT8	flags;

	/// Obtenir un pointeur sur la partie variable du champ
	PUINT8	beg_of_fields() const {
		return adaptation_field_length>0 ? PUINT8(this+1) : PUINT8(&flags); }

	/// Obtenir les flags, s'ils existent
	UINT8	get_flags() const {
		return adaptation_field_length==0 ? 0 : flags; }

	/// Tester si une valeur PCR existe
	bool has_pcr() const {
		return (get_flags() & AF_PCR_FLAG)!=0; }

	/// Obtenir la valeur du PCR (test pralable ncessaire)
	const TS_PCR & get_pcr() const {
		return *reinterpret_cast<const TS_PCR *>(beg_of_fields()); }

	/// Obtenir l'adresse du champ suivant (normalement "payload", s'il existe)
	PUINT8	next() const {
		return PUINT8(&flags)+adaptation_field_length; }
};

/// Structure de dballage d'un paquet TS
struct TS_unwrapper
{
	const TS_hdr &	hdr;
	PUINT8			p_next;			//!< Pointeur sur la fin du bloc "adaptation field"
	TS_AF_hdr *		p_afield;		//!< Pointeur sur le champ "adaptation field", ou NULL s'il n'y en a pas
	PUINT8			p_payload;		//!< Pointeur sur 1er octet "payload", ou NULL s'il n'y en a pas
	UINT8			payload_size;	//!< Taille du bloc "payload"

	TS_unwrapper(const TS_hdr & h) :
		hdr(h),
		p_afield(NULL),
		p_payload(NULL),
		payload_size(0)
	{
		PUINT8						p_nxt = h.beg_of_fields();

		if (h.has_af())
			get_and_skip(p_nxt, p_afield);	// Un champ "adaptation field" existe

		if (h.has_payload()) {
			// Un champ "payload" existe
			p_payload = p_nxt;
			payload_size = UINT8((PUINT8(&h)+TS_SIZE)-p_payload);
		}
		p_next = p_nxt;
	}
};

struct TS_AF_unwrapper_base
{
	PUINT8		p_next;
	TS_PCR *	p_pcr;
	TS_PCR *	p_opcr;

	bool parse(const TS_AF_hdr * paf);

	TS_AF_unwrapper_base() :
		p_next(NULL),
		p_pcr(NULL),
		p_opcr(NULL)
	{
	}

	TS_AF_unwrapper_base(const TS_AF_hdr * paf) {
		parse(paf); }

	TS_AF_unwrapper_base(const TS_AF_hdr & af) {
		parse(&af); }
};

struct TS_AF_PrivData
{
	UINT8	length;

	PUINT8	next() const {
		return PUINT8(this+1) + length; }
};

struct TS_AF_extension
{
	UINT8	length;
	UINT8	flags;

	PUINT8	next() const {
		return PUINT8(this+1); }
};

struct TS_AF_pw_rate
{
	UINT8	piecewise_rate[3];

	PUINT8	next() const {
		return PUINT8(this+1); }
};

struct TS_AF_SlSplice
{
	UINT8	seamless_splice[5];

	PUINT8	next() const {
		return PUINT8(this+1); }
};

struct TS_AF_unwrapper_full : public TS_AF_unwrapper_base
{
	UINT8uw *			p_splicing_point;
	TS_AF_PrivData *	p_priv_data;
	//
	TS_AF_extension *	p_ext;		// pointeur dbut champ d'extension optionnel
	UBE<15> *			p_ltw;
	TS_AF_pw_rate *		p_piecewise_rate;
	TS_AF_SlSplice *	p_seamless_splice;

	bool parse(const TS_AF_hdr * paf);

	TS_AF_unwrapper_full() :
		p_splicing_point(NULL),
		p_priv_data(NULL),
		p_ext(NULL),
		p_ltw(NULL),
		p_piecewise_rate(NULL),
		p_seamless_splice(NULL)
	{
	}

	TS_AF_unwrapper_full(const TS_AF_hdr * paf) {
		parse(paf); }
	TS_AF_unwrapper_full(const TS_AF_hdr & af) {
		parse(&af); }
};

#define TS_SYNC 0x47	// valeur de l'octet de synchronisation

struct TS_packet
{
	union {
		TS_hdr	hdr;
		UINT8	bytes[TS_SIZE];
	};

	PCUINT8				end() const {
							return bytes + _countof(bytes); }
	const TS_packet *	next() const {
							return this+1; }

	bool				isTS(bool single=false) const {
							return hdr.sync_byte == TS_SYNC && (single || next()->isTS(true)); }
};

/****************************************************************************************
 *									SERVICE INFORMATION									*
 ****************************************************************************************/

/*
*** ATTENTION  la remarque suivante, extraite de la documentation DirectShow de Microsoft,
*** concernant la mthode ISectionList::GetSectionData.
*** (ceci peut compliquer l'accs aux donnes de certaines sections via les structures
*** dcrites dans le prsent fichier) :

The section header is converted from network byte order to native byte order.
The number of header bytes that are converted depends on the header type.
The header types are short header (SECTION structure), long header (LONG_SECTION structure),
or DSM-CC header (DSMCC_SECTION structure).
If the section has a short header, the first three bytes are converted;
for a long header, the first eight bytes are converted;
and for a DSM-CC header, the first 20 bytes are converted.

The body of the section data, after the header, is left unparsed and unconverted.

http://msdn2.microsoft.com/en-us/library/ms786743(VS.85).aspx

*** Les structures suivantes ont pour but de pallier cette complication.
*/

struct LEN12M : public LEN12
{
	UINT16	operator () (bool microsoft) const {
		if (microsoft)
			return PMPEG_HEADER_BITS(this)->SectionLength;
		return LEN12::operator ()();
	}

	bool syntax_indicator(bool microsoft=false) const {
		if (microsoft)
			return PMPEG_HEADER_BITS(this)->SectionSyntaxIndicator != 0;
		return (flags() & 0x80)!=0;
	}
};

struct UBE16M : public UBE16
{
	UINT16	operator () (bool microsoft) const {
		if (microsoft)
			return *PUINT16(this);
		return UBE16::operator ()();
	}
};

// ############################################################################

// Codes des identifiants de tables :
enum SI_TableID : UINT8 {
	si_undef	= 0xff,
	si_PAT		= 0x00,
	si_CAT		= 0x01,
	si_PMT		= 0x02,
	si_DSMCC_a	= 0x3a,	// http://en.wikipedia.org/wiki/DSM-CC
	si_DSMCC_b	= 0x3b,
	si_DSMCC_c	= 0x3c,
	si_DSMCC_d	= 0x3d,
	si_DSMCC_e	= 0x3e,
	si_DSMCC_f	= 0x3f,
	si_NIT		= 0x40,
	si_SDT		= 0x42,
	si_EIT_act	= 0x4e,
	si_EIT_oth	= 0x4f,
	si_EIT_as0, si_EIT_as1, si_EIT_as2, si_EIT_as3, si_EIT_as4, si_EIT_as5, si_EIT_as6, si_EIT_as7,
	si_EIT_as8, si_EIT_as9, si_EIT_asa, si_EIT_asb, si_EIT_asc, si_EIT_asd, si_EIT_ase, si_EIT_asf,
	si_EIT_os0, si_EIT_os1, si_EIT_os2, si_EIT_os3, si_EIT_os4, si_EIT_os5, si_EIT_os6, si_EIT_os7,
	si_EIT_os8, si_EIT_os9, si_EIT_osa, si_EIT_osb, si_EIT_osc, si_EIT_osd, si_EIT_ose, si_EIT_osf
};

// Pour insertion dans un switch/case :
#define si_EIT_act_sched si_EIT_as0: \
	case si_EIT_as1: \
	case si_EIT_as2: \
	case si_EIT_as3: \
	case si_EIT_as4: \
	case si_EIT_as5: \
	case si_EIT_as6: \
	case si_EIT_as7: \
	case si_EIT_as8: \
	case si_EIT_as9: \
	case si_EIT_asa: \
	case si_EIT_asb: \
	case si_EIT_asc: \
	case si_EIT_asd: \
	case si_EIT_ase: \
	case si_EIT_asf

#define si_EIT_oth_sched si_EIT_os0: \
	case si_EIT_os1: \
	case si_EIT_os2: \
	case si_EIT_os3: \
	case si_EIT_os4: \
	case si_EIT_os5: \
	case si_EIT_os6: \
	case si_EIT_os7: \
	case si_EIT_os8: \
	case si_EIT_os9: \
	case si_EIT_osa: \
	case si_EIT_osb: \
	case si_EIT_osc: \
	case si_EIT_osd: \
	case si_EIT_ose: \
	case si_EIT_osf

/*
	System Information Header (SI)
	**** ISO IEC 13818-1 section 2.4.4
		Syntax                      No. of bits   Offset
	------------------------------------------------------
	system_information_header() {
		table_id                         8          0
		section_syntax_indicator         1          1
		reserved_future_use              1          1
		reserved                         2          1
		section_length                  12          1,2
		section_id                      16          3,4     -> variable selon type de table
		reserved                         2          5
		version_number                   5          5
		current_next_indicator           1          5
		section_number                   8          6
		last_section_number              8          7
		[...]
		[...]
		[...]
		CRC_32                          32
	}
*/

/// Partie commune  toutes les tables PSI
struct SI_hdr
{
	UINT8	_table_id;
	LEN12M	section_length;
	UBE16M	section_id;				// usage variable selon type de section
	UINT8	versnum_and_next_indic;
	UINT8	section_number;
	UINT8	last_section_number;
	//
	// Certaines des fonctions suivantes ont un paramtre 'microsoft' qui doit tre mis  'true'
	// lorsqu'on analyse une table PSI qui a t reue par le biais de certaines fonctions Microsoft.
	// En effet, ces fonctions ont pour caractristique de permuter **certains** octets du champ PSI
	// avec pour ide de les remettre dans l'ordre natif du systme d'exploitation ... ...
	// Le mme paramtre doit aussi tre donn  toutes les fonctions qui font indirectement usage
	// de l'une de ces trois fonctions.
	//
	SI_TableID table_id() const {
		return static_cast<SI_TableID>(_table_id); }
	bool section_syntax_indicator(bool microsoft=false) const {
		return section_length.syntax_indicator(microsoft); }
	UINT8 version_number() const {
		return (versnum_and_next_indic & 0x3e)>>1; }
	bool current_next_indicator() const {
		return (versnum_and_next_indic & 0x01)!=0; }

	PUINT8 begin() const {
		return PUINT8(this+1); }
	PUINT8 next(bool microsoft=false) const {
		return PUINT8(&section_id) + section_length(microsoft); }
	size_t size(bool microsoft=false) const {
		return next(microsoft)-PUINT8(this); }
	PUINT8 end(bool microsoft=false) const {
		return next(microsoft) - sizeof(CRC_32); }
	const CRC_32 & crc32(bool microsoft=false) const {
		return *reinterpret_cast<CRC_32 *>(end(microsoft)); }
	template <class T> const T & as() const {
		return *reinterpret_cast<const T *>(this); }
	bool check_crc32(bool microsoft=false) const {
		return ::crc32(PUINT8(this), int(size(microsoft)))==0; }

	bool is_valid(bool microsoft=false) const {
			UINT16 lng = section_length(microsoft);

			return lng >= 5 && lng <= 0x3fd &&
				last_section_number >= section_number && section_syntax_indicator(microsoft); }
};

/** \brief Structure dcrivant un pointeur sur table SI
 * \see ISO IEC 13818-1 section 2.4.4.1, table 2-24
 * <pre>
 *    Program specific information pointer
 *
 *      Syntax                        No. of bits   Offset
 *    ------------------------------------------------------
 *    program_specific information_pointer() {
 *        pointer_field                    8          0
 *    }
 * </pre>
 *
 * Cette structure prcde immdiatement toutes les tables SI.
 * Au niveau du "Transport Stream", elle se trouve exactement au dbut de
 * la zone "payload", lorsque celle-ci contient une table SI.
 * Comme l'octet qu'elle contient indique le dcalage de la table dans la
 * zone "payload" et que ce dcalage est le presque toujours nul, sa prsence
 * se rsume le plus souvent  intercaler un octet 0 juste devant la table.
 **/
struct SI_ptr
{
	UINT8 pointer_field;

	/// Obtenir une rfrence sur une table PSI  partir du dbut de la zone de donnes PSI
	const SI_hdr * getSI() const {
		return reinterpret_cast<const SI_hdr *>(&this[1]+pointer_field); }

	/// Fonction de "cast" ddie, normalement vers table PSI
	template <class T> const T & as() const {
		return getSI()->as<T>(); }

	/// "Cast" spcifique vers en-ttre de table PSI
	const SI_hdr & asSI() const {
		return as<SI_hdr>(); }
};

/*
// Obtenir une rfrence sur une table PSI  partir du dbut de la zone de donnes PSI :
template <class T> const T & asSI(LPCVOID ptr)
{
	return reinterpret_cast<const SI_ptr *>(ptr)->getSI()->as<T>();
}
*/

// ############################################################################

/*
	Program association section (PAT)
	**** ISO IEC 13818-1 section 2.4.4.4, table 2-25
		Syntax                      No. of bits   Offset
	------------------------------------------------------
	program_association_section() {                           _
		table_id                         8          0          \
		section_syntax_indicator         1          1           \
		reserved_future_use              1          1           |
		reserved                         2          1           |
		section_length                  12          1,2         |
		transport_stream_id             16          3,4         |-- en-tte SI
		reserved                         2          5           |
		version_number                   5          5           |
		current_next_indicator           1          5           |
		section_number                   8          6           /
		last_section_number              8          7         _/
		for (j=0;j< N;j++) {                        8
			program_number              16          +0,+1
			reserved                     3          +2
			if (program_number==0 ) {
				network_PID             13          +2,+3
			} else {
				program_map_PID         13          +2,+3
			}
		}
		CRC_32                          32
	}
*/

struct SI_PAT : public SI_hdr
{
	struct entry
	{
		UBE16	program_number;		// devrait peut-tre tre renomm en 'service_id' ??
		PID13	pid;
		//
		const entry * next() const {
			return this+1; }
	};

	typedef streamScanner<entry>	entryScanner;

	UINT16 tsid(bool microsoft=false) const {
		return section_id(microsoft); }
	int count(bool microsoft=false) const {
		return int((PUINT8(&crc32(microsoft))-begin())/sizeof(entry)); }
	const entry & operator [] (int index) const {
		return reinterpret_cast<const entry *>(begin())[index]; }
	entryScanner entryRange(bool microsoft=false) const {
		return entryScanner(begin(), end(microsoft)); }

	bool is_valid(bool microsoft=false) const {
			return __super::is_valid(microsoft) && (section_length(microsoft)-4-5)%4 == 0 && table_id() == si_PAT; }
};

// ############################################################################

/*
	Transport Stream Conditional Access section (CAT)
	**** ISO IEC 13818-1 section 2.4.4.6, table 2-27
		Syntax                      No. of bits   Offset
	------------------------------------------------------
	CA_section() {                                            _
		table_id                         8          0          \
		section_syntax_indicator         1          1           \
		'0'                              1          1           |
		reserved                         2          1           |
		section_length                  12          1,2         |
		reserved1                       16          3,4         |-- en-tte SI
		reserved2                        2          5           |
		version_number                   5          5           |
		current_next_indicator           1          5           |
		section_number                   8          6           /
		last_section_number              8          6         _/
		for (i = 0; i < N; i++) {
			descriptor()
		}
		CRC_32                          32 rpchof
	}
*/

/// Reprsentation table PSI : CAT = Conditional Access Table
struct SI_CAT : public SI_hdr
{
	MDESCR::D_scanner descrRange(bool microsoft=false) const {
		return MDESCR::D_scanner(this+1, end(microsoft)); }
};

// ############################################################################

/*
	Transport Stream Program Map section (PMT)
	**** ISO IEC 13818-1 section 2.4.4.8, table 2-28
		Syntax                      No. of bits   Offset
	------------------------------------------------------
	TS_program_map_section( ) {                               _
		table_id                         8          0          \
		section_syntax_indicator         1          1           \
		reserved_future_use              1          1           |
		reserved                         2          1           |
		section_length                  12          1,2         |
		program_number                  16          3,4         |-- en-tte SI
		reserved                         2          5           |
		version_number                   5          5           |
		current_next_indicator           1          5           |
		section_number                   8          6           /
		last_section number              8          7         _/
		reserved                         3          8
		PCR_PID                         13          8,9
		reserved                         4          10
		program_info_length             12          10,11
		for (i=0;i<N;i++) {                         12
			descriptor()
		}
		for (i=0;i<N1;i++) {
			stream_type                  8          +0
			reserved                     3          +1
			elementary_PID              13          +1,+2
			reserved                     4          +3
			ES_info_length              12          +3,+4
			for (j=0;j<N2;j++) {                    +5
				descriptor()
			}
		}
		CRC_32                          32
	}
*/

// PMT = Program Map Table
struct SI_PMT : public SI_hdr
{
	struct entry
	{
		StreamType	stream_type;
		PID13		elementary_PID;
		LEN12		ES_info_length;

		MDESCR::D_scanner descrRange() const {
			return MDESCR::D_scanner(this+1, PUINT8(this+1) + ES_info_length()); }
		const entry * next() const {
			return reinterpret_cast<const entry *>(PUINT8(this+1) + ES_info_length()); }
	};

	typedef streamScanner<entry>	entryScanner;

	PID13	PCR_PID;
	LEN12	program_info_length;

	UINT16 program_number(bool microsoft=false) const {
		return section_id(microsoft); }
	MDESCR::D_scanner descrRange() const {
		return MDESCR::D_scanner(this+1, PUINT8(this+1) + program_info_length()); }
	entryScanner entryRange(bool microsoft=false) const {
		return entryScanner(PUINT8(this+1) + program_info_length(), end(microsoft)); }

	bool is_valid(bool microsoft=false) const {
			return __super::is_valid(microsoft) &&
				section_length(microsoft) >= program_info_length()+9+4 && table_id()==si_PMT; }
};

// ############################################################################

/*
	Service description section (SDT)
	**** ETSI EN 300 468 v1.11.1 section 5.2.3, table 5
		Syntax                      No. of bits   Offset
	------------------------------------------------------
	service_description_section() {                           _
		table_id                         8          0          \
		section_syntax_indicator         1          1           \
		reserved_future_use              1          1           |
		reserved                         2          1           |
		section_length                  12          1,2         |
		transport_stream_id             16          3,4         |-- en-tte SI
		reserved                         2          5           |
		version_number                   5          5           |
		current_next_indicator           1          5           |
		section_number                   8          6           /
		last_section_number              8          7         _/
		original_network_id             16          8,9
		reserved_future_use              8          10
		for(i=0;i<N;i++) {                          11
			service_id                  16          +0,+1
			reserved_future_use          6          +2
			EIT_schedule_flag            1          +2
			EIT_present_following_flag   1          +2
			running_status               3          +3
			free_CA_mode                 1          +3
			descriptors_loop_length     12          +3,+4
			for(j=0;j<N;j++) {                      +5
				descriptor()
			}
		}
		CRC_32                          32
	}
*/

// SDT = Service Description Table
struct SI_SDT : public SI_hdr
{
	struct entry
	{
		UBE16	service_id;
		UINT8	eit_flags;
		LEN12	dloop_length;
		//
		// Mthodes d'extraction :
		MDESCR::D_scanner descrRange() const {
			return MDESCR::D_scanner(this+1, PUINT8(this+1) + dloop_length()); }
		const entry * next() const {
			return reinterpret_cast<const entry *>(PUINT8(this+1) + dloop_length()); }
	};

	typedef streamScanner<entry>	entryScanner;

	UBE16	onid;
	UINT8	reserved;
	//
	// Mthodes d'extraction :
	UINT16 tsid(bool microsoft=false) const {
		return section_id(microsoft); }
	entryScanner entryRange(bool microsoft=false) const {
		return entryScanner(this+1, end(microsoft)); }
};

// ############################################################################

/*
	Network Information section (NIT)
	**** ETSI EN 300 468 v1.11.1 section 5.2.1, table 3
		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	network_information_section() {                               _
		table_id                             8          0          \
		section_syntax_indicator             1          1           \
		reserved_future_use                  1          1           |
		reserved                             2          1           |
		section_length                      12          1,2         |
		network_id                          16          3,4         |-- en-tte SI
		reserved                             2          5           |
		version_number                       5          5           |
		current_next_indicator               1          5           |
		section_number                       8          6           /
		last_section number                  8          7         _/
		reserved_future_use                  4          8
		network_descriptors_length          12          8,9
		for(i=0;i<N;i++) {                              10
			descriptor()
		}
		reserved_future_use                  4          +0
		transport_stream_loop_length        12          +0,+1
		for(i=0;i<N;i++) {                              +2
			transport_stream_id             16          +0,+1
			original_network_id             16          +2,+3
			reserved_future_use              4          +4
			transport_descriptors_length    12          +4,+5
			for(j=0;j<N;j++) {                          +6
				descriptor()
			}
		}
		CRC_32                              32
	}
*/

/// Reprsentation table PSI : NIT = Network Information Table
struct SI_NIT : public SI_hdr
{
	LEN12	network_descriptors_length;

	/// Reprsentation d'une entre de la table SI_NIT
	struct entry
	{
		UBE16	transport_stream_id;
		UBE16	original_network_id;
		LEN12	transport_descriptors_length;
		//
		MDESCR::D_scanner descrRange() const {
			return MDESCR::D_scanner(this+1, PUINT8(this+1) + transport_descriptors_length()); }
		const entry * next() const {
			return reinterpret_cast<const entry *>(PUINT8(this+1) + transport_descriptors_length()); }
	};

	/// En-tte de la seconde partie d'une table SI_NIT
	struct entries
	{
		LEN12	transport_stream_loop_length;

		typedef streamScanner<entry> entryScanner;

		entryScanner entryRange() const {
			return entryScanner(this+1, PUINT8(this+1) + transport_stream_loop_length()); }
	};

	UINT16 network_id(bool microsoft=false) const {
		return section_id(microsoft); }
	MDESCR::D_scanner descrRange() const {
		return MDESCR::D_scanner(this+1, PUINT8(this+1) + network_descriptors_length()); }
	const entries & get_entries() const {
		return *reinterpret_cast<const entries *>(PUINT8(this+1) + network_descriptors_length()); }
};

// ############################################################################

/*
	Event Information Section (EIT)
	**** ETSI EN 300 468 v1.11.1 section 5.2.4, table 7
		Syntax                          No. of bits   Offset
	----------------------------------------------------------
	event_information_section( ) {                                _
		table_id                             8          0          \
		section_syntax_indicator             1          1           \
		reserved_future_use                  1          1           |
		reserved                             2          1           |
		section_length                      12          1,2         |
		service_id                          16          3,4         |-- en-tte SI
		reserved                             2          5           |
		version_number                       5          5           |
		current_next_indicator               1          5           |
		section_number                       8          6           /
		last_section_number                  8          7         _/
		transport_stream_id                 16          8,9
		original_network_id                 16          10,11
		segment_last_section_number          8          12
		last_table_id                        8          13
		for(i=0;i<N;i++) {                              14
			event_id                        16          +0,+1
			start_time                      40          +2..+6
			duration                        24          +7..+9
			running_status                   3          +10
			free_CA_mode                     1          +10
			descriptors_loop_length         12          +10,+11
			for (j=0;j<N;j++) {                         +12
				descriptor ()
			}
		}
		CRC_32                              32
	}
*/

/// Dfinition d'un octet BCD (Binary-coded decimal)
struct BCD8
{
	BYTE b;

	BYTE operator() () const {
		return (b>>4)*10 + (b & 0x0f); }

	BCD8 operator = (unsigned n) {
		b = BYTE(((n / 10) << 4) | n % 10); return *this; }

	bool operator == (BCD8 v2) const {
		return b == v2.b; }
	bool operator != (BCD8 v2) const {
		return b != v2.b; }
	bool operator < (BCD8 v2) const {
		return b != v2.b; }
	bool operator > (BCD8 v2) const {
		return b != v2.b; }
};

/// Reprsentation de l'heure sous forme de trois octets de type BCD
struct TIM24
{
	BCD8	h;
	BCD8	m;
	BCD8	s;

	operator UINT32() const {
		return ((h()*60+m())*60+s())*1000; }	// Rsultat en millisecondes

	bool operator == (TIM24 t2) const {
		return h == t2.h && m == t2.m && s == t2.s; }
	bool operator != (TIM24 t2) const {
		return h != t2.h || m != t2.m || s == t2.s; }
	bool operator < (TIM24 t2) const {
		return h != t2.h ?  h < t2.h : (m != t2.m ?  m < t2.m : s < t2.s); }
	bool operator > (TIM24 t2) const {
		return h != t2.h ?  h > t2.h : (m != t2.m ?  m > t2.m : s > t2.s); }
};

/// Reprsentation d'une date compacte sur 16 bits "big endian"
struct MJD16 : public UBE16
{
	void set(WORD year, WORD month, WORD day)
	{
		INT32	y	= year - 1900;
		INT32	l	= month <= 2 ? 1 : 0;
		INT32	mjd	= 14956 + day + INT32((y - l) * 365.25) + INT32((month + 1 + l * 12) * 30.6001);

		__super::operator = ((UINT16)mjd);
	}

	void get(WORD & year, WORD & month, WORD & day) const
	{ // algo: ETSI EN 300 468 - ANNEX C
		INT32	mjd	= (*this)();
		INT32	y	= INT32((mjd  - 15078.2) / 365.25);
		INT32	yd	= INT32(y * 365.25);
		INT32	m	= INT32((mjd - 14956.1 - yd) / 30.6001);
		INT32	k	= (m == 14 || m == 15) ? 1 : 0;

		day		= WORD(mjd - 14956 - yd - INT32(m * 30.6001));
		month	= WORD(m - 1 - k*12);
		year	= WORD(y + k + 1900);
	}
};

/// Reprsentation table PSI : EIT = Event Information Table
struct SI_EIT : public SI_hdr
{
	/// Identification du sous-type d'une table SI_EIT
	enum subtype {
		ses_unknown,
		ses_Actual,
		ses_Other,
		ses_Actual_Schedule,
		ses_Other_Schedule
	};

	/// Reprsentation heure et date dans la table SI_EIT
	struct e_time
	{
		MJD16	day;
		TIM24	time;

		UINT64 operator () () const {
			return UINT64(day)*24*60*60*1000 + time; }	// Conversion en millisecondes

		operator SYSTEMTIME () const
		{
			SYSTEMTIME st;

			day.get(st.wYear, st.wMonth, st.wDay);
			st.wDayOfWeek		= 0;
			st.wHour			= time.h();
			st.wMinute			= time.m();
			st.wSecond			= time.s();
			st.wMilliseconds	= 0;
			return st;
		}

		e_time() {
			memset(this, 0, sizeof(*this)); }

		e_time(const SYSTEMTIME & st)
		{
			day.set(st.wYear, st.wMonth, st.wDay);
			time.h = st.wHour;
			time.m = st.wMinute;
			time.s = st.wSecond;
		}

		bool operator == (const e_time t2) const {
			return day == t2.day && time == t2.time; }
		bool operator != (const e_time t2) const {
			return day != t2.day || time != t2.time; }
		bool operator < (const e_time t2) const {
			return day != t2.day ?  day < t2.day : time < t2.time; }
		bool operator > (const e_time t2) const {
			return day != t2.day ?  day > t2.day : time > t2.time; }
	};

	/// Reprsentation d'un vnement dans la table SI_EIT
	struct entry
	{
		UBE16	event_id;
		e_time	start_time;
		TIM24	duration;
		LEN12	dloop_length;

		UINT8 running_status() const {
			return UINT8(dloop_length.flags()>>5); }
		UINT8 free_CA_mode() const {
			return UINT8((dloop_length.flags()>>4)&0x01); }
		MDESCR::D_scanner descrRange() const {
			return MDESCR::D_scanner(this+1, PUINT8(this+1) + dloop_length()); }
		const entry * next() const {
			return reinterpret_cast<const entry *>(PUINT8(this+1) + dloop_length()); }
	};

	typedef streamScanner<entry>	entryScanner;

	UBE16 tsid;
	UBE16 onid;
	UINT8 segment_last_section_number;
	UINT8 last_table_id;

	UINT16 service_id(bool microsoft=false) const {
		return section_id(microsoft); }
	entryScanner entryRange(bool microsoft=false) const {
		return entryScanner(this+1, end(microsoft)); }

	subtype getSubtype() const
	{
		switch (table_id()) {
		case si_EIT_act:		return ses_Actual;
		case si_EIT_oth:		return ses_Other;
		case si_EIT_act_sched:	return ses_Actual_Schedule;
		case si_EIT_oth_sched:	return ses_Other_Schedule;
		}
		return ses_unknown;
	}
};

// ############################################################################

/**
 * \todo DIT et SIT (voir ETSI EN 300 468)
 **/