/*
 * $Id: sbuffer.h 348 2012-07-01 20:27:42Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion d'un tampon de donnes  croissance automatique (dfinitions)
 *
 *	Ce fichier dfinit des classes utilises pour la construction d'un tampon
 *	dont la taille crot automatiquement en cas de dpassement de longueur.
 *	Ce tampon est utilis pour la conversion interne TS -> PS.
 **/

#pragma once

#include <vector>
#include "memrng.h"

// ----------------------------------------------------------------------------

/** \brief Tampon  croissance automatique
 *
 * Drivation de la classe "vector" pour recevoir des donnes fragmentes et rassembler celles-ci,
 * en utilisant un tampon dont la taille crot automatiquement en cas de dpassement de longueur.
 * Afin d'viter des rallocations mmoire continuelles, la taille du tampon ne se rduit jamais
 * en cours d'utilisation (sauf usage explicite de la fonction membre \p clear).
 * Il est simplement entirement dtruit avec le destructeur de l'objet.
 **/
template <class _Ty, int _DefGran = 1, class _Ax = allocator<_Ty> >
	class auto_vector : private vector<_Ty, _Ax>
{
	size_type	_Granularity;	//!< granularit (valeur qui spcifie par multiples
								//!< de combien la taille du tampon doit tre augmente)
	size_type	_UsedSize;		//!< Nombre d'lments actuellement utiliss

public:
	auto_vector(size_type _Gran = _DefGran) :
		_Granularity(_Gran),
		_UsedSize(0)
	{
	}

	using vector<_Ty, _Ax>::reserve;		// ajustement d'accs (temporaire ?)
	using vector<_Ty, _Ax>::begin;			// ajustement d'accs
	using vector<_Ty, _Ax>::iterator;		// ajustement d'accs
	using vector<_Ty, _Ax>::const_iterator;	// ajustement d'accs

	/// S'assurer que le tempon a au moins la taille "_NewSize"
	_Ty * grow(size_type _NewSize)
	{
		if (_NewSize > __super::size()) {
			// Si le tampon actuel n'atteint pas cette taille, on l'agrandit
			size_type _GrowSize = ((_NewSize+(_Granularity-1))/_Granularity)*_Granularity;

			resize(_GrowSize);
		}
		if (_NewSize > _UsedSize)
			_UsedSize = _NewSize;
		return &front();		// Retourner l'adresse de dbut;
	}

	//iterator			begin() {
	//						return __super::begin(); }

	//const_iterator	begin() const {
	//						return __super::begin(); }

	iterator			end() {
							return __super::begin() + sVPt.siz; }

	const_iterator		end() const {
							return __super::begin() + sVPt.siz; }

	/// Remettre le pointeur d'criture au dbut du tampon
	void				reset() {
							_UsedSize = 0; }

	/// Tout rinitialiser et rcuprer la mmoire
	void				clear() {
							_UsedSize = 0; __super::clear(); }

	bool				empty() const {
							return _UsedSize == 0; }

	/// Obtenir le nombre d'lments couramment utiliss
	size_type			size() const {
							return _UsedSize; }

	reference			front() {
							if (__super::empty())
								resize(_Granularity);	// s'assurer que le vecteur n'est pas vide
							return __super::front(); }
	_Ty *				ptr() {
							return &front(); }

	const _Ty *			ptr() const {
							return &front(); }

	/// Obtenir un intervalle dlimitant le contenu actuel du tampon
	PtrRng<_Ty>			get() {
							return PtrRng<_Ty>(&front(), (UINT)_UsedSize); }

	// change des donnes avec celles d'un autre tampon de mme type :
	void				swap(auto_vector & b2swap) {
							std::swap(*dynamic_cast<vector<_Ty> *>(this), *dynamic_cast<vector<_Ty> *>(&b2swap));
							std::swap(_Granularity, b2swap._Granularity);
							std::swap(_UsedSize, b2swap._UsedSize); }

	// Accs  un lment par son indice (const)
	const _Ty &			operator [] (size_type _Pos) const {
							return __super::operator[] (_Pos); }

	// Accs  un lment par son indice (modifiable, croissance si ncessaire)
	_Ty &				operator [] (size_type _Pos) {
							grow(_Pos+1);
							return __super::operator [] (_Pos); }

	/// Rinitialise le tampon avec des donnes source
	void				set(const _Ty * ptr, size_type siz) {
							memcpy(grow(siz), ptr, siz*sizeof(_Ty)); _UsedSize = siz; }

	/// Rinitialise le tampon avec des donnes source
	void				set(PtrRng<_Ty> sMm) {
							set(sMm.ptr, sMm.siz); }

	/// Rinitialise le tampon avec un lment source
	void				set(_Ty _Val) {
							grow(1)[0] = _Val; _UsedSize = 1; }

	/// Ajout de donnes source dans le tampon
	void				append(const _Ty * ptr, size_type siz) {
							size_type	_OldSize = size();
							memcpy(grow(_OldSize + siz) + _OldSize, ptr, siz*sizeof(_Ty)); }

	/// Ajout de donnes source dans le tampon
	void				append(PtrRng<_Ty> sMm) {
							append(sMm.ptr, sMm.siz); }

	/// Ajout d'un lment source dans le tampon
	void				append(_Ty _Val) {
							size_type	_OldSize = size();
							grow(_OldSize+1)[_OldSize] = _Val; }

	/// limination de "_StripSize" lments, avec copie des lments suivants au dbut du tampon
	void strip(size_type _StripSize) {
		if (_StripSize > 0) {
			if (_StripSize < _UsedSize) {
				memmove(ptr(), ptr()+_StripSize, (_UsedSize - _StripSize)*sizeof(_Ty));
				_UsedSize -= _StripSize;
			} else
				_UsedSize = 0;
		}
	}
};

// ----------------------------------------------------------------------------

class StreamBuffer : public auto_vector<UINT8>
{
	virtual void	sample_start(PCUINT8 ptr) {
						}
	virtual void	sample_end(PCUINT8 ptr) {
						}
protected:
	/** \brief Retourne \p true si le paquet dbute une squence
	 *
	 * \param[in]	nPrevSize	Taille du tampon *avant* l'ajout
	 * \return		\p true si un lment de synchronisation permettant la vidange du tampon
	 *				est prsent depuis le nouvel ajout
	 **/
	virtual bool	test(UINT nPrevSize) {
						return false; }	// Par dfaut : pas de notion de squence

	/// liminer les donnes traites
	virtual void	discard() {
						reset(); }		// Si pas surcharg, rinitialiser le tampon

public:
	/// Constructeur
	StreamBuffer(size_t gr=0x200) :
		auto_vector((UINT)gr)
	{
	}

	/** \brief Ajout de donnes
	 * \param[in]	sMm		Intervalle (pointeur + longueur)  ajouter
	 * \param[in]	start	\p true si dbut de squence (provoque un \p flush pralable)
	 **/
	void			put_data(MemRng sMm, bool start=false) throw(...);

	/// Envoi et vidange du tampon
	virtual void	flush() throw(...);

	/// Prtraitement d'une squence  crire (excut avant \p send)
	virtual void	prepare(const MemRng & sMm) {
						}

	/// Envoi des donnes compltes
	virtual void	send(MemRng sMm) throw(...) = 0;

	/// change des donnes avec celles d'un autre tampon de mme type
	void			swap(StreamBuffer & b2swap);

	virtual ~StreamBuffer() {
				}
};

// ----------------------------------------------------------------------------
