/*
 * $Id: utils.cpp 398 2020-10-22 17:02:27Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Fonctions utilitaires gnriques (implmentations)
 *
 *	Ce module contient des dfinitions et structures utilitaires simples, et non
 *	obligatoirement lies au programme principal.
 *
 *	Les objets dfinis ici ne doivent faire rfrence qu' des bibliothques externes
 *	standard du C, du C++ ou de Windows.
 **/

#include "stdafx.h"
#include <errno.h>
#include <shellapi.h>	// Pour CommandLineToArgvW

// ====================================================================================
// Chanes de caractres
// ====================================================================================

/**
 * Version CHAR -> WCHAR (utilise \p MultiByteToWideChar).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[out] pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
errno_t strcpy_T(LPWSTR pDst, UINT nSizDst, LPCSTR pSrc)
{
	if (MultiByteToWideChar(CP_ACP, 0, pSrc, -1, pDst, nSizDst))
		return 0;
	return EINVAL;
}

/**
 * Version WCHAR -> CHAR (utilise \p WideCharToMultiByte).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[out] pDst		Pointeur sur le tampon de destination (CHAR)
 **/
errno_t strcpy_T(LPSTR pDst, UINT nSizDst, LPCWSTR pSrc)
{
	if (WideCharToMultiByte(CP_ACP, 0, pSrc, -1, pDst, nSizDst, NULL, NULL))
		return 0;
	return EINVAL;
}

/**
 * Copie vers chane STL, version CHAR -> std::wstring.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  strDst	Chane STL de destination (std::wstring)
 **/
errno_t strcpy_T(std::wstring & strDst, LPCSTR pSrc)
{
	COPY_STR_ON_STACK(WCHAR, pszTmp, pSrc, nLen);

	strDst = pszTmp;
	return 0;
}

/**
 * Copie vers chane STL, version WCHAR -> std::string.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  strDst	Chane STL de destination (std::string)
 **/
errno_t strcpy_T(std::string & strDst, LPCWSTR pSrc)
{
	COPY_STR_ON_STACK(CHAR, pszTmp, pSrc, nLen);

	strDst = pszTmp;
	return 0;
}

/**
 * Version CHAR -> WCHAR (utilise \p MultiByteToWideChar).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[out] pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
errno_t strcat_T(LPWSTR pDst, UINT nSizDst, LPCSTR pSrc)
{
	UINT nLen = strlen_T(pDst);

	if (nLen < nSizDst)
		return strcpy_T(pDst+nLen, nSizDst-nLen, pSrc);
	return EINVAL;
}

/**
 * Version WCHAR -> CHAR (utilise \p WideCharToMultiByte).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[out] pDst		Pointeur sur le tampon de destination (CHAR)
 **/
errno_t strcat_T(LPSTR pDst, UINT nSizDst, LPCWSTR pSrc)
{
	UINT nLen = strlen_T(pDst);

	if (nLen < nSizDst)
		return strcpy_T(pDst+nLen, nSizDst-nLen, pSrc);
	return EINVAL;
}

/**
 *	Conversion d'une chane de caractres WCHAR vers une chane std::string
 *  \param[in]	pszStr	Chane  convertir
 **/
std::string WChar2Char(LPCWSTR pszStr)
{
	COPY_STR_ON_STACK(CHAR, pszTmp, pszStr, nLen);

	return pszTmp;
}

/**
 *	Conversion d'une chane de caractres CHAR vers une chane std::wstring
 *  \param[in]	pszStr	Chane  convertir
 **/
std::wstring Char2WChar(LPCSTR pszStr)
{
	COPY_STR_ON_STACK(WCHAR, pszTmp, pszStr, nLen);

	return pszTmp;
}

/**
 * Nettoyage des espaces en dbut et fin d'une chane de caractres
 *
 * \param[in,out]	pszStr		Adresse du tampon contenant la chane  nettoyer
 * \param[in]		nBufLen		Taille du tampon
 * \param[in]		pszDefault	Chane  substituer si le rsultat est vide
 **/
void clean_string(LPTSTR pszStr, UINT32 nBufLen, LPCTSTR pszDefault)
{	// Note : le rsultat tant de taille au maximum gale  la chane d'origine,
	// il n'est pas utile de spcifier la taille du tampon.
	UINT	nLen = strlen_T(pszStr);
	TCHAR	nCh;

	// Nettoyage des espaces et codes de contrle  la fin
	while (nLen>0 && pszStr[nLen-1]<=' ')
		--nLen;
	pszStr[nLen] = 0;

	// Nettoyage des espaces et codes de contrle au dbut
	UINT	nInx = 0;

	while ((nCh=pszStr[nInx])!=0 && nCh<=' ')
		++nInx;
	if (nInx>0)
		strcpy_T(pszStr, nBufLen, pszStr+nInx);

	// Valeur par dfaut si rsultat vide :
	if (*pszStr == 0 && pszDefault)
		strcpy_T(pszStr, nBufLen, pszDefault);
}

/**
 * Assignation d'une chane sur une autre avec test modification
 *
 * \param[in]		pszSrc		Chane source
 * \param[in,out]	pszDst		Chane destination
 * \param[in]		nSize		Taille du tampon de cette chane
 * \retval			\p true si la chane destination a t modifie
 * \retval			\p false si la nouvelle valeur tait identique  l'ancienne
 **/
bool copy_check(LPTSTR pszDst, UINT nSize, LPCTSTR pszSrc)
{
	if (_tcscmp(pszDst, pszSrc)==0)
		return false;		// false = Aucun changement
	strcpy_T(pszDst, nSize, pszSrc);
	return true;			// true = Changement
}

/**
 * Gnration d'une chane formate (format \p printf) avec rsultat dans une chane \p tstring
 **/
tstring tstr_printf(LPCTSTR pszFmt, ...)
{
	FMT_BUF(szBuf, 1024, pszFmt, pszFmt);

	return szBuf;
}

/**
 * Nettoyage des espaces et caractres spciaux (tels CR/LF)  la fin d'une chane de caractres
 * \param[in,out]	pszStr	Chane de caractres  nettoyer, et rsultat
 * \return	Pointeur sur cette mme chane
 **/
LPTSTR RTrimStr(LPTSTR pszStr)
{
	UINT nLen = strlen_T(pszStr);

	while (nLen>0 && UINT(pszStr[--nLen])<=' ')
		pszStr[nLen]=0;
	return pszStr;
}

/// Conversion d'un entier en chane de caractres, avec sparateur des milliers
tstring thousandsSepFmt(UINT64 nVal)
{
	TCHAR	szWrk[32];

	if (nVal < 1000) {
		_ui64tot_s(nVal, szWrk, _countof(szWrk), 10);
		return szWrk;
	}

	TCHAR	szTh[8];

	GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, szTh, _countof(szTh));
	_stprintf_s(szWrk, _T("%s%s%03u"), thousandsSepFmt(nVal / 1000).c_str(),
		szTh, UINT(nVal % 1000));
	return szWrk;
}

/**
 * Formatage d'une date au format Unix
 **/
tstring strftime_T(__time64_t time, LPCTSTR fmt)
{
	struct tm	_tm;

	if (_localtime64_s(&_tm, &time) == 0) {
		_locale_t	loc = _create_locale(LC_ALL, "French");
		TCHAR		szBuf[128];

		if (loc) {
			_tcsftime_l(szBuf, _countof(szBuf), fmt, &_tm, loc);
			_free_locale(loc);
		} else
			_tcsftime(szBuf, _countof(szBuf), fmt, &_tm);
		return szBuf;
	}

	return TEXT("(erreur)");
}

// ====================================================================================
// Gestion de fentres
// ====================================================================================

/**
 * Vrifie que les valeurs stockes sont dans la surface d'affichage actuelle
 *
 * \retval  true  si la zone est comprise dans l'affichage actuel
 * \retval  false sinon
 **/
bool WindowPos::verifie() const
{
	RECT rect;
	//
	SetRect(&rect, X, Y, X+L, Y+16);
	if (
		X==CW_USEDEFAULT	||
		Y==CW_USEDEFAULT	||
		L==CW_USEDEFAULT	||
		H==CW_USEDEFAULT	||
		MonitorFromRect(&rect, MONITOR_DEFAULTTONULL)==NULL
	) {
		return false;
	}
	return true;
}

/// Restauration d'une position de fentre
BOOL WindowPos::RestoreWindowPos(HWND hWnd, UINT nShowCmd) const
{
	WINDOWPLACEMENT wpl = {
		sizeof(wpl),			// length
		0,						// flags
		nShowCmd,				// showCmd
		{-1, -1},				// ptMinPosition
		{-1, -1},				// ptMaxPosition
		{X, Y, X + L, Y + H}	// rcNormalPosition
	};

	return SetWindowPlacement(hWnd, &wpl);
}

/**
 * Conversion des coordonnes d'un rectangle contenant des coordonnes relatives  l'cran en un rectangle
 * contenant des coordonnes relatives au rectangle client d'une fentre.
 * \param[in]		hWnd	"Handle" d'une fentre dont le rectangle client sera utilis pour la conversion
 * \param[in,out]	sRect	Rectangle dont les coordonnes seront converties
 **/
BOOL ScreenToClient(HWND hWnd, RECT & sRect)
{
	POINT	sPnt = {0, 0};
	BOOL	bRes = ScreenToClient(hWnd, &sPnt);

	return bRes && OffsetRect(&sRect, sPnt.x, sPnt.y);
}

// ====================================================================================
// Conversion entre entiers et chanes par table d'association
// ====================================================================================

/**
 * \brief Trouver la valeur binaire associe  une chane de caractres fournie dans une table
 *
 * Si la chane ne se trouve pas dans la table, alors la valeur \p dwDefault est retourne.
 *
 * \param[in]  assocTable	Tableau associatif d'o l'on doit dduire la valeur numrique
 *							correspondant  la chane de caractres passe
 * \param[in]	pszStr		Chane de caractres recherche
 * \param[in]	dwDefault	Valeur retourne si on n'a rien trouv
 * \return					La valeur numrique correspondante
 **/
DWORD str2dword(const AssocElement * assocTable, LPCTSTR pszStr, DWORD dwDefault)
{
	while (assocTable->str != NULL && _tcscmp(pszStr, assocTable->str) != 0)
		assocTable++;

	if (assocTable->str==NULL)
		return dwDefault;
	return assocTable->bin;
}

/**
 * \brief Trouver la chane de caractres associe  une valeur binaire fournie dans une table
 *
 * Si la valeur binaire ne se trouve pas dans la table, alors la chane \p pszDefault est retourne.
 *
 * \param[in]	assocTable	Tableau associatif d'o l'on doit extraire la valeur
 * \param[in]	dwValue		Valeur devant tre convertie en chane de caractres
 * \param[in]	pszDefault	Chane retourne si aucune correspondance trouve
 * \return					Chane de caractres correspondant  \p dwValue, ou \p pszDefault si pas trouv
 **/
LPCTSTR dword2str(const AssocElement * assocTable, DWORD dwValue, LPCTSTR pszDefault)
{
	while (assocTable->str != NULL && assocTable->bin != dwValue)
		assocTable++;

	LPCTSTR str = assocTable->str;

	if (str==NULL)
		str = pszDefault;
	return str;
}

/**
 * Obtenir une chane de caractres "Oui" ou "Non" selon la valeur d'un boolan.
 *
 * \param[in]	bValue	valeur boolenne  convertir
 * \return	Chane de caractres "Oui" ou "Non"
 **/
LPCTSTR boolToYesNo(bool bValue)
{
	return bValue ? TEXT("Oui") : TEXT("Non");
}

// ====================================================================================
// Heure et date
// ====================================================================================

/**
 * Fonction qui retourne la diffrence time1 - time2, convertie en millisecondes.
 *
 * \param[in]  time1  temps devant tre compar
 * \param[in]  time2  temps soustrait au prcdent
 * \return           diffrence entre les 2 temps en millisecondes
 **/
long DiffTime(SYSTEMTIME time1, SYSTEMTIME time2)
{
	INT64 Utime1, Utime2;
	SystemTimeToFileTime(&time1, (LPFILETIME)&Utime1);
	SystemTimeToFileTime(&time2, (LPFILETIME)&Utime2);

	// Plafonner la diffrence  DUREE_INFINIE-1
	INT64 diff = (Utime1 - Utime2) / 10000;

	if (diff > DUREE_INFINIE-1)
		return DUREE_INFINIE-1;
	if (diff < -(DUREE_INFINIE-1))
		return -(DUREE_INFINIE-1);

	return long(diff);
}

/**
 * Fonction qui ajoute 'add' millisecondes au temps 'time.
 *
 * \param[in,out] time  date de dpart/finale
 * \param[in]     add   millisecondes  ajouter
 **/
void AddTime(SYSTEMTIME & time, long add)
{
	INT64 Utime;

	SystemTimeToFileTime(&time, LPFILETIME(&Utime));
	Utime += INT64(add)*10000;
	FileTimeToSystemTime(LPFILETIME(&Utime), &time);
}

/**
 * Conversion heure (sans date) en chane de caractres
 *
 * \param[in]  st        structure SYSTEMTIME contenant les donnes  intgrer
 * \param[out] str       chane de caractres rsultant
 * \param[in]  str_size  longeur max de la chane de caractres
 **/
void TimeToStr(SYSTEMTIME st, LPTSTR str, size_t str_size)
{
	_stprintf_s(str, str_size, TEXT("%02i:%02i:%02i"), st.wHour, st.wMinute, st.wSecond);
}

/**
 * Conversion heure et date en chane de caractres
 *
 * \param[in]  st        structure SYSTEMTIME contenant les donnes  intgrer
 * \param[out] str       chane de caractres rsultant
 * \param[in]  str_size  longeur max de la chane de caractres
 **/
void DateTimeToStr(SYSTEMTIME st, LPTSTR str, size_t str_size)
{
	_stprintf_s(str, str_size, TEXT("%") A2t TEXT(" %02i, %02i:%02i:%02i"),
		EnumToString(st.wDayOfWeek, pszDoW_short_ucf),
		st.wDay, st.wHour, st.wMinute, st.wSecond);
}

/**
 * Conversion temps GMT en temps local
 *
 * \param[in,out] st  structure SYSTEMTIME  convertir
 **/
void SystemTimeToLocalSystemTime(SYSTEMTIME & st)
{
	FILETIME		ft;
	FILETIME		lft;

	SystemTimeToFileTime(&st, &ft);
	FileTimeToLocalFileTime(&ft, &lft);
	FileTimeToSystemTime(&lft, &st);
}

/// Liste des jours du mois (longs, minuscules)
LPCSTR pszMon_lower =
	"\1\0"
	"janvier\0"
	"fvrier\0"
	"mars\0"
	"avril\0"
	"mai\0"
	"juin\0"
	"juillet\0"
	"aot\0"
	"septembre\0"
	"octobre\0"
	"novembre\0"
	"dcembre\0";

/// Liste des jours de la semaine (longs, minuscules)
LPCSTR pszDoW_lower =
	"dimanche\0"
	"lundi\0"
	"mardi\0"
	"mercredi\0"
	"jeudi\0"
	"vendredi\0"
	"samedi\0";

/// Liste des jours de la semaine (longs, avec majuscules initiales)
LPCSTR pszDoW_ucfirst =
	"Dimanche\0"
	"Lundi\0"
	"Mardi\0"
	"Mercredi\0"
	"Jeudi\0"
	"Vendredi\0"
	"Samedi\0";

/// Liste des jours de la semaine (courts, avec majuscules initiales)
LPCSTR pszDoW_short_ucf =
	"Dim\0"
	"Lun\0"
	"Mar\0"
	"Mer\0"
	"Jeu\0"
	"Ven\0"
	"Sam\0";

/**
 * \brief Reformatage de la date et de l'heure de compilation
 *
 * \param[out]	pszBuf		Tampon destin  recevoir le rsultat
 * \param[in]	nBufSiz		Taille du tampon
 * \param[in]	pszDtSep	Chane de sparation entre date et heure
 * \param[in]	pszRawTime	Date et heure au format natif du compilateur, spars par une virgule
 * \note	La chane contenant l'heure et la date source ne doit surtout pas tre assigne  une
 *			variable globale, car cela compromettrait la fonctionnalit "Edit & Continue" de Visual Studio.
 **/
void ReformatCompileTime(LPTSTR pszBuf, UINT nBufSiz, LPCTSTR pszDtSep, LPCSTR pszRawTime)
{
	CHAR		szMon[8];
	int			nDay, nYear, nHour, nMin, nSec;

	// Reformatage de la date de compilation
	if (
		sscanf_s(pszRawTime, "%s %u %u, %u:%u:%u",
			szMon, (UINT)_countof(szMon), &nDay, &nYear, &nHour, &nMin, &nSec) == 6
	)
		_stprintf_s(pszBuf, nBufSiz, TEXT("%04i-%02i-%02i%s%ih%02i"), nYear,
			StringToEnum("Jan\0Feb\0Mar\0Apr\0May\0Jun\0Jul\0Aug\0Sep\0Oct\0Nov\0Dec\0", szMon)+1, nDay,
			pszDtSep, nHour, nMin);
	else
		strcpy_T(pszBuf, nBufSiz, pszRawTime);
}

// ====================================================================================
// Conversion entre chanes et numrations
// ====================================================================================

/**
 * Conversion d'une valeur d'un type numr vers une chane de caractres
 * slectionne dans une chane fournissant une suite de rsultats possibles
 *
 * \param[in] nVal			Valeur  associer
 * \param[in] pszList		Chane contenant une suite d'items spars par des zros binaires ;
 *							un item dbutant par \p 0x01 ("un" binaire) est remplac par l'item par
 *							dfaut ; un item vide (donc deux zros successifs) DOIT terminer.
 * \param[in] pszDefault	Valeur par dfaut
 * \retval					Une copie de l'item de rang \p nVal (base 0), ou bien \p pszDefault
 *							si \p nVal excde le nombre d'items ou si l'item dbute par \p 0x01.
 * \note \p pszList doit tre une chane constante, ou bien au moins statique.
 * \note Cette fonction permet de ne pas avoir de tableau externe pour effectuer la conversion.
 * \note Aucun item ne peut tre vide, puisqu'il serait alors d'office terminateur.
 * \note tant donne la longueur possible de la chane \p pszList, on n'utilise
 * pas de caractres \p TCHAR.
 **/
LPCSTR EnumToString(int nVal, LPCSTR pszList, LPCSTR pszDefault)
{
	if (pszList && nVal>=0) {
		while (*pszList) {	// Tant qu'on n'a pas une chane vide
			if (nVal==0) {
				if (pszList[0]==1)
					break;
				return pszList;
			}
			pszList += strlen_T(pszList)+1;
			nVal--;
		}
	}
	return pszDefault;
}

/**
 * Fonction inverse de \p EnumToString : recherche d'une chane de caractres dans une liste.
 * La liste de chanes est fournie sous la forme d'une seule chane avec des \p \\0 
 * la fin de chaque item. Un double \p \\0 termine donc implicitement.
 * La comparaison n'est pas sensible  la casse.
 * \param[in] pszList	Chane contenant une suite d'items spars par des zros binaires ;
 *						un item vide (donc deux zros successifs) DOIT terminer.
 * \param[in] pszStr	Chane recherche.
 * \return	Index de la chane trouve, ou bien -1 si pas trouv.
 **/
int StringToEnum(LPCSTR pszList, LPCSTR pszStr)
{
	int nVal = 0;

	if (pszStr && pszList) {
		while (*pszList) {
			if (_stricmp(pszList, pszStr)==0)
				return nVal;
			pszList += strlen(pszList)+1;
			++nVal;
		}
	}
	return -1;
}

// ====================================================================================
// Fichiers et rpertoires
// ====================================================================================

/**
 * Recherche d'un fichier correspondant au nom pass en paramtre
 *
 * \param[in]	pszDir	Rpertoire initial pour la recherche
 * \param[in]	pszFile	Fichier, ou motif de recherche
 * \param[in]	dwFlags	Options de recherche (FF_RECURSIVE, FF_DIRECTORY)
 * \return	Nom du premier fichier trouv avec le chemin, ou une chane vide si pas trouv
 **/
tstring FindFile(LPCTSTR pszDir, LPCTSTR pszFile, DWORD dwFlags)
{
	vtstring vResult;

	UINT	nFound = FindFile(pszDir, pszFile, vResult, dwFlags|FF_STOPATFIRST);

	if (nFound)
		return vResult[0];
	return TEXT("");
}

/// Tester si \p w32fs rfrence un fichier normal
static inline bool isFile(const WIN32_FIND_DATA & w32fs)
{
	return (w32fs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)==0;
}

/// Tester si \p w32fs rfrence un rpertoire
static inline bool isDir(const WIN32_FIND_DATA & w32fs)
{
	return
		(w32fs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 &&
		!(w32fs.cFileName[0] == TEXT('.') && (
			w32fs.cFileName[1] == 0 ||
			w32fs.cFileName[1] == TEXT('.') && w32fs.cFileName[2] == 0));
}

/**
 * Recherche de tous les fichiers correspondant au motif pass en paramtre
 *
 * \param[in]	pszDir	Rpertoire initial pour la recherche
 * \param[in]	pszFile	Fichier, ou motif de recherche
 * \param[out]	vResult	Vecteur contenant les fichiers rpondant au nom pass en paramtre
 * \param[in]	dwFlags	Options de recherche (FF_RECURSIVE, FF_STOPATFIRST, FF_DIRECTORY)
 * \return	Le nombre de fichiers trouvs correspondant
 **/
UINT FindFile(LPCTSTR pszDir, LPCTSTR pszFile, vtstring & vResult, DWORD dwFlags)
{
	WIN32_FIND_DATA	w32fs;				// Structure recevant les informations pour la recherche
	SZPathName		szTmp;				// Chemin de recherche : szDir + \\*
	UINT			nFound	= 0;		// Indique le nombre de fichiers dj trouvs
	HANDLE			h;					// Handle utilis pour la recherche
	BOOL			bNext	= TRUE;

	// Recherche avec le motif pass
	_stprintf_s(szTmp, TEXT("%s\\%s"), pszDir, pszFile);
	for (
		h = FindFirstFile(szTmp, &w32fs);
		h != INVALID_HANDLE_VALUE && bNext;
		bNext = FindNextFile(h, &w32fs)
	) {
		if ((dwFlags&FF_DIRECTORY)!=0 ? !isDir(w32fs) : !isFile(w32fs))
			continue;

		// on dtermine le szTmp complet du fichier (dossier + fichier)
		_stprintf_s(szTmp, TEXT("%s\\%s"), pszDir, w32fs.cFileName);
		vResult.push_back(szTmp);
		++nFound;
		if ((dwFlags&FF_STOPATFIRST) != 0)
			break;
	}

	FindClose(h);	// Ferme le handle

	if ((dwFlags&FF_RECURSIVE) == 0 || (dwFlags&FF_STOPATFIRST) != 0 && nFound > 0)
		return nFound;

	bNext = TRUE;

	// Chemin de recherche pour le rcursif
	_stprintf_s(szTmp, TEXT("%s\\*"), pszDir);
	for (
		h = FindFirstFile(szTmp, &w32fs);
		h != INVALID_HANDLE_VALUE && bNext;
		bNext = FindNextFile(h, &w32fs)
	) {
		if (!isDir(w32fs))
			continue;

		// on part en rcursif
		_stprintf_s(szTmp, TEXT("%s\\%s"), pszDir, w32fs.cFileName);

		nFound += FindFile(szTmp, pszFile, vResult, dwFlags);
		if (nFound > 0 && (dwFlags&FF_STOPATFIRST) != 0)
			break;
	}

	// Ferme le handle
	FindClose(h);

	return nFound;
}

// ====================================================================================
// UTF-8
// ====================================================================================

/**
 * Convertit une chane de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres UTF-8
 * \return     chane de caractres convertie
 **/
tstring ConvertUTF82Char(LPCSTR aconvertir)
{
	LPTSTR	_tmp = TEXT("");

	if (aconvertir) {
		UINT	i, j;								// Itrateurs
		UCHAR	c;									// Caractres  convertir
		UINT	taille = strlen_T(aconvertir);		// La taille de la chane rsultante est forcment plus courte

		_tmp = STK_ALLOC(TCHAR, taille+1);	// Chane temporaire
		for( i=0, j=0; i<taille; i++) {
			c = aconvertir[i];

			UINT32 nTmp = '?'; // Caractre par dfaut si conversion impossible

			if (c <= 0x7F)
				nTmp = c;
			else if (c < 0xE0)
				nTmp = ((c & 0x1F) << 6) + (aconvertir[++i] & 0x3F);
#ifdef _UNICODE
			// Seulement en _UNICODE, car il faut plus de 8 bits pour produire ces caractres
			else if (c < 0xF0)
				nTmp = ((c & 0x0F) << 12) + ((aconvertir[++i] & 0x3F) << 6) + (aconvertir[++i] & 0x3F);
			else // c < 0xF8
				nTmp = ((c & 0x07) << 18) + ((aconvertir[++i] & 0x0F) << 12) + ((aconvertir[++i] & 0x3F) << 6) + (aconvertir[++i] & 0x3F);
#endif
			_tmp[j++] = TCHAR(nTmp);
		}
		_tmp[j] = 0;
	}

	return _tmp;	// Conversion implicite en tstring
}

/**
 * Convertir une chaine de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     chane de caractres convertie
 **/
std::string ConvertChar2UTF8(LPCSTR aconvertir, bool for_sgml)
{
	if (!aconvertir)
		return "";

	COPY_STR_ON_STACK(WCHAR, temp, aconvertir, taille);

	return ConvertChar2UTF8(temp, for_sgml);
}

/**
 * Convertir une chaine de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     chane de caractres convertie
 **/
std::string ConvertChar2UTF8(LPCWSTR aconvertir, bool for_sgml)
{
	LPSTR pszTmp = "";

	if (aconvertir) {
#if 0	// exprimental
		UINT j	= WideCharToMultiByte(CP_UTF8, 0, aconvertir, -1, NULL, 0, NULL, NULL);

		LPSTR pszTmp = STK_ALLOC(CHAR, j);
		j = WideCharToMultiByte(CP_UTF8, 0, aconvertir, -1, convertis, j, NULL, NULL);
#else
		UINT	i, j;									// Itrateurs
		UINT	taille	= strlen_T(aconvertir);

		// Allouer l'espace
		pszTmp = STK_ALLOC(CHAR, taille*5+1);	// Chane de carastres temporaire (la taille max de la
													//   chane est de 5 octets par caractre original)
		// Effectuer la conversion
		for(i=0, j=0; i<taille; i++)
			j += ConvertChar2UTF8(aconvertir[i], &pszTmp[j], for_sgml);

#endif
	}
	return pszTmp;	// Conversion implicite en std::string
}

/**
 * Convertir un caractre en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[out] convertis   chane de caractres convertie
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     le nombre de caractres convertis
 * \note \p aconvertir est de type \p UINT32 car :
 * - TCHAR pose des problmes tant donn qu'il s'agit d'une valeur *signe*
 * - La fonction est capable de convertir des caractres cods sur plus de 16 bits
 * \note \p convertis doit pointer sur une variable de taille suffisante
 * pour tre capable de recevoir au moins 6 caractres (terminateur compris).
 **/
int ConvertChar2UTF8(UINT32 aconvertir, LPSTR convertis, bool for_sgml)
{
	int	j;		// Itrateur

#if 0	// exprimental
	j = WideCharToMultiByte(
			CP_UTF8, 0, reinterpret_cast<LPCWSTR>(&aconvertir),
			1, convertis, 5, NULL, NULL);
#else

	if (aconvertir <= 0x7F) {
		convertis[0] = (char)aconvertir;
		j = 1;
		if (for_sgml) {
			LPCSTR	sgml = NULL;
			switch (aconvertir) {
			case '&':	sgml = "&amp;";	break;
			case '<':	sgml = "&lt;";	break;
			case '>':	sgml = "&gt;";
			}
			if (sgml) {
				strcpy_s(convertis, 6, sgml);
				j = (int)strlen(sgml);
			}
		}
	} else if (aconvertir <= 0x7FF) {
		convertis[0] = (char)(0xC0 | ((aconvertir>>6) & 0x1F));
		convertis[1] = (char)(0x80 | (aconvertir & 0x3F));
		j = 2;
	} else if (aconvertir <= 0xFFFF) {
		convertis[0] = (char)(0xE0 | ((aconvertir>>12) & 0x0F));
		convertis[1] = (char)(0x80 | ((aconvertir>>6) & 0x3F));
		convertis[2] = (char)(0x80 | (aconvertir & 0x3F));
		j = 3;
	} else { // aconvertir <= 0x10FFFF
#pragma warning(suppress: 4333)
		convertis[0] = (char)(0xF0 | ((aconvertir>>18) & 0x07));
		convertis[1] = (char)(0x80 | ((aconvertir>>12) & 0x3F));
		convertis[2] = (char)(0x80 | ((aconvertir>>6) & 0x3F));
		convertis[3] = (char)(0x80 | (aconvertir & 0x3F));
		j = 4;
	}
#endif
	convertis[j] = 0;	// zro terminateur

	return j;
}

// ====================================================================================
// Systme
// ====================================================================================

/**
 * Conversion d'un code d'erreur systme (retourn par \p GetLastError)
 * en une chane de caractres affichable.
 *
 * \param[in]	dwErr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 **/
tstring SystemErrorToString(DWORD dwErr)
{
	TCHAR	szBuffer[512];

	FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dwErr,
		0, // Langue par dfaut (pourrait tre 0x040c = franais)
		szBuffer,
		_countof(szBuffer),
		NULL);

	return RTrimStr(szBuffer);	// Conversion implicite en tstring
}

// ====================================================================================
// Menus
// ====================================================================================

/**
 * Rcupration du menu parent ainsi que de la position d'un item de menu donn dans une arborescence de menus
 * \param[in]	hMenu	Handle de la racine de l'arborescence de menus contenant l'item dans un sous-menu quelconque
 * \param[in]	uIDItem	Identifiant de l'item de menu recherch
 * \param[out]	pPos	Si != NULL, la variable pointe reoit la position (base 0) de l'item de menu
 * \return		Handle du menu parent, ou bien NULL si l'item n'existe pas
 **/
HMENU GetMenuParentAndPos(HMENU hMenu, UINT uIDItem, INT * pPos)
{
	INT	nCount = GetMenuItemCount(hMenu);

	for (INT nInx = 0; nInx < nCount; ++nInx) {
		INT nId = GetMenuItemID(hMenu, nInx);

		if (nId == (INT)uIDItem) {
			if (pPos)
				*pPos = nInx;
			return hMenu;
		}
		if (nId == -1) {
			HMENU hSubMenu = GetSubMenu(hMenu, nInx);

			if (hSubMenu) {
				HMENU hRes = GetMenuParentAndPos(hSubMenu, uIDItem, pPos);

				if (hRes)
					return hRes;
			}
		}
	}
	return NULL;
}

/**
 * Cration d'un sous-menu en remplacement d'un autre, ou bien d'un item
 * \param[in]		hMenu	Handle du menu contenant l'item et/ou le sous-menu
 * \param[in, out]	nPos	Position du sous-menu, ou -1 si on rfrence un item.
 *							En sortie, reoit la position trouve si elle n'tait pas dfinie.
 * \param[in]		uIDItem	Item de menu rfrenc si la position est gale = -1
 * \return			Handle du nouveau sous-menu
 **/
HMENU CreatePopupMenuFromItem(HMENU hMenu, INT & nPos, UINT uIDItem)
{
	if (hMenu && nPos < 0)
		hMenu = GetMenuParentAndPos(hMenu, uIDItem, &nPos);

	MENUITEMINFO	sMii = {
		sizeof(MENUITEMINFO),		// cbSize
		MIIM_SUBMENU|MIIM_STATE,	// fMask
		0,							// fType
		MFS_ENABLED					// fState
		// Le reste  zro (implicite)
	};

	sMii.hSubMenu = CreatePopupMenu();
	if (hMenu && sMii.hSubMenu && SetMenuItemInfo(hMenu, nPos, TRUE, &sMii))
		return sMii.hSubMenu;

	// Ajout du sous-menu impossible : on libre la mmoire cre
	if (sMii.hSubMenu)
		DestroyMenu(sMii.hSubMenu);
	return NULL;
}

/**
 * Ajout d'un item de menu avec icne optionnelle
 * \param[in]	hMenu	Handle du menu contenant l'item et/ou le sous-menu
 * \param[in]	uIDItem	Item de menu rfrenc si la position est gale = -1
 * \param[in]	pszName	Texte de l'item de menu
 * \param[in]	fState	tat
 * \param[in]	hImage	Si != NULL, icne associe
 * \return		Handle du nouveau sous-menu
 **/
bool AppendMenuItemExt(HMENU hMenu, UINT uIDItem, LPCTSTR pszName, UINT fState, HBITMAP hImage)
{
	MENUITEMINFO	sMii = {
		sizeof(sMii),					// cbSize
		MIIM_STRING|MIIM_ID|MIIM_FTYPE|MIIM_STATE,	// fMask
		MFT_STRING,						// fType
		fState,							// fState
		uIDItem,						// wID
		NULL,							// hSubMenu
		NULL,							// hbmpChecked
		NULL,							// hbmpUnchecked
		0,								// dwItemData
		const_cast<LPTSTR>(pszName),	// dwTypeData,
		strlen_T(pszName),				// cch
		// Le reste  zro (implicite)
	};

	if (hImage) {
		sMii.fMask |= MIIM_BITMAP;
		sMii.hbmpItem = hImage;
	}

	bool	bRes = InsertMenuItem(hMenu, 0xffff, TRUE, &sMii) != 0;

	myprintf(CMyP ERRX(TEXT("chec InsertMenuItem(%s), err=%s")),
		!bRes, pszName, SystemErrorToString().c_str());
	return bRes;
}

/**
 * Modification de l'tat d'un item de menu
 * \param[in]	hMenu		Menu dans lequel se trouve l'item  modifier.
 * \param[in]	uItem		Identificateur de l'item  modifier.
 * \param[in]	fStateMask	Masque indiquant l'tat  modifier.
 * \param[in]	fState		Nouvel tat.
 **/
BOOL SetMenuItemState(HMENU hMenu, UINT uItem, UINT fStateMask, UINT fState)
{
	MENUITEMINFO info = {
		sizeof(info),	// cbSize
		MIIM_STATE		// fMask
		// le reste  zro (implicite)
	};

	if (!GetMenuItemInfo(hMenu, uItem, FALSE, &info))
		return FALSE;
	info.fState = (info.fState & (~fStateMask)) | fState;
	return SetMenuItemInfo(hMenu, uItem, FALSE, &info);
}

// ====================================================================================
// Donnes binaires
// ====================================================================================

/**
 * Enregistrement du contenu du vecteur dans un fichier
 *
 * \param[in]	pszFileName	Nom du fichier  sauvegarder
 * \param[in]	pSysTime	Date de modification  appliquer au fichier si != NULL
 * \retval		0 si tout s'est bien droul
 * \retval		Code d'erreur systme en cas d'erreur
 **/
DWORD ByteVector::SaveAsFile(LPCTSTR pszFileName, const SYSTEMTIME * pSysTime) const
{
	HANDLE	hFile = CreateFile(pszFileName, GENERIC_WRITE, FILE_SHARE_READ,
						NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile != INVALID_HANDLE_VALUE) {
		DWORD	dwWritten;	// Variable recevant le nombre d'octets crits
		bool	bOk = WriteFile(hFile, get(), (DWORD)size(), &dwWritten, NULL) != 0;

		if (bOk && pSysTime) {
			FILETIME sFTime;

			// Ajustement de la date du fichier
			if (SystemTimeToFileTime(&*pSysTime, &sFTime))
				SetFileTime(hFile, &sFTime, NULL, &sFTime);
		}

		CloseHandle(hFile);
		if (bOk)
			return 0;	// Ok
	}

	return GetLastError();
}

// ====================================================================================
// Ressources
// ====================================================================================

/**
 * Extrait une resource du fichier. Charge galement la taille alloue.
 * \param[in]	hInstance	Instance de l'application.
 * \param[in]	idRessource	Identifiant de la resource.
 * \param[in]	lpType		Type de la resource.
 * \retval		\p true en cas de succs, \p false en cas d'chec.
 **/
bool RawResource::Extraction(HINSTANCE hInstance, UINT idRessource, LPCTSTR lpType)
{
	HRSRC	hRcdataResource = FindResource(hInstance, MAKEINTRESOURCE(idRessource), lpType);
	bool	bRes			= false;

	if (hRcdataResource != NULL) {
		size_t nSize = SizeofResource(hInstance, hRcdataResource);
		// Charge la ressource identifie
		HGLOBAL hRcdataCharge = LoadResource(hInstance, hRcdataResource);

		if (hRcdataCharge != NULL) {
			// Bloque la ressource en mmoire pour utilisation
			LPBYTE RcdataRes = (LPBYTE)LockResource(hRcdataCharge);

			if (RcdataRes != NULL) {
				// Tout s'est bien pass: on peut utiliser la resource
				resize(nSize);

				LPBYTE pContent = get();

				memcpy(pContent, RcdataRes, nSize);
				bRes = true;
			}

			// Libre pour viter les fuites mmoires
			FreeResource(hRcdataCharge);
		}
	}

	return bRes;
}

// ====================================================================================
// Threads
// ====================================================================================

/// Point d'entre (statique) du thread
DWORD WINAPI WorkingThread::ThreadProc_static(LPVOID lpParameter)
{
	WorkingThread *	pWThread = reinterpret_cast<WorkingThread *>(lpParameter);

	if (pWThread) {
		DWORD	dwRes =	pWThread->ThreadProc();

		if (pWThread->bDeleteOnExit)
			delete pWThread;
		return dwRes;
	}
	return 0;
}

/**
 * Dmarrage du thread
 * \return \p true si le thread a dmarr correctement, \p false sinon
 **/
bool WorkingThread::StartThread()
{
	if (hThread==NULL)
		hThread = CreateThread(NULL, 0, ThreadProc_static, this, 0, &dwThreadId);
	return hThread != NULL;
}

/**
 * Attente de la terminaison du thread
 * \param[in]	dwMilliseconds	Temps d'attente maximal
 * \return \p true si le thread s'est termin correctement, \p false sinon
 **/
bool WorkingThread::WaitThreadExit(DWORD dwMilliseconds)
{
	if (WaitForSingleObject(hThread, dwMilliseconds) != WAIT_OBJECT_0)
		return false;
	// Mis  NULL pour indiquer que le thread n'est plus actif, et qu'on peut appeler
	// StartThread pour le relancer.
	/// \todo Chercher un autre moyen, car il existe apparemment des cas o on arrive
	/// ici alors que l'objet "WorkingThread" a dj t dtruit.
	hThread = NULL;
	return true;
}

/// Destructeur
WorkingThread::~WorkingThread()
{
	/// Attendre un peu la terminaison (prcaution, le thread devrait dj avoir t arrt)
	if (hThread && GetCurrentThreadId() != dwThreadId)
		WaitThreadExit(5000);	/// 5 secondes
}

// ====================================================================================
// Ligne de commande
// ====================================================================================

/** \brief Analyse des options de ligne de commande
 *	\param[in]	pOptTable	Table des options reconnues
 *	\param[in]	nSize		Taille de la table
 *	\param[out]	pDest		Pointeur sur la structure recevant les valeurs
 *	\param[out]	pOthers		Si non \p NULL, tous les lments prsents sur la ligne de commande qui ne
 *							sont pas spcifis dans la table sont ajouts  ce vecteur.
 **/
void ProcessCmdLine(OneCmdOption * pOptTable, int nSize, void * pDest, vtstring * pOthers)
{
	int			nArgs		= 0;	// Nombre d'arguments
	int			nOpt;

	// Pointeur sur un tableau de pointeur sur chanes de caractres, chacune reprsentant
	// un argument de ligne de commande
	LPWSTR *	papszArgList = CommandLineToArgvW(GetCommandLineW(), &nArgs);

	for (int nInx = 1; nInx < nArgs; ++nInx) {
		tstring	strArg;

		strcpy_T(strArg, papszArgList[nInx]);

		for (nOpt = 0; nOpt < nSize && _tcscmp(strArg.c_str(), pOptTable[nOpt].pszStr) != 0; ++nOpt);

		if (nOpt < nSize) {
			const OneCmdOption &	sOpt = pOptTable[nOpt];
			BYTE *					pOpt = reinterpret_cast<BYTE *>(pDest) + sOpt.nOffs;

			switch (sOpt.eType) {

			case cot_BOOL:
				*reinterpret_cast<bool *>(pOpt) = true;
				break;

			case cot_INT:
				if (++nInx < nArgs)
					*reinterpret_cast<int *>(pOpt) = _wtoi(papszArgList[nInx]);
				break;

			case cot_STR:
				if (++nInx < nArgs)
					strcpy_T(*reinterpret_cast<tstring *>(pOpt), papszArgList[nInx]);
			}
		} else if (pOthers)
			pOthers->push_back(strArg);
	}

	LocalFree(papszArgList);
}

// ====================================================================================
// Registre
// ====================================================================================

/// Ouverture d'une cl, dfinie par une cl racine et un chemin
bool CRegKey_tstr::Open(HKEY hParKey, LPCTSTR pszSubKey, REGSAM samDesired)
{
	if (!hParKey)
		return false;

	HKEY hNewKey;
	LONG nRes = RegOpenKeyEx(hParKey, pszSubKey, 0, samDesired, &hNewKey);

	if (nRes==ERROR_SUCCESS) {
		hKey = hNewKey;
		return true;
	}
	return false;
}

/// Lecture de la valeur d'index \p nInx prsente sous une cl
bool CRegKey_tstr::EnumKey(DWORD nInx, tstring & strRes) const
{
	DWORD	nSize = 0;
	DWORD	nIOSize;	// Variable d'change de longueurs de tampon (nombre de TCHARs)
	LONG	nRes;

	// Lecture du nom de la sous-cl demande, avec incrmentation automatique et
	// progressive de la taille du tampon si celui-ci est trop court
	do {
		nSize += 0x100;
		strRes.resize(nSize+1);	// +1 car la taille d'un 'tstring' n'inclut pas le terminateur ;
								// comme RegEnumKeyEx en inclut un, il faut bien l'crire
								// quelque part
		nIOSize = nSize;
		nRes = RegEnumKeyEx(hKey, nInx, &strRes[0], &nIOSize, NULL, NULL, NULL, NULL);
	} while (nRes==ERROR_MORE_DATA);

	if (nRes==ERROR_SUCCESS && nIOSize>0) {
		strRes[nIOSize] = 0;	// Terminateur (prcaution)
		nSize = nIOSize;
		strRes.resize(nSize);	// Rajuster prcisment la taille
		return true;
	}

	strRes.clear();
	return false;
}

/// Lecture d'une valeur chane
bool CRegKey_tstr::QueryStringValue(LPCTSTR pszName, tstring & strRes) const
{
	DWORD	nSize = 0;
	DWORD	nIOSize;	// Variable d'change de longueurs de tampon (nombre d'octets)
	DWORD	nType;
	LONG	nRes;

	// Lecture du nom de la valeur demande, avec incrmentation automatique et
	// progressive de la taille du tampon si celui-ci est trop court
	do {
		nSize += 0x100;
		strRes.resize(nSize+1);	// +1 car la taille d'un 'tstring' n'inclut pas le terminateur ;
								// comme RegQueryValueEx *peut* en ajouter un, il faut bien
								//  l'crire quelque part
		nIOSize = nSize*sizeof(TCHAR);
		nRes = RegQueryValueEx(hKey, pszName, NULL, &nType,
			reinterpret_cast<LPBYTE>(&strRes[0]), &nIOSize);
	} while (nRes==ERROR_MORE_DATA && nType==REG_SZ);

	if (nRes==ERROR_SUCCESS && nType==REG_SZ && nIOSize>=sizeof(TCHAR)) {
		nSize = nIOSize/sizeof(TCHAR);
		strRes[nSize] = 0;					// Terminateur
		strRes.resize(_tcslen(&strRes[0]));	// Rajuster prcisment la taille, sans utiliser
											// tstring::length car ceci incluerait le terminateur
		return true;
	}

	// En cas d'chec, ou bien si la cl demande ne correspond pas  une
	// chane de caractres, retourner une chane vide.
	strRes.clear();
	return false;
}

/// criture d'une valeur chane
bool CRegKey_tstr::SetStringValue(LPCTSTR pszName, LPCTSTR pszValue)
{
	return
		RegSetValueEx(hKey, pszName, 0, REG_SZ, (LPBYTE)pszValue,
			static_cast<DWORD>((_tcslen(pszValue)+1)*sizeof(TCHAR))) == ERROR_SUCCESS;
}
