/*
 * $Id$
 * Copyright  2010-2020 gingko - https://www.gingko.ovh/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See https://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentations de fonctions lies  mmorisation des missions
 *
 *	Ce fichier implmente les lments ncessaires au stockage et  la gestion
 *	des informations lies aux informations recueillies via l'EPG.
 **/

#include "stdafx.h"
#include "Filters\Misc\epgfilter.h"

/**
 * \brief Extraction d'une chane de l'EPG
 *
 * Les codes spciaux (codes 0x80  0x9f) sont nettoys,
 * ainsi que les espaces de dbut, de fin et multiples
 **/
static void get_epg_string(const MDESCR::S_str & sMs, tstring & strStr)
{
	TCHAR		str[256];

	sMs.get(str);

	LPCTSTR	src = str;
	LPTSTR	dst = str;
	UINT	ch;	// non sign !

	while ((ch=*src++) != 0) {
		switch (ch) {
		case DCTL_(CRLF):								// Saut de ligne
			ch = '\n';
			break;

		case '\t':										// Tabulation
			ch = ' ';									// remplacer par espace
			break;

		case DCTL_(EMPH_ON):							// Codes d'emphase : ignorer
		case DCTL_(EMPH_OFF):
			continue;

#ifdef _UNICODE
		default:
			if (ch>=DCTL_(MIN) && ch<=DCTL_(MAX)) {		// Autre code de contrle :
				// On va supposer que c'est une erreur de la chane TV, et le traiter comme
				// de l'encodage natif Windows (Windows 1252) (ce genre d'erreur a t constat).
				// En mode UNICODE, le caractre d'origine a juste t augment de la valeur
				// 0xE000, donc en pointer l'octet de poids faible doit permettre de le retrouver.
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, LPCSTR(&src[-1]), 1, LPWSTR(&ch), 1);
			}
#endif
		}

		*dst++ = (TCHAR)ch;
	}

	// Enlever les espaces de fin :
	while (dst > str && (unsigned)dst[-1]<= (unsigned)' ')
		--dst;
	*dst = 0;
	strStr += str;
}

/// Obtenir le nom de l'mission
/// \retval Nom de l'mission si valide, chane vide si non valide
tstring Emission::getName() const
{
	if (isOk())
		return strNom;
	return TEXT("");
}

static void append_trimmed(tstring & strRes, LPCTSTR pszStr)
{
	if (*pszStr) {
		int nTrimBeg = 0;
		int nTrimEnd = strlen_T(pszStr);

		while (nTrimEnd > 0 && (UINT)pszStr[nTrimEnd-1] <= ' ')
			--nTrimEnd;
		while (nTrimBeg < nTrimEnd && (UINT)pszStr[nTrimBeg] <= ' ')
			++nTrimBeg;
		if (nTrimBeg < nTrimEnd) {
			if (!strRes.empty())
				strRes += EOL EOL;
			strRes.append(pszStr+nTrimBeg, nTrimEnd-nTrimBeg);
		}
	}
}

/// Obtenir une description courte, limite  une ligne
const tstring & Emission::getShortDescr() const
{
	if (strSDCache.empty()) {
		#define MAX_SHORTDESC 160

		append_trimmed(strSDCache, strDesc.c_str());

		if (strSDCache.empty()) {
			TCHAR	szWrk[MAX_SHORTDESC+20];

			_tcsncpy_s(szWrk, _countof(szWrk), strExtDesc.c_str(), _countof(szWrk)-1);
			szWrk[_countof(szWrk)-1] = 0;
			append_trimmed(strSDCache, szWrk);
		}

		size_t nPos;

		// Nettoyage des fins de lignes :
		while ((nPos = strSDCache.find(EOL)) != strSDCache.npos)
			strSDCache[nPos] = ' ';
		// Nettoyage des espaces doubls (qui peuvent ventuellement rsulter du nettoyage prcdent) :
		while ((nPos = strSDCache.find(TEXT("  "))) != strSDCache.npos)
			strSDCache.erase(nPos, 1);
		// Couper au-del de 80 caractres, en mettant des points de suspension si coup :
		if (strSDCache.length() > MAX_SHORTDESC) {
			strSDCache.resize(MAX_SHORTDESC-3);
			strSDCache += TEXT("...");
		}
		// Si le rsultat est vide, mettre un espace pour qu'il ne soit pas calcul  nouveau :
		if (strSDCache.empty())
			strSDCache = TEXT(" ");
	}

	return strSDCache;
}

/// Obtenir une description longue, ventuellement multilignes
const tstring & Emission::getLongDescr(LPCSTR pszNomChaine) const
{
	if (strLDCache.empty()) {
		TCHAR szWrk[160];

		// Synthse de la date et de l'heure de dbut et de fin, de la manire la plus naturelle possible
		int nLen = _stprintf_s(szWrk, TEXT("Sur %") A2t TEXT(", le %") A2t TEXT(" %i%s %") A2t TEXT("  %ih%02i, fin "),
			pszNomChaine, EnumToString(debut.wDayOfWeek, pszDoW_lower), debut.wDay, debut.wDay==1 ? TEXT("er"): TEXT(""),
			EnumToString(debut.wMonth, pszMon_lower), debut.wHour, debut.wMinute);

		if (debut.wDay != fin.wDay) {
			nLen += _stprintf_s(szWrk+nLen, _countof(szWrk)-nLen, TEXT("le %") A2t TEXT(" %i%s "),
				EnumToString(fin.wDayOfWeek, pszDoW_lower), fin.wDay, fin.wDay==1 ? TEXT("er"): TEXT(""));
			if (debut.wMonth != fin.wMonth)
				nLen += _stprintf_s(szWrk+nLen, _countof(szWrk)-nLen, TEXT("%") A2t TEXT("i "),
					EnumToString(fin.wMonth, pszMon_lower));
		}
		_stprintf_s(szWrk+nLen, _countof(szWrk)-nLen, TEXT(" %ih%02i."), fin.wHour, fin.wMinute);

		strLDCache = szWrk;

		// Ajout de la description de base si diffrente du dbut de la description tendue :
		if (_tcsncmp(strDesc.c_str(), strExtDesc.c_str(), strDesc.length()) != 0)
			append_trimmed(strLDCache, strDesc.c_str());

		// Ajout de la description tendue :
		append_trimmed(strLDCache, strExtDesc.c_str());

		// Nettoyage de la chane
		int	nInx = 0;
		int	nMax = (int)strLDCache.length();

		while (nInx < nMax) {
			LPCTSTR	pszStr	= strLDCache.c_str();
			TCHAR	nCh		= pszStr[nInx];

			switch (nCh) {

			case ' ':
				if (nInx + 1 < nMax) {
					switch (pszStr[nInx+1]) {

					case ' ':
					case '\n':
						// Nettoyage des espaces en fin de lignes et des espaces doubls :
						strLDCache.erase(nInx, 1);
						--nMax;
						// La suppression d'un espace en fin de ligne peut gnrer une fin de
						// ligne  supprimer :
						while (nInx > 0 && strLDCache[nInx-1] == '\n')
							--nInx;
						continue;

					case ':':
					case ';':
					case '!':
					case '?':
					case TCHAR(L''):	// cast vers TCHAR pour viter les problmes de signe
						// Un espace suivi de certains caractres gagne  tre inscable
						strLDCache[nInx] = TCHAR(L'');	// -> code 0xa0, ce n'est pas un espace
					}
				}
				break;

			case '\n':
				if (nInx + 2 < nMax && pszStr[nInx+1] == '\n' && pszStr[nInx+2] == '\n') {
					// Nettoyage des sauts de ligne excdentaires :
					strLDCache.erase(nInx, 1);
					--nMax;
					continue;
				}
				break;

			case TCHAR(L''):	// cast vers TCHAR pour viter les problmes de signe
				if (nInx + 1 < nMax) {
					switch (pszStr[nInx+1]) {

					case ' ':
						// L'espace suivant '' doit tre inscable
						strLDCache[nInx+1] = TCHAR(L'');	// -> code 0xa0, ce n'est pas un espace
						break;
					}
				}
			}
			++ nInx;
		}

		if (nRating > 1) {
			// si nRating == 1, cela veut dire en principe 4 ans. Mais on ignore cette
			// valeur, car les usages constats ne semblent pas pertinents.
			_stprintf_s(szWrk, EOL EOL TEXT("Enfants: minimum %u ans."), nRating + 3);
			strLDCache += szWrk;
		}

		// Si le rsultat est vide, mettre un espace pour qu'il ne soit pas calcul  nouveau :
		if (strLDCache.empty())
			strLDCache = TEXT(" ");
	}

	return strLDCache;
}

/*
/// Constructeur
Emission::Emission() :
	nUtcBegT(0),
	nUtcEnd10(0),
	nEvtId(0),
	nContent(0),
	ixChaine(-1),
	bUsed(false),
	nFlags(0),
	nRating(0),
	cleChaine(0)
{
	SYSTEMTIME st = {0};

	debut = fin = st;
}
*/

/// Constructeur
Emission::Emission(const SI_EIT::entry & sEitEn, int ixch, UINT32 cle, BYTE flags) :
	EmissionBase(sEitEn),
	nContent(0),
	ixChaine(ixch),
	bUsed(false),
	nFlags(flags),
	nRating(0),
	cleChaine(cle)
{
	debut	= sEitEn.start_time;	// (format SYSTEMTIME)

	// transforme l'heure GMT du programme en local
	SystemTimeToLocalSystemTime(debut);

	// on ajoute la dure pour trouver l'heure de fin
	fin = debut;
	AddTime(fin, sEitEn.duration);

	// Ajouter le nom et le descriptif
	for (MDESCR::D_scanner sed(sEitEn.descrRange()); !sed; ++sed) {
		MDESCR::D_tag eTag = sed().tag;

		switch (eTag) {

		case MDESCR::mdt_Short_Event: {
			const MDESCR::D_4d & d4d = sed().as<MDESCR::D_4d>();

			// Extraction et nettoyage des chanes :
			get_epg_string(d4d.event_name(), strNom);
			if (strNom.empty())
				strNom = TEXT("<mission sans nom>");
			get_epg_string(d4d.text(), strDesc);
			break; }

		case MDESCR::mdt_Extended_Event: {
			// Attention, les "mdt_Extended_Event" peuvent tre multiples,
			// et les donnes qu'on en extrait doivent tre concatnes.
			const MDESCR::D_4e & d4e = sed().as<MDESCR::D_4e>();

			get_epg_string(d4e.text(), strExtDesc);
			break; }

		case MDESCR::mdt_Parental_Rating: {
			for (MDESCR::D_55::sdScanner s55 = sed().as<MDESCR::D_55>().subDescrRange(); !s55; ++s55) {
				// S'il y a plusieurs limites d'ge, on prend la plus leve.
				// On ne tient pas compte du code de pays pour le moment.
				if (nRating < s55().rating)
					nRating = s55().rating;
			}
			break; }

		case MDESCR::mdt_Content:	// Exprimental
			for (MDESCR::D_54::sdScanner s54 = sed().as<MDESCR::D_54>().subDescrRange(); !s54; ++s54) {
				if (nContent == 0) {
					nContent = MAKELONG(
						MAKEWORD(s54().content_nibble_level_1(), s54().content_nibble_level_2()),
						MAKEWORD(s54().user_nibble_1(), s54().user_nibble_2()));
					break;
				}
			}
			break;

#ifdef _DEBUG
		case MDESCR::mdt_Component:
			break;

		default:
			eTag = eTag;
#endif
		}
	}
}

/**
 * \brief Obtenir l'mission courante et la suivante, par rapport  une heure donne
 *
 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
 * \param[in]	bStrict	si \p true, les horaires de l'mission doivent strictement englober
 *				le temps courant ; si \p false, un nom d'mission non commence peut
 *				ventuellement tre retourn.
 * \retval		Structure contenant deux pointeurs (avec compteurs de rfrence),
 *				l'un sur l'mission en cours, l'autre sur l'mission suivante.
 *				Les deux pointeurs peuvent tre NULL si l'mission n'est pas dfinie.
 * \todo		tant donn que la spcification \p bStrict ne semble pas faire l'unanimit parmi
 *				les utilisateurs, tudier ventuellement un paramtre de configuration pour dcider.
 **/
ActAndNxtProg ChanEventList::getCurrentAndNext(const SYSTEMTIME & localtime, bool bStrict) const
{
	ActAndNxtProg	sRes;
	CAutoLock		cLock(crit_sec());

	for (ITERATE_CONST_LIST(*this, EmissionPtr, it)) {
		const Emission &	cEmis = **it;

		if (cEmis.isOk() && (cEmis.nFlags & DEF_IN_ACTUAL)!= 0 && DiffTime(localtime, cEmis.fin)<0) {
			if (!bStrict || DiffTime(localtime, cEmis.debut) >= 0)
				sRes.apEmi[0] = *it++;
			if (it != end())
				sRes.apEmi[1] = *it;
			break;
		}
	}
	return sRes;
}

/**
 * Obtenir le temps restant avant le prochain changement d'tat EPG concernant cette chane.
 *
 * \param[in]	pLocalTime	Pointeur sur temps courant (rcupr en interne si NULL)
 * \return		Temps restant, exprim en millisecondes, ou bien #DUREE_INFINIE si aucun temps
 *				ne peut tre dtermin (aucune mission dfinie dans l'EPG ...).
 **/
long ChanEventList::timeBeforeNextEvent(LPSYSTEMTIME pLocalTime) const
{
	SYSTEMTIME	localtime;
	long		nResult = DUREE_INFINIE;

	if (pLocalTime == NULL) {
		GetLocalTime(&localtime);
		pLocalTime = &localtime;
	}

	EmissionPtr	pEmi = getCurrentAndNext(*pLocalTime).apEmi[0];

	if (pEmi) {
		long nTempTime;

		nTempTime = DiffTime(pEmi->debut, *pLocalTime);
		if (nTempTime>0 && nTempTime<nResult) {
			nResult = nTempTime;
		} else {
			// Si le temps de dbut tait infrieur, on va prsumer que le temps de fin
			// ne l'est pas encore davantage ...
			nTempTime = DiffTime(pEmi->fin, *pLocalTime);
			if (nTempTime>0 && nTempTime<nResult)
				nResult = nTempTime;
		}
	}

	return nResult;
}

void ChanEventList::clear()
{
	CAutoLock cLock(crit_sec());

	__super::clear();
}

/**
 * Recherche, dans les informations d'une chane, de l'mission correspondant  une cl d'mission donne
 *
 * \param[in]	nEvtId	ID d'mission recherche
 * \return		Pointeur sur l'mission dans la liste, ou NULL si pas trouv
 **/
const Emission * ChanEventList::cherche_par_id(UINT16 nEvtId) const
{
	CAutoLock	cLock(crit_sec());

	// Retrouver l'item qui a l'heure de dbut qui nous intresse
	for (ITERATE_CONST_LIST(*this, EmissionPtr, it)) {
		const Emission & cEmis = **it;

		if (cEmis.nEvtId == nEvtId)
			return &cEmis;
	}
	return NULL;
}
