/*
 * $Id$
 * Copyright  2010-2020 gingko - https://www.gingko.ovh/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See https://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions de fonctions lies  mmorisation des missions
 *
 *	Ce fichier implmente les lments ncessaires au stockage et  la gestion
 *	des informations lies aux informations recueillies via l'EPG.
 **/

#include "mpeg2defs.h"

class EmissionBase
{
public:
	UINT64			nUtcBegT;	//!< Reprsentation simplifie de l'heure de dbut pour tests (0 si pas dfini)
	UINT64			nUtcEnd10;	//!< Heure d'expiration (= fin +10mn) pour tests
	UINT16			nEvtId;		//!< "event id"

	EmissionBase(const SI_EIT::entry & sEitEn) :
		nUtcBegT(sEitEn.start_time()),
		nEvtId(sEitEn.event_id())
	{
		nUtcEnd10	= nUtcBegT + sEitEn.duration + 10*(60+1000);	// Heure de fin + 10 minutes
	}
};

/// Structure dcrivant une mission pour l'EPG
class Emission : public EmissionBase
{
public:
	SYSTEMTIME		debut;		//!< Heure de dbut
	SYSTEMTIME		fin;		//!< Heure de fin
	tstring			strNom;		//!< Nom de l'mission (natif)
	tstring			strDesc;	//!< Description (native)
	tstring			strExtDesc;	//!< Description tendue (native)
private:
	mutable tstring	strSDCache;	//!< Description courte reformate (cache)
	mutable tstring	strLDCache;	//!< Description longue reformate (cache)
public:
	DWORD			nContent;	//!< Type de contenu \test Exprimental \todo Voir comment prsenter a
	mutable int		ixChaine;	//!< Index de la chane dans la table \p Canaux
	mutable bool	bUsed;		//!< Utilis pour marquer les missions dans l'EPG en vue de suppression
	BYTE			nFlags;		//!< #DEF_IN_ACTUAL / #DEF_IN_SCHEDULE (combinaison possible)
	UINT8			nRating;	//!< ge limite pour les mineurs, moins 3
	UINT32			cleChaine;	//!< Cl chane (pour retrouver une chane dont la position aurait chang)

	/// Constructeur
	Emission(const SI_EIT::entry & sEitEn, int ixch, UINT32 cle, BYTE flags);

	/// Vrifie si l'mission est valide
	/// \retval \p true si valide, \p false si non valide
	/// \deprecated Probablement plus ncessaire
	bool	isOk() const {
		return nUtcBegT != 0; }

	/// Obtenir le nom de l'mission
	/// \retval Nom de l'mission si valide, chane vide si non valide
	tstring getName() const;

	/// Tester si l'mission a reu un nom
	bool hasName() const {
		return isOk() && !strNom.empty(); }

	/// Obtenir une description courte, limite  une ligne
	const tstring & getShortDescr() const;

	/// Obtenir une description longue, ventuellement multilignes
	const tstring & getLongDescr(LPCSTR pszNomChaine) const;

	bool operator == (const Emission & e2) const {
			return
				nEvtId == e2.nEvtId || (
					cleChaine == e2.cleChaine &&
					nUtcBegT == e2.nUtcBegT &&
					nUtcEnd10 == e2.nUtcEnd10); }
};

/// Liste des missions rapportes par l'EPG
struct EmissionPtr : public shared_ptr<Emission>
{
	/// Constructeur
	EmissionPtr(Emission * pNewObj = NULL) :
		shared_ptr<Emission>(pNewObj)
	{
	}

	Emission & operator * () const {
				return *get(); }

	tstring getName(LPCTSTR pszDefault) const {
		if (*this)
			return (*this)->getName();
		return pszDefault; }

	/// Tester si l'mission a reu un nom
	bool hasName() const {
		return (*this) && (*this)->hasName(); }
};

struct  ActAndNxtProg
{
	EmissionPtr		apEmi[2];

	const Emission * get(int nEmis) {
		return apEmi[nEmis].get(); }
};

class ChanEventList : public std::list<EmissionPtr>
{
	/**
	 * \brief Section critique pour la synchronisation
	 *
	 * Sous forme de pointeur avec compteur de rfrence, car la liste entire est
	 * susceptible d'tre copie au moment de la construction de la liste des chanes.
	 **/
	mutable shared_ptr<CCritSec>	pCritSec;
public:

	ChanEventList() :
		pCritSec(new CCritSec)
	{
	}

	/// Obtenir un pointeur non const sur le descripteur de section critique
	CCritSec * crit_sec() const {
				return pCritSec.get(); }

	/**
	 * Recherche, dans les informations d'une chane, de l'mission correspondant  une cl d'mission donne
	 *
	 * \param[in]	nEvtId	ID d'mission recherche
	 * \return		Pointeur sur l'mission dans la liste, ou NULL si pas trouv
	 **/
	const Emission * cherche_par_id(UINT16 nEvtId) const;

	/**
	 * \brief Obtenir l'mission courante et la suivante, par rapport  une heure donne
	 * 
	 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
	 * \param[in]	bStrict	si \p true, les horaires de l'mission doivent strictement englober
	 *				le temps courant ; si \p false, un nom d'mission non commence peut
	 *				ventuellement tre retourn.
	 * \retval		Structure contenant deux pointeurs (avec compteurs de rfrence),
	 *				l'un sur l'mission en cours, l'autre sur l'mission suivante.
	 *				Les deux pointeurs peuvent tre NULL si l'mission n'est pas dfinie.
	 * \todo	tant donn que la spcification \p bStrict ne semble pas faire l'unanimit parmi
	 *			les utilisateurs, tudier ventuellement un paramtre de configuration pour dcider.
	 **/
	ActAndNxtProg getCurrentAndNext(const SYSTEMTIME & localtime, bool bStrict=false) const;

	/**
	 * \brief Obtenir l'mission courante et la suivante
	 * 
	 * \param[in]	bStrict	si \p true, les horaires de l'mission doivent strictement englober
	 *				le temps courant ; si \p false, un nom d'mission non commence peut
	 *				ventuellement tre retourn.
	 * \retval		Structure contenant deux pointeurs (avec compteurs de rfrence),
	 *				l'un sur l'mission en cours, l'autre sur l'mission suivante.
	 *				Les deux pointeurs peuvent tre NULL si l'mission n'est pas dfinie.
	 **/
	ActAndNxtProg getCurrentAndNext(bool bStrict=false) const {
					SYSTEMTIME localtime;
					GetLocalTime(&localtime);
					return getCurrentAndNext(localtime, bStrict); }

	/**
	 * Obtenir l'mission couramment en cours de diffusion
	 *
	 * \param[in]	bStrict	si \p true, les horaires de l'mission doivent strictement englober
	 *				le temps courant ; si \p false, un nom d'mission non commence peut
	 *				ventuellement tre retourn.
	 * \retval	Pointeur (avec compteurs de rfrence) sur l'mission
	 **/
	EmissionPtr getCurrentProgram(bool bStrict=false) const {
					return getCurrentAndNext(bStrict).apEmi[0]; }

	/**
	 * Obtenir le temps restant avant le prochain changement d'tat EPG concernant cette chane.
	 *
	 * \param[in]	pLocalTime	Pointeur sur temps courant (rcupr en interne si NULL)
	 * \return		Temps restant, exprim en millisecondes, ou bien #DUREE_INFINIE si aucun temps
	 *				ne peut tre dtermin (aucune mission dfinie dans l'EPG ...).
	 **/
	long timeBeforeNextEvent(LPSYSTEMTIME pLocalTime = NULL) const;

	/// Effacement de la liste
	void clear();
};
