/*
 * $Id: graph.h 348 2012-07-01 20:27:42Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions ncessaires  la contruction du graphe
 *
 *	Ce fichier implmente toutes les classes et fonctions ncessaires  la construction
 *	et  la gestion des graphes \p DirectShow grs par Pouchin TV Mod, dans toutes leurs
 *	variations.
 **/

#pragma once

#include "filters.h"
#include "rendering.h"

#if USE_PMTFILTER
#include "Filters\Misc\pmtfilter.h"
#endif

// ====================================================================================

extern CLSID MEDIASUBTYPE_DOLBY_DDPLUS;
extern CLSID MEDIASUBTYPE_MPEG_LOAS;

// ====================================================================================

/// Identifiants de types de filtres pour accs aux proprits
enum FilterType
{
	ft_Video_Mp2v,		//!< Vido MPEG2
	ft_Video_H264,		//!< Vido H264
	ft_Audio_Mp2a,		//!< Audio MPEG2
	ft_Audio_Ac3,		//!< Audio AC3
	ft_Audio_Eac3,		//!< Audio E-AC3
	ft_Audio_Aac,		//!< Audio AAC
	ft_AudioRenderer,	//!< Rendu audio
	ft_VideoRenderer	//!< Rendu vido
};

/// Slecteur de connexions aux sorties du dmultiplexeur
struct MappingData
{
	VideoStreamType eVideoType;	//!< Type de flux vido
	TrackStreamType eAudioType;	//!< Type de flux audio
	UINT16			video_pid;	//!< PID vido
	UINT16			audio_pid;	//!< PID audio
	UINT16			pmt_pid;	//!< PID PMT

	MappingData(UINT16 ppid=0, VideoStreamType eVTyp=vst_Unknown, TrackStreamType eATyp=tst_Unknown, UINT16 vpid=0, UINT16 apid=0) :
		eVideoType(eVTyp),
		eAudioType(eATyp),
		video_pid(vpid),
		audio_pid(apid),
		pmt_pid(ppid)
	{
	}

	// Retourne \p true si tous les PIDs (sauf PMT) sont  zro (impliquant suspension)
	bool noMediaPids() const {
			return video_pid == 0 && audio_pid == 0; }
};

// ====================================================================================
// Interfaces d'accs au graphe
// ====================================================================================

#pragma warning(disable: 4250)	// 'classe1' : hrite de 'classe2::membre' via la dominance

DeclareInterface(IPTvM_GraphBase)
	virtual HRESULT			Build() = 0;

	/**
	 * Traitement d'un vnement du graphe
	 * \param[in] hWnd Handle de la fentre principale
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			HandleGraphEvent(HWND hWnd) = 0;

	virtual HRESULT			AddEventNotifier(HWND hWnd, UINT uMsg) = 0;

	/**
	 * Ajout d'un filtre au graphe
	 *
	 * \param[in]	pFilter	Filtre	 ajouter
	 * \param[in]	pszName	Nom de ce filtre
	 **/
	virtual HRESULT			AddFilterToGraph(IBaseFilter * pFilter, LPCTSTR pszName) = 0;

	/**
	 * Cration et ajout d'une instance de filtre dfini par son "clsid" (le paramtre "nom"
	 * n'est ici qu'informatif,  l'usage des messages d'erreur et de l'exportation du graphe).
	 * \param[in,out]	sFData	Donnes d'aprs lesquelles crer le filtre
	**/
	virtual HRESULT			create_and_add_filter(FilterData & sFData) = 0;

	/**
	 * Connexion d'une broche de sortie  la premire broche d'entre disponible d'un filtre.
	 * \param[in] pOutPin	Broche de sortie  connecter
	 * \param[in] pFilter	Filtre prsentant la broche d'entre  connecter
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			connect_pin_to_filter(IPin * pOutPin, IBaseFilter * pFilter) = 0;

	/**
	 * Connexion de la premire broche de sortie d'un filtre  la premire broche
	 * d'entre d'un autre.
	 * \param[in] source	Filtre prsentant la broche de sortie  connecter
	 * \param[in] dest		Filtre prsentant la broche d'entre  connecter
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			connect_filters(IBaseFilter * source, IBaseFilter * dest) = 0;

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/**
	 * Dconnexion d'une broche de la broche correspondante du filtre adjacent.
	 * \param[in]	pPin	Broche  dconnecter
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			disconnect_pin(IPin * pPin) = 0;
#endif

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/**
	 * Dconnexion d'une broche d'un filtre de la broche correspondante du filtre adjacent.
	 * \param[in]	pFilter	Filtre prsentant la broche  dconnecter
	 * \param[in]	dir		Direction de la broche  dconnecter (entre ou sortie)
	 * \return Code d'erreur DirectShow
	 **/
	 virtual HRESULT		disconnect_filter(IBaseFilter * pFilter, PIN_DIRECTION dir) = 0;
#endif

	/**
	 * \brief Retrait d'un filtre du graphe.
	 * \param[in]	pFilter	Filtre  retirer
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			remove_filter(IBaseFilter * pFilter) = 0;

	/**
	 * Mise en route du graphe
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			graph_Run() = 0;

	/**
	 * Mise en pause du graphe
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			graph_Pause() = 0;

	/**
	 * Arrt du graphe
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			graph_Stop() = 0;

	/**
	 * Interrogation de l'tat du graphe
	 * \return Valeurs de l'numration FILTER_STATE (strmif.h)
	 * \retval 0 (State_Stopped) Le graphe est arrt
	 * \retval 1 (State_Paused) Le graphe est en mode "pause"
	 * \retval 2 (State_Running) Le graphe est en marche
	 * \retval -1 Le graphe ne s'est pas initialis correctement
	 **/
	virtual OAFilterState	graph_GetState() = 0;

	/**
	 * Vrifie si le graphe est en route
	 * \return \p true (en route), ou \p false (arrte)
	 * \note	Se base sur une variable interne  la classe, alors que
				\p graph_GetState interroge le graphe lui-mme.
	 **/
	virtual bool			graph_isRunning() const = 0;

#ifdef _DIAGNOSTIC
	virtual tstring			get_graph_state() = 0;
#endif
EndInterface

// ====================================================================================

DeclareBasedInterface(IPTvM_TunerGraph, virtual IPTvM_GraphBase)
	/**
	 * Syntonisation d'une frquence dans le tuner (bas niveau).
	 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
	 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
	 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
	 **/
	virtual HRESULT			LowLevelTune(LONG nFreq, LONG nBandwidth, tstring * pstrErr=NULL) = 0;

	/**
	 * Syntonisation d'une frquence dans le tuner.
	 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
	 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
	 * \param[in]	pChIDs		Autres informations optionnelles concernant la chane
	 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
	 **/
	virtual HRESULT			change_frequence(LONG nFreq, LONG nBandwidth, const ChaineIDs * pChIDs=NULL, tstring * pstrErr=NULL) = 0;

	/**
	 * Ajout d'un filtre  la fin du tronc principal du graphe.
	 * \param[in]	pFilter	Filtre  ajouter
	 * \param[in]	pszName	Nom qui lui sera attribu
	 * \return		Code d'erreur DirectShow
	 **/
	virtual HRESULT			PushFilter(IBaseFilter * pFilter, LPCTSTR pszName) = 0;

	/**
	 * Retrait du dernier filtre du tronc principal du graphe, avec restauration de l'tat prcdent.
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT			PopFilter() = 0;

	/**
	 * Identification du type de graphe.
	 * \retval	false	Normal
	 * \retval	true	Alternatif
	 **/
	virtual bool			isLegacyGraph() const = 0;

EndInterface

// ====================================================================================

DeclareBasedInterface(IPTvM_FullGraph, virtual IPTvM_TunerGraph)

	/**
	 * Changement de volume sonore
	 * \param[in]	volume	Nouveau volume sonore, valeur de 0  100
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT			set_volume(long volume) = 0;

	/**
	 * Branchement de tous les mdias de sortie d'une chane.
	 *
	 * \param[in]	sMData	Informations sur les mdias  connecter
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT			branche_medias(const MappingData & sMData) = 0;

	/**
	 * \brief Vrification de l'existence d'un dialogue de proprits pour filtre
	 *
	 * \param[in]	eFltType	Type du filtre pour lequel l'existence du dialogue doit tre vrifie
	 * \return	\p true si un dialogue de proprits existe pour ce filtre
	 * \note	La mthode cherchera  crer une instance du filtre s'il n'y en a pas encore
	 **/
	 virtual bool			check_property_dialog(FilterType eFltType) = 0;

	/**
	 * \brief Appel du dialogue de proprits d'un filtre
	 * \param[in]	eFltType	Type du filtre dont le dialogue de proprits doit tre appel
	 **/
	virtual void			create_property_dialog(FilterType eFltType) = 0;

	/// Obtenir l'interface du grabber
	virtual GrabberData &	GetGrabberInterface() = 0;

	/**
	 * Obtenir l'interface \p IMpeg2Data (pour la recherche de chanes)
	 **/
	virtual CComPtr<IMpeg2Data>
							GetMpeg2Data() = 0;

#if USE_PMTFILTER
	/**
	 * Obtenir l'interface \p IPMTMonitor (pour la surveillance des changements de PMT)
	 **/
	virtual CComPtr<IPMTMonitor>
							GetPMTMonitor() = 0;
#endif
EndInterface

// ====================================================================================
/// Pointeur d'accs global au graphe
// ====================================================================================

extern IPTvM_FullGraph *	pGraph;

// ====================================================================================
// Fonctions globales
// ====================================================================================

/// Construction d'un graphe partiel (pour essais tuners), selon le type de graphe requis
IPTvM_TunerGraph * new_TunerGraph(bool bLegacy);

/**
 * Construction et mise en route du graphe DirectShow
 * \return Code d'erreur DirectShow
 **/
HRESULT build_graph();

/**
 * Arrt du graphe avec nettoyage de toutes les ressources utilises par celui-ci
 **/
void clean_dshow(void);

/**
 * Changement de volume sonore
 * \param[in]	volume	Nouveau volume sonore, valeur de 0  100
 * \return	Code d'erreur DirectShow
 **/
HRESULT set_volume(long volume);

/**
 * Branchement de tous les mdias de sortie d'une chane.
 *
 * \param[in]	sMData	Informations sur les mdias  connecter
 * \return	Code d'erreur DirectShow
 **/
HRESULT branche_medias(const MappingData & sMData);

/**
 * Dbranchement de TOUS les mdias
 **/
void debranche_medias();

/**
 * \brief Appel du dialogue de proprits d'un filtre
 * \param[in]	eFltType	Type du filtre dont le dialogue de proprits doit tre appel
 **/
inline void create_property_dialog(FilterType eFltType)
{
	if (pGraph)
		pGraph->create_property_dialog(eFltType);
}

/**
 * \brief Vrification de l'existence d'un dialogue de proprits pour filtre
 *
 * \param[in]	eFltType	Type du filtre pour lequel l'existence du dialogue doit tre vrifie
 * \return	\p true si un dialogue de proprits existe pour ce filtre
 * \note	La fonction cherchera  crer une instance du filtre s'il n'y en a pas encore
 **/
inline bool check_property_dialog(FilterType eFltType)
{
	if (pGraph)
		return pGraph->check_property_dialog(eFltType);
	return false;
}

/**
 * Conversion d'un code d'erreur DirectShow en une chane de caractres affichable.
 *
 * \param[in]	hr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 **/
tstring DShowErrorToString(HRESULT hr);

/**
 * Conversion d'un code d'erreur DirectX en une chane de caractres affichable.
 *
 * \param[in]	hr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 **/
tstring DXErrorToString(HRESULT hr);

// ====================================================================================
// Qualit du signal
// ====================================================================================

/// Intervalle (recommand) pour l'chantillonnage de la qualit du signal
#define DUREE_SIGNAL_STATS 200

enum Stat_BOOL : BYTE {
	sb_NO	= 0,		///< Non
	sb_YES	= 1,		///< Oui
	sb_UNK	= 0xff		///< Inconnu (pas chantillonn ou erreur
};

#define STATS_UNKNOWN MAXLONG	///< Valeur numrique inconnue
#define SIGNAL_QUALITY_MINI 20	///< Qualit de signal minimale accepte

/// Informations de qualit du signal
class SignalStatistics
{
	static AssocElement	aSSBool[];

	Stat_BOOL	ePresent;	///< Prsence du signal
	Stat_BOOL	eLocked;	///< TRUE si signal verrouill
	LONG		nStrength;	///< Force du signal (selon tuner ....)
	LONG		nQuality;	///< Qualit du signal (0  100 %, en principe)

public:
	HRESULT		hr_1st;		///< 1er code d'erreur recontr
	BYTE		nErrStep;	///< flags pointant les indicateurs en erreur

	/// Constructeur :
	SignalStatistics() :
		ePresent(sb_UNK),
		eLocked(sb_UNK),
		nStrength(STATS_UNKNOWN),
		nQuality(STATS_UNKNOWN),
		hr_1st(S_OK),
		nErrStep(0)
	{
	}

	/// Dterminer si l'tat de prsence du signal valide la chane
	bool	isPresenceOk() const {
				return ePresent!=sb_NO || sTunProps.scan_ignore_presence;}

	/// Dterminer si l'tat de signal verrouill valide la chane
	bool	isLockedOk() const {
				return eLocked!=sb_NO || sTunProps.scan_ignore_lock;}

	/// Dterminer si l'tat de qualit du signal valide la chane
	bool	isQualityOk() const {
				return nQuality >= SIGNAL_QUALITY_MINI ||
					nQuality==STATS_UNKNOWN || sTunProps.scan_ignore_quality;}

	/// Chargement depuis l'interface de qualit du signal
	void	getStats();

	/// Chargement de la force et de la qualit du signal dans des barres de niveau.
	void	SetLevelBars(HWND hSBar, HWND hQBar) const;

	/// Retourne l'tat de prsence du signal sous forme de chane
	LPCTSTR	getPresentStr() const {
				return dword2str(aSSBool, ePresent);}

	/// Retourne l'tat de signal verrouill sous forme de chane
	LPCTSTR	getLockedStr() const {
				return dword2str(aSSBool, eLocked);}

	/// Renvoie une chane de caractres reprsentant la force du signal
	void	getStrengthStr(bool bPct, LPTSTR pszBuf, UINT32 nBufLen, LPCTSTR pszPctSign) const;

	/// Renvoie une chane de caractres reprsentant la force du signal
	void	getQualityStr(LPTSTR pszBuf, UINT32 nBufLen, LPCTSTR pszPctSign) const;
};

// ====================================================================================
// Chargement des tables SI
// ====================================================================================

/// Classe de soutien  la lecture des tables SI
class Mpeg2SectionsHandler
{
	CComPtr<IMpeg2Data>		pMpeg2Data;
	PID						_pid;
	TID						_tid;
	DWORD					_dwTimeout;
	CComPtr<ISectionList>	pSectionList;
	MPEG2_FILTER			filter;
	FOURCC					tableName;
	BYTE					nextSection;
	BYTE					lastSection;
public:
	Mpeg2SectionsHandler(CComPtr<IMpeg2Data> pMpeg2Dat) :
		pMpeg2Data(pMpeg2Dat)
	{
	}

	PSECTION				data;
	DWORD					length;
	bool					complete;		//!< true si toute la table a t lue avec succs
	//
	void Init(PID pid, TID tid, DWORD dwTimeout, FOURCC tblName);
	bool GetNextSection();
};

// ====================================================================================
// Utilitaire d'arrt temporaire du graphe
// ====================================================================================

class GraphSuspender
{
	IPTvM_GraphBase &	iGraph;
	bool				bRunning;
	bool				bCanRestart;
	HRESULT				hr;

public:
	/// Constructeur
	GraphSuspender(IPTvM_GraphBase & iGrph);

	/// Validation (requise) du redmarrage
	void Ok() {
			bCanRestart = true; }

	/// Destructeur
	~GraphSuspender();
};

// ====================================================================================
