/*
 * $Id: recprog.cpp 401 2020-10-24 06:38:08Z gingko $
 * Copyright (C) 2007 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion des enregistrements programms (implmentations)
 *
 *	Ce fichier contient l'implmentation des fonctionnalits
 *	ncessaires  la programmation des enregistrements.
 **/

#include "stdafx.h"
#include "resource.h"
#include "dlgutils.h"
#include "recprog.h"
#include "channels.h"
#include "record.h"
#include <Lmcons.h>	// pour UNLEN

/**
 * \brief Dlai entre arrts de l'enregistrement et de l'application
 *
 * En cas d'arrt programm de l'application et/ou de l'ordinateur  l'issue d'un enregistrement,
 * le dlai suivant est appliqu avant l'excution effective de l'arrt, afin de garantir une
 * terminaison propre de l'enregistrement.
 **/
#define			DUREE_QUIT					2000	// Valeur en millisecondes

/**
 * \brief  Marge pour le calcul des recouvrements de programmations
 *
 * Deux programmations doivent tre spares par au moins ce temps
 * (en millisecondes) pourqu'il soit considr qu'elles ne se superposent pas.
 **/
#define			DUREE_MARGE_RECOUVREMENT	2500	// Valeur en millisecondes

/// Liste des enregistrements programms
ProgramList		Programmes;

/// Waitable timer (permettant ici de faire sortir du mode veille)
static CWaitTimer	cWaitTimer;

///**
// * \brief Obtenir une frquence  partir d'une cl
// *
// * On passe une cl de chane en paramtre,
// * et on retourne la frquence du multiplex correspondant
// **/
//static INT32 getFrequence_par_cle(UINT32 cleChaine)
//{
//	int ixChaine = Canaux.trouve_par_cle(cleChaine);
//
//	if (ixChaine >= 0)
//		return Canaux[ixChaine].khz;
//
//	return -1;
//}

#if USE_TASKMAN_PASSWORD
/**
 * \brief Classe de gestion de la bote de dialogue de la saisie du mot de passe
 *
 * Ce mot de passe est requis lors de l'ajout d'une tche dans le planificateur de tches
 **/
class CPaswdDialog : public CModalDialogBase
{
	LPWSTR	pszPassword;
	UINT32	nPwBufSize;

	/// Traitement des messages de la bote de dialogue
	virtual INT_PTR DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
	/// Constructeur :
	CPaswdDialog(LPWSTR	pszPasswrd, UINT32 nPwBufSiz) :
		CModalDialogBase(IDD_PASWD),
		pszPassword(pszPasswrd),
		nPwBufSize(nPwBufSiz)
	{
	}
};

/// Traitement des messages de la bote de dialogue
INT_PTR CPaswdDialog::DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG:
		return TRUE;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {

		case IDOK: {
			TCHAR buffer[128];

			*reinterpret_cast<LPWORD>(&buffer[0]) = _countof(buffer);

			UINT nLen = (UINT)SendDlgItemMessage(hDlg, IDC_PASWD, EM_GETLINE, 0, (LPARAM)buffer);

			buffer[nLen] = 0;	// (selon la doc, EM_GETLINE n'ajouterait pas toujours le zro terminateur)
			strcpy_T(pszPassword, nPwBufSize, buffer);
			EndDialog(hDlg, 0);
			return TRUE; }

		case IDCANCEL:
			EndDialog(hDlg, 1);
			return TRUE;
		}
		return FALSE;

	default:
		return FALSE;
	}
}
#endif	// #if USE_TASKMAN_PASSWORD

#if USE_TASKMAN_PASSWORD
/**
 * Saisie du mot de passe requis lors de l'ajout d'une tche dans le planificateur de tches
 **/
bool CRecordDialog::taskLogon(LPCWSTR username)
{
	// Vrifie le login, et demande le mot de passe si besoin est
	bool identifie = false;

	HANDLE session;
	do {
		identifie = LogonUserW(username, NULL, szPassword,
			LOGON32_LOGON_NETWORK, LOGON32_PROVIDER_DEFAULT, &session)==TRUE;
		if (identifie)
			break;

		// Saisie du mot de passe, et reboucler si pas d'annulation :
		szPassword[0] = 0;	// (vider la saisie prcdente)
	} while (CPaswdDialog(szPassword, _countof(szPassword)).Do(hDlg) == 0);

	CloseHandle(session);

	return identifie;
}
#endif

/**
 * Tableau de chanes affichables correspondant aux diffrents tats possibles
 * d'un enregistrement programm
 **/
const LPCTSTR Programme::t_states[] =
{
	TEXT("Attente"),		//!< epr_actif
	TEXT("En cours"),		//!< epr_encours
	TEXT("Dsactiv"),		//!< epr_inactif
	TEXT("Interrompu"),		//!< epr_interrompu
	TEXT("Supplant"),		//!< epr_supplante
	TEXT("Termin"),		//!< epr_termine
	TEXT("Erreur"),			//!< epr_erreur
	TEXT("Erreur fichier")	//!< epr_fileerr
};

/// Constructeur
Programme::Programme()
{
	apres = apr_rien;
	memset(&debut, 0, LPBYTE(this)+sizeof(*this)-LPBYTE(&debut));
	etat = epr_inactif;
}


#ifdef _XMLUTILS_H_INCLUDED_

/**
 * Extraire la date et l'heure d'un noeud XML
 * \param[in]	pXmlElm		lment (parent) des tags recherchs
 * \param[in]	pszTagList	List de tags  utiliser
 * \return		Date et heure retourns
 **/
SYSTEMTIME Programme::XmlToTime(const TiXmlElement * pXmlElm, LPCSTR pszTagList)
{
	SYSTEMTIME	sTime = {0};

	for (ITERATE_XML_ELEMENTS(pXmlElm, pXmlChild)) {
		LPCSTR	pszTag = pXmlChild->Value();
		LPCSTR	pszTxt = pXmlChild->GetText();

		if (pszTxt == NULL)
			pszTxt = "0";

		// Conversion en numrique pour les champs qui le requirent.
		TagListTime eTag = (TagListTime)StringToEnum(pszTagList, pszTag);

		if (eTag >= tlt_Annee && eTag <= tlt_Seconde)
			reinterpret_cast<WORD *>(&sTime)[eTag] = (WORD)atol(pszTxt);
	}
	return sTime;
}

/**
 * Gnrer le noeud XML correspondant  une structure date et heure
 * \param[in]	sTime	Date et heure  inclure
 * \param[in]	pszTag	Tag du noeud parent
 * \return		Noeud XML gnr
 **/
TiXmlElement * Programme::TimeToXml(SYSTEMTIME sTime, LPCSTR pszTag)
{
	TiXmlElement *	pXmlTime(new TiXmlElement(pszTag));
	CHAR			szTmp[16];

	if (pXmlTime) {
		for (int eTag = tlt_Annee; eTag <= tlt_Seconde; ++eTag) {
			pszTag = EnumToString(eTag, pszTimeTagStrList);
			_itoa_s(reinterpret_cast<WORD *>(&sTime)[eTag], szTmp, 10);
			SetXmlContent(pXmlTime, pszTag, szTmp);
		}
	}
	return pXmlTime;
}

/**
 * Extraire les informations de rptitions d'un noeud XML
 * \param[in]	pXmlElm		lment (parent) des tags recherchs
 **/
BYTE Programme::XmlToRepeat(const TiXmlElement * pXmlElm)
{
	BYTE nResult = 0;

	for (ITERATE_XML_ELEMENTS(pXmlElm, pXmlChild)) {
		LPCSTR	pszTag	= pXmlChild->Value();
		LPCSTR	pszTxt	= pXmlChild->GetText();

		if (pszTxt == NULL)
			pszTxt = "0";

		int		nDoW	= StringToEnum(pszDoW_ucfirst, pszTag);

		if (nDoW>= 0 && atol(pszTxt))
			nResult |= 1<<nDoW;
	}
	return nResult;
}

/**
 * Gnrer le noeud XML correspondant  variable de rptition hebdomadaire
 * \param[in]	nRep	Flags contenant les jours de rptition
 * \param[in]	pszTag	Tag du noeud parent
 * \return		Noeud XML gnr
 **/
TiXmlElement * Programme::RepeatToXml(BYTE nRep, LPCSTR pszTag)
{
	TiXmlElement *	pXmlRept(new TiXmlElement(pszTag));

	if (pXmlRept) {
		for (int j=0; j<7; j++) {
			SetXmlContent(pXmlRept, EnumToString(j, pszDoW_ucfirst),
				(nRep & (1<<j))!=0 ? "1" : "0");
		}
	}
	return pXmlRept;
}

/// Noms des tags XML utiliss pour la sauvegarde d'une programmation
LPCSTR Programme::pszTagStrList =
	"Nom\0"				// tlp_Nom
	"NomAuto\0"			// tlp_NomAuto
	"Chaine\0"			// tlp_Chaine
	"ServiceID\0"		// tlp_ServiceID
	"Cl\xC3\xA9\0"		// tlp_Cle			("Cl" (UTF-8))
	"D\xC3\xA9""but\0"	// tlp_Debut		("Dbut" (UTF-8))
	"Fin\0"				// tlp_Fin
	"Methode\0"			// tlp_Methode
	"Audio\0"			// tlp_Audio
	"R\xC3\xA9""veil\0"	// tlp_Reveil		("Rveil" (UTF-8))
	"After\0"			// tlp_After
	"Tache\0"			// tlp_Tache
	"Etat\0"			// tlp_Etat
	"Repetitions\0"		// tlp_Repetitions
	"S_Annee\0"			// tlp_S_Annee		(compatibilit anciennes versions)
	"E_Annee\0";		// tlp_E_Annee		(compatibilit anciennes versions)


/// Noms des tags XML utiliss pour la sauvegarde de l'heure et de la date
LPCSTR Programme::pszTimeTagStrList =
	"Annee\0"
	"Mois\0"
	"Jour_Semaine\0"
	"Jour\0"
	"Heure\0"
	"Minute\0"
	"Seconde\0";

/**
 * Chargement d'une programmation  partir d'un noeud TinyXML construit
 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
 * \return		\p true ou \p false selon succs ou chec
 **/
bool Programme::SetFromXMLElement(const TiXmlElement * pXmlElm)
{
	// Anciens noms des tags heure et date de dbut d'enregistrement
	static LPCSTR pszOldSTimeTagStrList =
		"S_Annee\0"
		"S_Mois\0"
		"S_Jour_semaine\0"
		"S_Jour\0"
		"S_Heure\0"
		"S_Minute\0"
		"S_Seconde\0";

	// Anciens noms des tags heure et date de fin d'enregistrement
	static LPCSTR pszOldETimeTagStrList =
		"E_Annee\0"
		"E_Mois\0"
		"E_Jour_semaine\0"
		"E_Jour\0"
		"E_Heure\0"
		"E_Minute\0"
		"E_Seconde\0";

	for (ITERATE_XML_ELEMENTS(pXmlElm, pXmlChild)) {
		LPCSTR	pszTag = pXmlChild->Value();
		LPCSTR	pszTxt = pXmlChild->GetText();

		if (pszTxt == NULL)
			pszTxt = "0";

		// Conversion en numrique pour les champs qui le requirent.
		UINT32	nVal = strtoul(pszTxt, NULL, 0);	// Gre les valeurs en 0xNN

		switch (StringToEnum(pszTagStrList, pszTag)) {
		case tlp_Nom:			strcpy_T(nom, ConvertUTF82Char(pszTxt).c_str());				break;
		case tlp_NomAuto:		nomAuto			= nVal !=0;										break;
		case tlp_Chaine:		nNumeroChaine	= (WORD)nVal;									break;
		case tlp_ServiceID:		sidChaine		= (WORD)nVal;									break;
		case tlp_Cle:			cleChaine		= nVal;											break;
		case tlp_Debut:			debut			= XmlToTime(pXmlChild, pszTimeTagStrList);		break;
		case tlp_Fin:			fin				= XmlToTime(pXmlChild, pszTimeTagStrList);		break;
		case tlp_Methode:		methode			= (MethodeEnregistrement)nVal;					break;
		case tlp_Reveil:		reveil			= nVal != 0;									break;
		case tlp_Audio:			audio			= (AudioMode)nVal;								break;
		case tlp_After:			apres			= (ApresEnregistrement)nVal;					break;
		case tlp_Tache:			tache			= (PlanificateurTache)nVal;						break;
		case tlp_Etat:			etat			= (EtatProgramme)nVal;							break;
		case tlp_Repetitions:	repetition		= XmlToRepeat(pXmlChild);						break;
		// compatibilit anciennes versions :
		case tlp_S_Annee:		debut			= XmlToTime(pXmlElm, pszOldSTimeTagStrList);	break;
		case tlp_E_Annee:		fin				= XmlToTime(pXmlElm, pszOldETimeTagStrList);
		}
	}

	// Ajustement de compatibilit permettant de rcuprer correctement les programmations
	// faites avec les anciennes versions qui ne sauvegardaient pas le SID.
	if (sidChaine==0)
		sidChaine = indexToSID(Canaux.trouve_par_no(nNumeroChaine));

	// Ajustement de compatibilit permettant de rcuprer correctement les programmations
	// faites avec les anciennes versions qui ne sauvegardaient pas la cl.
	if (cleChaine==0)
		cleChaine = indexToCle(Canaux.trouve_par_sid(sidChaine));

	return true;
}

/**
 * Construction d'un noeud TinyXML reprsentant la dfinition de la programmation
 * \return	Noeud XML construit
 **/
TiXmlElement * Programme::GetAsXMLElement() const
{
	TiXmlElement *	pXmlProg(new TiXmlElement("Programme"));

	if (pXmlProg) {
		for (int eTag = tlp_Nom; eTag <= tlp_Repetitions; ++eTag) {
			CHAR			szTmp[_countof(nom)*2] = "0";	// * 2 car la conversion en UTF-8 peut faire augmenter la longueur, en
															// principe dans cette limite, bien qu'il soit fort improbable d'aller jusque l.
			LPCSTR			pszTag = EnumToString(eTag, pszTagStrList);
			TiXmlElement *	pXmlElm = NULL;

			switch (eTag) {
			case tlp_Nom:			strcpy_T(szTmp, ConvertChar2UTF8(nom).c_str());		break;
			case tlp_NomAuto:		_itoa_s(nomAuto, szTmp, 10);						break;
			case tlp_Chaine:		_itoa_s(nNumeroChaine, szTmp, 10);					break;
			case tlp_ServiceID:		_itoa_s(sidChaine, szTmp, 10);						break;
			case tlp_Cle:			sprintf_s(szTmp, "0x%08X", cleChaine);				break;
			case tlp_Debut:			pXmlElm = TimeToXml(debut, pszTag);					break;
			case tlp_Fin:			pXmlElm = TimeToXml(fin, pszTag);					break;
			case tlp_Methode:		_itoa_s(methode, szTmp, 10);						break;
			case tlp_Reveil:		_itoa_s(reveil, szTmp, 10);							break;
			case tlp_Audio:			_itoa_s(audio, szTmp, 10);							break;
			case tlp_After:			_itoa_s(apres, szTmp, 10);							break;
			case tlp_Tache:			_itoa_s(tache, szTmp, 10);							break;
			case tlp_Etat:
				// L'tat 'epr_encours' n'est pas sauvegardable
				_itoa_s(etat == epr_encours ? epr_actif : etat, szTmp, 10);				break;
			case tlp_Repetitions:	pXmlElm = RepeatToXml(repetition, pszTag);
			}

			if (pXmlElm)
				pXmlProg->LinkEndChild(pXmlElm);
			else
				SetXmlContent(pXmlProg, pszTag, szTmp);
		}
	}
	return pXmlProg;
}

#endif	// #ifdef _XMLUTILS_H_INCLUDED_

/// Retourne \p true si intersection temporelle avec la programmation \p p2
bool Programme::overlap(const Programme & p2) const
{
	// avec 1000 millisecondes de marge
	return
		DiffTime(fin, p2.debut) >= DUREE_MARGE_RECOUVREMENT &&
		DiffTime(p2.fin, debut) >= DUREE_MARGE_RECOUVREMENT;
}

/**
 * \brief Comparaison en vue de tri sur l'heure de dbut
 *
 * Utilis par la fonction std::sort
 * \retval \p true si cette programmation se classe avant la programmation \p p2 dans le temps.
 **/
bool Programme::operator < (const Programme & p2) const
{
	return DiffTime(debut, p2.debut)<0;
}

/// Obtenir le nom de la chane
tstring	Programme::nomChaine() const
{
	TCHAR	szWrk[64]	= TEXT("??");
	int		ixChaine	= Canaux.trouve_par_cle(cleChaine);

	if (ixChaine_ok(ixChaine))
		strcpy_T(szWrk, Canaux[ixChaine].nom);
	return szWrk;
}

/// Obtenir une chane de caractres avec le nom de la programmation et le nom de la chane
tstring	Programme::progDescr() const
{
	TCHAR	szWrk[128]	= TEXT("");

	_stprintf_s(szWrk, TEXT("%s sur %s"), nom, nomChaine().c_str());
	return szWrk;
}

/**
 * \brief Calcul du temps restant avant dbut enregistrement en millisecondes.
 *
 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
 **/
long Programme::temps_restant(const SYSTEMTIME & localtime) const
{
	if (!active() || DiffTime(localtime, fin)>=0)
		return DUREE_INFINIE; // La programmation est inactive ou l'heure de fin est dpasse

	long reste = DiffTime(debut, localtime);

	return max(reste, 0);
}

/**
 * \brief Ajustement des donnes si rptitions programmes et horaire de fin dpass
 *
 * \param[in]	localtime	Heure de rfrence (en principe l'heure systme)
 * \retval		\p true si un ajustement a eu lieu
 **/
bool Programme::ajustementDesRepetitions(const SYSTEMTIME & localtime)
{
	bool res = false;

	if (repetition != 0 && etat != epr_inactif) {
		// En cas de rptition, tous les enregistrements non explicitement marqus comme
		// "inactifs" sont reconduits et ractivs ds lors que l'heure de fin est dpasse

		while (DiffTime(localtime, fin) >= 0) {	// Tant que l'heure de fin est dpasse
			int nbJours;

			// Trouver le nombre de jours avant la rptition suivante :
			for (nbJours=1; (repetition & (1<<((debut.wDayOfWeek + nbJours) % 7)))==0; nbJours++);

			long to_add = nbJours * DUREE_24H;	// Conversion en millisecondes

			AddTime(debut, to_add);
			AddTime(fin, to_add);

			/// \todo Il faudrait vrifier d'une faon ou d'une autre qu'une transition
			/// heure d't <-> heure d'hiver intempestive ne risque pas d'introduire un
			/// dcalage d'une heure dans la programmation  cet endroit.

			etat = epr_actif;	// ractivation
			programmation_modifiee = true;
			res = true;

			// Note : on revrifie en bouclant sur le test de diffrence de temps pour parer au cas
			// o celui-ci serait effectu alors que le programme n'a pas t excut pendant un dlai
			// excdant plusieurs rptitions conscutives.
		}
	}
	return res;
}

/**
 * Affichage du message d'avertissement en cas de conflit dans une programmation
 * \param[in]	pszMessage	Texte du message  afficher
 * \return		\p true ou \p false selon que l'utilisateur souhaite nanmoins poursuive ou pas
 **/
bool Programme::warn_prog(LPCTSTR pszMessage)
{
	tstring strMsg(pszMessage);

	strMsg += EOL TEXT("Voulez-vous quand mme effectuer cette programmation ?");

	if (MessageBox(hMainWnd, strMsg.c_str(), TEXT("Avertissement"), MB_YESNO|MB_ICONWARNING)==IDNO)
		return false;
	return true;
}

/**
 * \brief Vrification de la validit de la programmation
 *
 * \param[in]	prog_modif	Si non NULL, pointeur sur la programmation qu'on modifie (pour
 *							viter de la comparer  elle-mme lors des tests de recouvrement)
 * \retval		\p true ou \p false selon succs ou chec, respectivement
 **/
bool Programme::verifie(const Programme * prog_modif) const
{
	int			ixChaine = Canaux.trouve_par_cle(cleChaine);

	if (!ixChaine_ok(ixChaine)) {
		affiche_erreurs(TEXT("La chane est incorrecte"));
		return false;
	}

	SYSTEMTIME	localtime;

	GetLocalTime(&localtime);

	long duree = DiffTime(fin, debut);

	if (duree < 0) {
		affiche_erreurs(TEXT("Le programme commence aprs avoir fini !"));
		return false;
	}

	if (duree == 0) {
		affiche_erreurs(TEXT("La dure de l'enregistrement est nulle !"));
		return false;
	}

	if (duree > 12*DUREE_1H) {
		affiche_erreurs(TEXT("La dure de l'enregistrement dpasse 12 heures !"));
		return false;
	}

	if (DiffTime(fin, localtime) < 0 &&
		!warn_prog(TEXT("L'heure de fin du programme est dpasse !")) )
		return false;

	if (nom[0] == 0) {
		affiche_erreurs(TEXT("Veuillez donner un nom  ce programme"));
		return false;
	}

	// Rfrencer la chane concerne :
	const Chaine &	canal = Canaux[ixChaine];

	if (methode==meth_PS && canal.video_type!=vst_MPEG2) {
		affiche_erreurs(
			TEXT("L'enregistrement d'une chane H264 (cas de la haute dfinition) ")
			TEXT("en mode PS n'est pas implment"));
		return false;
	}

	for (ITERATE_CONST_VECTOR(Programmes, Programme, it)) {
		if (prog_modif && (&*it == prog_modif))
			continue;	// Ne pas comparer la programmation  celle en cours de modification

		if (_tcscmp(it->nom, nom) == 0) {
			affiche_erreurs(TEXT("Ce nom est dj utilis"));
			return false;
		}

		if (!prog_modif || !prog_modif->active() || !it->active())
			continue;	// La programmation modifie ou bien l'autre programmation n'est pas active
		if (!overlap(*it))
			continue;	// Il n'y a pas de chevauchement des horaires

		// Chevauchement des horaires :

		// Vrifier qu'on n'a pas programm plusieurs fois la mme chane au mme moment :
		if (
			cleChaine == it->cleChaine &&					// mme chane ?
			isMultiplexProg() == it->isMultiplexProg() &&	// mme type ? (multiplex ou pas multiplex)
			!warn_prog(
				tstr_printf(
					TEXT("Chevauchement des horaires avec \"%s\", qui concerne la mme chane !"),
					it->nom).c_str()) )
			return false;

		int	ixChaine2 = Canaux.trouve_par_cle(it->cleChaine);

		if (!ixChaine_ok(ixChaine2))
			continue;	// L'autre chane est incorrecte : on saute

		// Rfrencer l'autre chane :
		const Chaine &	canal2 = Canaux[ixChaine2];

		// Vrifie que le programme chevauchant est sur le mme multiplex :
		if (
			!canal.sameMultiplex(canal2) &&
			!warn_prog(
				tstr_printf(
					TEXT("Chevauchement des horaires avec \"%s\" sur %") A2t
					TEXT(", qui ne partage pas le mme multiplex !"),
					it->nom,
					Canaux[ixChaine2].nom).c_str()) )
			return false;
	}

	return true;	// pas d'erreur
}

/**
 * \brief Gnration du nom de tche programme correspondant
 *
 * \param[in]	pstr	Pointeur sur le tampon qui va recevoir le nom
 * \param[in]	bufSize	Taille de ce tampon
 **/
void Programme::genereNomTache(LPWSTR pstr, UINT bufSize) const
{
	TCHAR	szTmp[256];

	if (*prefix_conf)
		_stprintf_s(szTmp, TEXT("PTV[%s]-%s") , prefix_conf, nom);
	else
		_stprintf_s(szTmp, TEXT("PTV-%s"), nom);

	// Remplace les caractres incorrects :
	NomProtege<TCHAR, _countof(szTmp)> nom_tache(szTmp, TCHAR('_'));

	// On convertit le nom de la tche en Unicode (s'il ne l'est pas dj) :
	nom_tache.Extract(pstr, bufSize);
}

#if USE_TASKMAN_PASSWORD
/**
 * \brief Ajout de cette programmation aux tches programmes
 *
 * \param[in]	username	Nom d'utilisateur
 * \param[in]	motdepasse	Mot de passe
 * \retval		\p true ou \p false selon succs ou chec, respectivement
 **/
bool Programme::ajouterTacheProgrammee(LPCWSTR username, LPCWSTR motdepasse) const
#else
/**
 * \brief Ajout de cette programmation aux tches programmes
 *
 * \retval		\p true ou \p false selon succs ou chec, respectivement
 **/
bool Programme::ajouterTacheProgrammee() const
#endif
{
	WCHAR			taskName[_countof(nom)+4];
	SYSTEMTIME		date = debut;
	CComPtr<ITask>	pTask;
	HRESULT			hr;

	AddTime(date, -DUREE_PLANIF);	// Soustraire le temps de dcalage

	// On gnre le nom de la tche :
	genereNomTache(taskName, _countof(taskName));

	{
		CComPtr<ITaskScheduler> pITS;

		hr = pITS.CoCreateInstance(CLSID_CTaskScheduler, NULL, CLSCTX_INPROC_SERVER);
		if (FAILED(hr)) {
			myprintf(TEXT("Erreur init task scheduler") EOL);
			return false;
		}
		// Supprime la tche si elle existe
		hr = pITS->Delete(taskName);
		check_hr(hr, "ITaskScheduler::Delete");

		// Cre la nouvelle tche
		hr = pITS->NewWorkItem(taskName, CLSID_CTask, IID_ITask, (IUnknown**)&pTask);
	}

	if (FAILED(hr)) {
		myprintf(TEXT("Erreur cration tche %s") EOL, taskName);
		return false;
	}

	wchar_t szTmp[MAX_PATH];

	// Dfinition de la tche
	GetModuleFileNameW(NULL, szTmp, _countof(szTmp));
	hr = pTask->SetApplicationName(szTmp);
	check_hr(hr, "ITask::SetApplicationName");

	swprintf_s(szTmp, L"--minimize%" T2w, GetInstancePrefix().c_str());
	hr = pTask->SetParameters(szTmp);
	check_hr(hr, "ITask::SetParameters");

	strcpy_T(szTmp, pouchindir_curr);
	hr = pTask->SetWorkingDirectory(szTmp);
	check_hr(hr, "ITask::SetWorkingDirectory");

#if USE_TASKMAN_PASSWORD
	hr = pTask->SetAccountInformation(username, motdepasse);
#else
	hr = pTask->SetAccountInformation(L"", NULL);
#endif
	check_hr(hr, "ITask::SetAccountInformation");

	if (SUCCEEDED(hr)) {
		WORD					piNewTrigger;
		CComPtr<ITaskTrigger>	pTaskTrigger;

		hr = pTask->SetFlags(TASK_FLAG_DELETE_WHEN_DONE | TASK_FLAG_SYSTEM_REQUIRED | TASK_FLAG_RUN_ONLY_IF_LOGGED_ON);
		check_hr(hr, "ITask::SetFlags");
		hr = pTask->SetMaxRunTime(INFINITE);
		check_hr(hr, "ITask::SetMaxRunTime");
		hr = pTask->CreateTrigger(&piNewTrigger, &pTaskTrigger);
		check_hr(hr, "ITask::CreateTrigger");

		TASK_TRIGGER Trig = {
			sizeof(Trig),			// cbTriggerSize
			0,						// Reserved1
			date.wYear,				// wBeginYear
			date.wMonth,			// wBeginMonth
			date.wDay,				// wBeginDay
			0,						// wEndYear
			0,						// wEndMonth
			0,						// wEndDay
			date.wHour,				// wStartHour
			date.wMinute,			// wStartMinute
			7,						// MinutesDuration		(ressayer toutes les deux minutes
			2,						// MinutesInterval		 pendant sept minutes)
			0,						// rgFlags
			TASK_TIME_TRIGGER_ONCE	// TriggerType (Pas de rptition de la tche, par dfaut)
			// le reste  zro (implicite)
		};

		if (repetition!=0) {
			// On a demand  ce que la tche soit rpte
			Trig.TriggerType = TASK_TIME_TRIGGER_WEEKLY;

			WEEKLY & jours = Trig.Type.Weekly;

			jours.WeeksInterval = 1; // Rpte toutes les semaines
			jours.rgfDaysOfTheWeek = 0;

			static const BYTE repFlgs[7] =
				{TASK_SUNDAY, TASK_MONDAY, TASK_TUESDAY, TASK_WEDNESDAY,
				TASK_THURSDAY, TASK_FRIDAY, TASK_SATURDAY};

			for (int i=0; i<7; i++) {
				if ((repetition & (1<<i))!=0)
					jours.rgfDaysOfTheWeek |= repFlgs[i];
			}
		}

		hr = pTaskTrigger->SetTrigger(&Trig);
		check_hr(hr, "ITask::SetTrigger");

		CComQIPtr<IPersistFile> pPersistFile(pTask);

		if (pPersistFile) {
			hr = pPersistFile->Save(NULL, TRUE);
			check_hr(hr, "IPersistFile::Save");
		}
	}

	return true;
}

/**
 * \brief Ajout de la programmation si ncessaire, incluant identification
 *
 * \param[in]	cProg	Programmation  ajouter
 * \retval		\p true ou \p false selon succs ou chec, respectivement
 **/
bool CRecordDialog::ajoutTacheAvecLogon(Programme & cProg)
{
	// Par prcaution, dtruire d'abord toute tche antrieurement existante portant
	// le nom demand :
	cProg.supprimerTacheProgrammee();

	if (cProg.etat==epr_actif && cProg.tache == planif_ajout) {
#if USE_TASKMAN_PASSWORD
		WCHAR username[UNLEN+1];
		DWORD iLen = _countof(username);

		GetUserNameW(username, &iLen);	// Obtenir le nom de l'utilisateur courant

		if (taskLogon(username)) {
			if (cProg.ajouterTacheProgrammee(username, szPassword)) {
				cProg.tache = planif_avec;
				return true;
			}
		} else {
			cProg.tache = planif_sans;
			CheckDlgButton(hDlg, IDC_TACHE, BST_UNCHECKED);
			MessageBox(hDlg, TEXT("Vous n'avez pas saisi de mot de passe valide :") EOL
				TEXT("Cet enregistrement va tre programm sans ajout au gestionnaire des tches."),
				TEXT("Saisie de mot de passe annule"), MB_ICONINFORMATION | MB_OK);
		}
#else
		if (cProg.ajouterTacheProgrammee()) {
			cProg.tache = planif_avec;
			MessageBoxCheck(hDlg,
				TEXT("Une programmation vient d'tre ajoute au gestionnaire des tches, afin d'assurer")
				TEXT("le lancement du logiciel peu de temps avant le dbut de l'enregistrement.") EOL
				TEXT("Note: la saisie du mot de passe de session Windows, ncessaire dans les versions prcdentes, a t supprime."),
				TEXT("Ajout au gestionnaire des tches"), MB_OK|MB_ICONINFORMATION,
				MB_OK, TEXT("Note gestionnaire des tches"));
			return true;
		}
#endif
	}
	return false;
}

/**
 * \brief Suppression de cette programmation de la liste des tches programmes
 *
 * \retval		\p true ou \p false selon succs ou chec, respectivement
 **/
bool Programme::supprimerTacheProgrammee() const
{
	// Nom de la tche (rserver la taille du nom de l'mission plus une marge  cause de la taille
	// du prfixe, lequel peut tre assez long si un nom d'instance de PTVM y est inclus).
	WCHAR					taskName[_countof(nom)+32];

	// On gnre le nom de la tche :
	genereNomTache(taskName, _countof(taskName));

	CComPtr<ITaskScheduler>	pITS;

	HRESULT hr = pITS.CoCreateInstance(CLSID_CTaskScheduler, NULL, CLSCTX_INPROC_SERVER);
	if (FAILED(hr)) {
		myprintf(TEXT("Erreur init task scheduler") EOL);
		return false;
	}

	// Supprime la tche
	hr = pITS->Delete(taskName);
	check_hr(hr, "suppression tche programme");
	return true;
}

#ifdef _XMLUTILS_H_INCLUDED_

/**
 * Chargement de la liste des programmations  partir d'un noeud TinyXML construit
 * \param[in]	pXmlElm	Noeud XML contenant les donnes  charger
 * \return		\p true ou \p false selon succs ou chec
 **/
bool ProgramList::SetFromXMLElement(const TiXmlElement * pXmlElm)
{
	if (_stricmp(pXmlElm->Value(), "Programmes")!=0)
		return false;
	for (ITERATE_XML_ELEMENTS_NAMED(pXmlElm, pXmlChild, "Programme")) {
		Programme	prog;

		if (prog.SetFromXMLElement(pXmlChild))
			push_back(prog);
	}

	// Tri de la liste des programmations (en cas de chargement
	// d'ancienne version, ou bien de modification extrieure) :
	std::sort(begin(), end());
	programmation_modifiee = false;
	return true;
}

/**
 * Construction d'un noeud TinyXML reprsentant la dfinition de la liste des programmations
 * \return	Noeud XML construit
 **/
TiXmlElement * ProgramList::GetAsXMLElement() const
{
	TiXmlElement *	pXmlPrLst(new TiXmlElement("Programmes"));

	if (pXmlPrLst) {
		for (ITERATE_CONST_VECTOR(*this, Programme, it)) {
			TiXmlElement *	pXmlProg(it->GetAsXMLElement());

			if (pXmlProg)
				pXmlPrLst->LinkEndChild(pXmlProg);
		}
	}

	return pXmlPrLst;
}

#endif	// #ifdef _XMLUTILS_H_INCLUDED_

/**
 * \brief Traitement du chargement de la liste des programmes au dmarrage
 *
 * Inclut purge des enregistrements prims
 **/
void init_programmation()
{
	// Charge la liste des programmes
	lit_programmes();

	SYSTEMTIME localtime;
	GetLocalTime(&localtime);
	int i = 0;
	while (i < (int)Programmes.size()) {
		Programme & prog = Programmes[i];
		long diff = DiffTime(prog.fin, localtime);

		if (diff <= 0) {
			// On efface les programmes sans rptitions dont l'heure de fin est passe
			if (prog.repetition==0) {
				Programmes.erase(Programmes.begin()+i);
				programmation_modifiee = true;
				continue;	// L'enregistrement courant ayant t supprim, on n'incrmente pas 'i'
			}

			// Le cas chant, on calcule la rptition suivante
			// (mais a a probablement dj t fait ailleurs)
			prog.ajustementDesRepetitions(localtime);
		}
		i++;
	}

	// Mises  jour ventuelles au cas o les oprations
	// prcdentes auraient modifi la programmation
	record_list_updated();
}

/**
 * Trouver l'index d'un enregistrement programm  partir de son nom.
 *
 * \param[in]	nom	Nom de l'enregistrement recherch
 * \retval	 Index numrique de l'enregistrement
 * \retval	 -1 si nom vide ou rien trouv.
 **/
int trouve_prog_par_nom(LPCTSTR nom)
{
	if (nom && nom[0] != 0) {
		int nbProgrammes = (int)Programmes.size();

		for (int i=0; i<nbProgrammes; i++) {
			if (_tcscmp(Programmes[i].nom, nom)==0)
				return i;
		}
	}

	return -1;
}

/**
 * Trouver le temps restant avant le prochain enregistrement  effectuer.
 *
 * \param[in]		pLocaltime	Si non NULL, pointeur sur temps systme de rfrence (actuel)
 * \param[in,out]	pProg		Pointeur qui, s'il est diffrent de \p NULL, pointe sur une variable
 *								destine  recevoir l'adresse du descripteur de la programmation correspondante.
 * \param[out]		pFirstWake	Si != NULL, la variable pointe reoit le temps avant le 1er enregistrement pour
 *								lequel la sortie de mode veille a t demande, ou bien #DUREE_INFINIE si aucun
 * \return	Temps restant, exprim en millisecondes, ou bien #DUREE_INFINIE s'il n'y en a pas ou plus.
 **/
LONG temps_prochain_enregistrement(const SYSTEMTIME * pLocaltime, Programme ** pProg, LPLONG pFirstWake)
{
	int			nbProgrammes	= (int)Programmes.size();
	LONG		minReste		= DUREE_INFINIE;	// maximum possible au dpart
	SYSTEMTIME	localtime;

	if (pLocaltime)
		localtime = *pLocaltime;
	else
		GetLocalTime(&localtime);

	if (pFirstWake)
		*pFirstWake = DUREE_INFINIE;
	if (pProg)
		*pProg = NULL;	// par dfaut

	// On recherche le premier programme  enregistrer
	for (int i=0; i<nbProgrammes; i++) {
		Programme & prog = Programmes[i];

		// Au passage, on regarde s'il y a des enregistrements dont l'heure de fin est
		// dpasse mais qui ont des rptitions programmes, et si c'est le cas,
		// on ajuste leur date :
		prog.ajustementDesRepetitions(localtime);

		if (prog.etat==epr_actif && enregistrements_actuels.trouve_enreg_par_nom(prog.nom)<0) {
			// Seulement les enregistrements pas dj en cours et aucune erreur dclare
			LONG reste = prog.temps_restant(localtime);

			// Seulement les programmations actives et non termines
			if (reste < minReste) {
				minReste = reste;
				if (pProg)
					*pProg = &prog;
			}

			if (pFirstWake && prog.reveil && reste < *pFirstWake)
				*pFirstWake = reste;
		}
	}

	return minReste;
}

/**
 * (Re)calculer le timer de la prochaine transition d'enregistrement
 **/
void set_timer_record()
{
	SYSTEMTIME	localtime;

	GetLocalTime(&localtime);

	LONG		nTimeBeforeNextAwake		= DUREE_INFINIE;
	LONG		nTimeBeforeNextRecording	= temps_prochain_enregistrement(&localtime, NULL, &nTimeBeforeNextAwake);
	LONG		nTimeBeforeNextEvent		= min(enregistrements_actuels.temps_restant(localtime),
												nTimeBeforeNextRecording);
	bool		bNeedAwake					= false;

	if (nTimeBeforeNextEvent == DUREE_INFINIE) {
		KillTimer(hMainWnd, TIMER_DELAYED_RECORD);
		myprint(TEXT("set_timer_record: Aucun vnement d'enregistrement en attente") EOL);
	} else {
		if (nTimeBeforeNextAwake != DUREE_INFINIE) {
			nTimeBeforeNextAwake -= wakeup_delay * DUREE_1MN;
			bNeedAwake = nTimeBeforeNextAwake > 0;
		}

		// Si l'heure de dbut est dj passe, on lance un timer qui aboutit immdiatement
		long timer = min(max(nTimeBeforeNextEvent, USER_TIMER_MINIMUM), DUREE_RECALCUL_TIMERS*1000);

		myprintf(TEXT("set_timer_record: Prochain vnement dans %i mS (timer = %i mS)") EOL, nTimeBeforeNextEvent, timer);
		SetTimer(hMainWnd, TIMER_DELAYED_RECORD, timer, NULL);
	}

	update_record_protect(nTimeBeforeNextRecording);

	if (bNeedAwake) {
		if (allow_wakeup)
			cWaitTimer.Set(nTimeBeforeNextAwake);
	} else {
		cWaitTimer.Cancel();
	}
	cShutBlockReason.Update();
}

/**
 * Obtenir la date et l'heure d'une paire de contrles de type \p SysDateTimePick32
 * \param[in]	hDlg	Handle du dialogue contenant ces contrles
 * \param[in]	dateID	Identifiant du contrle de date
 * \param[in]	timeID	Identifiant du contrle d'heure
 * \param[out]	result	Date et heure retournes
 **/
static void GetCtlTimeDate(HWND hDlg, int dateID, int timeID, SYSTEMTIME & result)
{
	SYSTEMTIME time_buffer;

	SendDlgItemMessage(hDlg, dateID, DTM_GETSYSTEMTIME, 0, (LPARAM)&time_buffer);
	result = time_buffer;
	SendDlgItemMessage(hDlg, timeID, DTM_GETSYSTEMTIME, 0, (LPARAM)&time_buffer);
	result.wHour	= time_buffer.wHour;
	result.wMinute	= time_buffer.wMinute;
	result.wSecond	= time_buffer.wSecond;
	result.wMilliseconds = 0;
}

/**
 * Dfinir la date et l'heure dans une paire de contrles de type \p SysDateTimePick32
 * \param[in]	hDlg	Handle du dialogue contenant ces contrles
 * \param[in]	dateID	Identifiant du contrle de date
 * \param[in]	timeID	Identifiant du contrle d'heure
 * \param[out]	value	Date et heure  charger
 **/
static void SetCtlTimeDate(HWND hDlg, int dateID, int timeID, const SYSTEMTIME & value)
{
	SendDlgItemMessage(hDlg, dateID, DTM_SETSYSTEMTIME, GDT_VALID, (LPARAM)&value);
	SendDlgItemMessage(hDlg, timeID, DTM_SETSYSTEMTIME, GDT_VALID, (LPARAM)&value);
}

/**
 * Obtenir la date et l'heure d'une paire de contrles de type "SysDateTimePick32"
 * avec gestion du report des retenues des secondes vers les minutes, des minutes
 * vers les secondes et des secondes vers les heures.
 * Le bon fonctionnement de cette procdure implique que 'result' contienne la
 * valeur prcdente du contrle en entre.
 * \param[in]	hDlg	Handle du dialogue contenant ces contrles
 * \param[in]	dateID	Identifiant du contrle de date
 * \param[in]	timeID	Identifiant du contrle d'heure
 * \param[out]	result	Date et heure retournes
 **/
static void GetCtlTimeDateChg(HWND hDlg, int dateID, int timeID, SYSTEMTIME & result)
{
	SYSTEMTIME tempres;
	long toAdd = 0;

	GetCtlTimeDate(hDlg, dateID, timeID, tempres);

	long diff = DiffTime(tempres, result)/DUREE_1SEC;

	myprintf(TEXT("GetCtlTimeDateChg, diff = %i S") EOL, diff);

	// Note : on ne traite le dfilement rapide sur les heures que par unit,  cause de la
	// difficult  discriminer les sauts de 20 heures que le contrle peut induire

	if (diff < 0) {
		// On a recul dans le temps

		if (GetAsyncKeyState(VK_LBUTTON)<0 || GetAsyncKeyState(VK_ADD)<0 || GetAsyncKeyState(VK_UP)<0) {
			// Traitement seulement si le bouton gauche de la souris ou une des touches
			// '+' ou fleche en haut est couramment press :
			if (diff >= -(60-1) && diff <= -(60-20)) {
				// On a incrment les secondes mais recul de (60-20)  (60-1) secondes
				if (tempres.wSecond < 20)
					toAdd = 60;
			} else if (diff >= -(60-1)*60 && diff <= -(60-20)*60) {
				// On a incrment les minutes mais recul de (60-20)  (60-1) minutes
				if (tempres.wMinute < 20)
					toAdd = 60*60;
			} else if (diff == -(24-1)*60*60) {
				// On a incrment les heures mais recul de (24-1) heures
				if (tempres.wHour < 1)
					toAdd = 24*60*60;
			}
		}
	} else {
		// On a avanc dans le temps

		if (GetAsyncKeyState(VK_LBUTTON)<0 || GetAsyncKeyState(VK_SUBTRACT)<0 || GetAsyncKeyState(VK_DOWN)<0) {
			// Traitement seulement si le bouton gauche de la souris ou une des touches
			// '-' ou fleche en bas est couramment press :
			if (diff >= (60-20) && diff <= (60-1)) {
				// On a dcrment les secondes mais avanc de (60-20)  (60-1) secondes
				if (tempres.wSecond >= 60-20)
					toAdd = -60;
			} else if (diff >= (60-20)*60 && diff <= (60-1)*60) {
				// On a dcrment les minutes mais avanc de (60-20)  (60-1) minutes
				if (tempres.wMinute >= 60-20)
					toAdd = -60*60;
			} else if (diff == (24-1)*60*60) {
				// On a dcrment les heures mais avanc de (24-1) heures
				if (tempres.wHour >= 24-1)
					toAdd = -24*60*60;
			}
		}
	}

	if (toAdd != 0) {
		AddTime(tempres, toAdd*DUREE_1SEC);
		SetCtlTimeDate(hDlg, dateID, timeID, tempres);
	}

	result = tempres;
}

#define REC_COL_NOM			0	///< Colonne indiquant le nom de l'enregistrement
#define REC_COL_CHAINE		1	///< Colonne indiquant le nom de la chane
#define REC_COL_DEBUT		2	///< Colonne indiquant l'heure de dbut
#define REC_COL_FIN			3	///< Colonne indiquant l'heure de fin
#define REC_COL_ETAT		4	///< Colonne indiquant l'tat de l'enregistrement

/**
 * Chargement de la liste des enregistrements programms.
 * \param[in]	hListItem	Handle du contrle de liste
 **/
static void remplit_liste_programmes(HWND hListItem)
{
	TCHAR buffer[64];

	LVITEM item = {
		LVIF_PARAM | LVIF_TEXT,	// mask
		0, 0, 0, 0,				// iItem, iSubItem, state, stateMask
		buffer,					// buffer
		_countof(buffer)		// cchTextMax
		// le reste  zro (implicite)
	};

	int nbProgrammes = (int)Programmes.size();

	for(; item.iItem<nbProgrammes; item.iItem++) {
		const Programme &	prog		= Programmes[item.iItem];
		int					ixChaine	= Canaux.trouve_par_cle(prog.cleChaine);

		if (!ixChaine_ok(ixChaine)) {
			ixChaine = Canaux.trouve_par_no(prog.nNumeroChaine);
			if (!ixChaine_ok(ixChaine))
				continue;
		}

		item.lParam = reinterpret_cast<LPARAM>(& prog);

		item.iSubItem = REC_COL_NOM;
		strcpy_T(buffer, prog.nom);
		ListView_InsertItem(hListItem, &item);

		item.iSubItem = REC_COL_CHAINE;
		strcpy_T(buffer, Canaux[ixChaine].nom);
		SendMessage(hListItem, LVM_SETITEMTEXT, item.iItem, (LPARAM)&item);

		item.iSubItem = REC_COL_DEBUT;
		prog.formate_heure_debut(buffer, _countof(buffer));
		SendMessage(hListItem, LVM_SETITEMTEXT, item.iItem, (LPARAM)&item);

		item.iSubItem = REC_COL_FIN;
		prog.formate_heure_fin(buffer, _countof(buffer));
		SendMessage(hListItem, LVM_SETITEMTEXT, item.iItem, (LPARAM)&item);

		item.iSubItem = REC_COL_ETAT;
		prog.formate_etat(buffer, _countof(buffer));
		SendMessage(hListItem, LVM_SETITEMTEXT, item.iItem, (LPARAM)&item);
	}
}

static void InitCtlAfter(HWND hDlg, int ctlID)
{
	static LPCTSTR aAfterList[] = {
		TEXT("Ne rien faire"),
		TEXT("Quitter l'application"),
		TEXT("teindre l'ordinateur"),
		TEXT("teindre l'ordinateur (forc)"),
		TEXT("Mettre l'ordinateur en veille"),
		TEXT("Mettre l'ordinateur en veille prolonge"),
		NULL
	};

	ComboBox_Init(GetDlgItem(hDlg, ctlID), aAfterList, apr_rien);
}

/**
 * Rcupration de l'action "aprs enregistrement" selon l'tat de la combo box.
 **/
static ApresEnregistrement GetCtlAfter(HWND hDlg, int ctlID)
{
	LRESULT lRes = SendMessage(GetDlgItem(hDlg, ctlID), CB_GETCURSEL, 0, NULL);

	if (lRes != CB_ERR)
		return (ApresEnregistrement)lRes;
	return apr_rien;
}

/**
 * Mise  jour de la combo box correspondant  l'action "aprs enregistrement"
 * selon le contenu de la variable 'after'.
 **/
static void SetCtlAfter(HWND hDlg, int ctlID, ApresEnregistrement after)
{
	SendMessage(GetDlgItem(hDlg, ctlID), CB_SETCURSEL, after, NULL);
}

/**
 * Coche ou dcoche les cases lies aux rptitions :
 **/
static void SetCtlRepeat(HWND hDlg, int repeatBaseID, BYTE repetition)
{
	for (int i=0; i<7; i++)
		CheckDlgButton(hDlg, repeatBaseID+i, (repetition & (1<<i))!=0);
}

/**
 * Slection de la chane 'ixChaine' dans la combo box
 **/
static void CtlSelChannel(HWND hCtl, int ixChaine)
{
	if (ixChaine_ok(ixChaine)) {
		TCHAR str[64];

		Canaux[ixChaine].toMenuString_tsid(str, _countof(str));
		SendMessage(hCtl, CB_SELECTSTRING, WPARAM(-1), LPARAM(str));
	}
}

/**
 * Rcupration du nom d'une programmation avec nettoyage.
 **/
static void getDlgProgName(HWND hDlg, LPTSTR pnom, UINT bufSize)
{

	GetWindowText(GetDlgItem(hDlg, IDC_NAME), pnom, bufSize);

	// Supprimer les espaces de dbut :
	LPTSTR pwrk = pnom;
	while (*pwrk>0 && *pwrk<=' ')	// attention, type CHAR sign !!
		pwrk++;
	if (pwrk>pnom)
		strcpy_T(pnom, bufSize, pwrk);

	// Supprimer les espaces de fin :
	UINT strSize = strlen_T(pnom);
	while (strSize>0 && pnom[strSize-1]>0 && pnom[strSize-1]<=' ')
		pnom[--strSize] = 0;
}

/**
 * Rcupre les donnes de l'enregistrement dans la fentre.
 * Retourne 'true' si ces donnes sont valides, 'false' sinon.
 **/
bool CRecordDialog::ToProg(Programme & prog, bool modif)
{
	UINT32	cleChaine;
	int		ixChaine;

	// Rcuprer la chane :
	if ((ixChaine = (int)SendDlgItemMessage(hDlg, IDC_CHANNEL, CB_GETCURSEL, 0, 0))==CB_ERR)		// index combo
		return false;
	if ((cleChaine = (UINT32)SendDlgItemMessage(hDlg, IDC_CHANNEL, CB_GETITEMDATA, ixChaine, 0))==0)// Cl
		return false;
	if (!ixChaine_ok(ixChaine = Canaux.trouve_par_cle(UINT32(cleChaine))))							// index chane
		return false;

	// Rcuprer d'abord les dates, car un ventuel nom gnr automatiquement se sert
	// de la date de dbut :
	GetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, prog.debut);
	GetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, prog.fin);

	// Ces deux valeurs ont t extraites de la bote de dialogue au moment de leur slection :
	prog.methode	= epg_typerecvideo;
	prog.audio		= epg_typerecaudio;

	const Chaine & canal = Canaux[ixChaine];
	NomProg nom;

	// Rcuprer le nom :
	getDlgProgName(hDlg, nom, _countof(nom));

	if (nom[0]==0) {
		// le nom de la programmation est vide :
		if (prog.nom[0]==0 || prog.nomAuto) {
			// En gnrer un s'il n'y avait pas d'ancien nom,
			// ou bien si celui-ci avait dj t gnr automatiquement :
			unsigned ixSuffNom = 0;
			//
			do {
				ixSuffNom++;
				_stprintf_s(nom, _countof(nom),
					ixSuffNom>1 ? TEXT("%") A2t TEXT("-%i-%02i_%u") : TEXT("%") A2t TEXT("-%i-%02i"),
					prog.isMultiplexProg() ? "Multiplex" : canal.nom,
					prog.debut.wDay, prog.debut.wMonth, ixSuffNom);
			} while (_tcscmp(nom, prog.nom)!=0 && trouve_prog_par_nom(nom)>=0);
			strcpy_T(prog.nom, nom);
			prog.nomAuto = true;
		}
	} else {
		strcpy_T(prog.nom, nom);
		prog.nomAuto = false;
	}

	prog.nNumeroChaine = canal.nNumeroChaine;
	prog.cleChaine = canal.cle();
	prog.sidChaine = canal.SID;	/// \deprecated \p sidChaine remplac par \p cleChaine

	prog.apres = GetCtlAfter(hDlg, IDC_AFTER);
	prog.etat = IsDlgButtonChecked(hDlg, IDC_INACTIVE) == BST_CHECKED ? epr_inactif : epr_actif;

	// Ajoute l'indicatif pour la rptition
	prog.repetition = 0;
	for (int i=0; i<7; i++) {
		if (IsDlgButtonChecked(hDlg, IDC_REP_BASE+i) == BST_CHECKED)
			prog.repetition |= 1<<i;
	}

	prog.tache = IsDlgButtonChecked(hDlg, IDC_TACHE) == BST_CHECKED ? planif_ajout: planif_sans;
	prog.reveil	= IsDlgButtonChecked(hDlg, IDC_WAKEUP) == BST_CHECKED;

	return true;
}

/**
 * Transfert des donnes d'un enregistrement programm dans les items appropris de la bote
 * de dialogue correspondante.
 **/
void CRecordDialog::progToDlg(const Programme & prog)
{
	// (en fait, uniquement les items susceptibles d'tre contenus dans le guide EPG)
	CtlSelChannel(GetDlgItem(hDlg, IDC_CHANNEL), Canaux.trouve_par_cle(prog.cleChaine));

	SendDlgItemMessage(hDlg, IDC_NAME, WM_SETTEXT, 0, LPARAM(prog.nomAuto ? TEXT("") : prog.nom));
	SetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut = prog.debut);
	SetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin = prog.fin);

	CheckDlgButton(hDlg, IDC_INACTIVE, prog.active() ? BST_UNCHECKED : BST_CHECKED);
	dtc_debut = prog.debut;
	dtc_fin = prog.fin;
}

/**
 * Transfert d'une description d'mission dans une structure d'enregistrement programm
 * \param[in]	canal	Rfrence sur la description de la chane
 * \param[in]	cEmis	Rfrence sur l'mission dans la liste des missions de cette chane
 * \param[out]	prog	Structure recevant la description
 **/
bool chaineToProg(const Chaine & canal, const Emission & cEmis, Programme & prog)
{
	_tcsncpy_s(prog.nom, cEmis.getName().c_str(), _countof(prog.nom)-1);	// Tronquer si la longueur dpasse

	prog.nom[_countof(prog.nom)-1] = 0;
	prog.cleChaine		= canal.cle();
	prog.sidChaine		= canal.SID;	/// \deprecated \p sidChaine remplac par \p cleChaine
	prog.nNumeroChaine	= canal.nNumeroChaine;
	prog.apres			= apr_rien;
	prog.methode		= meth_TS;
	prog.audio			= am_Mpeg_1;

	// Ajout de l'heure de dbut et de l'heure de fin, avec leurs dcalages respectifs :
	prog.fin			= cEmis.fin;
	AddTime(prog.fin,	retard_enreg *  60000);
	prog.debut			= cEmis.debut;
	AddTime(prog.debut,	avance_enreg * -60000);

	prog.repetition		= 0;
	prog.tache			= planif_sans;
	prog.etat			= epr_actif;

	return true;
}

// Chargement des donnes de dmarrage initialisant une nouvelle programmation :
void CRecordDialog::InitNewRecord()
{
	SYSTEMTIME localtime;

	GetLocalTime(&localtime);

	localtime.wSecond = 0;
	localtime.wMilliseconds = 0;
	SetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut = localtime);
	SetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin = localtime);
	CtlSelChannel(GetDlgItem(hDlg, IDC_CHANNEL), ixChaineCourante);
	SendDlgItemMessage(hDlg, IDC_NAME, WM_SETTEXT, 0, (LPARAM)TEXT(""));

	SetCtlAfter(hDlg, IDC_AFTER, apr_rien);
	CheckDlgButton(hDlg, IDC_TACHE, BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_WAKEUP, BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_INACTIVE, BST_UNCHECKED);

	EnableDlgItem(IDC_WAKEUP, allow_wakeup);

	SetCtlRepeat(hDlg, IDC_REP_BASE, 0);	// Rptitions
}

/**
 * Remplissage d'une combo-box de mthodes d'enregistement
 * \param[in]	hCtl	"Handle" de la combo-box
 * \param[in]	eSel	Item slectionn
 **/
void remplit_liste_methodes(HWND hCtl, MethodeEnregistrement eSel)
{
	LPCTSTR aMethodsInitList[] = {	// non static
		TEXT("TS"),
		TEXT("PS"),
		allow_stream_record ? TEXT("Multiplex") : NULL,
		NULL
	};

	ComboBox_Init(hCtl, aMethodsInitList, eSel);
}

/**
 * Remplissage d'une combo-box d'items audio
 * \param[in]	hCtl	"Handle" de la combo-box
 * \param[in]	eSel	Item slectionn
 **/
void remplit_liste_audio(HWND hCtl, AudioMode eSel)
{
	static LPCTSTR aAudioInitList[] = {
		TEXT("1re piste"),
		TEXT("Mpeg (1re)"),
		TEXT("Mpeg (2me)"),
		TEXT("(E-)AC3"),
		TEXT("(E-)AC3 2.0"),
		TEXT("(E-)AC3 HC"),
		TEXT("AAC"),
		TEXT("AAC 2.0"),
		TEXT("AAC HC"),
		TEXT("V.O."),
		NULL
	};

	ComboBox_Init(hCtl, aAudioInitList, eSel);
}

/**
 * Initialisation des items de mthodes d'enregistrement d'un dialogue d'enregistrements programms
 * \param[in]	hDlg	"Handle" de la bote de dialogue
 **/
void init_methodes(HWND hDlg)
{
	HWND hCtlMtd = GetDlgItem(hDlg, IDC_METHOD);
	HWND hCtlAud = GetDlgItem(hDlg, IDC_AUDIO);

	remplit_liste_methodes(hCtlMtd, epg_typerecvideo);
	remplit_liste_audio(hCtlAud, epg_typerecaudio);

	EnableWindow(hCtlAud, epg_typerecvideo == meth_PS);
}

/**
 * Mise  jour des mthodes d'enregistrement aprs changement d'un des items de dialogue
 * qui les contrlent
 * \param[in]	hDlg	"Handle" de la bote de dialogue
 **/
void update_methodes(HWND hDlg)
{
	HWND hCtlMtd = GetDlgItem(hDlg, IDC_METHOD);
	HWND hCtlAud = GetDlgItem(hDlg, IDC_AUDIO);

	epg_typerecvideo = (MethodeEnregistrement)SendMessage(hCtlMtd, CB_GETCURSEL, 0, NULL);
	epg_typerecaudio = (AudioMode)SendMessage(hCtlAud, CB_GETCURSEL, 0, NULL);

	EnableWindow(hCtlAud, epg_typerecvideo == meth_PS);
}

void CRecordDialog::update_edit_state(int sel)
{
	NomProg nom;
	bool edit = sel >= 0 && Programmes[sel].etat != epr_encours;

	EnableDlgItem(IDC_MODIFY, edit);
	EnableDlgItem(IDC_REMOVE, edit);

	// Activation ou dsactivation du bouton "Ajouter" selon le contenu de l'item
	// d'dition de nom (ne doit pas tre vide et ne doit pas dj exister) :
	getDlgProgName(hDlg, nom, _countof(nom));
	EnableDlgItem(IDC_ADD, trouve_prog_par_nom(nom) < 0);
}

int CALLBACK CRecordDialog::LVRecCompareProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	CLViewHelper &		cLVhlpr	= *reinterpret_cast<CLViewHelper *>(lParamSort);
	const Programme &	cProg1	= *reinterpret_cast<const Programme *>(lParam1);
	const Programme &	cProg2	= *reinterpret_cast<const Programme *>(lParam2);
	int					res		= 0; // rsultat

	switch (cLVhlpr.nColon) {

	case REC_COL_NOM:
		res = _tcsicmp(cProg1.nom, cProg2.nom);
		break;

	case REC_COL_CHAINE:
		res = cProg1.nNumeroChaine - cProg2.nNumeroChaine;
		break;

	case REC_COL_ETAT:
		res = cProg1.compare_etat(cProg2.etat);
		break;

	case REC_COL_DEBUT:
		res = DiffTime(cProg1.debut, cProg2.debut);
		if (res == 0)
			res = cProg1.nNumeroChaine - cProg2.nNumeroChaine;
		break;

	case REC_COL_FIN:
		res = DiffTime(cProg1.fin, cProg2.fin);
		if (res == 0)
			res = cProg1.nNumeroChaine - cProg2.nNumeroChaine;
	}

	// Tri descendant si cLVhlpr.bDescending != 0
	if (cLVhlpr.bDescending)
		res = -res;
	return res;
}

/// Traitement des messages de la bote de dialogue
INT_PTR CRecordDialog::DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG: {

		selProgNom[0] = 0;

		static const struct Tab_cols t_cols[] = {
			{TEXT("Nom"),		160},
			{TEXT("Chane"),	86},
			{TEXT("Dbut"),		140},
			{TEXT("Fin"),		48},
			{TEXT("tat"),		78},
			{NULL, 0}			// Terminateur
		};

		cLVhlpr.Init(GetDlgItem(hDlg, IDC_LIST_PROGRAMMES), t_cols);
		SubClass(cLVhlpr.hLView);
		remplit_liste_programmes(cLVhlpr.hLView);

		// Chargement du combo de choix des chanes
		HWND	hCtlChn = GetDlgItem(hDlg, IDC_CHANNEL);

		for (ITERATE_CONST_VECTOR(Canaux, Chaine, it)) {
			const Chaine &	cChan = *it;
			TCHAR			szWrk[64];

			if (cChan.etat == ec_inactive)
				continue;
			cChan.toMenuString_tsid(szWrk, _countof(szWrk));

			// Ajout de l'item, avec association de la cl de la chane pour identifier la slection.
			AddComboItemWithData(hCtlChn, szWrk, cChan.cle());
		}

		init_methodes(hDlg);
		InitCtlAfter(hDlg, IDC_AFTER);
		InitNewRecord();
		update_edit_state(selection);
		return TRUE; }

	case WM_NOTIFY:
		switch (wParam) {

		case IDC_LIST_PROGRAMMES:
			switch (_ncode_(lParam)) {

			case NM_CLICK:
				selection = ListView_GetSelectionMark(cLVhlpr.hLView);
				if (selection>=0) {
					// Clic dans la liste des programmations
					const Programme & prog = Programmes[selection];

					progToDlg(prog);

					SendDlgItemMessage(hDlg, IDC_METHOD,	CB_SETCURSEL, epg_typerecvideo = prog.methode, NULL);
					SendDlgItemMessage(hDlg, IDC_AUDIO,		CB_SETCURSEL, epg_typerecaudio = prog.audio, NULL);
					EnableDlgItem(IDC_AUDIO, epg_typerecvideo == meth_PS);

					SetCtlAfter(hDlg, IDC_AFTER, prog.apres);

					CheckDlgButton(hDlg, IDC_TACHE,
						prog.tache!=planif_sans ? BST_CHECKED : BST_UNCHECKED);
					CheckDlgButton(hDlg, IDC_WAKEUP, prog.reveil ? BST_CHECKED : BST_UNCHECKED);

					// Coche les cases pour les rptitions :
					SetCtlRepeat(hDlg, IDC_REP_BASE, prog.repetition);

					strcpy_T(selProgNom, prog.nom);
					return TRUE;
				}
				update_edit_state(selection);
				break;

			case LVN_KEYDOWN: {
				LPNMLVKEYDOWN pnkd = (LPNMLVKEYDOWN) lParam;
				if (pnkd->wVKey == VK_DELETE)
					SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_REMOVE, 0), 0);
				break; }

			case LVN_COLUMNCLICK:		// Clic sur la colonne => tri
				cLVhlpr.ProcessColumnClick(LVRecCompareProc, lParam);
				break;

			case NM_CUSTOMDRAW: {
				NMLVCUSTOMDRAW &	sNlcd = *reinterpret_cast<LPNMLVCUSTOMDRAW>(lParam);

				switch(sNlcd.nmcd.dwDrawStage) {

				case CDDS_PREPAINT:
					SetWindowLongPtr(hDlg, DWLP_MSGRESULT, CDRF_NOTIFYITEMDRAW);
					return TRUE;

				case CDDS_ITEMPREPAINT: {
					const Programme & prog = Programmes[sNlcd.nmcd.dwItemSpec];

					switch (prog.etat) {

					case epr_actif:
						sNlcd.clrText = RGB(0,128,255);
						break;

					case epr_encours:
						sNlcd.clrText = RGB(255,0,0);
						break;

					case epr_interrompu:
					case epr_termine:
					case epr_inactif:
						sNlcd.clrText = RGB(128,128,128);
						break;

					case epr_supplante:
					case epr_erreur:
					case epr_fileerr:
						sNlcd.clrText = RGB(128,0,128);
					}
					return TRUE; }
				}
				break; }
			}
			break;

		case IDC_DATE_START:
		case IDC_TIME_START:
			if (_ncode_(lParam)==DTN_DATETIMECHANGE) {
				GetCtlTimeDateChg(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
				GetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
				if (DiffTime(dtc_debut, dtc_fin) > 0)
					SetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin = dtc_debut);
			}
			break;

		case IDC_DATE_END:
		case IDC_TIME_END:
			if (_ncode_(lParam)==DTN_DATETIMECHANGE) {
				GetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
				GetCtlTimeDateChg(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
				if (DiffTime(dtc_debut, dtc_fin) > 0)
					SetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut = dtc_fin);
			}
		}
		return FALSE;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {

		case IDC_ADD_24:	// Ajouter 24h des deux cts
			GetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
			AddTime(dtc_debut, DUREE_24H);
			SetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
			GetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
			AddTime(dtc_fin, DUREE_24H);
			SetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
			break;

		case IDC_SUB_24:	// Soustraire 24h des deux cts
			GetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
			AddTime(dtc_debut, -DUREE_24H);
			SetCtlTimeDate(hDlg, IDC_DATE_START, IDC_TIME_START, dtc_debut);
			GetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
			AddTime(dtc_fin, -DUREE_24H);
			SetCtlTimeDate(hDlg, IDC_DATE_END, IDC_TIME_END, dtc_fin);
			break;

		case IDC_METHOD:
		case IDC_AUDIO:
			if (HIWORD(wParam) == CBN_SELCHANGE) {
				update_methodes(hDlg);
				return TRUE;
			}
			break;

		case IDC_WAKEUP:
			if (HIWORD(wParam) == BN_CLICKED && eWinVers >= wv_Seven && get_check(hDlg, IDC_WAKEUP)) {
				MessageBoxCheck(hDlg,
					TEXT("Pour que la sortie programme du mode veille puisse fonctionner sous Windows 7/8/10, il est")
					TEXT(" ncessaire d'activer une option dans les Options d'alimentation du Panneau de Configuration:") EOL
					TEXT("Aller dans Modifier les paramtres du mode ->Modifier les paramtres d'alimentation avancs")
					TEXT(" ->Veille ->Autoriser les minuteurs de sortie de veille.") EOL
					TEXT(" Il est recommand de n'y activer l'option que pour le fonctionnement Sur secteur."),
				TEXT("Note pour les utilisateurs de Windows 7/8/10"),
				MB_OK|MB_ICONINFORMATION, MB_OK, TEXT("Rveil sous Windows 7/8/10"), &bWarnAwake);
			}
			break;

		case IDC_NAME:
			if (HIWORD(wParam) == EN_CHANGE) {
				update_edit_state(selection);
				return TRUE;
			}
			break;

		case IDC_REMOVE:
			if (selection >= 0 && Programmes[selection].etat!=epr_encours) {
				ListView_DeleteItem(cLVhlpr.hLView, selection);

				// Suppression ventuelle tche antrieure portant ce nom :
				Programmes[selection].supprimerTacheProgrammee();

				Programmes.erase(Programmes.begin() + selection);
				finalize_prog_change();

				// Transfrer le focus  un autre contrle
				// Voir http://www.olivierlanglois.net/trapped_focus.html
				PostDlgMessage(WM_NEXTDLGCTL, (WPARAM)GetItem(IDC_NEW), TRUE);
			}
			return TRUE;

		case IDC_MODIFY:
			if (selection >= 0) {
				Programme & prog = Programmes[selection];
				Programme newprog = prog;

				if (ToProg(newprog, true) && newprog.verifie(&prog)) {
					// Rcupration et vrification donnes du dialogue

					if (prog.etat == epr_encours)
						MessageBox(hDlg, TEXT("Cet enregistrement est actuellement en cours.") EOL
							TEXT("Utilisez la commande \"Arrter l'enregistrement dans...\"")
							TEXT(" pour y apporter des modifications maintenant."),
							TEXT("Erreur lors de la modification"), MB_ICONERROR);
					else {
						// Suppression ventuelle tche antrieure portant l'ancien nom :
						prog.supprimerTacheProgrammee();

						prog = newprog;

						strcpy_T(selProgNom, prog.nom);
						ajoutTacheAvecLogon(prog);
						finalize_prog_change();
					}
				}
			}
			return TRUE;

		case IDC_ADD: {
			Programme prog;

			// Rcupration des donnes de l'enregistrement depuis la bote de dialogue
			if (!ToProg(prog, false) || !prog.verifie())
				return FALSE;

			strcpy_T(selProgNom, prog.nom);
			ajoutTacheAvecLogon(prog);
			Programmes.push_back(prog);

			finalize_prog_change();

			// Transfrer le focus  un autre contrle
			// Voir http://www.olivierlanglois.net/trapped_focus.html
			PostDlgMessage(WM_NEXTDLGCTL, (WPARAM)GetItem(IDC_NEW), TRUE);
			return TRUE; }

		case IDC_NEW:
			// Rinitialiser les donnes de programmation
			InitNewRecord();
			selection = -1;
			selProgNom[0] = 0;
			update_edit_state(selection);
			return TRUE;

		case IDC_CURRENT_PROGRAM:
			// Charger les donnes de l'mission en cours
			if (ixChaine_ok(ixChaineCourante)) {
				const Chaine &	canal = Canaux[ixChaineCourante];
				Programme		prog;
				EmissionPtr		pEmis = canal.getCurrentAndNext().apEmi[0];

				if (pEmis && chaineToProg(canal, *pEmis, prog)) {
					progToDlg(prog);
					selection = -1;
					selProgNom[0] = 0;
					update_edit_state(selection);
				}
			}
			return TRUE;

		case IDOK:
			DestroyWindow(hDlg);
			return TRUE;
		}
		break;

	case WM_CLOSE:
		DestroyWindow(hDlg);
		return TRUE;

	case WM_NCDESTROY:
		// Fermeture de la bote de dialogue : on efface le mot de passe (mme si le champ va
		// tre immdiatement dtruit,  cause des rsidus qui pourraient traner en mmoire)
#if USE_TASKMAN_PASSWORD
		SecureZeroMemory(szPassword, sizeof(szPassword));
#endif
		return TRUE;

	case WM_APP_PROG_CHANGED:
		// Reconstruction de la liste des enregistrements programms aprs changement :
		ListView_DeleteAllItems(cLVhlpr.hLView);
		remplit_liste_programmes(cLVhlpr.hLView);
		selection = trouve_prog_par_nom(selProgNom);
		if (selection>=0)
			ListView_SetItemState(cLVhlpr.hLView, selection, LVIS_SELECTED, LVIS_SELECTED);
		update_edit_state(selection);

		// Activation ou dsactivation de l'option de sortie de veille, si celle-ci
		// est change dans la configuration pendant que ce dialogue est ouvert.
		if (!allow_wakeup)
			CheckDlgButton(hDlg, IDC_WAKEUP, BST_UNCHECKED);
		EnableDlgItem(IDC_WAKEUP, allow_wakeup);

		return TRUE;

	case WM_APP_ADD_EPG_PROG: {
		// Ajout d'un enregistrement depuis l'EPG
		int	ixChaine = Canaux.trouve_par_cle((UINT32)lParam);

		if (ixChaine_ok(ixChaineCourante)) {
			const Chaine &		cChan = Canaux[ixChaine];
			const Emission *	pEmis = cChan.cherche_par_id((UINT16)wParam);
			Programme			prog;

			if (pEmis && chaineToProg(cChan, *pEmis, prog)) {
				progToDlg(prog);
				selection = -1;
				selProgNom[0] = 0;
				update_edit_state(selection);
			}
		}

		return TRUE; }
	}

	return FALSE;
}

/**
 * \brief Transformation d'un horaire de fin en temps restant
 * \param[in]	time	Temps  transformer
 * \return		Temps rsultant
 **/
static SYSTEMTIME EndTimeToRemTime(SYSTEMTIME time)
{
	SYSTEMTIME	localtime;
	LONG		wMillisec;

	GetLocalTime(&localtime);
	wMillisec = DiffTime(time, localtime);

	// S'assurer que l'heure de fin tombe bien dans les 24 heures qui suivent :
	while (wMillisec < 0)
		wMillisec += DUREE_24H;
	while (wMillisec >= DUREE_24H)
		wMillisec -= DUREE_24H;

	time.wMilliseconds = WORD(wMillisec % 1000);
	wMillisec /= DUREE_1SEC;
	time.wSecond = WORD(wMillisec % 60);
	wMillisec /= 60;
	time.wMinute = WORD(wMillisec % 60);
	wMillisec /= 60;
	time.wHour = WORD(wMillisec);
	return time;
}

/**
 * \brief Transformation d'un temps restant en horaire de fin
 **/
static SYSTEMTIME RemTimeToEndTime(SYSTEMTIME time)
{
	LONG wMillisec = time.wHour*DUREE_1H + time.wMinute*DUREE_1MN + time.wSecond*DUREE_1SEC + time.wMilliseconds;

	GetLocalTime(&time);
	AddTime(time, wMillisec);
	return time;
}

enum DlgEditState {
	des_notModified,	//!< Aucun changement  appliquer
	des_noConfirmNeed,	/*!< Un changement peut-tre appliqu mais son annulation n'a pas
							 besoin d'tre explicite */
	des_needConfirm		//!< Un changement a t effectu mais son annulation doit tre explicite
};

/**
 * \brief Classe de gestion du dialogue d'arrt programm d'enregistrement
 **/
class CDelayedStopDialog : public CModalDialogBase
{
	HWND				hEnrCombo;		//!< "handle" de la liste droulante des chanes
	HWND				hTimeCtl;		//!< "handle" de contrle de temps
	int					ixRecDescr;		//!< Index du descripteur d'enregistrement
	int					comboSel;		//!< Item couramment slectionn dans la combo box
	Enregistrement *	penreg;			//!< Pointeur enregistrement  modifier

	/** \brief Contenu du contrle de temps
	 *
	 * Ce contenu peut tre, soit l'heure de fin, soit le temps restant, selon que
	 * \p useEndTime est \p true ou \p false (il dfinit le temps de rfrence dans ce dernier cas)
	 **/
	SYSTEMTIME			refTime;

	bool				useEndTime;		//!< \p true si le temps mmoris (\p refTime) est l'heure de fin
	bool				dispEndTime;	//!< \p true si le contrle affiche l'heure de fin
	bool				delayActive;	//!< \p true si la temporisation est  activer
	DlgEditState		dlgEditState;	//!< Niveau de modifications effectues

	/// Chargement du combo des chanes en cours d'enregistrement
	void				remplitCombo();

	/**
	 * \brief Chargement des donnes de l'enregistrement dont l'index est \p ixRecDescr
	 *
	 * \param[in]	initActive	Si \p true, le dialogue sera prpar pour activation si aucune
	 *							temporisation n'existait auparavant pour cet enregistrement
	 *							(afin d'viter  l'utilisateur d'avoir  cliquer sur la case
	 *							d'activation lors de l'ouverture du dialogue). Il restera
	 *							nanmoins ncessaire de valider ce changement.
	 **/
	void				load(bool initActive);

	/// Transfrer le temps de rfrence dans le contrle, selon le mode courant
	void				setCtlTime();

	/// Fonction retournant \p true si un timer est ncessaire
	bool				needTimer() const {
							return delayActive && dispEndTime != useEndTime; }

	/// Conversion en mode "heure de fin"
	void				toEndTime();

	/// Application des modifications
	void				apply();

	/**
	 * \brief Changement de l'tat "modifi" du dialogue
	 * \param[in]	newState	Nouvel tat
	 **/
	void				setChangedState(DlgEditState newState);

	/// Modification de l'tat "actif" de la l'arrt programm d'enregistrement
	void				setActiveState(bool active);

	/// Fonction appele si l'enregistrement se termine pendant que le dialogue est ouvert
	void				recordEnded();

	/// Traitement des messages de la bote de dialogue
	virtual INT_PTR		DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
	/// Constructeur :
	CDelayedStopDialog() :
		CModalDialogBase(IDD_DELAYED_STOP),
		hEnrCombo(NULL),
		hTimeCtl(NULL),
		ixRecDescr(0),
		comboSel(0),
		penreg(NULL),
		refTime(SYSTEMTIME()),
		useEndTime(false),
		dispEndTime(false),
		delayActive(false),
		dlgEditState(des_notModified)
	{
	}
};


/// Chargement du combo des chanes en cours d'enregistrement
void CDelayedStopDialog::remplitCombo()
{
	int res, i, sel=0;

	comboSel = 0;
	for (i = 0; i < enregistrements_actuels.count(); i++) {
		const Enregistrement & enreg = enregistrements_actuels[i];

		if (!enreg.recording())
			continue;
		res = (int)AddComboItemWithData(hEnrCombo, enreg.dispChannelName().c_str(), i);
		if (i == ixRecDescr)
			sel = res;
	}

	comboSel = (int)SendMessage(hEnrCombo, CB_SETCURSEL, sel, 0);
}

/**
 * \brief Chargement des donnes de l'enregistrement dont l'index est \p ixRecDescr
 *
 * \param[in]	initActive	Si \p true, le dialogue sera prpar pour activation si aucune
 *							temporisation n'existait auparavant pour cet enregistrement
 *							(afin d'viter  l'utilisateur d'avoir  cliquer sur la case
 *							d'activation lors de l'ouverture du dialogue). Il restera
 *							nanmoins ncessaire de valider ce changement.
 **/
void CDelayedStopDialog::load(bool initActive)
{
	if (ixRecDescr<0)
		ixRecDescr = 0;
	penreg = &enregistrements_actuels[ixRecDescr];

	CheckDlgButton(hDlg, IDC_DELAY_BY_REMTIME, BST_CHECKED);
	dispEndTime = false;
	if (!penreg->recording()) {
		recordEnded();
	} else  {
		EnableDlgItem(IDC_DELAY_ACTIVE, TRUE);
		if (penreg->apres != apr_null) {
			setActiveState(true);
			SetCtlAfter(hDlg, IDC_AFTER, penreg->apres);
			refTime = penreg->fin;
			useEndTime = true;
			setChangedState(des_notModified);
		} else {
			setActiveState(initActive);
			GetLocalTime(&refTime);
			// Indiquer temps restant = 1/2 heure par dfaut.
			refTime.wHour = refTime.wSecond = refTime.wMilliseconds = 0;
			refTime.wMinute = 30;
			useEndTime = false;
			setChangedState(initActive ? des_noConfirmNeed : des_notModified);
		}
	}
	CheckDlgButton(hDlg, IDC_DELAY_ACTIVE, delayActive);
	setCtlTime();
}

/// Modification de l'tat "actif" de la l'arrt programm d'enregistrement
void CDelayedStopDialog::setActiveState(bool active)
{
	static const int t_id[] =
		{IDC_DELAYED_STOP, IDC_DELAY_LABEL, IDC_DELAY_BY_REMTIME, IDC_DELAY_BY_ENDTIME,
		 IDC_AFTER_RECORD, IDC_AFTER, IDC_AFTER_NOTE};

	delayActive = active;
	for (int i=0; i<_countof(t_id); i++)
		EnableWindow(GetDlgItem(hDlg, t_id[i]), delayActive);
}

/// Fonction appele si l'enregistrement se termine pendant que le dialogue est ouvert
void CDelayedStopDialog::recordEnded()
{
	setActiveState(false);
	EnableDlgItem(IDC_DELAY_ACTIVE, FALSE);
}

/**
 * \brief Changement de l'tat "modifi" du dialogue
 * \param[in]	newState	Nouvel tat
 **/
void CDelayedStopDialog::setChangedState(DlgEditState newState)
{
	bool modified = newState!=des_notModified;
	bool oldActive = penreg && penreg->apres!=apr_null;
	bool activeChanged = delayActive != oldActive;

	dlgEditState = newState;
	EnableDlgItem(IDC_APPLY, modified && (delayActive || activeChanged));
}

/// Conversion en mode "heure de fin"
void CDelayedStopDialog::toEndTime()
{
	if (useEndTime) {
		// Si on utilise dj l'heure de fin, on effectue une double conversion
		// pour s'assurer que celle-ci se trouve bien dans le futur, et non pas
		// plus tt dans la journe (tant donn que le dialogue ne peut pas
		// dfinir la date).
		refTime = EndTimeToRemTime(refTime);
		useEndTime = false;
	}
	if (!useEndTime) {
		refTime = RemTimeToEndTime(refTime);
		useEndTime = true;
	}
}

void CDelayedStopDialog::setCtlTime()
{
	if (dispEndTime==useEndTime)
		DateTime_SetSystemtime(hTimeCtl, GDT_VALID, &refTime);
	else {
		SYSTEMTIME tmptime = (dispEndTime ? RemTimeToEndTime : EndTimeToRemTime)(refTime);

		DateTime_SetSystemtime(hTimeCtl, GDT_VALID, &tmptime);
	}
};

/// Application des modifications
void CDelayedStopDialog::apply()
{
	if (penreg->recording() && dlgEditState!=des_notModified) {	// Toujours en train d'enregistrer ?
		ApresEnregistrement apres = delayActive ? GetCtlAfter(hDlg, IDC_AFTER) : apr_null;

		toEndTime();
		penreg->fin = refTime;
		penreg->apres = apres;

		// Rechercher si l'arrt retard rsultait d'une programmation,
		// afin de mettre  jour celle-ci en mme temps, le cas chant :
		int ixProg = trouve_prog_par_nom(penreg->nom);

		if (ixProg>=0) {
			Programme & prog = Programmes[ixProg];

			programmation_modifiee = true;
			if (prog.repetition != 0) {
				// La programmation comporte des rptitions :
				// on se contente de la "dconnecter" de l'enregistrement, lui permettant
				// ainsi d'tre reporte intgralement  son occurrence suivante.
				prog.etat = epr_supplante;
			} else {
				if (apres == apr_null) {
					// On a annul l'arrt d'enregistrement (celui-ci
					// peut donc se poursuivre indfiniment)
					prog.etat = epr_supplante;
				} else {
					// On a modifi l'heure ou l'action de fin : on reporte les
					// modifications dans la programmation, de sorte qu'un arrt
					// impromptu ventuel du programme suivi d'une relance
					// (plantage ?) reprennne le reliquat ainsi.
					prog.fin = refTime;
					prog.apres = apres;
				}
			}
			record_list_updated();
		}

		setChangedState(des_notModified);
		PostMessage(hMainWnd, WM_TIMECHANGE, 0, 0);
	}
}

/// Traitement des messages de la bote de dialogue
INT_PTR CDelayedStopDialog::DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG:
		InitCtlAfter(hDlg, IDC_AFTER);
		hEnrCombo = GetDlgItem(hDlg, IDC_CHANNEL);
		hTimeCtl = GetDlgItem(hDlg, IDC_DELAYED_STOP);
		ixRecDescr = enregistrements_actuels.getRecordingDescr(ixChaineCourante);
		load(false /*true*/);
		remplitCombo();	// Chargement du combo des enregistrements

		// Cration d'un timer battant  la seconde pour la mise  jour du contrle de saisie
		// de l'heure (selon son mode d'affichage), ainsi que la dtection des
		// enregistrements termins.
		SetTimer(hDlg, TIMER_DELAYED_STOP, 1000, NULL);
		return TRUE;

	case WM_TIMER:
		if (!penreg->recording()) // Tester si l'enregistrement se termine  ce moment
			recordEnded();
		if (needTimer())
			setCtlTime();
		return TRUE;

	case WM_COMMAND: {
		switch(wParam) {

		case _cmd_(IDC_CHANNEL, CBN_SELCHANGE): {
			int newSel = (int)SendMessage(hEnrCombo, CB_GETCURSEL, 0, 0);

			if (newSel != CB_ERR && newSel != comboSel) {
				if (dlgEditState==des_needConfirm) {
					switch (MessageBox(hDlg,
						TEXT("Vous avez apport des modifications  la programmation de la chane en cours.") EOL
						TEXT("Dsirez-vous valider ces modifications ?"),
						TEXT("Changement de chane"),
						MB_YESNOCANCEL)) {

					case IDCANCEL:
						SendMessage(hEnrCombo, CB_SETCURSEL, WPARAM(comboSel), 0);
						return TRUE;

					case IDYES:
						apply();
					case IDNO: ;
					}
				}
				comboSel = newSel;
				ixRecDescr = (int)SendMessage(hEnrCombo, CB_GETITEMDATA, WPARAM(newSel), 0);
				load(false /*true*/);
			}
			return TRUE; }

		case _cmd_(IDC_DELAY_BY_REMTIME, BN_CLICKED):
		case _cmd_(IDC_DELAY_BY_ENDTIME, BN_CLICKED):
			dispEndTime = IsDlgButtonChecked(hDlg, IDC_DELAY_BY_ENDTIME)==BST_CHECKED;
			setCtlTime();
			return TRUE;

		case _cmd_(IDC_DELAY_ACTIVE, BN_CLICKED):
			setActiveState(IsDlgButtonChecked(hDlg, IDC_DELAY_ACTIVE)==BST_CHECKED);
			setChangedState(des_needConfirm);
			return TRUE;

		case _cmd_(IDC_AFTER, CBN_SELCHANGE):
			setChangedState(des_needConfirm);
			return TRUE;

		case _cmd_(IDC_APPLY, BN_CLICKED):
			apply();

			// Transfrer le focus  un autre contrle
			// Voir http://www.olivierlanglois.net/trapped_focus.html
			PostDlgMessage(WM_NEXTDLGCTL, (WPARAM)GetItem(IDOK), TRUE);
			return TRUE;

		case _cmd_(IDOK, BN_CLICKED):
			apply();
			// ...
		case _cmd_(IDCANCEL, BN_CLICKED):
			KillTimer(hDlg, TIMER_DELAYED_STOP);
			EndDialog(hDlg, 0);
			return TRUE;
		}
		break; }

	case WM_NOTIFY:
		if (_ncode_(lParam)==DTN_DATETIMECHANGE) {
			DateTime_GetSystemtime(hTimeCtl, &refTime);
			// Le mode d'affichage utilis sert maintenant de rfrence :
			useEndTime = dispEndTime;
			setChangedState(des_needConfirm);
		}
		return TRUE;
	}

	return FALSE;
}

/**
 * Appel du dialogue de fin d'enregistrement retarde.
 **/
void delayed_stop_dialog()
{
	CDelayedStopDialog().Do(hMainWnd);
}

/**
 * Fonction  appeler en cas de changement externe possible de la liste des enregistrements programms.
 * Assure la sauvegarde des modification et la mise  jour de la liste dans la fentre de dialogue
 * si celle-ci est ouverte.
 **/
void record_list_updated()
{
	if (programmation_modifiee) {
		// Tri de la liste :
		std::sort(Programmes.begin(), Programmes.end());

		CRecordDialog::PostDlgMessage(WM_APP_PROG_CHANGED);	// Reconstruction de la liste si couramment affiche

		// Sauvegarde de la liste des programmes
		sauve_programmes();
	}
}

/**
 * Finalisation d'un changement dans les enregistrements programms.
 **/
void finalize_prog_change()
{
	programmation_modifiee = true;

	// Mise  jour du timer des enregistrements :
	set_timer_record();

	// Mise  jour de l'affichage, et sauvegarde dans le fichier :
	record_list_updated();
}

/**
 * Dmarre tous les enregistrements qui devraient l'tre  cet instant prcis.
 * Retourne 'true' si au moins un changement de nature  affecter les menus
 * ou la programmation d'enregistrement a eu lieu
 **/
static bool do_record_start(const SYSTEMTIME & localtime)
{
	const Chaine *	pRecordingChan = NULL;	// Pointeur sur une chane en cours d'enregistrement, s'il y en a une
	int				ixChaine = -1;
	bool			changement = false;

	// Dterminer si on enregistre quelque chose actuellement, et si oui, quelle est la
	// frquence utilise.
	UINT32 cleChaine = enregistrements_actuels.recording();

	if (cleChaine != 0) {
		ixChaine = cleChaine != STREAM_PSEUDO_CLE ? Canaux.trouve_par_cle(cleChaine) : ixChaineCourante;

		if (ixChaine_ok(ixChaine))
			pRecordingChan = &Canaux[ixChaine];
	}

	// Passer en revue les programmations dfinies afin de dmarrer TOUS les enregistrements qui
	// devraient tre en cours et qui peuvent tre dmarrs.
	for (ITERATE_VECTOR(Programmes, Programme, it)) {
		Programme & prog = *it;	// Description de l'enregistrement  effectuer

		if (prog.etat == epr_encours)
			continue;	// Cet enregistrement est dj en cours
		if (prog.temps_restant(localtime)!=0)
			continue;	// Cet enregistrement n'est pas  effectuer en ce moment

		// Chercher l'index de la chane dont la cl correspond  celle demande pour
		// l'enregistrement :
		ixChaine = Canaux.trouve_par_cle(prog.cleChaine);

		if (ixChaine < 0) {
			changement = true;
			prog.etat = epr_erreur;
			programmation_modifiee = true;
			continue;		// si on n'en a pas trouv (ne devrait pas se produire)
		}

		const Chaine & canal = Canaux[ixChaine];

		if (pRecordingChan) {
			// On ne peut pas effectuer d'enregistrements simultans sur deux frquences diffrentes
			if (!pRecordingChan->sameMultiplex(canal)) {
				changement = true;
				prog.etat = epr_erreur;
				programmation_modifiee = true;
				continue;
			}
		} else {
			// Il n'y avait pas d'autre enregistrement en cours
			if (!ixChaine_ok(ixChaineCourante) || !Canaux[ixChaineCourante].sameMultiplex(canal)) {
				// On change de chane, seulement si le multiplex est diffrent
				HRESULT hr = zappe_index(ixChaine);

				check_hr(hr, "zappe_index / dmarrage enregistement");
			}
		}

		if (!IsMinimized(hMainWnd) && suspend_minimized)
			rebranche(false);

		// Voir si la chane choisie n'est pas dj en cours d'enregistrement :
		int ixRecDescr = prog.methode == meth_multiplex ?
			enregistrements_actuels.getRecordingStream() :
			enregistrements_actuels.getRecordingDescr(ixChaine);

		if (ixRecDescr < 0) {
			// Si elle ne l'est pas, on dmarre l'enregistrement :
			int ixSon = canal.selection_son(prog.audio);// utilis seulement en PS pour le moment

			switch(prog.methode) {
			case meth_TS:
				ixRecDescr = enregistrements_actuels.start_record_ts(ixChaine, &prog);
				break;
			case meth_PS:
				ixRecDescr = enregistrements_actuels.start_record_ps(ixChaine, ixSon, &prog);
				break;
			case meth_multiplex:
				ixRecDescr = enregistrements_actuels.start_record_stream(&prog);
			}

			if (ixRecDescr < 0) {
				// L'enregistrement n'a pas pu dmarrer
				prog.etat = epr_erreur;
				programmation_modifiee = true;
			}

			pRecordingChan = & canal;	// Retenir canal enregistr pour itration suivante de la boucle
		} else
			// Programmation sur une chane dj en cours d'enregistrement :
			enregistrements_actuels[ixRecDescr].modify_prog(&prog);

		changement = true;
	}

	return changement;
}

/**
 * Vrifier si le fait de quitter l'application interromprait un enregistrement en cours,
 * ou bien empcherait un enregistrement imminent.
 * \param[in]	pLocalTime	Pointeur optionnel sur le temps systme courant
 **/
bool can_power_down(const SYSTEMTIME * pLocalTime)
{
	return
		!enregistrements_actuels.recording() &&
		temps_prochain_enregistrement(pLocalTime) >= duree_grace_quit * DUREE_1MN;
}

/**
 * Rechercher tous les enregistrements en cours qui devraient tre stopps maintenant
 * Retourne apr_null si aucun enregistrement n'a t stopp, ou bien une valeur de
 * type 'ApresEnregistrement' reprsentant l'action d'arrt du programme et/ou de
 * l'ordinateur la plus importante qui ait t programme  l'isssue des arrts effectus
 **/
static ApresEnregistrement do_record_stop(const SYSTEMTIME & localtime)
{
	ApresEnregistrement after = apr_null;

	for (int i = 0; i < enregistrements_actuels.count(); i++) {
		Enregistrement & enreg = enregistrements_actuels[i];

		// Voir si l'heure de fin d'enregistrement est atteinte :
		if (enreg.temps_restant(localtime)==0) {
			ApresEnregistrement apres = enreg.stop(epr_termine);

			after = max(after, apres);
		}
	}

	if (after > apr_rien) {
		// Vrifier si d'autres instances de l'application ne seraient pas en train d'enregistrer
		// ou bien auraient un enregistrement programm imminent
		DWORD_PTR	dwResult = 0;

		bPowerStateChangeDenied = false;
		// 'SendMessageTimout' et pas 'SendMessage' car il peut arriver que certaines applications ne rpondent pas, causant un deadlock.
		SendMessageTimeout(HWND_BROADCAST, nWM_APP_INSTANCES, iit_RequestPowerStateChange, (LPARAM)hMainWnd, SMTO_ABORTIFHUNG, 500, &dwResult);

		if (bPowerStateChangeDenied || !can_power_down(&localtime)) {
			// si d'autres enregistrements sont en cours, ou bien si le temps restant avant le
			// prochain enregistrement est infrieur  celui dfini par la variable \p duree_grace_quit,
			// alors on annule toute action d'arrt du programme et/ou de l'ordinateur dfinie  l'issue
			// du ou des arrts effectus
			myprint(TEXT("Extinction ou veille impossible, car enregistrement(s) concurrent(s).") EOL);
			after = min(after, apr_rien);
		}
	}

	return after;
}

/**
 * Excution de toutes les transitions d'enregistrements (dmarrage,
 * arrt, erreur...) qui devraient tre assures  cet instant.
 *
 * \retval	\p true si au moins une transition a eu lieu dans la fonction
 **/
bool do_record_events()
{
	// myprintf(TEXT("Appel do_record_events()") EOL);

	SYSTEMTIME localtime;

	GetLocalTime(&localtime);

	// Traiter tous les enregistrements en cours qui devraient tre arrts maintenant :
	ApresEnregistrement after = do_record_stop(localtime);

	// Traiter tous les enregistrements en cours qui devraient tre dmarrs maintenant :
	if (do_record_start(localtime))
		after = apr_rien;

	// Traiter l'arrt du programme et/ou de l'ordinateur s'il y a lieu :
	switch (after) {

	case apr_null:
		return false;

	case apr_quitter:
		// Arrt de pouchinTV
		SetTimer(hMainWnd, TIMER_QUIT, DUREE_QUIT, NULL);
		break;

	case apr_eteindrePC:
		// Arrt de l'ordinateur
		SetTimer(hMainWnd, TIMER_SHUTDOWN, DUREE_QUIT, NULL);
		break;

	case apr_eteindrePF:
		// Arrt forc de l'ordinateur
		SetTimer(hMainWnd, MAKEWPARAM(TIMER_SHUTDOWN, 1), DUREE_QUIT, NULL);
		break;

	case apr_veille:
		// Mise en veille de l'ordinateur
		SetTimer(hMainWnd, TIMER_SLEEP, DUREE_QUIT, NULL);
		break;

	case apr_veille_pr:
		// Mise en veille prolonge de l'ordinateur
		SetTimer(hMainWnd, MAKEWPARAM(TIMER_SLEEP, 1), DUREE_QUIT, NULL);
	}

	return true;
}

/**
 * Formatage d'une heure de dbut, pour affichage
 * \param[in]	sTime	Heure  formater
 * \param[in]	pszBuf	Tampon recevant la chane formate
 * \param[in]	nCount	Taille du tampon
 * \retval		Nombre de caractres crits dans le tampon.
 **/
UINT32 formate_heure_debut(const SYSTEMTIME & sTime, LPTSTR pszBuf, UINT32 nCount)
{
	return _stprintf_s(pszBuf, nCount, TEXT("%") A2t TEXT(" %i/%02i  %ih%02i"),
		EnumToString(sTime.wDayOfWeek, pszDoW_ucfirst), sTime.wDay, sTime.wMonth, sTime.wHour, sTime.wMinute);
}

/// Chargement de la liste des pistes audio d'une chane
void CRecOtherDialog::FillAudioList(INT nCbIdx)
{
	UINT32	cleChaine	= (UINT32)SendMessage(hCtlChn, CB_GETITEMDATA, nCbIdx, 0);
	int		ixChaine	= cleChaine ? Canaux.trouve_par_cle(cleChaine) : -1;

	if (ixChaine_ok(ixChaine)) {
		Chaine & sChan = Canaux[ixChaine];

		// Insrer les nouveaux items de pistes audio
		SendMessage(hCtlAud, CB_RESETCONTENT, 0, NULL);
		for (int i=0; i<sChan.nbr; i++) {
			TCHAR	szWrk[64];

			sChan.toTrackMenuString(i, szWrk, _countof(szWrk), i+1);
			SendMessage(hCtlAud, CB_ADDSTRING, 0, (LPARAM)szWrk);
		}
		if (ixChaine == ixChaineCourante)
			SendMessage(hCtlAud, CB_SETCURSEL, ixSonCourant, NULL);
		else
			SendMessage(hCtlAud, CB_SETCURSEL, sChan.selection_son(audio_pref), NULL);
	}
}

/// Mise  jour de l'tat d'activation de certains items
void CRecOtherDialog::UpdateStates(MethodeEnregistrement eMeth)
{
	EnableWindow(hCtlChn, nChanCount > 0 && eMeth != meth_multiplex);
	EnableWindow(hCtlAud, nChanCount > 0 && eMeth == meth_PS);
	EnableWindow(GetItem(IDOK), nChanCount > 0 || eMeth == meth_multiplex);
}

/// Traitement des messages de la bote de dialogue
INT_PTR CRecOtherDialog::DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG: {
		hCtlChn = GetDlgItem(hDlg, IDC_CHANNEL);
		hCtlMtd = GetDlgItem(hDlg, IDC_METHOD);
		hCtlAud = GetDlgItem(hDlg, IDC_AUDIO);

		if (epg_typerecvideo == meth_multiplex)
			epg_typerecvideo = meth_TS;			// Ne pas mmoriser le mode multiplex

		const Chaine *	pCurChan	= ixChaine_ok(ixChaineCourante) ? & Canaux[ixChaineCourante] : NULL;

		// Chargement du combo de choix des chanes
		for (ITERATE_CONST_VECTOR(Canaux, Chaine, it)) {
			const Chaine &	cChan = *it;
			TCHAR			szWrk[64];

			if (cChan.etat == ec_inactive)
				continue;
			if (pCurChan && !cChan.sameMultiplex(*pCurChan))
				continue;
			if (enregistrements_actuels.getRecordingDescr(Canaux.iterToIndex(it)) >= 0)
				continue;	// Dj en cours d'enregistrement
			cChan.toMenuString(szWrk, _countof(szWrk));

			// Ajout de l'item, avec association de la cl de la chane pour identifier la slection.
			if (AddComboItemWithData(hCtlChn, szWrk, cChan.cle()) != CB_ERR)
				++nChanCount;
		}

		// Slection de la mthode ; s'il n'y a aucune chane, que seul reste possible le multiplex
		// et que ce mode est autoris dans la configuration, on le slectionne
		if (nChanCount == 0 && allow_stream_record)
			epg_typerecvideo = meth_multiplex;

		remplit_liste_methodes(hCtlMtd, epg_typerecvideo);
		UpdateStates(epg_typerecvideo);
		SendMessage(hCtlChn, CB_SETCURSEL, 0, NULL);
		FillAudioList(0);
		return TRUE; }

	case WM_COMMAND:
		// Commande reue (clic, etc...)
		switch(wParam) {

		case _cmd_(IDC_CHANNEL, CBN_SELCHANGE):
			FillAudioList((INT)SendMessage(hCtlChn, CB_GETCURSEL, 0, NULL));
			return TRUE;

		case _cmd_(IDC_METHOD, CBN_SELCHANGE):
			epg_typerecvideo = (MethodeEnregistrement)SendMessage(hCtlMtd, CB_GETCURSEL, 0, NULL);
			UpdateStates(epg_typerecvideo);
			return TRUE;

		case _cmd_(IDOK, BN_CLICKED): {
			// Clic sur le bouton OK
			UINT32	cleChaine	= (UINT32)SendMessage(hCtlChn, CB_GETITEMDATA,
									SendMessage(hCtlChn, CB_GETCURSEL, 0, NULL), 0);
			int		ixChaine	= cleChaine ? Canaux.trouve_par_cle(cleChaine) : -1;

			switch ((MethodeEnregistrement)SendMessage(hCtlMtd, CB_GETCURSEL, 0, NULL)) {

			case meth_TS:
				if (ixChaine_ok(ixChaine))
					enregistrements_actuels.start_record_ts(ixChaine);
				break;

			case meth_PS:
				if (ixChaine_ok(ixChaine))
					enregistrements_actuels.start_record_ps(
						ixChaine, (int)SendMessage(hCtlAud, CB_GETCURSEL, 0, NULL));
				break;

			case meth_multiplex:
				if (enregistrements_actuels.getRecordingStream() >= 0) {
					MessageBox(hDlg, TEXT("Le multiplex est dj en cours d'enregistrement"),
						TEXT("Erreur nouvel enregistrement"), MB_ICONERROR);
					return TRUE;
				}
				enregistrements_actuels.start_record_stream();
			}
			EndDialog(hDlg, IDOK);
			return TRUE;
		}

		case _cmd_(IDCANCEL, BN_CLICKED):
			// Clic sur le bouton Annuler
			EndDialog(hDlg, IDCANCEL);
			return TRUE;
		}
	}

	// Autre message, trait par Windows
	return FALSE;
}

// ====================================================================================
// Timer de rveil
// ====================================================================================

/** \brief Dfinir  un temps donn
 * \param[in]	nDelay	Temps avant rveil (millisecondes d'attente)
 * \return	\p true ou \p false selon succs ou chec
 **/
bool CWaitTimer::Set(UINT nDelay)
{
	// 1 = timer absolu, 0 = timer relatif
	#define WT_ABSOLUTE 1

	if (!hWTimer) {
		myprint(TEXT("CWaitTimer::Set: CreateWaitableTimer") EOL);
		hWTimer = CreateWaitableTimer(NULL, FALSE, NULL);
		if (!hWTimer) {
			myprint(TEXT("CWaitTimer::Set: Timer non cr") EOL);
			return false;
		}
	}

	LARGE_INTEGER	sTime;
#if WT_ABSOLUTE
	SYSTEMTIME		sSysTime;
	FILETIME		sFilTime;

	GetSystemTime(&sSysTime);
	SystemTimeToFileTime(&sSysTime, &sFilTime);
	sTime.LowPart = sFilTime.dwLowDateTime;
	sTime.HighPart = sFilTime.dwHighDateTime;
	sTime.QuadPart += LONGLONG(nDelay) * 10000I64;

	//INT64	nDebugTime = sTime.QuadPart/10000I64;

	//myprintf(TEXT("CWaitTimer::Set: SetWaitableTimer = %iJ, %ih, %imn, %i.%03iS") EOL,
	//	INT32(nDebugTime / DUREE_24H),
	//	INT32((nDebugTime / DUREE_1H) % 24),
	//	INT32((nDebugTime / DUREE_1MN) % 60),
	//	INT32((nDebugTime / DUREE_1SEC) % 60),
	//	INT32(nDebugTime % DUREE_1SEC));
#else
	// Temps restant spcifi en tant que valeur ngative en units de 100 nanosecondes.
	// La valeur ngative indique un temps relatif plutt qu'un temps absolu (utiliser un
	// temps absolu aurait pos le problme de la conversion en temps universel (UTC)).

	/// \todo	Attention : le temps restant limite est de 24 jours, 20 heures, 31 minutes,
	///			23 secondes et 646 millisecondes (= 0x7ffffffe). Voir si on peut amliorer a.
	sTime.QuadPart = LONGLONG(nDelay) * -10000I64;

	INT32	nDebugTime = INT32(-sTime.QuadPart/10000I64);

	myprintf(TEXT("CWaitTimer::Set: SetWaitableTimer = %iJ, %ih, %imn, %i.%03iS") EOL,
		nDebugTime / DUREE_24H,
		(nDebugTime / DUREE_1H) % 24,
		(nDebugTime / DUREE_1MN) % 60,
		(nDebugTime / DUREE_1SEC) % 60,
		nDebugTime % DUREE_1SEC);
#endif

	if (!SetWaitableTimer(hWTimer, &sTime, 0, NULL, NULL, TRUE)) {
		myprintf(ERRX(TEXT("CWaitTimer::Set: chec SetWaitableTimer : %s")), SystemErrorToString().c_str());
		return false;
	}
	nLastDelay = nDelay;
	return true;
#undef WT_ABSOLUTE
}

/// Annulation du timer de rveil
void CWaitTimer::Cancel()
{
	if (!hWTimer) {
		// myprint(TEXT("CWaitTimer::Cancel: Timer inexistant") EOL);
		return;
	}
	myprint(TEXT("CWaitTimer::Cancel: Suppression du timer de rveil") EOL);
	CancelWaitableTimer(hWTimer);
	CloseHandle(hWTimer);
	hWTimer = NULL;
	nLastDelay = DUREE_INFINIE;
}

// ====================================================================================
// Informations
// ====================================================================================

static tstring AddAndFormat(SYSTEMTIME sTime, LONG nToAdd)
{
	TCHAR		szTmp[64];

	AddTime(sTime, nToAdd);
	formate_heure_debut(sTime, szTmp, _countof(szTmp));
	return szTmp;
}

/// Affichage d'informations concernant les prochains enregistrements
void DisplayRecordInfo(HWND hParentWnd)
{
	SYSTEMTIME	sCurrTime;

	GetLocalTime(&sCurrTime);

	Programme *	pProg = NULL;
	LONG		nTimeBeforeNextRecord = temps_prochain_enregistrement(&sCurrTime, &pProg);
	tstring		strMsg;
	tstring		strRecInfo;
	int			nRec = 0;

	for (int nInx=0; nInx < enregistrements_actuels.count(); ++nInx) {
		const Enregistrement & cRec = enregistrements_actuels[nInx];

		if (cRec.recording()) {
			TCHAR	szTmp[128];

			if (cRec.apres == apr_null) {
				_stprintf_s(szTmp, EOL TEXT("%s, aucune fin dfinie."),
					cRec.dispChannelName().c_str());
			} else {
				_stprintf_s(szTmp, EOL TEXT("%s, fin  %ih%02i."),
					cRec.dispChannelName().c_str(), cRec.fin.wHour, cRec.fin.wMinute);
			}
			strRecInfo += szTmp; 
			++nRec;
		}
	}
	if (nRec > 0) {
		if (nRec == 1)
			strMsg = TEXT("Un enregistrement est en cours:");
		else if (nRec > 1)
			strMsg = tstr_printf(TEXT("%u enregistrements sont en cours:"), nRec);
		strMsg += strRecInfo;
	}

	if (nTimeBeforeNextRecord != DUREE_INFINIE) {
		strMsg += EOL TEXT("Le prochain enregistrement");
		if (pProg)
			strMsg += tstr_printf(TEXT(" (%s)"), pProg->progDescr().c_str());
		strMsg += TEXT(" dmarrera ")
			+ AddAndFormat(sCurrTime, nTimeBeforeNextRecord) + TEXT(".");
	}
	if (pProg && pProg->tache == planif_avec)
		strMsg += EOL TEXT("Pour l'effectuer, une tche programme relancera ventuellement le programme ")
			+ AddAndFormat(sCurrTime, nTimeBeforeNextRecord-DUREE_PLANIF) + TEXT(".");
	if (cWaitTimer.GetCurrent() != DUREE_INFINIE) {
		strMsg +=
			tstr_printf(
				EOL TEXT("Si l'ordinateur est laiss en mode veille sans quitter l'application,")
				TEXT(" un rveil automatique aura lieu %s."),
			AddAndFormat(sCurrTime, cWaitTimer.GetCurrent()).c_str());
	}
	if (strMsg.empty())
		strMsg = TEXT("Aucun enregistrement n'est actuellement en cours ou planifi.");
	if ((BYTE)strMsg[0] < BYTE(' '))
		strMsg.erase(0, 1);
	MessageBox(hParentWnd, strMsg.c_str(), TEXT("Information enregistements"), MB_OK);
}