/*
 * $Id: search.cpp 370 2020-09-09 12:43:56Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentations des classes ncessaires  la recherche de filtres
 *
 *	Implmentations des classes qui sont utilises pour rechercher quels
 *	sont les tuners et les codecs qui sont disponibles sur l'ordinateur.
 **/

#include "stdafx.h"
#include "search.h"

// ====================================================================================
// Descripteur de configuration tuner
// ====================================================================================

bool DeviceFilterInfo::Valid() const
{
	return
		! IsNullName(strName.c_str()) &&
		! IsNullName(strPath.c_str());
}

/// Ajout d'un rsultat valide
void DeviceFilterInfo::Set(LPCTSTR pszName, LPCTSTR pszPath)
{
	strName = pszName;

	// Si des espaces existent en fin de la chane "device path" (ce qui est vaguement
	// un bug de driver de certains tuners, cf par exemple Hauppauge WinTV-NOVA-T-500 version
	// driver 4.3.27240.0), ceux-ci DOIVENT tre prservs. La lecture du fichier
	// "config.ini" les liminant automatiquement, il convient donc d'encadrer la chane
	// avec des crochets (les guillemets ne sont pas adquats, car ils sont aussi limins)
	// qui devront tre retirs plus tard.
	strPath = TEXT("[");
	strPath += pszPath;
	strPath += TEXT("]");
}

/// Comparaison pour tri
bool DeviceInfo::operator < (const DeviceInfo & s2) const
{
	if (asDev[dt_Tuner].strName == s2.asDev[dt_Tuner].strName)
		return nDevCount < s2.nDevCount;
	return asDev[dt_Tuner].strName < s2.asDev[dt_Tuner].strName;
}

void DeviceInfoList::FindDuplicates()
{
	for (ITERATE_VECTOR(*this, DeviceInfo, it1)) {
		for (iterator it2 = it1+1;  it2 != end();  it2++) {
			if (it1->asDev[dt_Tuner].strName == it2->asDev[dt_Tuner].strName) {
				// Doublon trouv : lui donner le numro du prcdent + 1 et passer au suivant
				it2->nDevCount = it1->nDevCount + 1;
				break;
			}
		}
	}
}

// ====================================================================================
// Recherche de codecs
// ====================================================================================

CSearchByType::CSearchByType(GUID type, GUID subtype, bool bExactMatch) :
	hr(E_FAIL)
{
	hr = pMapper.CoCreateInstance(CLSID_FilterMapper2, NULL, CLSCTX_INPROC);
	if (check_hr(hr, "CSearchByType / CoCreateInstance"))
		return;

	GUID arrayInTypes[2] = {type, subtype};

	hr = pMapper->EnumMatchingFilters(
			&pEnum,
			0,					// Reserved.
			bExactMatch,		// Use exact match?
			MERIT_DO_NOT_USE,	// Minimum merit.
			TRUE,				// At least one input pin?
			1,					// Number of major type/subtype pairs for input.
			arrayInTypes,		// Array of major type/subtype pairs for input.
			NULL,				// Input medium.
			NULL,				// Input pin category.
			FALSE,				// Must be a renderer?
			TRUE,				// At least one output pin?
			0,					// Number of major type/subtype pairs for output.
			NULL,				// Array of major type/subtype pairs for output.
			NULL,				// Output medium.
			NULL);				// Output pin category.

	if (check_hr(hr, "CSearchByType / EnumMatchingFilters"))
		return;
}

HRESULT CSearchByType::Do()
{
	if (FAILED(hr))
		return hr;		// Il y a eu une erreur  l'initialisation

	// numration des "monikers".
	for (ENUM_ITF(pEnum, IMoniker, pMoniker)) {
		MonikerBindings	cMnkb(pMoniker, &hr);

		if (!cMnkb.isOk())
			continue;
		hr = ApplyOne(cMnkb);
		if (hr==S_FALSE)
			return S_OK;	// On a trouv quelque chose
		if (FAILED(hr))
			return hr;		// Il y a eu une erreur
	}

	return E_FAIL;			// Aucun filtre trouv
}

// ====================================================================================
// Recherche de filtres BDA
// ====================================================================================

CSearchByCategory::CSearchByCategory(CLSID clsid) :
	hr(E_FAIL)
{
	hr = pDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER);
	if (check_hr(hr, "CSearchByCategory / ICreateDevEnum"))
		return;

	hr = pDevEnum->CreateClassEnumerator(clsid, &pEnumMoniker, 0);
	if (check_hr(hr, "CSearchByCategory / cration Enum Moniker") || pEnumMoniker == NULL)
		return;
}

/**
 * Ajout d'un nom de filtre  une liste de noms, en s'assurant que ce nom n'existe pas dj dans
 * la liste. Si c'est le cas, le nom est modifi afin de devenir unique.
 * \param[in, out]	vStrFiltres	Vecteur, liste de noms  modifier
 * \param[in]		pszName		Nom  vrifier et  ajouter
 * \return			Nom ventuellement modifi effectivement ajout.
 **/
LPCTSTR CSearchByCategory::AddToVector(vector<tstring> & vStrFiltres, LPCTSTR pszName)
{
	tstring						strTmp(pszName);
	UINT						nCount = 1;

	while (find(vStrFiltres.begin(), vStrFiltres.end(), strTmp)!=vStrFiltres.end()) {
		TCHAR	szNum[8];

		_stprintf_s(szNum, TEXT(" #%u"), ++nCount);
		strTmp = pszName;
		strTmp += szNum;
	}
	vStrFiltres.push_back(strTmp);
	return vStrFiltres.rbegin()->c_str();
}

HRESULT CSearchByCategory::Do()
{
	if (hr != S_OK)
		return hr;		// Il y a eu une erreur  l'initialisation

	if (!pEnumMoniker)
		return S_FALSE;

	// numration des "monikers".
	for (ENUM_ITF(pEnumMoniker, IMoniker, pMoniker)) {
		MonikerBindings	cMnkb(pMoniker, &hr);

		if (!cMnkb.isOk())
			continue;
		hr = cMnkb.BindToFilter();
		if (FAILED(hr))
			continue;
		hr = ApplyOne(cMnkb);
		if (FAILED(hr) || hr==S_FALSE)
			break;
	}
	return hr;
}

// ====================================================================================
// Recherche de rcepteur avance
// ====================================================================================

/**
 * Constructeur : prpare l'numration pour la recherche d'un filtre rcepteur
 *  associer  un tuner donn.
 * \param[in]		pTstGraph	Graphe dj construit jusqu'au tuner.
 * \param[in]		bFrst		\p true au 1er niveau de profondeur, \p false au second
 * \param[in,out]	vRslt		Donnes tuner  complter.
 **/
CSearchTuner::CSearchReceiver::CSearchReceiver(IPTvM_TunerGraph * pTstGraph, bool bFrst, DeviceInfo & vRslt) :
	CSearchByCategory(KSCATEGORY_BDA_RECEIVER_COMPONENT),
	vResult(vRslt),
	pTestGraph(pTstGraph),
	bFirst(bFrst)
{
}

/**
 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
 * \param[in]	cMnkb	Structure contenant les informations de recherche
 * \return		\p S_OK car tous les filtres doivent tre numrs
 **/
HRESULT CSearchTuner::CSearchReceiver::ApplyOne(MonikerBindings & cMnkb)
{
	HRESULT hr2;	// Pour les codes d'erreur qu'on compte ne pas retourner
	tstring strPath;

	hr2 = cMnkb.GetDevicePath(strPath);	// Obtenir le "device path"
	if (FAILED(hr2))
		return S_OK;	// Invalide : S_OK pour continuer  chercher

	// Tenter d'ajouter le filtre au graphe
	myprintf(TEXT("%*sEssai %s\n"), bFirst ? 4 : 6, TEXT(""), cMnkb.strName.c_str());
	hr2 = pTestGraph->PushFilter(cMnkb.pFilter, cMnkb.strName.c_str());
	myprintf(TEXT("%*s-- rsultat essai, hr=0x%x\n"), bFirst ? 4 : 6, TEXT(""), hr2);
	if (FAILED(hr2))
		return S_OK;	// Invalide : S_OK pour continuer  chercher

	DeviceType eDevType = dt_Capture;	// dt_Capture par dfaut

	if (bFirst) {
		// Premier filtre rcepteur trouv : rechercher s'il n'y en a pas un second
		hr2 = CSearchReceiver(pTestGraph, false, vResult).Do();
		myprintf(TEXT("    -- rsultat recherche second rcepteur, hr=0x%x\n"), hr2);

		// Si un filtre de capture a t trouv en 3me position, alors celui-ci est un filtre
		// dmodulateur
		if (vResult.asDev[dt_Capture].Valid())
			eDevType = dt_Demod;
	}

	// Ajout du filtre  la position qui lui a t assigne
	vResult.asDev[eDevType].Set(cMnkb.strName.c_str(), strPath.c_str());
	myprintf(TEXT("%*sDevice path=\"%s\"\n"), bFirst ? 4 : 6, TEXT(""), vResult.asDev[eDevType].strPath.c_str());

	// Retirer le filtre
	hr2 = pTestGraph->PopFilter();
	myprintf(TEXT("%*sRetrait %s, hr=0x%x\n"), bFirst ? 4 : 6, TEXT(""), cMnkb.strName.c_str(), hr2);

	// Retour avec S_FALSE pour indiquer qu'on arrte de chercher
	return S_FALSE;
}

// ====================================================================================
// Recherche de tuner avance
// ====================================================================================

/**
 * Constructeur : prpare l'numration pour le remplissage d'une
 * liste droulante avec une liste de filtres tuner.
 * \param[out]	vRslts	Tableau de rsultats  remplir.
 * \param[in]	bLegacy	\p true si ancien graphe  utiliser.
 **/
CSearchTuner::CSearchTuner(DeviceInfoList & vRslts, bool bLegacy) :
	CSearchByCategory(KSCATEGORY_BDA_NETWORK_TUNER),
	vResults(vRslts),
	pTestGraph(new_TunerGraph(bLegacy))
{
	myprintf(TEXT("Dbut recherche tuners, creation du graphe\n"));
	if (pTestGraph.get()) {
		hr = pTestGraph->Build();

		check_hr(hr, "construction du graphe");
	}
}

/**
 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
 * \param[in]	cMnkb	Structure contenant les informations de recherche
 * \return		\p S_OK car tous les filtres doivent tre numrs
 **/
HRESULT CSearchTuner::ApplyOne(MonikerBindings & cMnkb)
{
	if (!pTestGraph.get())
		return S_FALSE;		// Pas de graphe

	if (FAILED(hr))
		return hr;			// Erreur  la construction

	HRESULT hr2;	// Pour les codes d'erreur qu'on compte ne pas retourner

	// Envoi d'un "tune request" (sauf si graphe alternatif, dans lequel
	// cette fonctionnalit n'est pas exploitable  ce stade)
	/** \todo Probablement plus ncessaire, maintenant que cette action a t ajoute
		dans CMicrosoftTunerGraph::AddProvider. Vrifier quand mme qu'il n'est pas requis
		qu'elle soit refaite  chaque insertion d'un nouveau tuner. */
	if (!pTestGraph->isLegacyGraph()) {
		hr2 = pTestGraph->change_frequence(-1, -1);
		if (FAILED(hr2)) {
			myprintf(ERRX(TEXT("Erreur : slection frquence impossible, hr=0x%x")), hr2);
			// return hr2;
		}
	}

	tstring strPath;

	hr2 = cMnkb.GetDevicePath(strPath);	// Obtenir le "device path"
	if (FAILED(hr2))
		return S_OK;	// Invalide : S_OK pour continuer  chercher

	// Tenter d'ajouter le filtre au graphe
	myprintf(TEXT("  Essai %s\n"), cMnkb.strName.c_str());
	hr2 = pTestGraph->PushFilter(cMnkb.pFilter, cMnkb.strName.c_str());
	myprintf(TEXT("  -- rsultat essai, hr=0x%x\n"), hr2);
	if (FAILED(hr2))
		return S_OK;	// chec : on retourne nanmoins avec S_OK (pour continuer  chercher)

	// Construction du rsultat
	DeviceInfo sInf;

	sInf.asDev[dt_Tuner].Set(cMnkb.strName.c_str(), strPath.c_str());
	myprintf(TEXT("  Device path=\"%s\"\n"), sInf.asDev[dt_Tuner].strPath.c_str());
	// Chercher les filtres matriels supplmentaires qui pourraient tre requis

	hr2 = CSearchReceiver(pTestGraph.get(), dt_Demod, sInf).Do();
	myprintf(TEXT("  -- rsultat recherche rcepteur, hr=0x%x\n"), hr2);

	// Retirer le filtre tuner
	hr2 = pTestGraph->PopFilter();
	myprintf(TEXT("  Retrait %s, hr=0x%x\n"), cMnkb.strName.c_str(), hr2);

	// Ajouter le rsultat  l'ensembles des rsultats de la liste
	vResults.push_back(sInf);

	return S_OK;
}

/**
 * Application de la mthode virtuelle \p ApplyOne  chaque filtre trouv.
 * \return Code d'erreur DirectShow
 **/
HRESULT CSearchTuner::Do()
{
	HRESULT hr2 = __super::Do();

	if (hr2 == S_OK) {
		// Recherche et marquage des filtres qui auraient des noms de tuners identiques
		vResults.FindDuplicates();

		// Tri de la liste
		sort(vResults.begin(), vResults.end());
	}
	return hr2;
}

// ====================================================================================
// Recherche d'informations dans la base de registre
// ====================================================================================

/// Comparaison des "device paths" entre eux pour dterminer comment ils sont groups
UsedDevices	NamesAndDevPaths::DevPathCompare() const
{
	if (strCapturePath.empty()) {
		// Note : il est considr impossible qu'un filtre dmodulateur existe mais qu'il
		// n'y ait pas de filtre de capture. Donc s'il n'y a pas de filtre de capture,
		// on n'a qu'un filtre tuner. Esprons que ce ne soit pas une erreur.
		return ud_t;
	}

	if (strDemodPath.empty()) {
		// Il n'y a pas de filtre dmodulateur (ce qui est sans doute le cas le plus courant).
		if (strTunerPath == strCapturePath)
			return ud_tc;
		return ud_t_plus_c;
	}

	if (strTunerPath == strDemodPath) {
		// Le filtre tuner et le filtre dmodulateur correspondent au mme pilote
		if (strDemodPath == strCapturePath)
			return ud_tdc;
		return ud_td_plus_c;
	}

	if (strDemodPath == strCapturePath) {
		// Le filtre dmodulateur et le filtre de capture correspondent au mme pilote
		return ud_t_plus_dc;
	}

	// On a carrment trois pilotes diffrents
	return ud_t_plus_d_plus_c;
}

/// Constructeur
CRegControlSet::CRegControlSet()
{
	regCtlSetRoot.Open(HKEY_LOCAL_MACHINE, TEXT("SYSTEM\\CurrentControlSet"));
}

/// Extraire tous les noms de filtres et "device path" (courts) de la configuration du tuner
bool CRegControlSet::GetAllNamesAndPaths(const TunerProps & sProps, NamesAndDevPaths & sNnP)
{
	sNnP.strTunerName	= MonikerBindings(sProps.szPathTuner).strName.c_str();

	bool	bRes	= GetDevicePath(KSCATEGORY_BDA_NETWORK_TUNER, sNnP.strTunerName.c_str(), sNnP.strTunerPath);

	if (!IsNullName(sProps.szPathDemod)) {
		sNnP.strDemodName = MonikerBindings(sProps.szPathDemod).strName.c_str();
		bRes &= GetDevicePath(KSCATEGORY_BDA_RECEIVER_COMPONENT, sNnP.strDemodName.c_str(), sNnP.strDemodPath);
	}

	if (!IsNullName(sProps.szPathCapture)) {
		sNnP.strCaptureName = MonikerBindings(sProps.szPathCapture).strName.c_str();
		bRes &= GetDevicePath(KSCATEGORY_BDA_RECEIVER_COMPONENT, sNnP.strCaptureName.c_str(), sNnP.strCapturePath);
	}

	return bRes;
}

/// Obtenir un "device path" (court)  partir d'une catgorie et d'un nom de filtre
bool CRegControlSet::GetDevicePath(const CLSID & clsidCategory, LPCTSTR pszFilterName, tstring & strDevicePath)
{
	CRegKey_tstr regDeviceClasses;

	strDevicePath.clear();
	/// \todo S'assurer que a marche dans les comptes limits !
	if (
		!regDeviceClasses.Open(regCtlSetRoot,
			tstr_printf(
				TEXT("Control\\DeviceClasses\\") GUIDFMT,
				GUIDPARM(clsidCategory)).c_str())
	)
		return false;

	tstring	strDeviceClass;

	for (ENUM_REG_KEYS(regDeviceClasses, strDeviceClass)) {
		CRegKey_tstr regDeviceClass;

		if (!regDeviceClass.Open(regDeviceClasses, strDeviceClass.c_str()))
			continue;

		tstring	strDeviceSubClass;

		for (ENUM_REG_KEYS(regDeviceClass, strDeviceSubClass)) {
			CRegKey_tstr regDeviceSubClass;

			if (strDeviceSubClass[0] != '#')	// Seules les cls ne commenant pas par "#"
				continue;						// nous concernent

			strDeviceSubClass += TEXT("\\Device Parameters");
			if (regDeviceSubClass.Open(regDeviceClass, strDeviceSubClass.c_str())) {
				tstring	strFriendlyName;

				regDeviceSubClass.QueryStringValue(TEXT("FriendlyName"), strFriendlyName);

				// _tcsstr et pas _tcscmp, car le nom du filtre peut avoir t nettoy
				// de certains espaces excdentaires.
				if (_tcsstr(strFriendlyName.c_str(), pszFilterName)) {
					regDeviceClass.QueryStringValue(TEXT("DeviceInstance"), strDevicePath);
					return true;
				}
			}
		}
	}

	return false;
}

/// Obtenir le chemin d'accs au pilote
bool CRegControlSet::GetDriverPath(LPCTSTR pszDevicePath, tstring & strDriverPath)
{
	CRegKey_tstr regDevice;

	strDriverPath.clear();
	if (
		!regDevice.Open(
			regCtlSetRoot,
			tstr_printf(TEXT("Enum\\%s"), pszDevicePath).c_str())
	)
		return false;

	//regDevice.QueryStringValue(TEXT("DeviceDesc"), strDeviceDesc);
	//regDevice.QueryStringValue(TEXT("Mfg"), strMfg);
	regDevice.QueryStringValue(TEXT("Driver"), strDriverPath);
	return true;
}

/// Obtenir toutes les informations d'un pilote
bool CRegControlSet::GetDeviceInfo(LPCTSTR pszDevicePath, RegFilterInfo & sInfo)
{
	LPCTSTR			pszCnxSearch = _tcschr(pszDevicePath, TCHAR('\\'));

	if (pszCnxSearch) {
		// Extraction du nom de connexion du chemin de priphrique
		TCHAR		szCnx[16];
		size_t		nCnxLen = pszCnxSearch-pszDevicePath;

		if (nCnxLen<_countof(szCnx)) {
			/// Extraction du type de connexion (pas officiel et apparemment trs
			/// souvent erron) \deprecated Supprimer ?
			_tcsncpy_s(szCnx, pszDevicePath, nCnxLen);
			szCnx[nCnxLen] = 0;
			sInfo.strConnection = szCnx;
		}
	}

	tstring	strDriverPath;

	if (!GetDriverPath(pszDevicePath, strDriverPath))
		return false;

	CRegKey_tstr	regDriver;

	if (
		!regDriver.Open(
			regCtlSetRoot,
			tstr_printf(TEXT("Control\\Class\\%s"), strDriverPath.c_str()).c_str())
	)
		return false;

	regDriver.QueryStringValue(TEXT("NTMPDriver"),		sInfo.strDriver);
	if (sInfo.strDriver.empty())
		regDriver.QueryStringValue(TEXT("Driver"),		sInfo.strDriver);
	regDriver.QueryStringValue(TEXT("DriverVersion"),	sInfo.strDriverVersion);
	regDriver.QueryStringValue(TEXT("DriverDate"),		sInfo.strDriverDate);
	regDriver.QueryStringValue(TEXT("DriverDesc"),		sInfo.strDriverDesc);
	regDriver.QueryStringValue(TEXT("ProviderName"),	sInfo.strProviderName);
	return true;
}
