/*
 * $Id: search.h 275 2011-02-27 18:22:05Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions des classes ncessaires  la recherche de filtres
 *
 *	Dfinitions des classes qui sont utilises pour rechercher quels
 *	sont les tuners et les codecs qui sont disponibles sur l'ordinateur.
 **/

#pragma once

// ====================================================================================
// Recherche de codecs
// ====================================================================================

/**
 * Classe d'assistance  l'numration de codecs ayant des
 * caractristiques donnes.
 **/
class CSearchByType
{
	CComPtr<IFilterMapper2> pMapper;
	CComPtr<IEnumMoniker>	pEnum;
public:
	HRESULT					hr;

	/**
	 * Constructeur : prpare l'numration pour une paire de types donns.
	 * \param[in] type			CLSID pour le type
	 * \param[in] subtype		CLSID pour le sous-type
	 * \param[in] bExactMatch	true si correspondance exacte requise
	 **/
	CSearchByType(GUID type, GUID subtype, bool bExactMatch);

	/**
	 * Mthode virtuelle "pure" appele pour chaque filtre trouv.
	 * \param[in]	cMnkb		Structure contenant les informations de recherche
	 * \retval		\p S_OK		si l'numration doit ensuite continuer
	 * \retval		\p S_FALSE	si l'numration doit s'arrter (par exemple parce qu'on
	 *							a trouv ce qu'on cherchait)
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb) = 0;

	/**
	 * Application de la mthode virtuelle \p ApplyOne  chaque filtre trouv.
	 * \return Code d'erreur DirectShow
	 **/
	HRESULT Do();
};

/**
 * Classe d'assistance au remplissage d'un tableau de chanes de caractres
 * avec une liste de codecs ayant des caractristiques donnes.
 */
class CSearchByType_FillVector : public CSearchByType
{
	vector<tstring>	& vStrFiltres;
public:
	/**
	 * Constructeur : prpare l'numration pour le remplissage d'un
	 * vecteur de chanes de caractres avec une liste de codecs ayant des
	 * caractristiques donnes.
	 * \param[in]	type		CLSID pour le type
	 * \param[in]	subtype		CLSID pour le sous-type
	 * \param[in]	bExactMatch	true si correspondance exacte requise
	 * \param[in]	vStrFlt		tableau de chanes de caractres  remplir.
	 **/
	CSearchByType_FillVector(GUID type, GUID subtype, bool bExactMatch, vector<tstring>	& vStrFlt) :
		CSearchByType(type, subtype, bExactMatch),
		vStrFiltres(vStrFlt)
	{
	}

	/**
	 * Implmentation de mthode virtuelle appele pour chaque codec trouv.
	 * \param[in]	cMnkb	Structure contenant les informations de recherche
	 * \return		\p S_OK car tous les filtres doivent tre numrs
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb)
	{
		vStrFiltres.push_back(cMnkb.strName);
		return S_OK;
	}
};

/**
 * Classe d'assistance  la recherche d'un codec ayant des caractristiques
 * donnes,  partir de son nom,
 */
class CSearchByType_Get : public CSearchByType
{
	LPCTSTR					pszSearch;
	CComPtr<IBaseFilter> &	pFilter;
public:
	/**
	 * Constructeur : prpare l'numration pour le remplissage d'une
	 * liste droulante avec une liste de filtres dans une catgorie donne.
	 * \param[in]	type		CLSID pour le type
	 * \param[in]	subtype		CLSID pour le sous-type
	 * \param[in]	bExactMatch	true si correspondance exacte requise
	 * \param[in]	pszSrch		Nom du codec recherch.
	 * \param[out]	pFltr		Pointeur ATL destin recevoir l'interface du filtre trouv, s'il existe.
	 **/
	CSearchByType_Get(GUID type, GUID subtype, bool bExactMatch, LPCTSTR pszSrch, CComPtr<IBaseFilter> & pFltr) :
		CSearchByType(type, subtype, bExactMatch),
		pszSearch(pszSrch),
		pFilter(pFltr)
	{
		myprintf(
			TEXT("CSearchByType_Get: nom=%s") EOL
			TEXT("  type=") GUIDFMT_N EOL
			TEXT("  subtype=") GUIDFMT_N EOL,
			pszSrch, GUIDPARM_N(type), GUIDPARM_N(subtype));

		pFltr.Release();	// prcaution
	}

	/**
	 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
	 * \param[in]	cMnkb		Structure contenant les informations de recherche
	 * \retval		\p S_OK		tant que le codec n'a pas t trouv
	 * \retval		\p S_FALSE	quand le codec a t trouv, afin d'arrter la recherche
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb)
	{
		if (cMnkb.strName == pszSearch) {
			if (SUCCEEDED(hr = cMnkb.BindToFilter())) {
				pFilter = cMnkb.pFilter;
				return S_FALSE;
			}
		}

		return S_OK;
	}
};

// ====================================================================================
// Recherche de filtres BDA
// ====================================================================================

/**
 * Classe d'assistance  l'numration des filtres dans une catgorie donne.
 **/
class CSearchByCategory
{
	CComPtr<ICreateDevEnum> pDevEnum;
	CComPtr<IEnumMoniker>	pEnumMoniker;

protected:

	HRESULT					hr;

	/**
	 * Ajout d'un nom de filtre  une liste de noms, en s'assurant que ce nom n'existe pas dj dans
	 * la liste. Si c'est le cas, le nom est modifi afin de devenir unique.
	 * \param[in, out]	vStrFiltres	Vecteur, liste de noms  modifier
	 * \param[in]		pszName		Nom  vrifier et  ajouter
	 * \return			Nom ventuellement modifi effectivement ajout.
	 **/
	static LPCTSTR AddToVector(vector<tstring> & vStrFiltres, LPCTSTR pszName);

public:

	/**
	 * Constructeur : prpare l'numration pour une catgorie donne.
	 * \param[in] clsid CLSID pour la catgorie
	 **/
	CSearchByCategory(CLSID clsid);

	/**
	 * Mthode virtuelle "pure" appele pour chaque filtre trouv.
	 * \param[in]	cMnkb		Structure contenant les informations de recherche
	 * \retval		\p S_OK		si l'numration doit ensuite continuer
	 * \retval		\p S_FALSE	si l'numration doit s'arrter (par exemple parce qu'on
	 *							a trouv ce qu'on cherchait)
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb) = 0;

	/**
	 * Application de la mthode virtuelle \p ApplyOne  chaque filtre trouv.
	 * \return Code d'erreur DirectShow
	 **/
	HRESULT Do();
};

/**
 * Classe d'assistance  la recherche d'un filtre  partir de son nom, dans une catgorie donne.
 */
class CSearchByCategory_Get : public CSearchByCategory
{
	vector<tstring>			vStrFiltres;	//!< Liste de noms contruite pendant la recherche, afin que les noms
											//!< dupliqus puissent tre gnrs et recherchs correctement
	LPCTSTR					pszSearch;
	CComPtr<IBaseFilter> &	pFilter;
public:
	/**
	 * Constructeur : prpare l'numration pour le remplissage d'une
	 * liste droulante avec une liste de filtres dans une catgorie donne.
	 * \param[in]	clsid	CLSID pour la catgorie
	 * \param[in]	pszSrch	Nom du filtre recherch.
	 * \param[out]	pFltr	Pointeur ATL destin recevoir l'interface du filtre trouv, s'il existe.
	 **/
	CSearchByCategory_Get(CLSID clsid, LPCTSTR pszSrch, CComPtr<IBaseFilter> & pFltr) :
		CSearchByCategory(clsid),
		pszSearch(pszSrch),
		pFilter(pFltr)
	{
		myprintf(
			TEXT("CSearchByCategory_Get: nom=%s") EOL
			TEXT("  clsid=") GUIDFMT_N EOL,
			pszSrch, GUIDPARM_N(clsid));

		pFltr.Release();	// prcaution
	}

	/**
	 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
	 * \param[in]	cMnkb		Structure contenant les informations de recherche
	 * \retval		\p S_OK		tant que le filtre n'a pas t trouv
	 * \retval		\p S_FALSE	quand le filtre a t trouv, afin d'arrter la recherche
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb)
	{
		if (_tcscmp(AddToVector(vStrFiltres, cMnkb.strName.c_str()), pszSearch)==0) {
			pFilter = cMnkb.pFilter;
			return S_FALSE;
		}
		return S_OK;
	}
};

// ====================================================================================
// Recherche de tuner avance
// ====================================================================================

/**
 * Types de filtres pouvant composer le tuner.
 * When enum'ing the device filters in BDACardCollection look for two KSCATEGORY_BDA_RECEIVER_COMPONENT
 * filters that will connect in sequence. If you only find one then you need mark the demodDevice as null.
 * And obviously flow all this through to the graph building code
 * Voir http://forums.dvbowners.com/index.php?showtopic=2302
 **/
enum DeviceType {
	dt_Null = -1,
	dt_Tuner,		//!< KSCATEGORY_BDA_NETWORK_TUNER => Tuner
	dt_Demod,		//!< KSCATEGORY_BDA_RECEIVER_COMPONENT => Dmodulateur (optionnel)
	dt_Capture,		//!< KSCATEGORY_BDA_RECEIVER_COMPONENT => Capture
	dt_END			//!< Fictif (nombre d'lments)
};

struct DeviceFilterInfo
{
	tstring	strName;			//!< Nom du filtre
	tstring	strPath;

	DeviceFilterInfo() :
		strName(NULL_NAME),
		strPath(NULL_NAME)
	{
	}

	DeviceFilterInfo & operator = (const DeviceFilterInfo & sSrc) {
		strName = sSrc.strName;
		strPath = sSrc.strPath;
		return *this;}

	/// Ajout d'un rsultat valide
	void	Set(LPCTSTR pszName, LPCTSTR pszPath);

	/**
	 * Gnration du nom du priphrique, avec numro d'ordre optionnel
	 * \param[in]	nDevCount	Numro d'ordre ajout  la fin si >  1
	 * return		Nom du priphrique gnr pour affichage
	 **/
	tstring	DisplayName(int nDevCount) const {
		if (nDevCount > 1)
			return tstr_printf(TEXT("%s #%u"), strName.c_str(), nDevCount);
		return strName; }

	bool Valid() const;
};


struct DeviceInfo
{
	int					nDevCount;		//!< N d'instance du nom tuner (si plusieurs tuners de mme nom)
	DeviceFilterInfo	asDev[dt_END];

	DeviceInfo() :
		nDevCount(1)
	{
	}

	/// Test validit de la dfinition (au moins un filtre tuner et un filtre de capture)
	bool	Valid() const {
				return asDev[dt_Tuner].Valid() &&  asDev[dt_Capture].Valid(); }

	/// Comparaison pour tri
	bool	operator < (const DeviceInfo & s2) const;

	/// Obtenir le nom du tuner pour affichage (avec numro ajout si plusieurs tuners de mme nom)
	tstring TunerDisplayName() const {
				return asDev[dt_Tuner].DisplayName(nDevCount); }

	/// Obtenir le nom du rcepteur pour affichage (avec numro ajout si plusieurs tuners de mme nom)
	tstring ReceiverDisplayName() const {
				return asDev[dt_Capture].DisplayName(nDevCount); }
};

class DeviceInfoList : public vector<DeviceInfo>
{
public:
	void FindDuplicates();
};

class CSearchTuner : public CSearchByCategory
{
	DeviceInfoList &			vResults;
	auto_ptr<IPTvM_TunerGraph>	pTestGraph;

	class CSearchReceiver : public CSearchByCategory
	{
		DeviceInfo &		vResult;
		IPTvM_TunerGraph *	pTestGraph;
		bool				bFirst;
	public:
		/**
		 * Constructeur : prpare l'numration pour la recherche d'un filtre rcepteur
		 *  associer  un tuner donn.
		 * \param[in]		pTstGraph	Graphe dj construit jusqu'au tuner.
		 * \param[in]		bFrst		\p true au 1er niveau de profondeur, \p false au second
		 * \param[in,out]	vRslt		Donnes tuner  complter.
		 **/
		CSearchReceiver(IPTvM_TunerGraph * pTstGraph, bool bFrst, DeviceInfo & vRslt);

		/**
		 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
		 * \param[in]	cMnkb	Structure contenant les informations de recherche
		 * \return		\p S_OK car tous les filtres doivent tre numrs
		 **/
		virtual HRESULT ApplyOne(MonikerBindings & cMnkb);
	};

public:
	/**
	 * Constructeur : prpare l'numration pour le remplissage d'une
	 * liste droulante avec une liste de filtres tuner.
	 * \param[out]	vRslts	Tableau de rsultats  remplir.
	 * \param[in]	bLegacy	\p true si ancien graphe  utiliser.
	 **/
	CSearchTuner(DeviceInfoList & vRslts, bool bLegacy);

	/**
	 * Implmentation de mthode virtuelle appele pour chaque filtre trouv.
	 * \param[in]	cMnkb	Structure contenant les informations de recherche
	 * \return		\p S_OK car tous les filtres doivent tre numrs
	 **/
	virtual HRESULT ApplyOne(MonikerBindings & cMnkb);

	/**
	 * Application de la mthode virtuelle \p ApplyOne  chaque filtre trouv.
	 * \return Code d'erreur DirectShow
	 **/
	HRESULT			Do();

	virtual ~CSearchTuner() {
		myprintf(TEXT("Fin recherche tuners, destruction du graphe\n")); 
		/* Note : c'est le destructeur de auto_ptr qui dtruit le graphe */ }
};

// ====================================================================================
// Recherche d'informations dans la base de registre
// ====================================================================================

/// Informations concernant un filtre, extraites de la base de registre
struct RegFilterInfo
{
	tstring	strDriverDesc;
	tstring	strProviderName;
	tstring	strDriver;
	tstring	strDriverVersion;
	tstring	strDriverDate;
	tstring	strConnection;
};

/// numration indiquant quels filtres sont implments pour un tuner, et comment ces filtres
/// sont regroups au niveau du "device path".
enum UsedDevices
{
	ud_t,				//!< Seul un filtre tuner existe
	ud_tc,				//!< Un filtre tuner+capture existe
	ud_tdc,				//!< Un filtre tuner+dmodulateur+capture existe
	ud_t_plus_c,		//!< Un filtre tuner existe, plus un filtre de capture ayant un "device path" diffrent
	ud_t_plus_dc,		//!< Un filtre tuner existe, plus un filtre dmodulateur+capture ayant un "device path" diffrent
	ud_t_plus_d_plus_c,	//!< Trois filtres tuner, dmodulateur et capture existent avec des "device path" diffrents
	ud_td_plus_c,		//!< Un filtre tuner+dmodulateur existe, plus un filtre de capture ayant un "device path" diffrent
};

/// Rsultat de l'extraction d'informations concernant un tuner  partir de ses "device paths" (long)
struct NamesAndDevPaths
{
	tstring	strTunerName;	//!< Nom du filtre tuner
	tstring	strTunerPath;	//!< "Device path" (court) du filtre tuner
	tstring	strDemodName;	//!< Nom du filtre dmodulateur
	tstring	strDemodPath;	//!< "Device path" (court) du filtre dmodulateur
	tstring	strCaptureName;	//!< Nom du filtre de capture
	tstring	strCapturePath;	//!< "Device path" (court) du filtre de capture

	/// Comparaison des "device paths" entre eux pour dterminer comment ils sont groups
	UsedDevices	DevPathCompare() const;
};

/// Classe d'assistance  l'obtention d'information depuis la base de registres,
/// sous la cl "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet".
class CRegControlSet
{
	CRegKey_tstr regCtlSetRoot;
public:
	CRegControlSet();

	/// Obtenir un "device path" (court)  partir d'une catgorie et d'un nom de filtre
	bool	GetDevicePath(const CLSID & clsidCategory, LPCTSTR pszFilterName, tstring & strDevicePath);

	/// Extraire tous les noms de filtres et "device path" (courts) de la configuration du tuner
	bool	GetAllNamesAndPaths(const TunerProps & sProps, NamesAndDevPaths & sNnP);

	/// Obtenir le chemin d'accs au pilote
	bool	GetDriverPath(LPCTSTR pszDevicePath, tstring & strDriverPath);

	/// Obtenir toutes les informations d'un pilote
	bool	GetDeviceInfo(LPCTSTR pszDevicePath, RegFilterInfo & sInfo);
};
