/*
 * $Id: update.cpp 399 2020-10-22 17:12:06Z gingko $
 * Copyright (C) 2008 lolo_32 <pouchintv-dev@baysse.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Traitement des mises  jour par Internet (implmentation)
 *
 *	Ce fichier contient les fonctionnalits ncessaires au
 *	traitement des mises  jour par Internet.
 **/

#include "stdafx.h"
#include "resource.h"
#include "dlgutils.h"
#include "internet.h"		// Pour tlcharger les mises  jour
#include "update.h"
#include "record.h"			// Si des enregistrements sont en cours
#include "version.h"		// Pour avoir la version compile
#include "xmlutils.h"		// Pour le XML
#include "lzma.h"
#include <shellapi.h>
#include <shlwapi.h>

/// Version maximale supporte pour les mises  jour
#define UPDATE_VERSION_MAX	1

/// URL pour la mise  jour.
#ifdef _DEBUG
	// Prtendre tre une version plus ancienne pour tests
	#undef PRODUCT_VERSION_STR
	#define PRODUCT_VERSION_STR "0.6.2.295"

	#define UPDATE_URL	"http://www.pouchintv.fr/version_debug.php"
#else
	#define UPDATE_URL	"http://www.pouchintv.fr/version.php"
#endif
// Note : Ces mises  jour pourraient tre faites en https://, et le serveur distant est mme configur
// pour accepter cela. Mais certains systmes anciens (XP, peut-tre Vista ) ne savent pas le grer
// correctement, et de toutes faons a pourrait tre instable, donc on ne va pas le faire pour le moment.

#ifdef _WIN64
	#define REQ_APP_TYPE	"app64"
#else
	#define REQ_APP_TYPE	"app32"
#endif

/// Inclure les versions beta dans la mise  jour
bool			update_include_beta = false;

/// Autoriser la recherche mensuelle de mise  jour
INT				update_check_auto = -1;

// Temps o t effectue la dernire recherche de mise  jour
UINT32			update_check_time = 0;	// Timestamp Unix / 60

enum UpdateState
{
	us_Undef,
	us_NeedCheck,
	us_Checking,
	us_NeedUpdate,
	us_Updating,
	us_NeedRestart
};

class CUpdateState
{
public:
	TiXmlDocument			xmlVersionDocument;		//!< Objet TinyXML pour traiter le fichier XML reu
	UpdateState				eCurrentState;			//!< tat courant de la mise  jour

	/// Constructeur
	CUpdateState() :
		eCurrentState(us_Undef)
	{
	}

	virtual void	InitProgress(int nMaxPos) {
						}
	virtual void	AddProgress(int iNewDeltaPos=1) {
						}

	virtual void	SetState(UpdateState eNewState, WPARAM wParam = 0) {
						if (eNewState == us_NeedUpdate)
							PostMessage(hMainWnd, WM_APP_UPDATEFOUND, wParam, 0); }

	virtual LRESULT	AddLineToList(LPCTSTR str) {
						return 0; }

	virtual HWND	GetParentWindow() {
						return hMainWnd; }

	/// Destructeur
	virtual ~CUpdateState() {
		}
};

class CUpdateThread : public WorkingThread
{
protected:

	/// Structure associant un document tlcharg  sa date de dernire modification.
	struct InternetFile : public ByteVector
	{
		SYSTEMTIME sTime;	// Date et heure du fichier

		/**
		 * Enregistre un fichier depuis un tableau d'octets
		 *
		 * \param[in]	nom	Nom du fichier  sauvegarder
		 * \retval		0 si tout s'est bien droul
		 * \retval		Code d'erreur systme en cas d'erreur
		 **/
		DWORD SauveFichier(LPCTSTR nom) const;
	};

	shared_ptr<CUpdateState>
						pUpdSta;	// Gestion d'un CUpdateState de substitution si aucun n'est fourni
	bool				bWithBeta;

	/// Arrt du tlchargement des fichiers demand
	bool				stop_maj;

	/**
	 * \brief Ajoute le message dans la liste
	 *
	 * \param[in]  pszFmt		message ou chane de formatage devant tre affich dans la liste de statut
	 **/
	void AddMsg(LPCTSTR pszFmt, ...);

	/**
	 * Lance la rcupration du fichier, dont l'url est passe en paramtre,
	 * et stocke le fichier rcupr dans \p vData
	 *
	 * \param[in]	url		Adresse contenant les fichiers  rcuprer
	 * \param[out]	vData	Vecteur d'octets contenant le fichier rcupr, initialis dans la fonction
	 * \return	Taille du fichier rcupr ( = vData.size() )
	 **/
	DWORD GetInternetFile(LPCTSTR url, InternetFile & vData);

	/**
	 * Lance la rcupration du fichier compress, dont l'url est passe en paramtre,
	 * et stocke le fichier rcupr dans \p vData
	 *
	 * \param[in]	url		Adresse contenant les fichiers  rcuprer
	 * \param[out]	vData	Vecteur d'octets contenant le fichier rcupr, initialis dans la fonction
	 * \return		Taille du fichier rcupr ( = vData.size() )
	 **/
	INT GetCompressedInternetFile(LPCTSTR url, InternetFile & vData);

	/**
	 * \brief Traitement pour une mise  jour possible du fichier d'application lui-mme
	 * \param[in]	sTiXml		Noeud XML du dcrivant le fichier
	 * \param[in]	pszDispType	Type de noeud recherch ("stable" ou "beta") pour affichage
	 **/
	virtual bool UpdateTraiteApp(const TiXmlElement & sTiXml, LPCTSTR pszDispType) = 0;

	/**
	 * \brief Traitement pour une mise  jour d'un fichier secondaire (autre qu'icne)
	 * \param[in]	strUrl		URL de tlchargement du fichier
	 * \param[in]	strPath		Chemin complet pour enregistrer le fichier
	 * \param[in]	strFile		Nom du fichier, seul
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	virtual bool UpdateTraiteFile(const tstring & strUrl, const tstring & strPath, const tstring & strFile) = 0;

	/**
	 * \brief Traitement pour une icne dont la mise  jour est requise
	 * \param[in]	strUrl		URL de tlchargement
	 * \param[in]	strPays		Pays concern par l'icne
	 * \param[in]	strDirPays	Rpertoire (relatif) des icnes de ce pays
	 * \param[in]	strIcone	Nom de base de l'icne
	 * \param[in]	strIconFile	Nom de fichier de l'icne
	 * \param[in]	strFound	Chemin de l'icne dj existante  mettre  jour, le cas chant
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	virtual bool UpdateTraiteIcon(const tstring & strUrl, const tstring & strPays, const tstring & strDirPays,
								  const tstring & strIcone, const tstring & strIconFile, const tstring & strFound) = 0;

	/**
	 * \brief Vrification de la date d'un fichier
	 *
	 * (en vue d'un remplacement si plus rcent disponible)
	 * \param[in]	pszPath	Chemin d'accs au fichier  tester
	 * \param[in]	sTiXml	Noeud XML contenant un attribut de date
	 * \return		\p true si une version plus rcente ( tlcharger) existe
	 **/
	static bool CheckFileDate(LPCTSTR pszPath, const TiXmlElement & sTiXml);

	/**
	 * \brief Traitement d'un noeud d'application
	 * \param[in]	pszReqType	Type de noeud recherch ("stable" ou "beta")
	 * \param[in]	pszDispType	Type de noeud recherch (pour affichage)
	 **/
	bool UpdateTraiteAppNode(LPCSTR pszReqType, LPCTSTR pszDispType);

	bool UpdateTraiteFileNode(const TiXmlElement & sTiXml);

	bool UpdateTraiteIcons(const TiXmlElement & sTiXml);

	/**
	 * Fonction traitant le fichier XML dcrivant la mise  jour
	 *
	 * \return	Cumul de \p UPDATE_APP_FLAG (si une mise  jour d'application a t dtecte ou ralise)
				ou \p UPDATE_ICON_FLAG (si une mise  jour d'icne(s) a t dtecte ou ralise)
	 **/
	BYTE UpdateTraiteXml();

public:
	CUpdateThread(shared_ptr<CUpdateState> pUpSta, bool bWBeta, bool bDelOnExit) :
		WorkingThread(bDelOnExit),
		pUpdSta(pUpSta),
		bWithBeta(bWBeta),
		stop_maj(false)
	{
		if (! pUpdSta.get()) {
			pUpdSta.reset(new CUpdateState());
		}
	}

	/// Ordonne au thread de s'arrter
	void Stop() {
		stop_maj = true;
		// Patienter 4 secondes pour l'arrt du thread
		WaitThreadExit(4 * 1000); }

	virtual ~CUpdateThread() {
				}
};

class CWebUpdateDialog : public CModalDialogBase, public CUpdateState
{
	friend class CUpdateThread;
	friend class CUpdateCheckThread;
	friend class CUpdateTelechargeThread;

	static weak_ptr<CWebUpdateDialog>
							pInstance;
	HWND					hUpdateStatut;			//!< Handle de la liste indiquant la progression
	HWND					hProgressStatut;		//!< Handle de la barre de progression
	HWND					hUpdateButton;			//!< Handle du bouton "Mettre  jour"
	auto_ptr<CUpdateThread>	pUpdateThread;			//!< Pointeur sur le "thread" actif
public:
	bool					bSearchNow;
	bool					bOpen;
private:
	/// Constructeur
	CWebUpdateDialog() :
		CModalDialogBase(IDD_UPDATE),
		hUpdateStatut(NULL),
		hProgressStatut(NULL),
		hUpdateButton(NULL),
		bSearchNow(false),
		bOpen(false)
	{
	}

	virtual HWND	GetParentWindow() const {
						return hDlg; }

	/**
	 * \brief Passage  un nouvel tat
	 *
	 * \param[in]  eNewState	tat auquel passer
	 **/
	virtual void	SetState(UpdateState eNewState, WPARAM wParam = 0);

	/**
	 * \brief Incrmente la barre de progression
	 *
	 * \param[in]  nMaxPos	Index maximal
	 **/
	virtual void	InitProgress(int nMaxPos);

	/**
	 * \brief Incrmente la barre de progression
	 *
	 * \param[in]  iNewDeltaPos	index indiquant la nouvelle position de la barre de progression
	 *							(utiliser -1 pour aller  la valeur maximale)
	 **/
	virtual void	AddProgress(int iNewDeltaPos=1);

	virtual LRESULT	AddLineToList(LPCTSTR str) {
						return hUpdateStatut ? ::AddLineToList(hUpdateStatut, str) : NULL; }

	virtual INT_PTR	DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam);

	void			EndDialog(INT_PTR nResult);
public:
	/// Destructeur
	~CWebUpdateDialog() {
		//if (pUpdateThread.get())
		//	pUpdateThread->Stop(); 
		}

	static shared_ptr<CWebUpdateDialog>
					getInstance();

	void			RequestUpdate() {
						PostDlgMessage(WM_COMMAND, IDC_UPDATE_CHECK); }
};

class CUpdateCheckThread : public CUpdateThread
{
	/// Fonction d'excution principale effective du 'thread'.
	virtual DWORD ThreadProc();

	/**
	 * \brief Traitement pour une mise  jour possible du fichier d'application lui-mme
	 * \param[in]	sTiXml		Noeud XML du dcrivant le fichier
	 * \param[in]	pszDispType	Type de noeud recherch ("stable" ou "beta") pour affichage
	 **/
	virtual bool UpdateTraiteApp(const TiXmlElement & sTiXml, LPCTSTR pszDispType);

	/**
	 * \brief Traitement pour une mise  jour d'un fichier secondaire (autre qu'icne)
	 * \param[in]	strUrl		URL de tlchargement du fichier
	 * \param[in]	strPath		Chemin complet pour enregistrer le fichier
	 * \param[in]	strFile		Nom du fichier, seul
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	virtual bool UpdateTraiteFile(const tstring & strUrl, const tstring & strPath, const tstring & strFile);

	/**
	 * \brief Traitement pour une icne dont la mise  jour est requise
	 * \param[in]	strUrl		URL de tlchargement
	 * \param[in]	strPays		Pays concern par l'icne
	 * \param[in]	strDirPays	Rpertoire (relatif) des icnes de ce pays
	 * \param[in]	strIcone	Nom de base de l'icne
	 * \param[in]	strIconFile	Nom de fichier de l'icne
	 * \param[in]	strFound	Chemin de l'icne dj existante  mettre  jour, le cas chant
	 * \return		toujours \p true
	 **/
	virtual bool UpdateTraiteIcon(const tstring & strUrl, const tstring & strPays, const tstring & strDirPays,
								  const tstring & strIcone, const tstring & strIconFile, const tstring & strFound);

public:
	CUpdateCheckThread(shared_ptr<CUpdateState> pUpSta, bool bWBeta, bool bDelOnExit = false) :
		CUpdateThread(pUpSta, bWBeta, bDelOnExit)
	{
		StartThread();
	}

	virtual ~CUpdateCheckThread() {
				}
};

class CUpdateTelechargeThread : public CUpdateThread
{
	tstring			strArgs;	//!< Liste de mouvements de fichiers  effectuer

	/**
	 * \brief Sauvegarde d'un fichier  mettre  jour dans un fichier temporaire
	 *
	 * Le contenu du fichier est sauvegarder dans un fichier temporaire, et les informations
	 * ncessaires pour le mettre ensuite en place sont ajoutes  la liste
	 * \param[in]	vData		Donnes  sauvegarder
	 * \param[in]	pszFileName	Nom de fichier final (aprs mise en place)
	 *`\return		\p true ou \p false selon succs ou erreur
	 **/
	bool			SaveUpdateInTemp(InternetFile & vData, LPCTSTR pszFileName);

	/**
	 * \brief Appel du processus externe de finalisation de la mise  jour
	 *
	 * Extraction du processus des ressources, dcompression, et lancement avec lvation
	 * de privilges (cette lvation tant paramtre dans le manifeste du processus externe).
	 *
	 * \param[in]	bAppIncluded	\p true si la mise  jour inclut l'application elle-mme
	 * \return		0 si succs, ou bien code d'erreur systme
	 **/
	DWORD			CallMoveProcess(bool bAppIncluded);

	/// Fonction d'excution principale effective du 'thread'.
	virtual DWORD	ThreadProc();

	/**
	 * \brief Traitement pour une mise  jour possible du fichier d'application lui-mme
	 * \param[in]	sTiXml		Noeud XML du dcrivant le fichier
	 * \param[in]	pszDispType	Type de noeud recherch ("stable" ou "beta") pour affichage
	 **/
	virtual bool	UpdateTraiteApp(const TiXmlElement & sTiXml, LPCTSTR pszDispType);

	/**
	 * \brief Traitement pour une mise  jour d'un fichier secondaire (autre qu'icne)
	 * \param[in]	strUrl		URL de tlchargement du fichier
	 * \param[in]	strPath		Chemin complet pour enregistrer le fichier
	 * \param[in]	strFile		Nom du fichier, seul
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	virtual bool	UpdateTraiteFile(const tstring & strUrl, const tstring & strPath, const tstring & strFile);

	/**
	 * \brief Traitement pour une icne dont la mise  jour est requise
	 * \param[in]	strUrl		URL de tlchargement
	 * \param[in]	strPays		Pays concern par l'icne
	 * \param[in]	strDirPays	Rpertoire (relatif) des icnes de ce pays
	 * \param[in]	strIcone	Nom de base de l'icne
	 * \param[in]	strIconFile	Nom de fichier de l'icne
	 * \param[in]	strFound	Chemin de l'icne dj existante  mettre  jour, le cas chant
	 * \return		\p true ou \p false selon succs ou chec
	 **/
	virtual bool UpdateTraiteIcon(const tstring & strUrl, const tstring & strPays, const tstring & strDirPays,
								  const tstring & strIcone, const tstring & strIconFile, const tstring & strFound);

public:
	CUpdateTelechargeThread(shared_ptr<CUpdateState> pUpSta, bool bWBeta, bool bDelOnExit = false) :
		CUpdateThread(pUpSta, bWBeta, bDelOnExit)
	{
		StartThread();
	}

	virtual ~CUpdateTelechargeThread() {
				}
};

class CUpdateElapsingThread : public WorkingThread
{
public:
	CUpdateElapsingThread() :
		WorkingThread(true)
	{
		StartThread();
	}

	void TSleep(DWORD nSeconds)
	{
		myprintf(TEXT("Dbut du dlai de %u secondes avant test mise  jour") EOL, nSeconds);
		Sleep(nSeconds*1000);	// Attente 10 secondes
		myprintf(TEXT("Fin du dlai de %u secondes") EOL, nSeconds);
	}

	/// Fonction d'excution principale effective du 'thread'.
	virtual DWORD ThreadProc()
	{
		__time64_t	nTim = _time64(NULL);
		UINT32		nCurMinute = static_cast<UINT32>(nTim/60);

		myprintf(TEXT("ge de la dernire vrification de mise  jour: %u minutes") EOL, nCurMinute-update_check_time);

		if (update_check_auto == 1) {
			// La mise  jour automatique est dj active
			if (nCurMinute > update_check_time + AUTO_UPDATE_DELAY) {
				TSleep(10);	// Attente 10 secondes

				new CUpdateCheckThread(shared_ptr<CUpdateState>(), update_include_beta, true);
				update_check_time = nCurMinute;
				save_config();
			}
		} else if (update_check_auto == -1) {
			TSleep(10);	// Attente 10 secondes

			update_check_auto =
				MessageBox(hMainWnd,
					APPLICATION_NAME
					TEXT(" vous offre maintenant la possibilit de rechercher une fois par mois si une mise  jour a t publie.") EOL EOL
					TEXT("Dsirez-vous activer cette recherche mensuelle automatique ?") EOL EOL
					TEXT("Vous pouvez modifier ce choix  tout moment  partir du dialogue de mise  jour (Menu Aide)"),
					TEXT("Note concernant les mises  jour automatiques"), MB_YESNO|MB_ICONQUESTION) == IDYES ? 1 : 0;
			if (update_check_auto == 1) {
				// On vient juste d'activer la mise  jour automatique au premier dmarrage : on fait le test immdiatement
				TSleep(5);	// Attente 5 secondes

				new CUpdateCheckThread(shared_ptr<CUpdateState>(), update_include_beta, true);
				update_check_time = nCurMinute;
			}
			save_config();
		}
		return 0;
	}
};

/// Indique si une mise  jour est en attente de redmarrage
static bool bUpdateNeedRestart		= false;

/// Structure dcrivant une version de l'application
struct AppVersion
{
	int		nMajor;
	int		nMinor;
	int		nRev;
	CHAR	szSvn[8];	// Des caractres alphanumriques peuvent exister sur ce champ

	/// Constructeur chargeant la version avec une chane de version
	AppVersion(LPCSTR pszVersion) :
		nMajor(0),
		nMinor(0),
		nRev(0)
	{
		szSvn[0] = '0';	szSvn[1] = 0;
		// On rcupre la version de PTVM fonctionnant actuellement
		sscanf_s(PRODUCT_VERSION_STR, "%i.%i.%i.%s", &nMajor, &nMinor, &nRev, szSvn, (UINT)_countof(szSvn));
	}

	/// Constructeur chargeant la version avec les attributs d'un lment TiXML
	AppVersion(const TiXmlElement & sTiXml) :
		nMajor(0),
		nMinor(0),
		nRev(0)
	{
		szSvn[0] = '0';	szSvn[1] = 0;
		sTiXml.QueryIntAttribute("major", &nMajor);
		sTiXml.QueryIntAttribute("minor", &nMinor);
		sTiXml.QueryIntAttribute("rev", &nRev);

		LPCSTR pszSvn = sTiXml.Attribute("svn");

		if (pszSvn && *pszSvn)
			strcpy_T(szSvn, pszSvn);
	}

	bool operator > (const AppVersion & sV2) const {
		return nMajor>sV2.nMajor || nMajor==sV2.nMajor && (nMinor>sV2.nMinor || nMinor==sV2.nMinor && nRev > sV2.nRev || nRev == sV2.nRev && atoi(szSvn) > atoi(sV2.szSvn)); }

	bool operator < (const AppVersion & sV2) const {
		return ! operator > (sV2); }

	/// Conversion en chane (le rsultat tant maintenu sur une variable statique,
	/// ne pas utiliser cette fonction deux fois dans la mme instruction)
	LPCTSTR toString() const {
		static TCHAR szTmp[20];

		_stprintf_s(szTmp, TEXT("%d.%d.%d.%") A2t, nMajor, nMinor, nRev, szSvn);
		return szTmp;}
};

/**
 * Enregistre un fichier depuis un tableau d'octets
 *
 * \param[in]	nom	Nom du fichier  sauvegarder
 * \retval		0 si tout s'est bien droul
 * \retval		Code d'erreur systme en cas d'erreur
 **/
DWORD CUpdateThread::InternetFile::SauveFichier(LPCTSTR nom) const
{
	myprintf(TEXT("Sauvegarde %s depuis un tlchargement de %u octets") EOL, nom, size());

	return SaveAsFile(nom, &sTime);
}

/**
 * \brief Ajoute le message dans la liste
 *
 * \param[in]  pszFmt		message ou chane de formatage devant tre affich dans la liste de statut
 **/
void CUpdateThread::AddMsg(LPCTSTR pszFmt, ...)
{
	FMT_BUF(szBuf, 512, pszFmt, pszFmt);

	pUpdSta->AddLineToList(szBuf);
}

/**
 * \brief Passage  un nouvel tat
 *
 * \param[in]  eNewState	tat auquel passer
 **/
void CWebUpdateDialog::SetState(UpdateState eNewState, WPARAM wParam)
{
	if (eNewState != eCurrentState) {
		eCurrentState = eNewState;

		if (! hDlg)
			return;

		switch (eCurrentState) {

		case us_NeedCheck:
			SetWindowText(hUpdateButton, TEXT("Vrifier"));
			EnableWindow(hUpdateButton, TRUE);	// Active le bouton
			break;

		case us_Checking:
			EnableWindow(hUpdateButton, FALSE);
			InitProgress(4);	// On met 4 tapes pour la vrification de la version
			break;

		case us_NeedUpdate:
			EnableDlgItem(IDC_UPDATE_AUTO, false);
			EnableDlgItem(IDC_UPDATE_BETA, false);
			SetWindowText(hUpdateButton, TEXT("Mettre  jour"));
			EnableWindow(hUpdateButton, TRUE);
			break;

		case us_Updating:
			EnableWindow(hUpdateButton, FALSE);
			InitProgress(1);	// Une seule tape, en attendant d'en connatre le nombre
			break;

		case us_NeedRestart:
			if (enregistrements_actuels.recording()) {
				// Peu vraisemblable qu'on arrive ici car on ne peut pas lancer une mise  jour
				// quand il y a des enregistrements en cours, mais dans certains cas limite 
				AddLineToList(TEXT("Vous avez des enregistrements en cours. Redmarrez plus tard."));
			} else {
				SetWindowText(hUpdateButton, TEXT("Redmarrer ") APPLICATION_NAME);
				EnableWindow(hUpdateButton, TRUE);	// Active le bouton
			}
		}
	}
}

/**
 * \brief Incrmente la barre de progression
 *
 * \param[in]  nMaxPos	Index maximal
 **/
void CWebUpdateDialog::InitProgress(int nMaxPos)
{
	if (hProgressStatut) {
		// On met les tapes
		SendMessage(hProgressStatut, PBM_SETRANGE, 0, MAKELPARAM(0, nMaxPos));

		// On se positionne au dbut de la barre de progression
		SendMessage(hProgressStatut, PBM_SETPOS, 0, 0);
	}
}

/**
 * \brief Incrmente la barre de progression
 *
 * \param[in]  iNewDeltaPos	index indiquant la nouvelle position de la barre de progression
 *							(utiliser -1 pour aller  la valeur maximale)
 **/
void CWebUpdateDialog::AddProgress(int iNewDeltaPos)
{
	if (! hProgressStatut)
		return;

	int nMax = (int)SendMessage(hProgressStatut, PBM_GETRANGE, false, 0);
	int nPos = (int)SendMessage(hProgressStatut, PBM_GETPOS, 0, 0);
	int	nDif = nMax-nPos;

	if (iNewDeltaPos < 0 || nDif < iNewDeltaPos)
		iNewDeltaPos = nMax-nPos;
	if (iNewDeltaPos > 0)
		SendMessage(hProgressStatut, PBM_DELTAPOS, iNewDeltaPos, 0);
}

/**
 * Lance la rcupration du fichier, dont l'url est passe en paramtre,
 * et stocke le fichier rcupr dans \p vData
 *
 * \param[in]	url		Adresse contenant les fichiers  rcuprer
 * \param[out]	vData	Vecteur d'octets contenant le fichier rcupr, initialis dans la fonction
 * \return	Taille du fichier rcupr ( = vData.size() )
 **/
DWORD CUpdateThread::GetInternetFile(LPCTSTR url, InternetFile & vData)
{
	// Si vData est renseign, on quitte
	if (!vData.empty())
		return 0;

	bool		stoppe			= false;

	// En version Unicode, la conversion en UTF-8 est implicite
	ConnexionInternet maj(	// Objet traitant de la connexion Internet
		tstr_printf(TEXT("%s / %s"), szAppName, szAppVersion).c_str(),
#ifdef _UNICODE
		url
#else
		ConvertChar2UTF8(url).c_str()
#endif
		);

	myprintf(TEXT("Tlchargement de %s") EOL, url);

	int http_statut = maj.EnvoieRequete();

	if (http_statut != maj.HTTP_ERR_OK) {
		if (http_statut == 0) {
			AddMsg(TEXT("Erreur lors de l'tablissement de la connexion."));
			myprintf(ERRX(TEXT("Impossible d'tablir la connexion avec %s")), url);
		} else {
			AddMsg(TEXT("Erreur %d lors du tlchargement du fichier."), http_statut);
			AddMsg(TEXT("Veuillez ressayer ultrieurement..."));
			myprintf(ERRX(TEXT("Erreur connexion  %s: erreur HTTP: %d")), url, http_statut);
		}
		stoppe = true;
	}

	if (stop_maj || stoppe)
		return 0;

	if (pUpdSta->eCurrentState == us_Checking)
		AddMsg(TEXT("Rcupration du fichier de version..."));

	// Tout est OK, on rcupre le fichier
	BYTE	temp[8*1024];	// Buffer de 8 Ko
	DWORD	lu = sizeof(temp), dwLuTotal = 0;

	if (pUpdSta->eCurrentState == us_Updating) {
		// On initialise la barre de progression, avec autant de morceaux que ncessaire
		pUpdSta->InitProgress(int(maj.GetTaille() / sizeof(temp))+1);
	}

	while (!stop_maj && maj.RecupereFichier(temp, &lu)) {
		vData.resize(dwLuTotal+lu);			// Redimensionnement
		memcpy_s(&vData.get()[dwLuTotal], lu, temp, lu);

		dwLuTotal += lu;

		// Remet le compteur de lecture  sa valeur max
		lu = sizeof(temp);

		if (stop_maj) // Arrt demand, on quitte
			break;

		// Si on tlcharge, on met  jour la barre de progression
		if (pUpdSta->eCurrentState == us_Updating)
			pUpdSta->AddProgress();
	}

	vData.sTime = maj.GetModifTime();
	return dwLuTotal;
}

/**
 * Lance la rcupration du fichier compress, dont l'url est passe en paramtre,
 * et stocke le fichier rcupr dans \p vData
 *
 * \param[in]	url		Adresse contenant les fichiers  rcuprer
 * \param[out]	vData	Vecteur d'octets contenant le fichier rcupr, initialis dans la fonction
 * \return		Taille du fichier rcupr ( = vData.size() )
 **/
INT CUpdateThread::GetCompressedInternetFile(LPCTSTR url, InternetFile & vData)
{
	INT				nSize = 0;
	InternetFile	vCompressedData;

	if (GetInternetFile(url, vCompressedData) > 0) {
		// Des donnes ont t rcupres
		bool		bResLzma = DecompressLzma(vData, vCompressedData);

		vData.sTime = vCompressedData.sTime;

		if (bResLzma)
			nSize = (INT)vData.size();
		else
			return -1;	// Erreur  la dcompression
	}
	return nSize;
}

/**
 * \brief Traitement pour une mise  jour possible du fichier d'application lui-mme
 * \param[in]	sTiXml		Noeud XML du dcrivant le fichier
 * \param[in]	pszDispType	Type de noeud recherch ("stable" ou "beta") pour affichage
 **/
bool CUpdateCheckThread::UpdateTraiteApp(const TiXmlElement & sTiXml, LPCTSTR pszDispType)
{
	AppVersion	sAppVers(PRODUCT_VERSION_STR);	// Version courante de PTVM
	AppVersion	sXmlVers(sTiXml);			// Version du noeud XML
	bool		bUpdate	= false;

	if (sXmlVers > sAppVers) {
		// Nouvelle version disponible

		// Affiche simplement le message informatif dans la fentre de statut
		AddMsg(
			TEXT("Nouvelle version %s de %s disponible : %s"),
			pszDispType, szAppName, sXmlVers.toString());
		AddMsg(TEXT("Version actuellement utilise : %s"), sAppVers.toString());
		bUpdate  = true;
	} else
		AddMsg(
			TEXT("Aucune nouvelle version %s de %s disponible"),
			pszDispType, szAppName);

	return bUpdate;
}

/**
 * \brief Sauvegarde d'un fichier  mettre  jour dans un fichier temporaire
 *
 * Le contenu du fichier est sauvegarder dans un fichier temporaire, et les informations
 * ncessaires pour le mettre ensuite en place sont ajoutes  la liste
 * \param[in]	vData		Donnes  sauvegarder
 * \param[in]	pszFileName	Nom de fichier final (aprs mise en place)
 *`\return		\p true ou \p false selon succs ou erreur
 **/
bool CUpdateTelechargeThread::SaveUpdateInTemp(InternetFile	& vData, LPCTSTR pszFileName)
{
	TempPathName sTmpFile;

	// Des donnes ont t rcupres
	// Tout semble bon, on sauvegarde le nouveau fichier
	DWORD nRes = vData.SauveFichier(sTmpFile);

	if (nRes!=0) {
		// Des erreurs d'criture du fichier se sont produites. L'indiquer.
		AddMsg(TEXT("Erreur %u : %s"), nRes, SystemErrorToString(nRes).c_str());
		AddMsg(TEXT("Impossible de sauvegarder %s."), sTmpFile.szPath);
		return false;
	}

	// Ajouter le fichier enregistr  la liste des fichiers qui doivent tre mis en place
	// par l'utilitaire de mise  jour qui va tre lanc avec des droits d'administration
	// levs.
	strArgs += tstr_printf(TEXT("\"%s\"=\"%s\"") EOL, sTmpFile.szPath, pszFileName);
	return true;
}

/**
 * \brief Traitement pour une mise  jour possible du fichier d'application lui-mme
 * \param[in]	sTiXml		Noeud XML du dcrivant le fichier
 * \param[in]	pszDispType	Type de noeud recherch ("stable" ou "beta") pour affichage
 **/
bool CUpdateTelechargeThread::UpdateTraiteApp(const TiXmlElement & sTiXml, LPCTSTR pszDispType)
{
	AppVersion	sAppVers(PRODUCT_VERSION_STR);	// Version courante de PTVM
	AppVersion	sXmlVers(sTiXml);			// Version du noeud XML
	bool		bUpdate	= false;

	if (sXmlVers > sAppVers) {
		// Nouvelle version disponible

		// Tlcharge la nouvelle version de PTVM
		AddMsg(TEXT("Tlchargement de la mise  jour de %s..."), szAppName);

		tstring			strUrl	= ConvertUTF82Char(sTiXml.FirstChild()->Value());
		InternetFile	vData;
		INT				nSize	= GetCompressedInternetFile(strUrl.c_str(), vData);

		if (nSize > 0)
			bUpdate = SaveUpdateInTemp(vData, szAppFileName);
		else if (nSize < 0)
			AddMsg(TEXT("Le fichier tlcharg n'est pas valide. Veuillez ressayer."));
		else
			AddMsg(TEXT("Le fichier n'a pas pu tre rcupr."));
	}

	return bUpdate;
}

/**
 * \brief Traitement pour une mise  jour d'un fichier secondaire (autre qu'icne)
 * \param[in]	strUrl		URL de tlchargement du fichier
 * \param[in]	strPath		Chemin complet pour enregistrer le fichier
 * \param[in]	strFile		Nom du fichier, seul
 * \return		\p true ou \p false selon succs ou chec
 **/
bool CUpdateCheckThread::UpdateTraiteFile(const tstring & strUrl, const tstring & strPath, const tstring & strFile)
{
	AddMsg(TEXT("Mise  jour de %s disponible"), strFile.c_str());
	return true;
}

/**
 * \brief Traitement pour une icne dont la mise  jour est requise
 * \param[in]	strUrl		URL de tlchargement
 * \param[in]	strPays		Pays concern par l'icne
 * \param[in]	strDirPays	Rpertoire (relatif) des icnes de ce pays
 * \param[in]	strIcone	Nom de base de l'icne
 * \param[in]	strIconFile	Nom de fichier de l'icne
 * \param[in]	strFound	Chemin de l'icne dj existante  mettre  jour, le cas chant
 * \return		toujours \p true
 **/
bool CUpdateCheckThread::UpdateTraiteIcon(const tstring & strUrl, const tstring & strPays, const tstring & strDirPays,
										  const tstring & strIcone, const tstring & strIconFile, const tstring & strFound)
{
	AddMsg(
		strFound.empty() ?
			TEXT("Nouvelle icne %s disponible: %s") :
			TEXT("Mise  jour disponible: icne %s %s"),
		strPays.c_str(), strIcone.c_str() );
	return true;
}

/**
 * \brief Traitement pour une mise  jour d'un fichier secondaire (autre qu'icne)
 * \param[in]	strUrl		URL de tlchargement du fichier
 * \param[in]	strPath		Chemin complet pour enregistrer le fichier
 * \param[in]	strFile		Nom du fichier, seul
 * \return		\p true ou \p false selon succs ou chec
 **/
bool CUpdateTelechargeThread::UpdateTraiteFile(const tstring & strUrl, const tstring & strPath, const tstring & strFile)
{
	InternetFile	vData;

	AddMsg(TEXT("Tlchargement du fichier %s"), strFile.c_str());

	// Envoie la connexion
	if (GetInternetFile(strUrl.c_str(), vData) > 0)
		return SaveUpdateInTemp(vData, strPath.c_str());

	return false;
}

/**
 * \brief Traitement pour une icne dont la mise  jour est requise
 * \param[in]	strUrl		URL de tlchargement
 * \param[in]	strPays		Pays concern par l'icne
 * \param[in]	strDirPays	Rpertoire (relatif) des icnes de ce pays
 * \param[in]	strIcone	Nom de base de l'icne
 * \param[in]	strIconFile	Nom de fichier de l'icne
 * \param[in]	strFound	Chemin de l'icne dj existante  mettre  jour, le cas chant
 * \return		\p true ou \p false selon succs ou chec
 **/
bool CUpdateTelechargeThread::UpdateTraiteIcon(const tstring & strUrl, const tstring & strPays, const tstring & strDirPays,
											   const tstring & strIcone, const tstring & strIconFile, const tstring & strFound)
{
	InternetFile	vData;

	AddMsg(TEXT("Tlchargement de l'icne %s"), strIconFile.c_str());

	// Envoie la connexion
	if (GetInternetFile(strUrl.c_str(), vData) > 0) {
		// Des donnes ont t rcupres
		if (!strFound.empty()) {
			// Mise  jour d'une icne existante par une version plus rcente :
			// Utiliser le dossier dans lequel elle a t trouve
			return SaveUpdateInTemp(vData, strFound.c_str());
		} else {
			// Enregistrement d'une nouvelle icne dans le premier rpertoire
			// trouv disponible et adquat :
			return SaveUpdateInTemp(vData, (strDirPays + TEXT("\\") + strIconFile).c_str());
		}
	}
	return false;
}

bool CUpdateThread::UpdateTraiteFileNode(const TiXmlElement & sTiXml)
{
	tstring		strUrl	= ConvertUTF82Char(sTiXml.GetText());
	tstring		strFile = ConvertUTF82Char(sTiXml.Attribute("nom"));;

	if (strUrl.empty() || strFile.empty())
		return false;

	tstring		strPath = pouchindir_prog + strFile;

	if (PathFileExists(strPath.c_str())) {
		// si trouv : vrifier si plus rcent disponible
		if (!CheckFileDate(strPath.c_str(), sTiXml))
			return false;	// Plus rcent non disponible
	}
	return UpdateTraiteFile(strUrl, strPath, strFile);
}

bool CUpdateThread::UpdateTraiteIcons(const TiXmlElement & sTiXml)
{
	bool		bUpdate	= false;
	// on parse les icones. tablissement de la liste locale

	// on liste les packs d'icnes
	for (ITERATE_XML_ELEMENTS_NAMED(&sTiXml, xmlIconesPack, "pack")) {
		// Pays du pack (pour classer les icnes)
		tstring	strPays		= ConvertUTF82Char(xmlIconesPack->Attribute("pays"));
		tstring	strDirPays; // 	= strPays==TEXT("France") ? TEXT(".") : (strPays + TEXT("\\")).c_str();
		tstring	strFound;	// Rceptacle gnrique pour les rsultats de recherches

		if (strPays==TEXT("France")) {
			// Si le sous-rpertoire de pays n'existe pas, c'est que l'utilisateur n'a pas
			// slectionn le pack correspondant  l'installation. On ignore donc.
			if (
				FindAppFile(TEXT(""), TEXT("Icones"),
					strDirPays, FF_DIRECTORY) == pd_notfound
			) {
				continue;
			}
		} else {
			// Si le sous-rpertoire de pays n'existe pas, c'est que l'utilisateur n'a pas
			// slectionn le pack correspondant  l'installation. On ignore donc.
			if (
				FindAppFile(TEXT("Icones"), strPays.c_str() /* strPays car on ne veut pas du "\" */,
					strDirPays, FF_DIRECTORY) == pd_notfound
			) {
				continue;
			}
		}

		// On parcourt les diffrentes icnes (au cas o certaines seraient manquantes)
		for (ITERATE_XML_ELEMENTS_NAMED(xmlIconesPack, xmlIconesIcone, "icone")) {
			tstring	strIcone	= ConvertUTF82Char(xmlIconesIcone->GetText());
			tstring	strIconFile	= strIcone + TEXT(".bmp");
			bool	bNewVers	= false;

			// Recherche dans les rpertoires de l'application :
			if (FindAppFile(TEXT("Icones"), strIconFile.c_str(), strFound)!=pd_notfound) {
				// si trouv : vrifier si plus rcent disponible
				if (!CheckFileDate(strFound.c_str(), *xmlIconesIcone))
					continue;	// Plus rcent non disponible
				bNewVers	= true;
			}

			tstring	strUrlBase	= ConvertUTF82Char(sTiXml.Attribute("base"));
			tstring	strUrlPath	= ConvertUTF82Char(xmlIconesPack->Attribute("path"));

			// Ajustement du rpertoire dans l'URL
			if (strUrlPath==TEXT("/"))
				strUrlPath.clear();
			else
				strUrlPath += TEXT("/");

			tstring	strUrl		= strUrlBase + strUrlPath + strIconFile;

			if (UpdateTraiteIcon(strUrl, strPays, strDirPays, strIcone, strIconFile, strFound))
				bUpdate  = true;
		} // For (icone)
	} // For (pack)
	// Plus d'cones  traiter
	return bUpdate;
}

/// Vrification de la version du noeud XML
/// Retourne \p true si le nud est valide
static bool CheckVersionVersion(const TiXmlElement & sTiXml)
{
	// Rcupre la version de la branche
	int updateversion;
	int	nQueryResult = sTiXml.QueryIntAttribute("version", &updateversion);

	if (
		nQueryResult == TIXML_NO_ATTRIBUTE ||	// anciennes versions de PTVM, on ignore
		updateversion > UPDATE_VERSION_MAX		// nud plus rcent, on le saute, il sera lu une fois PTVM mis  jour
	) {
		return false;
	}
	return true;
}

/**
 * \brief Vrification de la date d'un fichier
 *
 * (en vue d'un remplacement si plus rcent disponible)
 * \param[in]	pszPath	Chemin d'accs au fichier  tester
 * \param[in]	sTiXml	Noeud XML contenant un attribut de date
 * \return		\p true si une version plus rcente ( tlcharger) existe
 **/
bool CUpdateThread::CheckFileDate(LPCTSTR pszPath, const TiXmlElement & sTiXml)
{
	INT64		nRmtFTime	= 0;	// Date et heure du fichier distant
	INT64		nLocFTime	= 0;	// Date et heure du fichier local
	LPCSTR		pszUtime	= sTiXml.Attribute("utime");	// Timestamp Unix maintenant inclus

	if (pszUtime) {
		// http://msdn.microsoft.com/en-us/library/ms724228%28v=vs.85%29.aspx
		nRmtFTime = Int32x32To64(atoi(pszUtime), 10000000) + 116444736000000000I64;
	} else {
		// Plus vraiment utile mais on laisse quand mme pour le moment
		LPCSTR		pszDate		= sTiXml.Attribute("date");

		if (pszDate == NULL)
			return false;

		UINT nYear, nMonth, nDay;

		if (sscanf_s(pszDate, "%u-%u-%u", &nYear, &nMonth, &nDay) != 3)
			return false;

		SYSTEMTIME	sTmpTime	= {		// Date et heure du fichier distant
			(WORD)nYear,	// wYear
			(WORD)nMonth,	// wMonth
			0,				// wDayOfWeek
			(WORD)nDay		// wDay
			// Le reste  zro (implicite)
		};

		if (!SystemTimeToFileTime(&sTmpTime, reinterpret_cast<LPFILETIME>(&nRmtFTime)))
			return false;
	}

	// Ouverture du fichier pour rcupration de la date
	HANDLE		hFile		= CreateFile(pszPath, GENERIC_READ, FILE_SHARE_READ, NULL,
								OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile != INVALID_HANDLE_VALUE) {
		BOOL		bRes = GetFileTime(hFile, NULL, NULL, reinterpret_cast<LPFILETIME>(&nLocFTime));

		CloseHandle(hFile);	// Fermeture du fichier
		if (!bRes)
			return false;
	}

#ifdef _DIAGNOSTIC
	SYSTEMTIME	sDebugRmt, sDebugLoc;

	FileTimeToSystemTime((LPFILETIME)&nRmtFTime, &sDebugRmt);
	FileTimeToSystemTime((LPFILETIME)&nLocFTime, &sDebugLoc);
	myprintf(TEXT("CheckFileDate (%s): local=%04u-%02u-%02u %02u:%02u:%02u, distant=%04u-%02u-%02u %02u:%02u:%02u") EOL, _tcsrchr(pszPath, '\\')+1,
		sDebugLoc.wYear, sDebugLoc.wMonth, sDebugLoc.wDay, sDebugLoc.wHour, sDebugLoc.wMinute, sDebugLoc.wSecond,
		sDebugRmt.wYear, sDebugRmt.wMonth, sDebugRmt.wDay, sDebugRmt.wHour, sDebugRmt.wMinute, sDebugRmt.wSecond);
#endif
	return (nRmtFTime - nLocFTime) > 600000000I64;	// 60 secondes de marge d'erreur
}

/**
 * \brief Traitement d'un noeud d'application
 * \param[in]	pszReqType	Type de noeud recherch ("stable" ou "beta")
 * \param[in]	pszDispType	Type de noeud recherch (pour affichage)
 **/
bool CUpdateThread::UpdateTraiteAppNode(LPCSTR pszReqType, LPCTSTR pszDispType)
{
	bool bUpdateNodeFound = false;
	bool bUpdateFound = false;

	myprintf(TEXT("Vrification pour la version %s") EOL, pszDispType);
	for (ITERATE_XML_ELEMENTS_NAMED(pUpdSta->xmlVersionDocument.FirstChildElement("Versions"), xmlVersionElem, "version")) {
		if (!CheckVersionVersion(*xmlVersionElem))
			continue;

		// On ignore les versions dont le type de vrification n'a pas t choisi
		if (_stricmp(xmlVersionElem->Attribute("type"), pszReqType)!=0)
			continue;

		bUpdateNodeFound = true;		// On a trouv un descripteur de mise  jour (ce qui n'implique
										// pas ncessairement qu'on ait trouv une mise  jour ...)

		// Rcupration des mises  jour de l'application
		for (ITERATE_XML_ELEMENTS_NAMED(xmlVersionElem, xmlTypeMaj, REQ_APP_TYPE)) {
			if (UpdateTraiteApp(*xmlTypeMaj, pszDispType)) {
				bUpdateFound = true;	// Voil, maintenant on a trouv une mise  jour
				break;	// Une seule mise  jour possible, donc on sort
			}
		}
		if (bUpdateFound)	// toujours pas de "break" multi-niveaux en C/C++ ...
			break;
	}
	if (!bUpdateNodeFound)
		AddMsg(TEXT("Aucune information de mise  jour %s disponible"), pszDispType);
	myprintf(BOLD(TEXT("%s version %s")) EOL, (bUpdateFound ? TEXT("On a trouv une") : TEXT("On n'a pas trouv de")), pszDispType);
	return bUpdateFound;
}

/**
 * Fonction traitant le fichier XML dcrivant la mise  jour
 *
 * \return	Cumul de \p UPDATE_APP_FLAG (si une mise  jour d'application a t dtecte ou ralise)
 *			ou \p UPDATE_ICON_FLAG (si une mise  jour d'icne(s) a t dtecte ou ralise)
 *			ou \p UPDATE_HELP_FLAG (si une mise  jour de fichier d'aide a t dtecte ou ralise)
 **/
BYTE CUpdateThread::UpdateTraiteXml()
{
	// Rcupration de la version de PTVM fonctionnant actuellement, pour comparaison :
	AppVersion	sAppVers(PRODUCT_VERSION_STR);

	if (!pUpdSta->xmlVersionDocument.FirstChildElement("Versions")) {
		// Fichier bien XML, mais pas la bonne racine, on quitte donc
		AddMsg(TEXT("Fichier de version invalide. Ressayez ultrieurement..."));
		return 0;
	}

	// Commence le parcours des versions dans le fichier XML pars
	BYTE		retval			= 0;

	if (UpdateTraiteAppNode("stable", TEXT("stable")))
		retval |= UPDATE_APP_FLAG;

	myprintf(TEXT("retval=0x%02x, bWithBeta=%u") EOL, retval, bWithBeta);

	if (retval==0 && bWithBeta) {
		// Si (et seulement si) aucune version stable n'a t trouve, et que l'utilisateur
		// souhaite aussi les versions beta, alors on recherche celles-ci.
		// S'il y a  la fois une version stable et une version beta, on ne prend que la
		// stable. La version beta sera ventuellement mise  jour  la recherche suivante.
		if (UpdateTraiteAppNode("beta", TEXT("beta")))
			retval |= UPDATE_BETA_FLAG;
	}

	for (ITERATE_XML_ELEMENTS_NAMED(pUpdSta->xmlVersionDocument.FirstChildElement("Versions"), xmlVersionElem, "version")) {
		if (!CheckVersionVersion(*xmlVersionElem))
			continue;

		// On ignore les versions dont le type de vrification n'a pas t choisi
		if (_stricmp(xmlVersionElem->Attribute("type"), "common")!=0)
			continue;

		// Rcupration des mises  jour des icnes et du fichier d'aide
		for (ITERATE_XML_ELEMENTS(xmlVersionElem, xmlTypeMaj)) {
			if (_stricmp(xmlTypeMaj->Value(), "icones")==0) {
				if (UpdateTraiteIcons(*xmlTypeMaj))
					retval |= UPDATE_ICON_FLAG;
			} else if (_stricmp(xmlTypeMaj->Value(), "fichier")==0) {
				if (UpdateTraiteFileNode(*xmlTypeMaj))
					retval |= UPDATE_HELP_FLAG;
			}
		}
	}

	return retval;
}

/**
 * \brief Appel du processus externe de finalisation de la mise  jour
 *
 * Extraction du processus des ressources, dcompression, et lancement avec lvation
 * de privilges (cette lvation tant paramtre dans le manifeste du processus externe).
 *
 * \param[in]	bAppIncluded	\p true si la mise  jour inclut l'application elle-mme
 * \return		0 si succs, ou bien code d'erreur systme
 **/
DWORD CUpdateTelechargeThread::CallMoveProcess(bool bAppIncluded)
{
	TempPathName	sArgFile;
	DWORD			dwErr = 0, nWritten;
	HANDLE			hArgFile = CreateFile(sArgFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_DELETE,
								NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hArgFile == INVALID_HANDLE_VALUE)
		return GetLastError();

	if (!WriteFile(hArgFile, strArgs.c_str(), DWORD(strArgs.length()*sizeof(TCHAR)), &nWritten, NULL))
		dwErr = GetLastError();

	CloseHandle(hArgFile);
	if (dwErr)
		return dwErr;

	TempPathName	sExeFile(TEXT("PTvM_Autoupdate.exe"));

	{
		RawResource	vCompressed;
		ByteVector	vUpdApp;

		// Extraction de l'application de mise  jour
		if (
			vCompressed.Extraction(hAppInstance, IDR_AUTOUPDATE_EXE) &&
			DecompressLzma(vUpdApp, vCompressed)
		) {
			dwErr = vUpdApp.SaveAsFile(sExeFile);
		}

		// (les variables vCompressed et vUpdApp sont dtruites ici)
	}

	if (dwErr != 0) {
		AddMsg(TEXT("Erreur extraction de l'application de mise  jour, code %u"), dwErr);
		AddMsg(TEXT("-> %s"), SystemErrorToString(dwErr).c_str());
		CloseHandle(hArgFile);
		return dwErr;
	}

	tstring				strArgFile = tstr_printf(TEXT("\"%s\""), sArgFile.szPath);
	SHELLEXECUTEINFO	sShExInf = {
		sizeof(sShExInf),			// cbSize
		SEE_MASK_NOCLOSEPROCESS,	// fMask
		pUpdSta->GetParentWindow(),	// hwnd
		TEXT("open"),				// lpVerb
		sExeFile,					// lpFile
		strArgFile.c_str(),			// lpParameters
		NULL,						// lpDirectory
		SW_SHOW,					// nShow
		NULL,						// hInstApp
		NULL,						// lpIDList
		NULL,						// lpClass
		NULL,						// hkeyClass
		0,							// dwHotKey
		{NULL},						// hIcon /  hMonitor
		NULL,						// hProcess
	};

	if (ShellExecuteEx(&sShExInf)) {
		// Attente de sortie du processus de mise  jour
		if (WaitForSingleObject(sShExInf.hProcess, 10000) == WAIT_OBJECT_0) {
			// Processus termin : rcupration du code retour
			GetExitCodeProcess(sShExInf.hProcess, &dwErr);
			if (dwErr != 0) {
				AddMsg(TEXT("Erreur excution de la mise  jour, code %u"), dwErr);
				AddMsg(TEXT("-> %s"), SystemErrorToString(dwErr).c_str());
			} else {
				AddMsg(TEXT("La mise  jour a t effectue avec succs."), szAppName);
				if (bAppIncluded)
					AddMsg(TEXT("Veuillez relancer %s pour voir les changements."), szAppName);
			}
		} else {
			dwErr = WAIT_TIMEOUT;
			// Time out
			AddMsg(TEXT("Erreur : processus de mise--jour non termin"));
			if (!TerminateProcess(sShExInf.hProcess, 0)) {
				dwErr = GetLastError();
				AddMsg(TEXT("Erreur terminaison force processus, code %u"), dwErr);
				AddMsg(TEXT("-> %s"), SystemErrorToString(dwErr).c_str());
			}
		}
		CloseHandle(sShExInf.hProcess);
	} else {
		dwErr = GetLastError();
		AddMsg(TEXT("Erreur cration processus, code %u"), dwErr);
		AddMsg(TEXT("-> %s"), SystemErrorToString(dwErr).c_str());
	}

	DeleteFile(sArgFile);
	DeleteFile(sExeFile);
	return dwErr;
}

/**
 * On a lanc le tlchargement des nouveaux fichiers
 **/
DWORD CUpdateTelechargeThread::ThreadProc()
{
	DWORD			nRes = 0;

	pUpdSta->SetState(us_Updating);

	BYTE	ret			= UpdateTraiteXml();

	if (!strArgs.empty()) {
		// Appel du processus externe de finalisation de la mise  jour avec lvation de
		// privilges, si des arguments existent  lui passer
		nRes = CallMoveProcess((ret & UPDATE_APP_FLAG)!=0);
	}

	if (ret & (UPDATE_APP_FLAG|UPDATE_BETA_FLAG) && nRes == 0) {
		bUpdateNeedRestart = true;
		pUpdSta->SetState(us_NeedRestart);
	} else
		pUpdSta->SetState(us_NeedCheck, ret);

	// Librer le pointeur partag afin que le dialogue puisse se fermer, ce qui aura pour effet de librer la prsente classe.
	pUpdSta.reset();
	return 0;
}

/**
 * Fonction ralisant la connection internet, et les tests pour la vrification de la version,
 * le tout dans un nouveau Thread pour ne pas geler PTVM
 **/
DWORD CUpdateCheckThread::ThreadProc()
{
	InternetFile	vBytVersion;	// Tableau d'octets recevant le fichier XML de mise  jour

	pUpdSta->SetState(us_Checking);

#ifdef _DEBUG
	{	// Le debugging de modules d'accs Internet pose de gros problmes si Visual Studio a t
		// configur pour mettre  jour sa bibliothque de symboles via le serveur de symboles de
		// Microsoft, au fil du chargement des modules : cette mise  jour se faisant par Internet,
		// un "deadlock" peut en rsulter, provoquant un plantage de Visual Studio.
		// Les symboles ne sont normalement chargs qu'une fois, et rechargs seulement
		// si une mise  jour "Windows Update" a chang le module ".dll" correspondant, donc
		// le problme ne survient qu'au premier chargement, et ensuite trs occasionnellement.
		// Le code qui suit vise  prvenir ce blocage en forant le chargement des modules incrimins
		// (donc provoquant le chargement des symboles) pralablement  tout accs Internet rel.
		// Si d'autres modules provoquent le mme type de blocage (on voit lesquels dans la barre d'tat
		// de Visual Studio au moment du "deadlock"), il suffit de les rajouter dans la table qui suit.
		static LPCTSTR aszModules[] = {
			TEXT("rasapi32.dll"),
			TEXT("credssp.dll"),
			TEXT("schannel.dll"),
			TEXT("sensapi.dll"),
			TEXT("rtutils.dll"),
			TEXT("msapsspc.dll"),
			TEXT("msnsspc.dll"),
			TEXT("digest.dll"),
			TEXT("msv1_0.dll")
		};

		for (int nInx = 0; nInx < _countof(aszModules); ++nInx) {
			HMODULE hLib = LoadLibrary(aszModules[nInx]);

			if (hLib)
				FreeLibrary(hLib);
		}
	}
#endif

	pUpdSta->AddProgress();
	GetInternetFile(TEXT(UPDATE_URL), vBytVersion); // Objet traitant de la connexion Internet
#ifdef _DIAGNOSTIC
	if (!vBytVersion.empty()) {
		// Enregistrement du fichier XML reu, en compilation _DIAGNOSTIC seulement
		TempPathName sUpdateFileName("pouchin_update.xml");

		vBytVersion.SauveFichier(sUpdateFileName);
	}
#endif

	UpdateState	eNextState = us_NeedCheck;	// tat suivant par dfaut
	BYTE		retval = 0;

	if (! stop_maj && ! vBytVersion.empty()) {
		AddMsg(TEXT("Traitement des mises  jour..."));
		pUpdSta->AddProgress();
		// On ajoute un caractre '\0'  la fin (car c'est un fichier texte dont il s'agit)
		vBytVersion.resize(vBytVersion.size()+1, 0);

		// Analyse du fichier rcupr
		pUpdSta->xmlVersionDocument.Clear();
		pUpdSta->xmlVersionDocument.Parse(vBytVersion, 0, TIXML_ENCODING_UTF8);
		vBytVersion.clear();

		if (pUpdSta->xmlVersionDocument.Error()) {
			AddMsg(TEXT("Le fichier de version est invalide."));
			AddMsg(TEXT("Je ne peux dterminer la version !!!"));
		} else {
			retval = UpdateTraiteXml();

			if (retval != 0)
				eNextState = us_NeedUpdate;	// mise  jour trouve

			AddMsg(TEXT("Vrification termine"));

			__time64_t	nTim = _time64(NULL);

			update_check_time = static_cast<UINT32>(nTim/60);
		}
	}

	pUpdSta->AddProgress(-1);
	pUpdSta->SetState(eNextState, retval);

	// Librer le pointeur partag afin que le dialogue puisse se fermer, ce qui aura pour effet de librer la prsente classe.
	pUpdSta.reset();
	return 0;
}

void CWebUpdateDialog::EndDialog(INT_PTR nResult)
{
	::EndDialog(hDlg, nResult);
	hDlg = NULL;
	hProgressStatut = NULL;
	hUpdateButton = NULL;
	hUpdateStatut = NULL;
	bOpen = false;
}

/**
 * Fonction de gestion des vnements
 **/
INT_PTR CWebUpdateDialog::DialogProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG:
		bOpen = true;

		hUpdateStatut		= GetDlgItem(hDlg, IDC_UPDATE_LIST);
		hProgressStatut		= GetDlgItem(hDlg, IDC_UPDATE_PROGRESS);
		hUpdateButton		= GetDlgItem(hDlg, IDC_UPDATE_CHECK);

		set_check(hDlg, IDC_UPDATE_AUTO, update_check_auto == 1);
		set_check(hDlg, IDC_UPDATE_BETA, update_include_beta);
		SetState(us_NeedCheck);

		if (bSearchNow) {
			RequestUpdate();
		} else {
			if (update_check_time != 0) {
				AddLineToList(TEXT("La dernire recherche de mise  jour a eu lieu"));
				AddLineToList(strftime_T(static_cast<__time64_t>(update_check_time)*60ULL, TEXT("le %A %#d %B %Y  %Hh%M.")).c_str());
				AddLineToList(TEXT(""));
			}

			AddLineToList(TEXT("Statut inconnu, presser le bouton  Vrifier  ci-dessous."));
			AddLineToList(TEXT(""));
		}
		return TRUE; // WM_INITDIALOG

	case WM_COMMAND:

		switch (LOWORD(wParam)) {

		case IDC_UPDATE_CHECK: {
			SendMessage(hUpdateStatut, LB_RESETCONTENT, 0, 0);	// Efface les messages dans la liste

			if (enregistrements_actuels.recording()) {
				AddLineToList(TEXT("Vous avez un enregistrement en cours."));
				AddLineToList(TEXT("Vous devriez d'abord l'interrompre."));
				return TRUE;
			}

			bool bWithBeta = get_check(hDlg, IDC_UPDATE_BETA);

			switch (eCurrentState) {

			case us_NeedCheck:
				// On lance la recherche de mise  jour
				pUpdateThread.reset(new CUpdateCheckThread(pInstance.lock(), bWithBeta));
				break;

			case us_NeedUpdate:
				// On lance le tlchargement des fichiers
				pUpdateThread.reset(new CUpdateTelechargeThread(pInstance.lock(), bWithBeta));
				break;

			case us_NeedRestart:
				EndDialog(IDOK);
				// Redmarrage (complet) de l'application
				PostMessage(hMainWnd, WM_COMMAND, IDM_QUIT, em_FullRestart);
			}

			return TRUE; }	// IDC_UPDATE_CHECK

		case IDOK:
		case IDCANCEL:
			if (pUpdateThread.get()) {
				AddLineToList(TEXT("Arrt de la mise  jour"));
				// Ordonne au thread de s'arrter
				pUpdateThread->Stop();
			}
			update_check_auto	= get_check(hDlg, IDC_UPDATE_AUTO) ? 1 : 0;
			update_include_beta	= get_check(hDlg, IDC_UPDATE_BETA);
			save_config();	// sauvegarde la configuration
			EndDialog(LOWORD(wParam));
			return TRUE;
		}
		break; // case WM_COMMAND

	} // switch (uMsg)
	return FALSE;
}

weak_ptr<CWebUpdateDialog> CWebUpdateDialog::pInstance;

shared_ptr<CWebUpdateDialog> CWebUpdateDialog::getInstance()
{
	shared_ptr<CWebUpdateDialog> pRes = pInstance.lock();

	if (!pRes.get()) {
		pRes.reset(new CWebUpdateDialog());
		pInstance = pRes;
	}

	return pRes;
}

/**
 * Traitement de la mise  jour par Internet
 *
 * \param[in]  hOwner  handle de la fentre parente
 **/
void do_web_update(HWND hOwner, bool bSearch)
{
	if (bUpdateNeedRestart) {
		MessageBox(hOwner, TEXT("Une mise  jour a dj eu lieu, mais l'application n'a pas t redmarre"),
			TEXT("Mise  jour"), MB_ICONERROR);
		return;
	}

	shared_ptr<CWebUpdateDialog>	pDlg(CWebUpdateDialog::getInstance());

	if (pDlg->bOpen)
		pDlg->RequestUpdate();
	else {
		pDlg->bSearchNow = bSearch;
		pDlg->Do(hOwner);
	}
}

/**
 * Traitement de la mise  jour automatise
 *
 * \param[in]  hOwner  handle de la fentre parente
 **/
void do_web_update_check()
{
	new CUpdateElapsingThread();
}
